/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file BambuParameter.cpp
 * @brief This file contains the implementation of some methods for parameter parsing in Bambu tool
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/

///Autoheader
#include "config_HAVE_ALTERA.hpp"
#include "config_HAVE_BEAGLE.hpp"
#include "config_HAVE_COIN_OR.hpp"
#include "config_HAVE_CUDD.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_GLPK.hpp"
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"
#include "config_HAVE_ICARUS.hpp"
#include "config_HAVE_ILP_BUILT.hpp"
#include "config_HAVE_LP_SOLVE.hpp"
#include "config_HAVE_MODELSIM.hpp"
#include "config_HAVE_XILINX.hpp"
#include "config_HAVE_XILINX_VIVADO.hpp"
#include "config_SKIP_WARNING_SECTIONS.hpp"
#include "config_HAVE_MAPPING_BUILT.hpp"
#include "config_HAVE_VCD_BUILT.hpp"
#include "config_HAVE_I386_GCC45_COMPILER.hpp"
#include "config_HAVE_I386_GCC46_COMPILER.hpp"
#include "config_HAVE_I386_GCC47_COMPILER.hpp"
#include "config_HAVE_I386_GCC48_COMPILER.hpp"
#include "config_HAVE_I386_GCC49_COMPILER.hpp"
#include "config_HAVE_I386_GCC5_COMPILER.hpp"
#include "config_HAVE_FLOPOCO.hpp"
#include "config_HAVE_LIBRARY_CHARACTERIZATION_BUILT.hpp"
#include "config_HAVE_VERILATOR.hpp"
#include "config_PANDA_LIB_INSTALLDIR.hpp"
#include "config_PANDA_DATA_INSTALLDIR.hpp"

///algorithms/clique_covering
#include "clique_covering.hpp"

#if HAVE_HOST_PROFILING_BUILT
///frontend_flow/behavior_analysis
#include "profiling.hpp"
#endif

///HLS/binding/module/algorithms
#include "cdfc_module_binding.hpp"

///HLS/evaluation include
#include "evaluation.hpp"

///HLS/memory include
#include "memory_allocation.hpp"

///HLS/scheduling include
#include "parametric_list_based.hpp"
#if HAVE_ILP_BUILT
#include "sdc_scheduling.hpp"
#endif

#if HAVE_ILP_BUILT
///ilp includes
#include "meilp_solver.hpp"
#endif

///tree include
#include "tree_helper.hpp"

///Header include
#include "BambuParameter.hpp"

///Constants include
#include "constants.hpp"

///creation of the target architecture
#include "datapath_creator.hpp"
#include "chaining.hpp"

#if HAVE_BEAGLE
#if SKIP_WARNING_SECTIONS
#pragma GCC diagnostic ignored "-Wshadow"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Woverloaded-virtual"
#pragma GCC diagnostic ignored "-Wtype-limits"
#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wextra"
#pragma GCC diagnostic ignored "-Wignored-qualifiers"
#pragma GCC diagnostic ignored "-Wsign-compare"
#pragma GCC diagnostic ignored "-Wsign-promo"
#endif
#include "FitnessFunction.hpp"
#include "dse_hls.hpp"
#endif

///Constant option include
#include "constant_strings.hpp"

///STD include
#include <cstring>
#include <iostream>
#include <string.h>

///Technology include
#include "language_writer.hpp"
#include "parse_technology.hpp"
#include "technology_manager.hpp"
#include "target_device.hpp"

///technology/physical_library
#include "technology_node.hpp"

///Utility include
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include "cpu_time.hpp"
#include "dbgPrintHelper.hpp"
#include "fileIO.hpp"
#include <getopt.h>
#include "utility.hpp"

///Wrapper include
#include "gcc_wrapper.hpp"

///Design Space Exploration
#define OPT_ACCEPT_NONZERO_RETURN 256
#define INPUT_OPT_C_NO_PARSE                    1+OPT_ACCEPT_NONZERO_RETURN
#define INPUT_OPT_C_PYTHON_NO_PARSE             1+INPUT_OPT_C_NO_PARSE
#define OPT_ACO_FLOW                            1+INPUT_OPT_C_PYTHON_NO_PARSE
#define OPT_ACO_GENERATIONS                     1+OPT_ACO_FLOW
#define OPT_ADDITIONAL_TOP                      1+OPT_ACO_GENERATIONS
#define OPT_ALIGNED_ACCESS_PARAMETER            1+OPT_ADDITIONAL_TOP
#define OPT_AREA_WEIGHT                         1+OPT_ALIGNED_ACCESS_PARAMETER
#define OPT_BACKEND_SCRIPT_EXTENSIONS_PARAMETER 1+OPT_AREA_WEIGHT
#define OPT_BACKEND_SDC_EXTENSIONS_PARAMETER    1+OPT_BACKEND_SCRIPT_EXTENSIONS_PARAMETER
#define OPT_BRAM_HIGH_LATENCY                   1+OPT_BACKEND_SDC_EXTENSIONS_PARAMETER
#define OPT_CHANNELS_NUMBER                     1+OPT_BRAM_HIGH_LATENCY
#define OPT_CHANNELS_TYPE                       1+OPT_CHANNELS_NUMBER
#define OPT_CLOCK_PERIOD_RESOURCE_FRACTION      1+OPT_CHANNELS_TYPE
#define OPT_DEVICE_NAME                         1+OPT_CLOCK_PERIOD_RESOURCE_FRACTION
#define OPT_DISABLE_BOUNDED_FUNCTION            1+OPT_DEVICE_NAME
#define OPT_DISABLE_FUNCTION_PROXY              1+OPT_DISABLE_BOUNDED_FUNCTION
#define OPT_DISABLE_IOB                         1+OPT_DISABLE_FUNCTION_PROXY
#define OPT_DISTRAM_THRESHOLD                   1+OPT_DISABLE_IOB
#define OPT_DO_NOT_CHAIN_MEMORIES               1+OPT_DISTRAM_THRESHOLD
#define OPT_DO_NOT_EXPOSE_GLOBALS               1+OPT_DO_NOT_CHAIN_MEMORIES
#define OPT_DO_NOT_USE_ASYNCHRONOUS_MEMORIES     1+OPT_DO_NOT_EXPOSE_GLOBALS
#define OPT_DSE                                 1+OPT_DO_NOT_USE_ASYNCHRONOUS_MEMORIES
#define OPT_DSP_ALLOCATION_COEFFICIENT          1+OPT_DSE
#define OPT_DSP_MARGIN_COMBINATIONAL            1+OPT_DSP_ALLOCATION_COEFFICIENT
#define OPT_DSP_MARGIN_PIPELINED                1+OPT_DSP_MARGIN_COMBINATIONAL
#define OPT_DUMP_CONSTRAINTS                    1+OPT_DSP_MARGIN_PIPELINED
#define OPT_DYNAMIC_GENERATORS_DIR              1+OPT_DUMP_CONSTRAINTS
#define OPT_DISCREPANCY                         1+OPT_DYNAMIC_GENERATORS_DIR
#define OPT_DISCREPANCY_FORCE                   1+OPT_DISCREPANCY
#define OPT_DISCREPANCY_NO_LOAD_POINTERS        1+OPT_DISCREPANCY_FORCE
#define OPT_ENABLE_IOB                          1+OPT_DISCREPANCY_NO_LOAD_POINTERS
#define OPT_EVALUATION                          1+OPT_ENABLE_IOB
#define OPT_EVALUATION_MODE                     1+OPT_EVALUATION
#define OPT_EXPERIMENTAL_SETUP                  1+OPT_EVALUATION_MODE
#define OPT_FIXED_SCHED                         1+OPT_EXPERIMENTAL_SETUP
#define OPT_FLOPOCO                             1+OPT_FIXED_SCHED
#define OPT_GENERATE_VCD                        1+OPT_FLOPOCO
#define OPT_GENERATION                          1+OPT_GENERATE_VCD
#define OPT_HLS_DIV                             1+OPT_GENERATION
#define OPT_HOST_PROFILING                      1+OPT_HLS_DIV
#define OPT_ILP                                 1+OPT_HOST_PROFILING
#define OPT_ILP_NEWFORM                         1+OPT_ILP
#define OPT_ILP_SOLVER                          1+OPT_ILP_NEWFORM
#define INPUT_OPT_FILE_INPUT_DATA               1+OPT_ILP_SOLVER
#define OPT_INSERT_MEMORY_PROFILE               1+INPUT_OPT_FILE_INPUT_DATA
#define OPT_INSERT_VERIFICATION_OPERATION       1+OPT_INSERT_MEMORY_PROFILE
#define OPT_LIST_BASED                          1+OPT_INSERT_VERIFICATION_OPERATION
#define OPT_LOGICAL_OPTIMIZATION                1+OPT_LIST_BASED
#define OPT_MAX_EVALUATIONS                     1+OPT_LOGICAL_OPTIMIZATION
#define OPT_MAX_INHERITANCE                     1+OPT_MAX_EVALUATIONS
#define OPT_MAX_SIM_CYCLES                      1+OPT_MAX_INHERITANCE
#define OPT_MAX_ULP                             1+OPT_MAX_SIM_CYCLES
#define OPT_MEMORY_MAPPED_TOP                   1+OPT_MAX_ULP
#define OPT_MEM_DELAY_READ                      1+OPT_MEMORY_MAPPED_TOP
#define OPT_MEM_DELAY_WRITE                     1+OPT_MEM_DELAY_READ
#define OPT_MEMORY_BANKS_NUMBER                 1+OPT_MEM_DELAY_WRITE
#define OPT_MIN_INHERITANCE                     1+OPT_MEMORY_BANKS_NUMBER
#define OPT_MOSA_FLOW                           1+OPT_MIN_INHERITANCE
#define OPT_NO_MIXED_DESIGN                     1+OPT_MOSA_FLOW
#define OPT_NUM_THREADS                         1+OPT_NO_MIXED_DESIGN
#define OPT_PARALLEL_CONTROLLER                 1+OPT_NUM_THREADS
#define OPT_PERIOD_CLOCK                        1+OPT_PARALLEL_CONTROLLER
#define OPT_POWER_OPTIMIZATION                  1+OPT_PERIOD_CLOCK
#define OPT_PRAGMA_PARSE                        1+OPT_POWER_OPTIMIZATION
#define OPT_PRETTY_PRINT                        1+OPT_PRAGMA_PARSE
#define OPT_POST_RESCHEDULING                   1+OPT_PRETTY_PRINT
#define OPT_REGISTER_ALLOCATION                 1+OPT_POST_RESCHEDULING
#define OPT_REGISTERED_INPUTS                   1+OPT_REGISTER_ALLOCATION
#define OPT_RESET                               1+OPT_REGISTERED_INPUTS
#define OPT_LEVEL_RESET                         1+OPT_RESET
#define OPT_SCHEDULING_MUX_MARGINS              1+OPT_LEVEL_RESET
#define OPT_SDC_SCHEDULING                      1+OPT_SCHEDULING_MUX_MARGINS
#define OPT_SERIALIZE_MEMORY_ACCESSES           1+OPT_SDC_SCHEDULING
#define OPT_SILP                                1+OPT_SERIALIZE_MEMORY_ACCESSES
#define OPT_SIMULATE                            1+OPT_SILP
#define OPT_SKIP_PIPE_PARAMETER                 1+OPT_SIMULATE
#define OPT_SOFT_FLOAT                          1+OPT_SKIP_PIPE_PARAMETER
#define OPT_STG                                 1+OPT_SOFT_FLOAT
#define OPT_TESTBENCH                           1+OPT_STG
#define OPT_TESTBENCH_EXTRA_GCC_FLAGS           1+OPT_TESTBENCH
#define OPT_TIME_WEIGHT                         1+OPT_TESTBENCH_EXTRA_GCC_FLAGS
#define OPT_TIMING_MODEL                        1+OPT_TIME_WEIGHT
#define OPT_TIMING_VIOLATION                    1+OPT_TIMING_MODEL
#define OPT_TOP_FNAME                           1+OPT_TIMING_VIOLATION
#define OPT_TOP_RTLDESIGN_NAME                  1+OPT_TOP_FNAME
#define OPT_UNALIGNED_ACCESS_PARAMETER          1+OPT_TOP_RTLDESIGN_NAME
#define OPT_XML_CONFIG                          1+OPT_UNALIGNED_ACCESS_PARAMETER

/// constant correspond to the "parametric list based option"
#define PAR_LIST_BASED_OPT "parametric-list-based"
/// constant correspond to the fixed scheduling option
#define FIXED_SCHEDULING_OPT "fixed-scheduling"

static bool is_evaluation_objective_string(
      const std::vector<std::string> & obj_vec,
      const std::string & s)
{
   return std::find(obj_vec.begin(), obj_vec.end(), s) != obj_vec.end();
}

static void add_evaluation_objective_string(
      std::string & obj_string,
      const std::string & obj_to_add)
{
   if (obj_string.empty())
   {
      obj_string = obj_to_add;
      return;
   }
   std::vector<std::string> obj_vec =
      convert_string_to_vector<std::string>(obj_string, ",");
   const std::vector<std::string> obj_vec_to_add =
      convert_string_to_vector<std::string>(obj_to_add, ",");
   std::unordered_set<std::string> objs;

   for (const auto & s : obj_vec_to_add)
   {
      if (not is_evaluation_objective_string(obj_vec, s))
      {
         obj_string += "," + s;
         obj_vec.push_back(s);
      }
   }

   return;
}

void BambuParameter::PrintHelp(std::ostream &os) const
{
   os
   << "Usage:\n"
   << "       bambu [Options] <source_file> [<constraints_file>] [<technology_file>]\n\n"
   << "Options:\n\n";
   PrintGeneralOptionsUsage(os);
   PrintOutputOptionsUsage(os);
   os
   << "    --pretty-print=<file>\n"
   << "        C-based pretty print of the internal IR.\n\n"
   << "    --writer,-w<language>\n"
   << "        Output RTL language:\n"
   << "            V - Verilog (default)\n"
#if HAVE_EXPERIMENTAL
   << "            S - SystemC\n"
#endif
   << "            H - VHDL\n\n"
   << "    --no-mixed-design\n"
   << "        Avoid mixed design.\n\n"
   << "    --generate-tb=<file>\n"
   << "        Generate testbench for the input values defined in the specified XML\n"
   << "        file.\n\n"
   << "    --top-fname=<fun_name>\n"
   << "        Define the top function to be synthesized.\n\n"
   << "    --top-rtldesign-name=<top_name>\n"
   << "        Define the top module name for the RTL backend.\n\n"
   << "    --file-input-data=<file_list>\n"
   << "        A comma-separated list of input files used by the C specification.\n\n"
   << "    --C-no-parse=<file>\n"
   << "        Specify a comma-separated list of C files used only during the\n"
   << "        co-simulation phase.\n\n"
   << std::endl;

   PrintGccOptionsUsage(os);

   // Target options
   os
   << "  Target:\n\n"
   << "    --target-file=file, -b<file>\n"
   << "        Specify an XML description of the target device.\n\n"
   << "    --generate-interface=<type>\n"
   << "        Wrap the top level module with an external interface.\n"
   << "        Possible values for <type> and related interfaces:\n"
   << "            minimal  -  (minimal interface - default)\n"
   << "            WB4      -  (WishBone 4 interface)\n"
#if HAVE_EXPERIMENTAL
   << "            AXI4LITE -  (AXI4-Lite interface)\n"
   << "            FSL      -  (interface to the FSL bus)\n"
   << "            NPI      -  (interface to the NPI bus)\n"
#endif
   << "\n"
#if HAVE_EXPERIMENTAL
   << "    --edk-config <file>\n"
   << "        Specify the configuration file for Xilinx EDK.\n\n"
#endif
   << std::endl;
#if HAVE_EXPERIMENTAL

   // Frontend analysis options
   os
   << "  Frontend analysis:\n\n"
   << "    --pdg-reduction[=<algorithm>]\n"
   << "        Perform Program Dependence Graph reduction. The reduced PDG is used to\n"
   << "        build the parallel controller, if enabled.\n"
   << "        Possible values for <algorithm> are:\n"
   << "            GP - Girkar-Polychronopoulos algorithm (default)\n"
   << "            P  - Polimi algorithm (not supported yet)\n\n"
   << std::endl;
#endif

   // HLS options
   os
   << "  High Level Synthesis:\n\n"
   << "    --parametric-list-based[=<type>]\n"
   << "        Perform priority list-based scheduling. This is the default scheduling algorithm\n"
   << "        in bambu. The optional <type> argument can be used to set options for\n"
   << "        list-based scheduling as follows:\n"
   << "            0 - Dynamic mobility (default)\n"
   << "            1 - Static mobility\n"
   << "            2 - Priority-fixed mobility\n\n"
   << "    --post-rescheduling\n"
   << "        Perform post rescheduling to better distribute resources.\n\n"
#if HAVE_ILP_BUILT
#if HAVE_EXPERIMENTAL
   << "    --ilp-solver=<solver>\n"
   << "        Sets the ilp solver. Possible values for <solver> are:\n"
   << "            G - Solve the ilp problem using glpk solver (default)\n"
#if HAVE_COIN_OR
   << "            C - Solve the ilp problem using the COIN-OR solver\n"
#endif
#if HAVE_LP_SOLVE
   << "            L - Solve the ilp problem using lp_solve solver\n"
#endif
   << "\n"
   << "    --ilp\n"
   << "        Perform scheduling by using the integer linear programming formulation.\n"
   << "        Default: off.\n\n"
   << "    --ilp-newform\n"
   << "        Perform scheduling by using the integer linear programming with the new\n"
   << "        formulation. Default: off.\n\n"
   << "    --silp\n"
   << "        Perform scheduling by using the scheduling and allocation with ILP\n"
   << "        formulation. Default: off.\n\n"
#endif
   << "    --speculative-sdc-scheduling\n"
   << "        Perform scheduling by using speculative sdc.\n\n"
#endif
   << "    --fixed-scheduling=<file>\n"
   << "        Provide scheduling as an XML file.\n\n"
   << "    --no-chaining\n"
   << "        Disable chaining optimization.\n\n"
   << std::endl;

#if HAVE_EXPERIMENTAL
   // Controller style options
   os
   << "  Controller style:\n\n"
   << "    --stg[=<type>]\n"
   << "        Create the FSM-based controller. This is default for bambu. The\n"
   << "        optional argument <type> can be used to set options for FSM-based\n"
   << "        controller as follows:\n"
   << "            0 - Basic Block-based (exploiting scheduling information) (default)\n\n"
   << "    --parallel-controller\n"
   << "        Create the parallel controller\n\n"
   << std::endl;
#endif

   // Binding options
   os
   << "  Binding:\n\n"
#if HAVE_EXPERIMENTAL
   << "    --storage-value-insertion\n"
   << "        Specify the storage value insertion algorithm:\n"
   << "            0 - VARIABLE/VALUES (default)\n"
   << "            1 - VALUES_INSTANCES\n"
   << "            2 - LIMITED_VALUES_INSTANCES\n"
   << "            3 - PAULIN_SCHEME\n\n"
#endif
#if HAVE_EXPERIMENTAL
   << "    --explore-fu-reg=[string]\n"
   << "        Perform simultaneous FU/reg binding.\n\n"
#endif
   << "    --register-allocation=<type>\n"
   << "        Set the algorithm used for register allocation. Possible values for the\n"
   << "        <type> argument are the following:\n"
   << "            WEIGHTED_COLORING   - use weighted coloring algorithm (default)\n"
   << "            COLORING            - use simple coloring algorithm\n"
   << "            CHORDAL_COLORING    - use chordal coloring algorithm\n"
   << "            BIPARTITE_MATCHING  - use bipartite matching algorithm\n"
   << "            TTT_CLIQUE_COVERING - use a weighted clique covering algorithm\n"
   << "            UNIQUE_BINDING      - unique binding algorithm\n"
#if HAVE_EXPERIMENTAL
   << "            K_COFAMILY          - use k_cofamily algorithm\n"
   << "            LEFT_EDGE           - use left edge algorithm\n"
   << "            CYCLIC_ALLOCATION   - use cyclic allocation algorithm\n"
#endif
   << "\n"
   << "    --module-binding=<type>\n"
   << "        Set the algorithm used for module binding. Possible values for the\n"
   << "        <type> argument are one the following:\n"
   << "            WEIGHTED_TS        - solve the weighted clique covering problem by\n"
   << "                                 exploiting the Tseng&Siewiorek heuristics\n"
   << "                                 (default)\n"
   << "            WEIGHTED_COLORING  - solve the weighted clique covering problem\n"
   << "                                 performing a coloring on the conflict graph\n"
   << "            COLORING           - solve the unweighted clique covering problem\n"
   << "                                 performing a coloring on the conflict graph\n"
   << "            TTT_FAST           - use Tomita, A. Tanaka, H. Takahashi maxima\n"
   << "                                 weighted cliques heuristic to solve the clique\n"
   << "                                 covering problem\n"
   << "            TTT_FAST2          - use Tomita, A. Tanaka, H. Takahashi maximal\n"
   << "                                 weighted cliques heuristic to incrementally\n"
   << "                                 solve the clique covering problem\n"
   << "            TTT_FULL           - use Tomita, A. Tanaka, H. Takahashi maximal\n"
   << "                                 weighted cliques algorithm to solve the clique\n"
   << "                                 covering problem\n"
   << "            TTT_FULL2          - use Tomita, A. Tanaka, H. Takahashi maximal\n"
   << "                                 weighted cliques algorithm to incrementally\n"
   << "                                 solve the clique covering problem\n"
   << "            TS                 - solve the unweighted clique covering problem\n"
   << "                                 by exploiting the Tseng&Siewiorek heuristic\n"
   << "            BIPARTITE_MATCHING - solve the weighted clique covering problem\n"
   << "                                 exploiting the bipartite matching approach\n"
#if HAVE_EXPERIMENTAL
   << "            RANDOMIZED         - solve the weighted clique covering problem\n"
   << "                                 exploiting a randomized approach\n"
#endif
   << "            UNIQUE             - use a 1-to-1 binding algorithm\n\n"
   << std::endl;

   // Memory allocation options
   os
   << "  Memory allocation:\n\n"
   << "    --memory-allocation=<type>\n"
   << "        Set the algorithm used for memory allocation. Possible values for the\n"
   << "        type argument are the following:\n"
   << "            DOMINATOR          - all local variables, static variables and\n"
   << "                                 strings are allocated on BRAMs (default)\n"
   << "            XML_SPECIFICATION  - import the memory allocation from an XML\n"
   << "                                 specification\n\n"
   << "    --xml-memory-allocation=<xml_file_name>\n"
   << "        Specify the file where the XML configuration has been defined.\n\n"
   << "    --memory-allocation-policy=<type>\n"
   << "        Set the policy for memory allocation. Possible values for the <type>\n"
   << "        argument are the following:\n"
   << "            ALL_BRAM           - all objects that need to be stored in memory\n"
   << "                                 are allocated on BRAMs (default)\n"
   << "            LSS                - all local variables, static variables and\n"
   << "                                 strings are allocated on BRAMs\n"
   << "            GSS                - all global variables, static variables and\n"
   << "                                 strings are allocated on BRAMs\n"
   << "            NO_BRAM            - all objects that need to be stored in memory\n"
   << "                                 are allocated on an external memory\n"
   << "            EXT_PIPELINED_BRAM - all objects that need to be stored in memory\n"
   << "                                 are allocated on an external pipelined memory\n\n"
   << "   --base-address=address\n"
   << "        Define the starting address for objects allocated externally to the top\n"
   << "        module.\n\n"
   << "   --initial-internal-address=address\n"
   << "        Define the starting address for the objects allocated internally to the\n"
   << "        top module.\n\n"
   << "   --channels-type=<type>\n"
   << "        Set the type of memory connections.\n"
   << "        Possible values for <type> are:\n"
   << "            MEM_ACC_11 - the accesses to the memory have a single direct\n"
   << "                         connection or a single indirect connection (default)\n"
   << "            MEM_ACC_N1 - the accesses to the memory have n parallel direct\n"
   << "                         connections or a single indirect connection\n"
   << "            MEM_ACC_NN - the accesses to the memory have n parallel direct\n"
   << "                         connections or n parallel indirect connections\n\n"
   << "   --channels-number=<n>\n"
   << "        Define the number of parallel direct or indirect accesses.\n\n"
   << "   --memory-ctrl-type=type\n"
   << "        Define which type of memory controller is used. Possible values for the\n"
   << "        <type> argument are the following:\n"
   << "            D00 - no extra delay (default)\n"
   << "            D10 - 1 clock cycle extra-delay for LOAD, 0 for STORE\n"
   << "            D11 - 1 clock cycle extra-delay for LOAD, 1 for STORE\n"
   << "            D21 - 2 clock cycle extra-delay for LOAD, 1 for STORE\n\n"
#if HAVE_EXPERIMENTAL
   << "    --memory-banks-number=<n>\n"
   << "        Define the number of memory banks.\n\n"
#endif
   << "    --sparse-memory[=on/off]\n"
   << "        Control how the memory allocation happens.\n"
   << "            on - allocate the data in addresses which reduce the decoding logic (default)\n"
   << "           off - allocate the data in a contiguous addresses.\n\n"
   << "    --do-not-use-asynchronous-memories\n"
   << "        Do not add asynchronous memories to the possible set of memories used\n"
   << "        by bambu during the memory allocation step.\n\n"
   << "    --distram-threshold=value\n"
   << "        Define the threshold in bitsize used to infer DISTRIBUTED/ASYNCHRONOUS RAMs (default 256).\n\n"
   << "    --serialize-memory-accesses\n"
   << "        Serialize the memory accesses using the GCC virtual use-def chains\n"
   << "        without taking into account any alias analysis information.\n\n"
   << "    --unaligned-access\n"
   << "        Use only memories supporting unaligned accesses.\n\n"
   << "    --aligned-access\n"
   << "        Assume that all accesses are aligned and so only memories supporting aligned\n\n"
   << "        accesses are used.\n\n"
   << "    --do-not-chain-memories\n"
   << "        When enabled LOADs and STOREs will not be chained with other\n"
   << "        operations.\n\n"
   << "    --bram-high-latency\n"
   << "        Assume a 'high latency bram'-'faster clock frequency' block RAM memory\n"
   << "        based architecture: LOAD(II=1,L=3) STORE(1).\n\n"
   << "    --mem-delay-read=value\n"
   << "        Define the external memory latency when LOAD are performed (default 2).\n\n"
   << "    --mem-delay-write=value\n"
   << "        Define the external memory latency when LOAD are performed (default 1).\n\n"
   << "    --do-not-expose-globals\n"
   << "        All global variables are considered local to the compilation units.\n\n"
   << "    --data-bus-bitsize=<bitsize>\n"
   << "        Set the bitsize of the external data bus.\n\n"
   << "    --addr-bus-bitsize=<bitsize>\n"
   << "        Set the bitsize of the external address bus.\n\n"
   << std::endl;

   // Interconnection options
#if HAVE_EXPERIMENTAL
   os
   << "  Interconnection:\n\n"
   << "    --interconnection, -C[<type>]\n"
   << "        Perform interconnection binding. Possible values for <type> are:\n"
   << "            M - mux-based architecture (default)\n\n"
   << std::endl;

   // HLS Design Space Exploration Options
#if HAVE_BEAGLE
   os
   << "  High-Level Synthesis Design Space Exploration:\n\n"
   << "    --dse=<type>\n"
   << "        Perform design space exploration (default off). Possible values for\n"
   << "        <type> are the following:\n"
   << "            MOGA  - Multi-objective Genetic Algorithm with NSGA-II\n"
   << "            MOSA  - Multi-objective Simulated Annealing\n"
   << "            MOTS  - Multi-objective Tabu Search\n"
   << "            ACO   - Multi-objective Ant Colony Optimization\n"
   << "            GRASP - Grasp\n\n"
   << "    --max-evaluations=<num>\n"
   << "        Perform no more than the specified number of evaluations.\n\n"
   << "    --aco-flow[=<number of ants>]\n"
   << "        Use Ant Colony Optimization (default off).\n\n"
   << "    --generations[=<num>]\n"
   << "        Number ant colony's generations\n\n"
   << "    --mixed_synthesis, -H[<type>]\n"
   << "        Determine the exploration space. Possible values for <type> are:\n"
   << "            0 - based on operation binding information (default)\n"
   << "            1 - based on scheduling priority information\n"
   << "            2 - based on binding and scheduling information\n"
   << std::endl;
#endif
#endif

   // Options for Evaluation of HLS results
   os
   << "  Evaluation of HLS results:\n\n"
#if HAVE_ICARUS || HAVE_XILINX || HAVE_VERILATOR || HAVE_MODELSIM
   << "    --simulate\n"
   << "        Simulate the RTL implementation.\n\n"
   << "    --simulator=<type>\n"
   << "        Specify the simulator used in generated simulation scripts:\n"
#if HAVE_MODELSIM
   << "            MODELSIM - Mentor Modelsim\n"
#endif
#if HAVE_XILINX
   << "            XSIM - Xilinx XSim\n"
   << "            ISIM - Xilinx iSim\n"
#endif
#if HAVE_ICARUS
   << "            ICARUS - Verilog Icarus simulator\n"
#endif
#if HAVE_VERILATOR
   << "            VERILATOR - Verilator simulator\n"
#endif
   << "\n"
   << "    --max-sim-cycles=<cycles>\n"
   << "        Specify the maximum number of cycles a HDL simulation may run.\n"
   << "        (default 20000000).\n\n"
   << "    --accept-nonzero-return\n"
   << "        Do not assume that application main must return 0.\n\n"
   << "    --generate-vcd\n"
   << "        Enable .vcd output file generation for waveform visualization (requires\n"
   << "        testbench generation).\n\n"
#endif
   << "    --evaluation[=type]\n"
   << "        Perform evaluation of the results.\n"
   << "        The value of 'type' selects the objectives to be evaluated\n"
   << "        If nothing is specified all the following are evaluated\n"
   << "        The 'type' argument can be a string containing any of the following\n"
   << "        strings, separated with commas, without spaces:\n"
   << "            AREA            - Area usage\n"
   << "            AREAxTIME       - Area x Latency product\n"
   << "            TIME            - Latency for the average computation\n"
   << "            TOTAL_TIME      - Latency for the whole computation\n"
   << "            CYCLES          - n. of cycles for the average computation\n"
   << "            TOTAL_CYCLES    - n. of cycles for the whole computation\n"
   << "            BRAMS           - number of BRAMs\n"
   << "            CLOCK_SLACK     - Slack between actual and required clock period\n"
   << "            DSPS            - number of DSPs\n"
#if HAVE_EXPERIMENTAL
   << "            EDGES_REDUCTION - Performance evaluation for dependence reduction\n"
#endif
   << "            FREQUENCY       - Maximum target frequency\n"
#if HAVE_EXPERIMENTAL
   << "            NUM_AF_EDGES    - Number of incoming edges in the FAFG\n"
   << "                              techniques\n"
#endif
   << "            PERIOD          - Actual clock period\n"
   << "            REGISTERS       - number of registers\n"
   << "\n"
#if HAVE_EXPERIMENTAL
   << "    --evaluation-mode[=type]\n"
   << "        Perform evaluation of the results:\n"
   << "            EXACT:  based on actual synthesis and simulation (default)\n"
   << "            LINEAR: based on linear regression. Unlike EXACT it supports\n"
   << "                    only the evaluation of the following objectives:\n"
   << "                    - AREA\n"
   << "                    - CLOCK_SLACK\n"
   << "                    - TIME\n"
   << "\n"
   << "    --timing-simulation\n"
   << "        Perform a simulation considering the timing delays.\n\n"
   << "    --timing-violation\n"
   << "        Aborts if synthesized circuit does not meet the timing.\n\n"
#endif // HAVE_EXPERIMENTAL
   << std::endl;

   // Checks and debugging options
#if (HAVE_EXPERIMENTAL && (HAVE_XILINX || HAVE_MODELSIM || HAVE_VERILATOR || HAVE_ICARUS) && HAVE_VCD_BUILT) || HAVE_MODELSIM
   os
   << "  Checks and debugging:\n\n"
#if HAVE_MODELSIM
   << "    --assert-debug\n"
   << "        Enable assertion debugging performed by Modelsim.\n\n"
#endif
   << std::endl;
#endif

   // Export options
#if HAVE_EXPERIMENTAL
   os
   << "  Export:\n\n"
   << "    --export-core <type>\n"
   << "        Exports the generated accelerator:\n"
   << "            PCORE         - Xilinx XPS pcore\n\n"
   << std::endl;
#endif

   // RTL synthesis options
   os
   << "  RTL synthesis:\n\n"
   << "    Note: for a more complete evaluation you should use the option --evaluation\n\n"
   << "    --clock-period=value\n"
   << "        Specify the period of the clock signal (default = 10ns).\n\n"
   << "    --backend-script-extensions=file\n"
   << "        Specify a file that will be included in the backend specific synthesis\n"
   << "        scripts.\n\n"
   << "    --backend-sdc-extensions=file\n"
   << "        Specify a file that will be included in the Synopsys Design Constraints\n"
   << "        file (SDC).\n\n"
   << "    --device-name=value\n"
   << "        Specify the name of the device. Three different cases are foreseen:\n"
   << "            - Xilinx:  a comma separated string specifying device, speed grade\n"
   << "                       and package (e.g.,: \"xc7z020,-1,clg484,VVD\")\n"
   << "            - Altera:  a string defining the device string (e.g. EP2C70F896C6)\n"
   << "            - Lattice: a string defining the device string (e.g.\n"
   << "                       LFE335EA8FN484C)\n\n"
   << "    --power-optimization\n"
   << "        Enable Xilinx power based optimization (default no).\n\n"
   << "    --no-iob\n"
   << "        Disconnect primary ports from the IOB (the default is to connect\n"
   << "        primary input and outpur ports to IOBs).\n\n"
   << "    --soft-float\n"
   << "        Enable use of soft-based implementation of floating-point operations.\n"
   << "        This is the default for bambu.\n\n"
#if HAVE_FLOPOCO
   << "    --flopoco\n"
   << "        Enable use of flopoco-based implementation of floating-point operations\n\n"
#endif
   << "    --max-ulp\n"
   << "        Define the maximal ULP (Unit in the last place, i.e., is the spacing\n"
   << "        between floating-point numbers) accepted.\n\n"
   << "    --hls-div\n"
   << "        Perform the high-level synthesis of integer division and modulo\n"
   << "        operations starting from a C library based implementation.\n\n"
   << "    --skip-pipe-parameter=<value>\n"
   << "        Used during the allocation of pipelined units. <value> specifies how\n"
   << "        many pipelined units, compliant with the clock period, will be skipped.\n"
   << "        (default=0).\n\n"
   << "    --reset-type=value\n"
   << "        Specify the type of reset:\n"
   << "             no    - use registers without reset (default)\n"
   << "             async - use registers with asynchronous reset\n"
   << "             sync  - use registers with synchronous reset\n\n"
   << "    --reset-level=value\n"
   << "        Specify if the reset is active high or low:\n"
   << "             low   - use registers with active low reset (default)\n"
   << "             high  - use registers with active high reset\n\n"
   << "    --registered-inputs=value\n"
   << "        Specify if inputs are registered or not:\n"
   << "             auto  - inputs are registered only for proxy functions (default)\n"
   << "             yes   - all inputs are registered\n"
   << "             no    - none of the inputs is registered\n\n"
   << "    --cprf=value\n"
   << "        Clock Period Resource Fraction (default = 1.0).\n\n"
   << "    --DSP-allocation-coefficient=value\n"
   << "        During the allocation step the timing of the DSP-based modules is\n"
   << "        multiplied by value (default = 1.0).\n\n"
   << "    --DSP-margin-combinational=value\n"
   << "        Timing of combinational DSP-based modules is multiplied by value.\n"
   << "        (default = 1.0).\n\n"
   << "    --DSP-margin-pipelined=value\n"
   << "        Timing of pipelined DSP-based modules is multiplied by value.\n"
   << "        (default = 1.0).\n\n"
   << "    --mux-margins=n\n"
   << "        Scheduling reserves a margin corresponding to the delay of n 32 bit\n"
   << "        multiplexers.\n\n"
   << "    --timing-model=value\n"
   << "        Specify the timing model used by HLS:\n"
   << "             EC     - estimate timing overhead of glue logics and connections\n"
   << "                      between resources (default)\n"
   << "             SIMPLE - just consider the resource delay\n\n"
   << "    --experimental-setup=<setup>\n"
   << "        Specify the experimental setup. This is a shorthand to set multiple\n"
   << "        options with a single command.\n"
   << "        Available values for <setup> are the follwing:\n"
   << "             BAMBU-AREA           - this setup implies:\n"
   << "                                    -Os  -D'printf(fmt, ...)='\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    --DSP-allocation-coefficient=1.75\n"
   << "                                    --distram-threshold=256\n"
   << "             BAMBU-AREA-MP        - this setup implies:\n"
   << "                                    -Os  -D'printf(fmt, ...)='\n"
   << "                                    --channels-type=MEM_ACC_NN\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    --DSP-allocation-coefficient=1.75\n"
   << "                                    --distram-threshold=256\n"
   << "             BAMBU-BALANCED       - this setup implies:\n"
   << "                                    -O2  -D'printf(fmt, ...)='\n"
   << "                                    --channels-type=MEM_ACC_11\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    -fgcse-after-reload  -fipa-cp-clone\n"
   << "                                    -ftree-partial-pre  -funswitch-loops\n"
   << "                                    -finline-functions  -fno-ivopts\n"
   << "                                    --param max-inline-insns-auto=25\n"
   << "                                    -fno-tree-loop-ivcanon\n"
   << "                                    --distram-threshold=256\n"
   << "             BAMBU-BALANCED-MP    - (default) this setup implies:\n"
   << "                                    -O2  -D'printf(fmt, ...)='\n"
   << "                                    --channels-type=MEM_ACC_NN\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    -fgcse-after-reload  -fipa-cp-clone\n"
   << "                                    -ftree-partial-pre  -funswitch-loops\n"
   << "                                    -finline-functions  -fno-ivopts\n"
   << "                                    --param max-inline-insns-auto=25\n"
   << "                                    -fno-tree-loop-ivcanon\n"
   << "                                    --distram-threshold=256\n"
   << "             BAMBU-PERFORMANCE    - this setup implies:\n"
   << "                                    -O3  -D'printf(fmt, ...)='\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    --distram-threshold=512\n"
   << "             BAMBU-PERFORMANCE-MP - this setup implies:\n"
   << "                                    -O3  -D'printf(fmt, ...)='\n"
   << "                                    --channels-type=MEM_ACC_NN\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    --distram-threshold=512\n"
   << "             BAMBU                - this setup implies:\n"
   << "                                    -O0 --channels-type=MEM_ACC_11\n"
   << "                                    --memory-allocation-policy=LSS\n"
   << "                                    --distram-threshold=256\n"
   << "             BAMBU092             - this setup implies:\n"
   << "                                    -O3  -D'printf(fmt, ...)='\n"
   << "                                    --timing-model=SIMPLE\n"
   << "                                    --DSP-margin-combinational=1.3\n"
   << "                                    --cprf=0.9  -skip-pipe-parameter=1\n"
   << "                                    --channels-type=MEM_ACC_11\n"
   << "                                    --memory-allocation-policy=LSS\n"
   << "                                    --distram-threshold=256\n"
   << "             VVD                  - this setup implies:\n"
   << "                                    -O3  -D'printf(fmt, ...)='\n"
   << "                                    --channels-type=MEM_ACC_NN\n"
   << "                                    --memory-allocation-policy=ALL_BRAM\n"
   << "                                    --distram-threshold=256\n"
   << "                                    --DSP-allocation-coefficient=1.75\n"
   << "                                    --do-not-expose-globals --cprf=0.875\n\n"
   << std::endl;
#if HAVE_EXPERIMENTAL || HAVE_ILP_BUILT
   os
   << "  Other options:\n\n";
#endif
#if HAVE_EXPERIMENTAL
   os
   << "    --pragma-parse\n"
   << "        Perform source code parsing to extract information about pragmas.\n"
   << "        (default=no).\n\n";
#endif
#if HAVE_EXPERIMENTAL && HAVE_FROM_PRAGMA_BUILT && HAVE_BAMBU_BUILT
   os
   << "    --num-threads\n"
   << "        Set the number of threads in parallel sections (default=4).\n\n";
#endif
#if HAVE_EXPERIMENTAL
   os
   << "    --xml-config <file>\n"
   << "        Define the path to the XML configuration file for the synthesis.\n\n";
#endif
#if HAVE_EXPERIMENTAL
   os
   << "    --logical-optimization=<level>\n"
   << "        Enable logic optimization on EPDG:\n"
   << "             1  - enable optimizations based on CONDITIONAL instructions\n"
   << "                  information\n"
   << "             2  - enable optimizations based on ACTIVATION PATH information\n"
   << "             3  - enable CONDITIONAL and ACTIVATION PATH optimizations\n"
   << "             4  - enable DATA_TRANSITIVITY optimizations\n"
   << "             5  - enable CONDITIONAL and DATA_TRANSITIVITY optimizations\n"
   << "             6  - enable ACTIVATION PATH and DATA_TRANSITIVITY optimizations\n"
   << "             7  - enable CONDITIONAL, ACTIVATION PATH and DATA_TRANSITIVITY\n"
   << "                  optimizations\n"
   << "             8  - enable CONTROL_DATA_TRANSITIVITY optimizations\n"
   << "             9  - enable CONTROL_DATA_TRANSITIVITY and CONDITIONAL instructions\n"
   << "                  optimizations\n"
   << "             16 - enable CONTROL_DATA_CONTROL_FLOW_TRANSITIVITY and CONDITIONAL\n"
   << "                  instructions optimizations\n\n";
#endif
#if HAVE_ILP_BUILT
   os
   << "    --time, -t <time>\n"
   << "        Set maximum execution time (in seconds) for ILP solvers. (infinite).\n\n";
#endif
#if HAVE_HOST_PROFILING_BUILT
   os
   << "    --host-profiling\n"
   << "        Perform host-profiling.\n\n";
#endif
#if HAVE_EXPERIMENTAL || HAVE_ILP_BUILT
   os
   << std::endl;
#endif

   // Discrepancy analysis options
   os
   << "  Debug options:\n\n"
   << "    --discrepancy\n"
   << "           Performs automated discrepancy analysis between the execution\n"
   << "           of the original source code and the generated HDL (currently\n"
   << "           supports only Verilog). If a mismatch is detected reports\n"
   << "           useful information the user.\n"
   << "           Uninitialized variables in C are legal, but if they are used\n"
   << "           before initialization in HDL it is possible to obtain X values\n"
   << "           in simulation. This is not necessarily wrong, so these errors\n"
   << "           are not reported by default to avoid reporting false positives.\n"
   << "           If you can guarantee that in your C code there are no\n"
   << "           uninitialized variables and you want the X values in HDL to be\n"
   << "           reported use the option --discrepancy-force-uninitialized\n\n"
   << "    --discrepancy-force-uninitialized\n"
   << "           Reports errors due to uninitialized values in HDL.\n"
   << "           See the option --discrepancy for details\n\n"
   << "    --discrepancy-no-load-pointers\n"
   << "           Assume that the data loaded from memories in HDL are never used\n"
   << "           to represent addresses, unless they are explicitly assigned to\n"
   << "           pointer variables.\n"
   << "           The discrepancy analysis is able to compare pointers in software\n"
   << "           execution and addresses in hardware. By default all the values\n"
   << "           loaded from memory are treated as if they could contain addresses,\n"
   << "           even if they are integer variables. This is due to the fact that\n"
   << "           C code doing this tricks is valid and actually used in embedded\n"
   << "           systems, but it can lead to imprecise bug reports, because only\n"
   << "           pointers pointing to actual data are checked by the discrepancy\n"
   << "           analysis.\n"
   << "           If you can guarantee that your code always manipulates addresses\n"
   << "           using pointers and never using plain int, then you can use this\n"
   << "           option to get more precise bug reports.\n"
   << std::endl;
}

void BambuParameter::PrintProgramName(std::ostream & os) const
{
   os << std::endl;
   os << "********************************************************************************" << std::endl;
   os << "                    ____                  _" << std::endl;
   os << "                   | __ )  __ _ _ __ ___ | |_   _   _" << std::endl;
   os << "                   |  _ \\ / _` | \'_ ` _ \\| \'_ \\| | | |" << std::endl;
   os << "                   | |_) | (_| | | | | | | |_) | |_| |" << std::endl;
   os << "                   |____/ \\__,_|_| |_| |_|_.__/ \\__,_|" << std::endl;
   os << std::endl;
   os << "********************************************************************************" << std::endl;
   os << "                         High-Level Synthesis Tool" << std::endl;
   os << std::endl;
}

BambuParameter::BambuParameter(const std::string _program_name) : Parameter(_program_name)
{
   SetDefaults();
}

int BambuParameter::exec(int argc, char* const argv[])
{
   exit_code = PARAMETER_NOTPARSED;

   /// flag to check if scheduling algorithm has been already chosen
   bool scheduling_set_p = false;
   /// variable used into option parsing
   int option_index;
   int next_option;

   // Bambu short option. An option character in this string can be followed by a colon (`:') to indicate that it
   // takes a required argument. If an option character is followed by two colons (`::'), its argument is optional;
   // this is a GNU extension.
   const char* const short_options = COMMON_SHORT_OPTIONS_STRING
                                     "c:t:u:H:SC::b:w:"
                                     GCC_SHORT_OPTIONS_STRING;

   const struct option long_options[] =
      {
         COMMON_LONG_OPTIONS,
         ///General options
         {"top-fname",                  required_argument, nullptr, OPT_TOP_FNAME},
         {"top-rtldesign-name",         required_argument, nullptr, OPT_TOP_RTLDESIGN_NAME},
         {"xml-config",                 required_argument, nullptr, OPT_XML_CONFIG},
         {"time",                       required_argument, nullptr, 't'},
         { "file-input-data",           required_argument, nullptr, INPUT_OPT_FILE_INPUT_DATA},
         /// Frontend options
         {"circuit-dbg",                required_argument, nullptr,  0 },
#if HAVE_EXPERIMENTAL
         {"pdg-reduction",              optional_argument, nullptr,  0 },
         ///--- Flow options ---
         {"dump-constraints",           optional_argument, nullptr, OPT_DUMP_CONSTRAINTS},
         /// --- Design Space Exploration options ---
#if HAVE_BEAGLE
         {"dse",                        required_argument, nullptr, OPT_DSE},
         {"max_evaluations",            required_argument, nullptr, OPT_MAX_EVALUATIONS},
         //ACO
         {"aco-flow",                   optional_argument, nullptr, OPT_ACO_FLOW},
         {"generations",                optional_argument, nullptr, OPT_ACO_GENERATIONS},
         //Genetic algorithm
         {"mixed_synthesis",            optional_argument, nullptr, 'H'},
         {"seed",                       required_argument, nullptr, 'A'},
         {"run",                        required_argument, nullptr, 'B'},
         {"population",                 required_argument, nullptr, 'E'},
         {"generation",                 required_argument, nullptr, OPT_GENERATION},
         {"fitness_function",           required_argument, nullptr, 'P'},
         {"inheritance_rate",           required_argument, nullptr, 'F'},
         {"inheritance_mode",           required_argument, nullptr, 'G'},
         {"max_inheritance",            required_argument, nullptr, OPT_MAX_INHERITANCE},
         {"min_inheritance",            required_argument, nullptr, OPT_MIN_INHERITANCE},
         {"distance_rate",              required_argument, nullptr, 'M'},
         {"weighting_function",         required_argument, nullptr, 'N'},
         {"normalize",                  optional_argument, nullptr, 'Q'},
         {"time-weight",                required_argument, nullptr, OPT_TIME_WEIGHT}, //no short option
         {"area-weight",                required_argument, nullptr, OPT_AREA_WEIGHT}, //no short option
#endif
         /// --- Algorithms options ---
         {"explore-mux",                optional_argument,       nullptr,  0 },
         {"explore-fu-reg",             required_argument,       nullptr,  0 },
#endif
         /// Scheduling options
         {FIXED_SCHEDULING_OPT,         required_argument, nullptr, OPT_FIXED_SCHED},
#if HAVE_ILP_BUILT
         {"speculative-sdc-scheduling", no_argument,       nullptr, OPT_SDC_SCHEDULING},
#endif
         {"serialize-memory-accesses",  no_argument,       nullptr, OPT_SERIALIZE_MEMORY_ACCESSES},
         {PAR_LIST_BASED_OPT,           optional_argument, nullptr, OPT_LIST_BASED}, //no short option
         {"post-rescheduling",          no_argument,       nullptr, OPT_POST_RESCHEDULING},
#if HAVE_ILP_BUILT
         {"ilp-solver",                 required_argument, nullptr, OPT_ILP_SOLVER},
         {"ilp",                        no_argument,       nullptr, OPT_ILP},
         {"ilp-newform",                no_argument,       nullptr, OPT_ILP_NEWFORM},
         {"silp",                       no_argument,       nullptr, OPT_SILP},
#endif
         {"speculative",                no_argument,       nullptr, 'S'},
         {"no-chaining",                no_argument,       nullptr,  0 },
         /// Finite state machine options
         {"stg",                        required_argument, nullptr, OPT_STG},
         /// module binding
         {"module-binding",             required_argument, nullptr, 0 },
         /// register allocation
         {"register-allocation",        required_argument, nullptr, OPT_REGISTER_ALLOCATION },
         {"storage-value-insertion",    required_argument, nullptr, 0 },
         /// Memory allocation
         {"memory-allocation",          required_argument, nullptr,  0 },
         {"memory-allocation-policy",   required_argument, nullptr,  0 },
         {"xml-memory-allocation",      required_argument, nullptr,  0 },
         {"base-address",               required_argument, nullptr,  0 },
         {"initial-internal-address",   required_argument, nullptr,  0 },
         {"channels-type",              required_argument, nullptr,  0 },
         {"channels-number",            required_argument, nullptr,  0 },
         {"memory-ctrl-type",           required_argument, nullptr,  0 },
         {"sparse-memory",              optional_argument, nullptr,  0 },
         {"do-not-expose-globals",      no_argument,       nullptr, OPT_DO_NOT_EXPOSE_GLOBALS},
         //interconnections
         {"interconnection",            required_argument, nullptr, 'C' },
#if HAVE_EXPERIMENTAL
         {"parallel-controller",        no_argument,       nullptr, OPT_PARALLEL_CONTROLLER},
#if HAVE_CUDD
         {"logical-optimization",       required_argument, nullptr, OPT_LOGICAL_OPTIMIZATION},
#endif
#endif
         /// evaluation options
         {"evaluation",                 optional_argument, nullptr, OPT_EVALUATION},
#if HAVE_EXPERIMENTAL
         {"evaluation-mode",            required_argument, nullptr, OPT_EVALUATION_MODE},
         {"timing-simulation",          no_argument,       nullptr,  0 },
         {"timing-violation",           no_argument,       nullptr, OPT_TIMING_VIOLATION},
#endif
         {"assert-debug",               no_argument,       nullptr,  0 },
         {"device-name",                required_argument, nullptr, OPT_DEVICE_NAME},
         {"clock-period",               required_argument, nullptr, OPT_PERIOD_CLOCK},
         {"power-optimization",         no_argument,       nullptr, OPT_POWER_OPTIMIZATION},
         {"no-iob",                     no_argument,       nullptr, OPT_DISABLE_IOB},
         {"reset-type",                 required_argument, nullptr, OPT_RESET},
         {"reset-level",                 required_argument, nullptr, OPT_LEVEL_RESET},
         {"soft-float",                 no_argument,       nullptr, OPT_SOFT_FLOAT},
   #if HAVE_FLOPOCO
         {"flopoco",                    no_argument,       nullptr, OPT_FLOPOCO},
#endif
         {"hls-div",                    no_argument,       nullptr, OPT_HLS_DIV},
         {"max-ulp",                    required_argument, nullptr, OPT_MAX_ULP},
         {"skip-pipe-parameter",        required_argument, nullptr, OPT_SKIP_PIPE_PARAMETER},
         {"unaligned-access",           no_argument,       nullptr, OPT_UNALIGNED_ACCESS_PARAMETER},
         {"aligned-access",             no_argument,       nullptr, OPT_ALIGNED_ACCESS_PARAMETER},
         {"backend-script-extensions",  required_argument, nullptr, OPT_BACKEND_SCRIPT_EXTENSIONS_PARAMETER},
         {"backend-sdc-extensions",     required_argument, nullptr, OPT_BACKEND_SDC_EXTENSIONS_PARAMETER},
         {"do-not-use-asynchronous-memories",  no_argument, nullptr, OPT_DO_NOT_USE_ASYNCHRONOUS_MEMORIES},
         {"do-not-chain-memories",      no_argument,       nullptr, OPT_DO_NOT_CHAIN_MEMORIES},
         {"bram-high-latency",           no_argument,      nullptr, OPT_BRAM_HIGH_LATENCY},
         {"cprf",                       required_argument, nullptr, OPT_CLOCK_PERIOD_RESOURCE_FRACTION},
         {"experimental-setup",         required_argument, nullptr, OPT_EXPERIMENTAL_SETUP},
         {"distram-threshold",          required_argument, nullptr, OPT_DISTRAM_THRESHOLD},
         {"DSP-allocation-coefficient", required_argument, nullptr, OPT_DSP_ALLOCATION_COEFFICIENT},
         {"DSP-margin-combinational",   required_argument, nullptr, OPT_DSP_MARGIN_COMBINATIONAL},
         {"DSP-margin-pipelined",       required_argument, nullptr, OPT_DSP_MARGIN_PIPELINED},
         {"mux-margins",                required_argument, nullptr, OPT_SCHEDULING_MUX_MARGINS},
         {"timing-model",               required_argument, nullptr, OPT_TIMING_MODEL},
         {"registered-inputs",          required_argument, nullptr, OPT_REGISTERED_INPUTS},
         ///target options
         {"target-file",                required_argument, nullptr, 'b'},
#if HAVE_EXPERIMENTAL
         {"edk-config",                 required_argument, nullptr,  0 },
#endif
         {"export-core",                required_argument, nullptr,  0 },
         ///Output options
         {"writer",                     required_argument, nullptr, 'w'},
#if HAVE_EXPERIMENTAL
         {"no-mixed-design",            no_argument,       nullptr, OPT_NO_MIXED_DESIGN},
#endif
         {"dynamic-generators-dir",     required_argument, nullptr, OPT_DYNAMIC_GENERATORS_DIR},
         {"pretty-print",               required_argument, nullptr, OPT_PRETTY_PRINT},
#if HAVE_EXPERIMENTAL
         {"pragma-parse",              no_argument,       nullptr, OPT_PRAGMA_PARSE},
#endif
         {"generate-interface",         required_argument, nullptr,  0 },
         {"additional-top",             required_argument, nullptr,  OPT_ADDITIONAL_TOP },
         {"data-bus-bitsize",           required_argument, nullptr,  0 },
         {"addr-bus-bitsize",           required_argument, nullptr,  0 },
#if HAVE_EXPERIMENTAL
         {"resp-model",                 required_argument, nullptr,  0 },
#endif
         {"generate-tb",                required_argument, nullptr, OPT_TESTBENCH},
         {"testbench-extra-gcc-flags",  required_argument, nullptr, OPT_TESTBENCH_EXTRA_GCC_FLAGS},
         {"max-sim-cycles",             required_argument, nullptr, OPT_MAX_SIM_CYCLES},
         {"generate-vcd",               no_argument,       nullptr, OPT_GENERATE_VCD},
         {"simulate",                   no_argument, nullptr, OPT_SIMULATE},
         {"simulator",                  required_argument, nullptr,  0 },
         {"disable-function-proxy",     no_argument,       nullptr, OPT_DISABLE_FUNCTION_PROXY },
         {"disable-bounded-function",   no_argument, nullptr, OPT_DISABLE_BOUNDED_FUNCTION },
         {"memory-mapped-top",          no_argument,       nullptr, OPT_MEMORY_MAPPED_TOP },
         {"mem-delay-read",             required_argument, nullptr, OPT_MEM_DELAY_READ },
         {"mem-delay-write",            required_argument, nullptr, OPT_MEM_DELAY_WRITE },
         {"host-profiling",             no_argument,       nullptr, OPT_HOST_PROFILING},
         {"disable-function-proxy",     no_argument, nullptr, OPT_DISABLE_FUNCTION_PROXY },
         {"memory-mapped-top",          no_argument, nullptr, OPT_MEMORY_MAPPED_TOP },
         {"discrepancy",                no_argument, nullptr, OPT_DISCREPANCY},
         {"discrepancy-force-uninitialized", no_argument, nullptr, OPT_DISCREPANCY_FORCE},
         {"discrepancy-no-load-pointers",    no_argument, nullptr, OPT_DISCREPANCY_NO_LOAD_POINTERS},
#if HAVE_EXPERIMENTAL && HAVE_FROM_PRAGMA_BUILT && HAVE_BAMBU_BUILT
         {"num-threads",                required_argument, nullptr, OPT_NUM_THREADS},
#endif
#if HAVE_EXPERIMENTAL
         {"memory-banks-number",        required_argument, nullptr, OPT_MEMORY_BANKS_NUMBER},
#endif
         { "C-no-parse",                required_argument, nullptr, INPUT_OPT_C_NO_PARSE},
         { "C-python-no-parse",         required_argument, nullptr, INPUT_OPT_C_PYTHON_NO_PARSE},
         {"accept-nonzero-return",      no_argument, nullptr, OPT_ACCEPT_NONZERO_RETURN},
         GCC_LONG_OPTIONS,
         {nullptr, 0, nullptr, 0}
      };

   if (argc == 1)   // Bambu called without arguments, it simple prints help message
   {
      PrintUsage(std::cout);
      return EXIT_SUCCESS;
   }

   while (1)
   {
      next_option = getopt_long(argc, argv, short_options, long_options, &option_index);

      // no more options are available
      if (next_option == -1)
         break;

      switch (next_option)
      {
         ///general options
         case OPT_TOP_FNAME:
         {
            // set name of the function to be take into account as top function
            setOption(OPT_top_function_name, optarg);
            setOption(OPT_top_file, optarg);
            break;
         }
         case OPT_TOP_RTLDESIGN_NAME:
         {
            setOption(OPT_top_design_name, optarg);
            break;
         }
         case 't':
         {
            setOption(OPT_ilp_max_time, optarg);
            break;
         }
         case INPUT_OPT_FILE_INPUT_DATA :
         {
            setOption(OPT_file_input_data, optarg);
            break;
         }
#if HAVE_EXPERIMENTAL
         case OPT_XML_CONFIG:
         {
            setOption(OPT_synthesis_flow, HLSFlowStep_Type::XML_HLS_SYNTHESIS_FLOW);
            setOption(OPT_xml_input_configuration, optarg);
            break;
         }
         case OPT_DUMP_CONSTRAINTS:
         {
            setOption(OPT_hls_flow, HLSFlowStep_Type::DUMP_DESIGN_FLOW);
            setOption("dumpConstraints", true);
            if (optarg)
               setOption("dumpConstraints_file", optarg);
            break;
         }
         ///frontend options
         case 'u': // compute the unrolling degree of a loop
         {
            setOption("compute_unrolling_degree",true);
            if (optarg)
               setOption("IImax", optarg);
            else
               throw "BadParameters: compute unrolling degree needs a additional parameter";
            break;
         }
         case OPT_INSERT_VERIFICATION_OPERATION:
         {
            setOption("insert_verification_operation", true);
            break;
         }
         ///design space exploration options
#if HAVE_BEAGLE
         case OPT_DSE:
         {
            setOption(OPT_hls_flow, hls_flow::DSE);
            if (std::string(optarg) == ("MOSA"))
               setOption("dse_algorithm", dse_hls::MOSA);
            else if (std::string(optarg) == ("MOTS"))
               setOption("dse_algorithm", dse_hls::MOTS);
            else if (std::string(optarg) == ("ACO"))
               setOption("dse_algorithm", dse_hls::ACO);
            else if (std::string(optarg) == ("GRASP"))
               setOption("dse_algorithm", dse_hls::GRASP);
            else if (std::string(optarg) == ("MOGA"))
               setOption("dse_algorithm", dse_hls::MOGA);
            else
               THROW_ERROR("Not supported exploration algorithm: " + std::string(optarg));
            break;
         }
         case OPT_MAX_EVALUATIONS:
         {
            setOption("max_evaluations", optarg);
            break;
         }
         //ACO
         case OPT_ACO_FLOW:
         {
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::ACO);
            if (optarg)
               setOption("aco_ant_num", optarg);
            break;
         }
         case OPT_ACO_GENERATIONS:
         {
            if (optarg)
               setOption("aco_generations", optarg);
            break;
         }
         case OPT_MOSA_FLOW:
         {
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::MOSA);
            break;
         }
         //multi-objective genetic algorithm
         case 'H':  // enable mixed high level synthesis
         {
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::MOGA);
            if (optarg)
               setOption("exploration_technique", optarg);
            else
               setOption("exploration_technique", dse_hls::BINDING);
            break;
         }
         case 'Q':
         {
            if (optarg)
               setOption("to_normalize", optarg);
            else
               setOption("to_normalize", 1);
            break;
         }
         case 'P':  // mixed high level synthesis fitness function
         {
            setOption("fitness_function", optarg);
            break;
         }
         case 'A':
         {
            setOption("seed", optarg);
            break;
         }
         case 'B':
         {
            setOption("run", optarg);
            break;
         }
         case 'E':
         {
            setOption("population", optarg);
            break;
         }
         case OPT_GENERATION:
         {
            setOption("GA_generation", optarg);
            break;
         }
         case 'F':
         {
            setOption("fitness_inheritance_rate", optarg);
            break;
         }
         case 'G':
         {
            setOption("inheritance_mode", optarg);
            break;
         }
         case OPT_MAX_INHERITANCE:
         {
            setOption("max_for_inheritance", optarg);
            break;
         }
         case OPT_MIN_INHERITANCE:
         {
            setOption("min_for_inheritance", optarg);
            break;
         }
         case 'M':
         {
            setOption("distance_rate", optarg);
            break;
         }
         case 'N':
         {
            setOption("weighting_function", optarg);
            break;
         }
         case OPT_TIME_WEIGHT:
         {
            setOption("time_weight", optarg);
            break;
         }
         case OPT_AREA_WEIGHT:
         {
            setOption("area_weight", optarg);
            break;
         }
#endif
#endif
         ///scheduling options
         case OPT_FIXED_SCHED:
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::FIXED_SCHEDULING);
            if (optarg)
               setOption("fixed_scheduling_file", optarg);
            break;
         }
         case OPT_LIST_BASED: // enable list based scheduling
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::LIST_BASED_SCHEDULING);
            if (optarg)
               setOption(OPT_scheduling_priority, optarg);
            break;
         }
         case OPT_POST_RESCHEDULING:
         {
            setOption(OPT_post_rescheduling, true);
            break;
         }
#if HAVE_EXPERIMENTAL
#if HAVE_ILP_BUILT
         case OPT_ILP_SOLVER:
         {
#if HAVE_GLPK
            if (optarg[0] == 'G')
               setOption(OPT_ilp_solver, meilp_solver::GLPK);
            else
#endif
#if HAVE_COIN_OR
            if (optarg[0] == 'C')
               setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
            else
#endif
#if HAVE_LP_SOLVE
            if (optarg[0] == 'L')
               setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
            else
#endif
               THROW_ERROR("BadParameters: not recognized ilp solver");
            break;
         }
         case OPT_ILP:
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::ILP_SCHEDULING);
            setOption(OPT_chaining, true);/// ILP formulation tries to chain the operations
            break;
         }
         case OPT_ILP_NEWFORM:
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::ILP_NEW_FORM_SCHEDULING);
            break;
         }
         case OPT_SILP:
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::SILP_SCHEDULING);
            break;
         }
#endif
         case 'S': // enable scheduling with speculative computation
         {
            setOption(OPT_speculative, true);
            break;
         }
#endif
         case OPT_STG:
         {
            setOption(OPT_stg, true);
            if (optarg)
               setOption(OPT_stg_algorithm, optarg);
            break;
         }
         ///binding options
         //register
         case OPT_REGISTER_ALLOCATION: // enable register allocation
         {
            if (std::string(optarg) == "COLORING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::COLORING_REGISTER_BINDING);
            }
            else if(std::string(optarg) == "CHORDAL_COLORING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::CHORDAL_COLORING_REGISTER_BINDING);
            }
#if HAVE_EXPERIMENTAL
            else if(std::string(optarg) == "LEFT_EDGE")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::LEFT_EDGE_REGISTER_BINDING);
            }
            else if(std::string(optarg) == "K_COFAMILY")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::K_COFAMILY_REGISTER_BINDING);
            }
#endif
            else if(std::string(optarg) == "WEIGHTED_COLORING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::WEIGHTED_CLIQUE_REGISTER_BINDING);
               setOption(OPT_weighted_clique_register_algorithm, CliqueCovering_Algorithm::WEIGHTED_COLORING);
            }
            else if(std::string(optarg) == "BIPARTITE_MATCHING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::WEIGHTED_CLIQUE_REGISTER_BINDING);
               setOption(OPT_weighted_clique_register_algorithm, CliqueCovering_Algorithm::BIPARTITE_MATCHING);
            }
            else if(std::string(optarg) == "TTT_CLIQUE_COVERING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::WEIGHTED_CLIQUE_REGISTER_BINDING);
               setOption(OPT_weighted_clique_register_algorithm, CliqueCovering_Algorithm::TTT_CLIQUE_COVERING);
            }
            else if(std::string(optarg) == "UNIQUE_BINDING")
            {
               setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::UNIQUE_REGISTER_BINDING);
            }
            else
            {
               THROW_ERROR("BadParameters: register allocation not correctly specified");
            }
            break;
         }
#if HAVE_EXPERIMENTAL
         case OPT_TIMING_VIOLATION:
         {
            setOption(OPT_timing_violation_abort, true);
            break;
         }
#endif
         //interconnection
         case 'C':
         {
            if (std::string(optarg) == "M")
               setOption(OPT_datapath_interconnection_algorithm, HLSFlowStep_Type::MUX_INTERCONNECTION_BINDING);
            else
               throw "BadParameters: interconnection binding not correctly specified";
            break;
         }
#if HAVE_EXPERIMENTAL
         case OPT_PARALLEL_CONTROLLER:
         {
            setOption(OPT_controller_architecture, HLSFlowStep_Type::PARALLEL_CONTROLLER_CREATOR);
            break;
         }
#endif
         ///target options
         case 'b':
         {
            setOption(OPT_target_device_file, optarg);
            break;
         }
         ///evaluation
         case OPT_EVALUATION:
         {
            // set OPT_evaluation, because the evaluation has to be performed
            setOption(OPT_evaluation, true);
            /*
             * check if OPT_evaluation_mode has already been decided (for
             * example with OPT_EVALUATION_MODE). in cas it's already set, we
             * don't overwrite it since OPT_EVALUATION is meant to set the
             * objectives, not the mode, hence the mode set from other options
             * has precedence
             */
            if (not isOption(OPT_evaluation_mode) or
                  getOption<Evaluation_Mode>(OPT_evaluation_mode) == Evaluation_Mode::NONE)
            {
               setOption(OPT_evaluation_mode, Evaluation_Mode::EXACT);
            }
            std::string objective_string =
               getOption<std::string>(OPT_evaluation_objectives);
            std::vector<std::string> objective_vector =
               convert_string_to_vector<std::string>(objective_string, ",");
            if (optarg == nullptr)
            {
               if (isOption(OPT_evaluation_mode) and getOption<Evaluation_Mode>(OPT_evaluation_mode) == Evaluation_Mode::EXACT)
               {
                  std::string to_add =
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT
                        "AREAxTIME,"
                        "AREA,"
                        "TIME,"
                        "TOTAL_TIME,"
                        "CYCLES,"
                        "TOTAL_CYCLES,"
                        "FREQUENCY,"
                        "PERIOD,"
                        "CLOCK_SLACK,"
                        "REGISTERS,"
                        "DSPS,"
                        "BRAMS"
#else
                        "CYCLES"
#endif
                  ;
                  add_evaluation_objective_string(objective_string, to_add);
               }
#if HAVE_EXPERIMENTAL
               else if (isOption(OPT_evaluation_mode) and getOption<Evaluation_Mode>(OPT_evaluation_mode) == Evaluation_Mode::ESTIMATION)
               {
                  add_evaluation_objective_string(objective_string,
                        "AREA,"
                        "TIME,"
                        "CLOCK_SLACK");
               }
#endif
               else
               {
                  THROW_ERROR("BadParameters: invalid evaluation mode");
               }
            }
            else
            {
               add_evaluation_objective_string(objective_string, std::string(optarg));
            }
            setOption(OPT_evaluation_objectives, objective_string);
            break;
         }
#if HAVE_EXPERIMENTAL
         case OPT_EVALUATION_MODE:
         {
            // set OPT_evaluation, because the evaluation has to be performed
            setOption(OPT_evaluation, true);
            /*
             * here don't check if OPT_evaluation mode was already set by
             * someone else. if someone else has already defined
             * OPT_evaluation_mode the --evaluation-mode parameter should
             * overwrite it anyway.
             */
            if (optarg == nullptr)
            {
               throw "BadParameters: evaluation mode must be specified, use EXACT or LINEAR";
            }
            else if (std::string(optarg) == "EXACT")
            {
               setOption(OPT_evaluation_mode, Evaluation_Mode::EXACT);
            }
            else if (std::string(optarg) == "LINEAR")
            {
               setOption(OPT_evaluation_mode, Evaluation_Mode::ESTIMATION);
            }
            else
            {
               throw "BadParameters: evaluation mode not correctly specified, use EXACT or LINEAR";
            }
            break;
         }
#endif
         case OPT_SIMULATE:
         {
            /*
             * OPT_SIMULATE is a shorthand form OPT_EVALUATION with optarg =
             * CYCLES,TOTAL_CYCLES
             */
            // set OPT_evaluation, because the evaluation has to be performed
            setOption(OPT_evaluation, true);
            /*
             * check if OPT_evaluation_mode has already been decided (for
             * example with OPT_EVALUATION_MODE). in cas it's already set, we
             * don't overwrite it since OPT_SIMULATION is meant to set the
             * objectives, not the mode, hence the mode set from other options
             * has precedence
             */
            if (not isOption(OPT_evaluation_mode) or
                  getOption<Evaluation_Mode>(OPT_evaluation_mode) == Evaluation_Mode::NONE)
            {
               setOption(OPT_evaluation_mode, Evaluation_Mode::EXACT);
            }
            else if (isOption(OPT_evaluation_mode) and
                  getOption<Evaluation_Mode>(OPT_evaluation_mode) !=
                  Evaluation_Mode::EXACT)
            {
               THROW_ERROR("Simulation is only supported with EXACT evaluation mode");
            }
            std::string objective_string =
               getOption<std::string>(OPT_evaluation_objectives);
            std::vector<std::string> objective_vector =
               convert_string_to_vector<std::string>(objective_string, ",");
            /*
             * look among the objectives of the evaluation. if "CYCLES" is
             * already there, the simulation will be executed.
             * if it's not, it's enough to add it to enable simulation
             */
            if (objective_string.empty())
            {
               objective_string = "CYCLES";
            }
            else if (std::find(objective_vector.begin(), objective_vector.end(), "CYCLES") == objective_vector.end())
            {
               objective_string = objective_string + ",CYCLES";
            }
            setOption(OPT_evaluation_objectives, objective_string);
            break;
         }
         case OPT_DEVICE_NAME:
         {
            std::string tmp_string = optarg;
            std::vector<std::string> values = convert_string_to_vector<std::string>(tmp_string, std::string(","));
            setOption("device_name", "");
            setOption("device_speed", "");
            setOption("device_package", "");
            setOption("device_synthesis_tool", "");
            if(values.size() == 1)
            {
               setOption(OPT_device_string, values[0]);
            }
            else if (values.size() == 3)
            {
               setOption("device_name", values[0]);
               setOption("device_speed", values[1]);
               setOption("device_package", values[2]);
            }
            else if (values.size() == 4)
            {
               setOption("device_name", values[0]);
               setOption("device_speed", values[1]);
               setOption("device_package", values[2]);
               setOption("device_synthesis_tool", values[3]);
            }
            else
            {
                THROW_ERROR("Malformed device: " + tmp_string);
            }
            break;
         }
         case OPT_PERIOD_CLOCK:
         {
            /*if(boost::lexical_cast<double>(optarg) > 10)
               setOption(OPT_add_connection_offset, false);*/
            setOption(OPT_clock_period, optarg);
            break;
         }
         case OPT_POWER_OPTIMIZATION:
         {
            setOption("power_optimization", true);
            break;
         }
         case OPT_DISABLE_IOB:
         {
            setOption(OPT_connect_iob, false);
            break;
         }
         case OPT_RESET:
         {
            if (std::string(optarg) == "no")
               setOption(OPT_sync_reset, std::string(optarg));
            else if(std::string(optarg) == "async")
               setOption(OPT_sync_reset, std::string(optarg));
            else if(std::string(optarg) == "sync")
               setOption(OPT_sync_reset, std::string(optarg));
            else
               throw "BadParameters: reset type not correctly specified";

            break;
         }
         case OPT_LEVEL_RESET:
         {
            if (std::string(optarg) == "high")
              setOption(OPT_level_reset, true);
            else if (std::string(optarg) == "low")
              setOption(OPT_level_reset, false);
            else
               throw "BadParameters: reset edge type not correctly specified";
            break;
         }
#if HAVE_ILP_BUILT
         case OPT_SDC_SCHEDULING:
         {
            if (scheduling_set_p)
               THROW_ERROR("BadParameters: only one scheduler can be specified");
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::SDC_SCHEDULING);
            std::string defines;
            if(isOption(OPT_gcc_defines))
               defines = getOption<std::string>(OPT_gcc_defines) + STR_CST_string_separator;
            defines += std::string("PANDA_SDC");
            setOption(OPT_gcc_defines, defines);
            break;
         }
#endif
         case OPT_SERIALIZE_MEMORY_ACCESSES:
         {
            setOption(OPT_gcc_serialize_memory_accesses, true);
            break;
         }
         case OPT_SOFT_FLOAT:
         {
            setOption(OPT_soft_float, true);
            break;
         }
#if HAVE_FLOPOCO
         case OPT_FLOPOCO:
         {
            setOption(OPT_soft_float, false);
            break;
         }
#endif
         case OPT_HLS_DIV:
         {
            setOption(OPT_hls_div, true);
            break;
         }
         case OPT_CLOCK_PERIOD_RESOURCE_FRACTION:
         {
            setOption(OPT_clock_period_resource_fraction, optarg);
            break;
         }
         case OPT_SCHEDULING_MUX_MARGINS:
         {
            setOption(OPT_scheduling_mux_margins, optarg);
            break;
         }
         case OPT_TIMING_MODEL:
         {
            if(std::string(optarg)=="SIMPLE")
               setOption(OPT_estimate_logic_and_connections, false);
            else if(std::string(optarg)=="EC")
               setOption(OPT_estimate_logic_and_connections, true);
            else
               throw "BadParameters: unknown timing model";
            break;
         }
         case OPT_REGISTERED_INPUTS:
         {
            setOption(OPT_registered_inputs, optarg);
            break;
         }
         case OPT_MAX_ULP:
         {
            setOption(OPT_max_ulp, optarg);
            break;
         }
         case OPT_SKIP_PIPE_PARAMETER:
         {
            setOption(OPT_skip_pipe_parameter, optarg);
            break;
         }
         case OPT_UNALIGNED_ACCESS_PARAMETER:
         {
            setOption(OPT_unaligned_access, true);
            break;
         }
         case OPT_ALIGNED_ACCESS_PARAMETER:
         {
            setOption(OPT_aligned_access, true);
            break;
         }
         case OPT_BACKEND_SCRIPT_EXTENSIONS_PARAMETER:
         {
            setOption(OPT_backend_script_extensions, optarg);
            break;
         }
         case OPT_BACKEND_SDC_EXTENSIONS_PARAMETER:
         {
            setOption(OPT_backend_sdc_extensions, optarg);
            break;
         }
         case OPT_DO_NOT_USE_ASYNCHRONOUS_MEMORIES:
         {
            setOption(OPT_use_asynchronous_memories, false);
            break;
	 }
         case OPT_DISTRAM_THRESHOLD:
	 {
            setOption(OPT_distram_threshold, optarg);
            break;
         }
         case OPT_DO_NOT_CHAIN_MEMORIES:
         {
            setOption(OPT_do_not_chain_memories, true);
            break;
         }
         case OPT_BRAM_HIGH_LATENCY:
         {
            setOption(OPT_bram_high_latency, true);
            break;
         }
         case OPT_DO_NOT_EXPOSE_GLOBALS:
         {
            setOption(OPT_do_not_expose_globals, true);
            break;
         }
         case OPT_EXPERIMENTAL_SETUP:
         {
            setOption(OPT_experimental_setup, optarg);
            break;
         }
         case OPT_DSP_ALLOCATION_COEFFICIENT:
         {
            setOption(OPT_DSP_allocation_coefficient, optarg);
            break;
         }
         case OPT_DSP_MARGIN_COMBINATIONAL:
         {
            setOption(OPT_DSP_margin_combinational, optarg);
            break;
         }
         case OPT_DSP_MARGIN_PIPELINED:
         {
            setOption(OPT_DSP_margin_pipelined, optarg);
            break;
         }
         ///output options
         case 'w':
         {
            if (std::string(optarg) == "V")
               setOption(OPT_writer_language, static_cast<int>(HDLWriter_Language::VERILOG));
#if HAVE_EXPERIMENTAL
            else if (std::string(optarg) == "S")
               setOption(OPT_writer_language, static_cast<int>(HDLWriter_Language::SYSTEMC));
#endif
            else if (std::string(optarg) == "H")
               setOption(OPT_writer_language, static_cast<int>(HDLWriter_Language::VHDL));
            else
               throw "BadParameters: backend language not correctly specified";
            break;
         }
         case OPT_DYNAMIC_GENERATORS_DIR:
         {
            setOption("dynamic_generators_dir", optarg);
            break;
         }
         case OPT_PRETTY_PRINT:
         {
            setOption(OPT_pretty_print, optarg);
            break;
         }
#if HAVE_EXPERIMENTAL
         case OPT_PRAGMA_PARSE:
         {
            setOption(OPT_parse_pragma, true);
            break;
         }
#endif
         case OPT_TESTBENCH:
         {
            setOption(OPT_generate_testbench, true);
            const std::string arg = TrimSpaces(std::string(optarg));
            if (arg.size() >= 4 and arg.substr(arg.size() - 4) == ".xml")
            {
               setOption(OPT_testbench_input_xml, optarg);
            }
            else
            {
               setOption(OPT_testbench_input_string, optarg);
            }
            break;
         }
         case OPT_TESTBENCH_EXTRA_GCC_FLAGS:
         {
            setOption(OPT_testbench_extra_gcc_flags, optarg);
            break;
         }
         case OPT_MAX_SIM_CYCLES:
         {
            setOption(OPT_max_sim_cycles, optarg);
            break;
         }
         case OPT_GENERATE_VCD:
         {
            setOption(OPT_generate_vcd, true);
            break;
         }
#if HAVE_CUDD
         case OPT_LOGICAL_OPTIMIZATION:
         {
            if (optarg)
               setOption("logical_optimization", optarg);
            break;
         }
#endif
         case OPT_ADDITIONAL_TOP:
         {
            setOption(OPT_additional_top, optarg);
            break;
         }
         case OPT_DISABLE_FUNCTION_PROXY:
         {
            setOption(OPT_disable_function_proxy, true);
            break;
         }
         case OPT_DISABLE_BOUNDED_FUNCTION:
         {
            setOption(OPT_disable_bounded_function, true);
            break;
         }
         case OPT_MEMORY_MAPPED_TOP:
         {
            setOption(OPT_memory_mapped_top, true);
            break;
         }
         case OPT_MEM_DELAY_READ:
         {
            setOption(OPT_mem_delay_read, optarg);
            break;
         }
         case OPT_MEM_DELAY_WRITE:
         {
            setOption(OPT_mem_delay_write, optarg);
            break;
         }
#if HAVE_HOST_PROFILING_BUILT
         case OPT_HOST_PROFILING:
         {
            setOption(OPT_profiling_method, static_cast<int>(HostProfiling_Method::PM_BBP));
            break;
         }
#endif
         case OPT_NO_MIXED_DESIGN:
         {
            setOption<bool>(OPT_mixed_design, false);
            break;
         }
         case OPT_DISCREPANCY:
         {
            setOption(OPT_discrepancy, true);
            break;
         }
         case OPT_DISCREPANCY_NO_LOAD_POINTERS:
         {
            setOption(OPT_discrepancy, true);
            setOption(OPT_discrepancy_no_load_pointers, true);
            break;
         }
         case OPT_DISCREPANCY_FORCE:
         {
            setOption(OPT_discrepancy, true);
            setOption(OPT_discrepancy_force, true);
            break;
         }
#if HAVE_EXPERIMENTAL && HAVE_FROM_PRAGMA_BUILT && HAVE_BAMBU_BUILT
         case OPT_NUM_THREADS:
         {
            setOption(OPT_num_threads, std::string(optarg));
            break;
         }
#endif
#if HAVE_EXPERIMENTAL
         case OPT_MEMORY_BANKS_NUMBER:
         {
            setOption(OPT_memory_banks_number, std::string(optarg));
            break;
         }
#endif
         case OPT_ACCEPT_NONZERO_RETURN:
         {
            setOption(OPT_no_return_zero, true);
            break;
         }
         case INPUT_OPT_C_NO_PARSE:
         {
            std::vector<std::string> Splitted;
            std::string no_parse_files;
            boost::algorithm::split(Splitted, optarg, boost::algorithm::is_any_of(" ,"));
            for (unsigned int i = 0; i < Splitted.size(); i++)
            {
               boost::trim(Splitted[i]);
               no_parse_files += Splitted[i] + " ";
            }
            setOption(OPT_no_parse_files, no_parse_files);
            break;
         }
         case INPUT_OPT_C_PYTHON_NO_PARSE:
         {
            std::vector<std::string> Splitted;
            std::string no_parse_c_python_files;
            boost::algorithm::split(Splitted, optarg, boost::algorithm::is_any_of(" ,"));
            for (unsigned int i = 0; i < Splitted.size(); i++)
            {
               boost::trim(Splitted[i]);
               no_parse_c_python_files += Splitted[i] + " ";
            }
            setOption(OPT_no_parse_c_python, no_parse_c_python_files);
            break;
         }
         case 0:
         {
            if (strcmp(long_options[option_index].name, "module-binding") == 0)
            {
               if (std::string(optarg) == "TTT_FAST")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TTT_CLIQUE_COVERING_FAST);
               }
               else if (std::string(optarg) == "TTT_FAST2")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TTT_CLIQUE_COVERING_FAST2);
               }
               else if (std::string(optarg) == "TTT_FULL")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TTT_CLIQUE_COVERING);
               }
               else if (std::string(optarg) == "TTT_FULL2")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TTT_CLIQUE_COVERING2);
               }
               else if (std::string(optarg) == "TS")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TS_CLIQUE_COVERING);
               }
               else if (std::string(optarg) == "WEIGHTED_TS")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TS_WEIGHTED_CLIQUE_COVERING);
               }
               else if (std::string(optarg) == "COLORING")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::COLORING);
               }
               else if (std::string(optarg) == "WEIGHTED_COLORING")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::WEIGHTED_COLORING);
               }
               else if (std::string(optarg) == "BIPARTITE_MATCHING")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::BIPARTITE_MATCHING);
               }
#if HAVE_EXPERIMENTAL
               else if (std::string(optarg) == "RANDOMIZED")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
                  setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::RANDOMIZED);
               }
#endif
               else if (std::string(optarg) == "UNIQUE")
               {
                  setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::UNIQUE_MODULE_BINDING);
               }
               else
                  throw "BadParameters: module binding option not correctly specified";
               break;
            }
            if (strcmp(long_options[option_index].name, "memory-allocation") == 0)
            {
               if (std::string(optarg) == "DOMINATOR")
                  setOption(OPT_memory_allocation_algorithm, HLSFlowStep_Type::DOMINATOR_MEMORY_ALLOCATION);
               else if (std::string(optarg) == "XML_SPECIFICATION")
                  setOption(OPT_memory_allocation_algorithm, HLSFlowStep_Type::XML_MEMORY_ALLOCATOR);
               else
                  throw "BadParameters: memory allocation option not correctly specified";
               break;
            }
            if (strcmp(long_options[option_index].name, "xml-memory-allocation") == 0)
            {
               setOption(OPT_memory_allocation_algorithm, HLSFlowStep_Type::XML_MEMORY_ALLOCATOR);
               setOption(OPT_xml_memory_allocation, optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "memory-allocation-policy") == 0)
            {
               if (std::string(optarg) == "LSS")
                  setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::LSS);
               else if (std::string(optarg) == "GSS")
                  setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::GSS);
               else if (std::string(optarg) == "ALL_BRAM")
                  setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
               else if (std::string(optarg) == "NO_BRAM")
                  setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::NO_BRAM);
               else if (std::string(optarg) == "EXT_PIPELINED_BRAM")
                  setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::EXT_PIPELINED_BRAM);
               else
                  throw "BadParameters: memory allocation policy option not correctly specified";
               break;
            }
            if (strcmp(long_options[option_index].name, "base-address") == 0)
            {
               setOption(OPT_base_address, optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "initial-internal-address") == 0)
            {
               setOption(OPT_initial_internal_address, optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "channels-type") == 0)
            {
               if (std::string(optarg) == CHANNELS_TYPE_MEM_ACC_11)
                  setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_11);
               else if (std::string(optarg) == CHANNELS_TYPE_MEM_ACC_N1)
                  setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_N1);
               else if (std::string(optarg) == CHANNELS_TYPE_MEM_ACC_NN)
                  setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_NN);
               else
                  throw "BadParameters: memory accesses type not correctly specified";
               break;
            }
            if (strcmp(long_options[option_index].name, "channels-number") == 0)
            {
               setOption(OPT_channels_number, optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "memory-ctrl-type") == 0)
            {
               if (std::string(optarg) == "D00")
                  setOption(OPT_memory_controller_type, optarg);
               else if (std::string(optarg) == "D10")
                  setOption(OPT_memory_controller_type, optarg);
               else if (std::string(optarg) == "D11")
                  setOption(OPT_memory_controller_type, optarg);
               else if (std::string(optarg) == "D21")
                  setOption(OPT_memory_controller_type, optarg);
               else
                  throw "BadParameters: memory controller type not correctly specified";
               break;
            }
            if (strcmp(long_options[option_index].name, "sparse-memory") == 0)
            {
               if (!optarg || std::string(optarg) == "on")
                  setOption(OPT_sparse_memory, true);
               else if (std::string(optarg) == "off")
                  setOption(OPT_sparse_memory, false);
               else
                  throw "BadParameters: sparse-memory option not expected";
               break;
            }
#if HAVE_EXPERIMENTAL
            if (strcmp(long_options[option_index].name, "timing-simulation") == 0)
            {
               setOption(OPT_timing_simulation, true);
               break;
            }
            if (strcmp(long_options[option_index].name, "resp-model") == 0)
            {
               setOption(OPT_resp_model, optarg);
               break;
            }
            //front-end analysis option
            if (strcmp(long_options[option_index].name, "pdg-reduction") == 0)
            {
               //Set the default
               if(optarg)
                  setOption("pdg-reduction", optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "explore-mux") == 0)
            {
               setOption("explore-mux", true);
               if (optarg) setOption("worst_case_delay", optarg);
               break;
            }
            if (strcmp(long_options[option_index].name, "explore-fu-reg") == 0)
            {
               setOption("explore-fu-reg", true);
               setOption("explore-fu-reg-param", optarg);
               break;
            }
#endif
            if (strcmp(long_options[option_index].name, "assert-debug") == 0)
            {
               setOption(OPT_assert_debug, true);
               break;
            }
            if (strcmp(long_options[option_index].name, "circuit-dbg") == 0)
            {
               setOption(OPT_circuit_debug_level, optarg);
               break;
            }
            ///scheduling options
            if (strcmp(long_options[option_index].name, "no-chaining") == 0)
            {
               setOption(OPT_chaining, false);
               break;
            }
            if (strcmp(long_options[option_index].name, "generate-interface") == 0)
            {
               setOption(OPT_interface, true);
               if (std::string(optarg) == "MINIMAL")
               {
                  setOption(OPT_interface_type, HLSFlowStep_Type::MINIMAL_INTERFACE_GENERATION);
               }
               else if (std::string(optarg) == "WB4")
               {
                  setOption(OPT_interface_type, HLSFlowStep_Type::WB4_INTERFACE_GENERATION);
               }
#if HAVE_EXPERIMENTAL
               else if (std::string(optarg) == "AXI4LITE")
               {
                  setOption(OPT_interface_type, HLSFlowStep_Type::AXI4LITE_INTERFACE_GENERATION);
               }
               else if (std::string(optarg) == "FSL")
               {
                  setOption(OPT_interface_type, HLSFlowStep_Type::FSL_INTERFACE_GENERATION);
               }
               else if (std::string(optarg) == "NPI")
               {
                  setOption(OPT_interface_type, HLSFlowStep_Type::NPI_INTERFACE_GENERATION);
               }
#endif
               else
               {
                  THROW_ERROR("Not supported interface: " + std::string(optarg));
               }
               break;
            }
#if HAVE_EXPERIMENTAL
            if (strcmp(long_options[option_index].name, "export-core") == 0)
            {
               setOption(OPT_export_core, true);
               if (std::string(optarg) == "PCORE")
               {
                  setOption(OPT_export_core_mode, HLSFlowStep_Type::EXPORT_PCORE);
               }
               else
               {
                  THROW_ERROR("Not supported export mode: " + std::string(optarg));
               }
               break;
            }
#endif
            if (strcmp(long_options[option_index].name, "data-bus-bitsize") == 0)
            {
               setOption(OPT_data_bus_bitsize, boost::lexical_cast<int>(optarg));
               break;
            }
            if (strcmp(long_options[option_index].name, "addr-bus-bitsize") == 0)
            {
               setOption(OPT_addr_bus_bitsize, boost::lexical_cast<int>(optarg));
               break;
            }
#if HAVE_EXPERIMENTAL
            if (strcmp(long_options[option_index].name, "edk-config") == 0)
            {
               setOption("edk_wrapper", true);
               setOption("edk_config_file", std::string(optarg));
               break;
            }
#endif
            if (strcmp(long_options[option_index].name, "simulator") == 0)
            {
               setOption(OPT_simulator, std::string(optarg));
               break;
            }
         }
         ///other options
         default:
         {
            bool exit_success=false;
            bool res = ManageGccOptions(next_option, optarg);
            if (res)
               res = ManageDefaultOptions(next_option, optarg, exit_success);
            if(exit_success)
               return EXIT_SUCCESS;
            if(res)
            {
               return PARAMETER_NOTPARSED;
            }
         }
      }
   }

#if HAVE_EXPERIMENTAL
   if (isOption(OPT_gcc_write_xml))
   {
      const std::string filenameXML = getOption<std::string>(OPT_gcc_write_xml);
      write_xml_configuration_file(filenameXML);
      PRINT_MSG("Configuration saved into file \"" + filenameXML + "\"");
      return EXIT_SUCCESS;
   }
#endif

   std::string cat_args;

   for(int i = 0; i < argc; i++)
   {
      cat_args += std::string(argv[i]) + " ";
   }
   setOption(OPT_cat_args, cat_args);

   if (getOption<int>(OPT_output_level) >= OUTPUT_LEVEL_MINIMUM)
   {
      std::cerr << " ==  Bambu executed with: ";
      std::cerr << cat_args;
      std::cerr << std::endl << std::endl;
   }

   while (optind < argc)
   {
      const auto file_type = GetFileFormat(argv[optind], true);
      if(file_type == Parameters_FileFormat::FF_XML_CON)
      {
         setOption(OPT_constraints_file, argv[optind]);
      }
      else if(file_type == Parameters_FileFormat::FF_XML_TEC)
      {
         const auto tech_files = isOption(OPT_technology_file) ? getOption<std::string>(OPT_technology_file) + STR_CST_string_separator : "";
         setOption(OPT_technology_file, tech_files + argv[optind]);
      }
      else if(file_type == Parameters_FileFormat::FF_C or getOption<bool>(OPT_use_raw))
      {
         const auto input_file = isOption(OPT_input_file) ? getOption<std::string>(OPT_input_file) + STR_CST_string_separator : "";
         setOption(OPT_input_file, input_file + argv[optind]);
      }
      else
      {
         THROW_ERROR("Unexpected file extension: " + std::string(argv[optind]));
      }
      optind++;
   }

   CheckParameters();

   return PARAMETER_PARSED;

}

void BambuParameter::add_experimental_setup_gcc_options(bool kill_printf)
{
   if(kill_printf)
   {
   std::string defines;
   if(isOption(OPT_gcc_defines))
      defines = getOption<std::string>(OPT_gcc_defines) + STR_CST_string_separator;
   defines +=  "\'printf(fmt, ...)=\'";
   setOption(OPT_gcc_defines, defines);
   }
   if(isOption(OPT_top_function_name) && getOption<std::string>(OPT_top_function_name) == "main")
   {
      std::string optimizations;
      if(isOption(OPT_gcc_optimizations))
         optimizations = getOption<std::string>(OPT_gcc_optimizations);
      THROW_ASSERT(isOption(OPT_input_file), "Input file not specified");
      if(getOption<std::string>(OPT_input_file).find(STR_CST_string_separator) == std::string::npos)
      {
         if(optimizations != "")
            optimizations = optimizations + STR_CST_string_separator;
         optimizations = optimizations + "whole-program";
         setOption(OPT_gcc_optimizations, optimizations);
      }
      if(isOption(OPT_top_design_name))
      {
         if(optimizations != "")
            optimizations = optimizations + STR_CST_string_separator;
         optimizations = optimizations + "no-ipa-cp" + STR_CST_string_separator + "no-ipa-cp-clone";
         setOption(OPT_gcc_optimizations, optimizations);
      }
   }
}

void BambuParameter::CheckParameters()
{
   Parameter::CheckParameters();
   ///target device options
   if (not isOption(OPT_device_string))
   {
      std::string device_string = getOption<std::string>("device_name") + getOption<std::string>("device_speed")  + getOption<std::string>("device_package");
      if(isOption("device_synthesis_tool") && getOption<std::string>("device_synthesis_tool") != "")
         device_string += "-"+getOption<std::string>("device_synthesis_tool");
      setOption(OPT_device_string, device_string);
   }

   if ((getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_N1 or getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_NN) and not isOption(OPT_channels_number))
   {
      setOption(OPT_channels_number, 2);
   }

   ///DSE options
#if HAVE_EXPERIMENTAL
#if HAVE_BEAGLE
   if (getOption<int>("exploration_technique") == dse_hls::PRIORITY and getOption<int>("to_normalize"))
      setOption("to_normalize", false);
   if (isOption("dse_algorithm") and getOption<unsigned int>("dse_algorithm") == dse_hls::ACO)
   {
      if (!isOption("aco_ant_num"))
         setOption("aco_ant_num", 5);
      if (!isOption("aco_generations"))
         setOption("aco_generations", 10);
   }
#endif
#endif

   ///circuit debugging options
   if (isOption(OPT_generate_vcd) and getOption<bool>(OPT_generate_vcd))
   {
      setOption(OPT_assert_debug, true);
   }
   if (getOption<int>(OPT_debug_level)>=DEBUG_LEVEL_VERY_PEDANTIC)
   {
      setOption(OPT_assert_debug, true);
   }


   ///controller options
   if (getOption<HLSFlowStep_Type>(OPT_controller_architecture) == HLSFlowStep_Type::FSM_CONTROLLER_CREATOR)
      setOption(OPT_stg, true);

   ///chaining options
   setOption(OPT_chaining_algorithm, HLSFlowStep_Type::SCHED_CHAINING);
#if HAVE_EXPERIMENTAL
   if (getOption<HLSFlowStep_Type>(OPT_controller_architecture) == HLSFlowStep_Type::PARALLEL_CONTROLLER_CREATOR)
   {
      setOption(OPT_liveness_algorithm, HLSFlowStep_Type::CHAINING_BASED_LIVENESS);
      setOption(OPT_chaining_algorithm, HLSFlowStep_Type::EPDG_SCHED_CHAINING);
   }
#endif

   ///evaluation options
   if (getOption<bool>(OPT_evaluation))
   {
      THROW_ASSERT(isOption(OPT_evaluation_objectives),
            "missing evaluation objectives");
      std::string objective_string =
         getOption<std::string>(OPT_evaluation_objectives);
      THROW_ASSERT(not objective_string.empty(), "");
      std::vector<std::string> objective_vector =
         convert_string_to_vector<std::string>(objective_string, ",");

      if (getOption<Evaluation_Mode>(OPT_evaluation_mode) ==
            Evaluation_Mode::EXACT)
      {
         if (is_evaluation_objective_string(objective_vector,"AREAxTIME"))
         {
            add_evaluation_objective_string(objective_string, "AREA,TIME");
            setOption(OPT_evaluation_objectives, objective_string);
            objective_vector =
               convert_string_to_vector<std::string>(objective_string, ",");
         }

         if (is_evaluation_objective_string(objective_vector, "TIME") or
               is_evaluation_objective_string(objective_vector, "TOTAL_TIME") or
               is_evaluation_objective_string(objective_vector, "CYCLES") or
               is_evaluation_objective_string(objective_vector, "TOTAL_CYCLES"))
         {
            if (not getOption<bool>(OPT_generate_testbench))
            {
               setOption(OPT_generate_testbench, true);
               setOption(OPT_testbench_input_xml, "test.xml");
            }
         }
         const auto is_valid_evaluation_mode = [](const std::string & s) -> bool
         {
            return s == "AREA" or
               s == "AREAxTIME" or
               s == "TIME" or
               s == "TOTAL_TIME" or
               s == "CYCLES" or
               s == "TOTAL_CYCLES" or
               s == "BRAMS" or
               s == "CLOCK_SLACK" or
               s == "DSPS" or
#if HAVE_EXPERIMENTAL
               s == "EDGES_REDUCTION" or
#endif
               s == "FREQUENCY" or
#if HAVE_EXPERIMENTAL
               s == "NUM_AD_EDGES" or
#endif
               s == "PERIOD" or
               s == "REGISTERS";
         };
         if (not all_of(objective_vector.begin(), objective_vector.end(), is_valid_evaluation_mode))
         {
            THROW_ERROR("BadParameters: evaluation mode EXACT don't support the evaluation objectives");
         }
      }
#if HAVE_EXPERIMENTAL
      else if (getOption<Evaluation_Mode>(OPT_evaluation_mode) ==
            Evaluation_Mode::ESTIMATION)
      {
         const auto is_valid_evaluation_mode = [](const std::string & s) -> bool
         {
            return s == "AREA" or s == "TIME" or s == "CLOCK_SLACK";
         };
         if (not all_of(objective_vector.begin(), objective_vector.end(), is_valid_evaluation_mode))
         {
            THROW_ERROR("BadParameters: evaluation mode LINEAR don't support the evaluation objectives");
         }
      }
#endif
      else
      {
         THROW_ERROR("BadParameters: invalid evaluation mode");
      }
   }
#if HAVE_EXPERIMENTAL
   ///Export and interface generation
   if (getOption<bool>(OPT_export_core) && getOption<HLSFlowStep_Type>(OPT_export_core_mode) == HLSFlowStep_Type::EXPORT_PCORE)
   {
      setOption(OPT_interface, true);
      setOption(OPT_interface_type, HLSFlowStep_Type::AXI4LITE_INTERFACE_GENERATION);
   }
#endif

   if(getOption<bool>(OPT_do_not_expose_globals))
   {
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      else if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) != MemoryAllocation_Policy::ALL_BRAM)
         THROW_ERROR("BadParameters: --do-not-expose-globals implies --memory-allocation-policy=ALL_BRAM");
   }
   tree_helper::debug_level = get_class_debug_level("tree_helper");

   ///add experimental setup options
   if(getOption<std::string>(OPT_experimental_setup) == "VVD")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O3);
      setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_NN);
      setOption(OPT_channels_number, 2);
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      setOption(OPT_DSP_allocation_coefficient, 1.75);
      setOption(OPT_clock_period_resource_fraction, "0.875");
      setOption(OPT_do_not_expose_globals, true);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU092")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O0);
      setOption(OPT_estimate_logic_and_connections, false);
      setOption(OPT_clock_period_resource_fraction, "0.9");
      setOption(OPT_DSP_margin_combinational, 1.3);
      setOption(OPT_skip_pipe_parameter, 1);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-BALANCED" or
         getOption<std::string>(OPT_experimental_setup) == "BAMBU-BALANCED-MP")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O2);
      std::string optimizations;
      if(isOption(OPT_gcc_optimizations))
         optimizations = getOption<std::string>(OPT_gcc_optimizations) + STR_CST_string_separator;
      std::string tuning_optimizations =
         "gcse-after-reload" + STR_CST_string_separator +
         "ipa-cp-clone" + STR_CST_string_separator +
         "unswitch-loops" + STR_CST_string_separator +
         "inline-functions" + STR_CST_string_separator +
         "no-tree-loop-ivcanon";
     if(false
#if HAVE_I386_GCC48_COMPILER
           or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) ==
           GccWrapper_CompilerTarget::CT_I386_GCC48
#endif
#if HAVE_I386_GCC49_COMPILER
           or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) ==
           GccWrapper_CompilerTarget::CT_I386_GCC49
#endif
#if HAVE_I386_GCC5_COMPILER
           or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) ==
           GccWrapper_CompilerTarget::CT_I386_GCC5
#endif
      )
      {
         tuning_optimizations += STR_CST_string_separator + "tree-partial-pre";
      }
      setOption(OPT_gcc_optimizations, optimizations +  tuning_optimizations);
      std::string parameters;
      if(isOption(OPT_gcc_parameters))
         parameters = getOption<std::string>(OPT_gcc_parameters) + STR_CST_string_separator;
#if 0
      setOption(OPT_gcc_parameters, parameters + "max-inline-insns-auto=25");
#endif
      if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-BALANCED-MP")
      {
         setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_NN);
         setOption(OPT_channels_number, 2);
      }
      else
      {
         setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_11);
      }
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-PERFORMANCE-MP")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O3);
      setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_NN);
      setOption(OPT_channels_number, 2);
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 512);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-PERFORMANCE")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O3);
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 512);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-AREA-MP")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::Os);
      setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_NN);
      setOption(OPT_channels_number, 2);
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      setOption(OPT_DSP_allocation_coefficient, 1.75);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU-AREA")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::Os);
      if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
         setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::ALL_BRAM);
      setOption(OPT_DSP_allocation_coefficient, 1.75);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(true);
   }
   else if(getOption<std::string>(OPT_experimental_setup) == "BAMBU")
   {
      if(not isOption(OPT_gcc_opt_level))
         setOption(OPT_gcc_opt_level, GccWrapper_OptimizationSet::O0);
      if(not isOption(OPT_distram_threshold))
          setOption(OPT_distram_threshold, 256);
      add_experimental_setup_gcc_options(false);
   }
   else
   {
      THROW_ERROR("Experimental setup not recognized: " + getOption<std::string>(OPT_experimental_setup));
   }

   add_bambu_library("bambu");

   if(isOption(OPT_soft_float) && getOption<bool>(OPT_soft_float))
      add_bambu_library("softfloat");

   if(isOption(OPT_hls_div) && getOption<bool>(OPT_hls_div))
      add_bambu_library("hls-div");
   add_bambu_library("hls-cdiv");
#if HAVE_EXPERIMENTAL && HAVE_FROM_PRAGMA_BUILT && HAVE_BAMBU_BUILT
   if(getOption<bool>(OPT_parse_pragma))
   {
      setOption(OPT_disable_function_proxy, true);
      setOption(OPT_function_allocation_algorithm, HLSFlowStep_Type::OMP_FUNCTION_ALLOCATION);
      add_bambu_library("pthread");
   }
#endif

   if(isOption(OPT_gcc_libraries))
   {
      const std::unordered_set<std::string> libraries = getOption<const std::unordered_set<std::string> >(OPT_gcc_libraries);
      std::unordered_set<std::string>::const_iterator library, library_end = libraries.end();
      for(library = libraries.begin(); library != library_end; library++)
      {
         add_bambu_library(*library);
      }
   }

   /// default for memory allocation policy
   if(getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NONE)
      setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::LSS);

   ///base address and initial internal address checks
   if(isOption(OPT_initial_internal_address) && (getOption<unsigned int>(OPT_base_address)==0 || getOption<unsigned int>(OPT_initial_internal_address)==0))
   {
      std::string optimizations;
      if(isOption(OPT_gcc_optimizations))
         optimizations = getOption<std::string>(OPT_gcc_optimizations) + STR_CST_string_separator;
      setOption(OPT_gcc_optimizations, optimizations + "no-delete-null-pointer-checks");
   }

   /// Checks
   if (getOption<HLSFlowStep_Type>(OPT_memory_allocation_algorithm) == HLSFlowStep_Type::XML_MEMORY_ALLOCATOR and not !isOption(OPT_xml_memory_allocation))
   {
      if (!isOption(OPT_xml_input_configuration))
         THROW_ERROR("XML specification of the memory allocation has not been specified!");
   }
   if(isOption(OPT_memory_banks_number) && getOption<int>(OPT_memory_banks_number) > 1)
   {
      setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_P1N);
   }
   if(isOption(OPT_channels_number) and getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_11)
   {
      if(getOption<unsigned int>(OPT_channels_number) != 1)
         THROW_ERROR("the number of channels cannot be specified for MEM_ACC_11");
   }
   if(isOption(OPT_channels_number) and (getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_N1 or
                                         getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_NN))
   {
      if(getOption<unsigned int>(OPT_channels_number)>2 and getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) != MemoryAllocation_Policy::NO_BRAM and getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) != MemoryAllocation_Policy::EXT_PIPELINED_BRAM)
         THROW_ERROR("no more than two channels is supported for MEM_ACC_N1 and MEM_ACC_NN: try to add this option --memory-allocation-policy=NO_BRAM or --memory-allocation-policy=EXT_PIPELINED_BRAM");
   }
   if(getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_NN and isOption(OPT_interface_type) and getOption<HLSFlowStep_Type>(OPT_interface_type) == HLSFlowStep_Type::WB4_INTERFACE_GENERATION)
      THROW_ERROR("Wishbone 4 interface does not yet support multi-channel architectures (MEM_ACC_NN)");

   if(not getOption<bool>(OPT_gcc_include_sysdir))
   {
      if(not isOption(OPT_input_file))
      {
         THROW_ERROR("No input file specified");
      }
      if(isOption(OPT_gcc_optimizations) and getOption<std::string>(OPT_input_file).find(STR_CST_string_separator) != std::string::npos and getOption<std::string>(OPT_gcc_optimizations).find("whole-program") != std::string::npos and getOption<std::string>(OPT_gcc_optimizations).find("no-whole-program") == std::string::npos)
      {
         THROW_ERROR("-fwhole-program cannot be used with multiple input files");
      }
   }
#if HAVE_EXPERIMENTAL
   if(isOption(OPT_interface_type) && getOption<HLSFlowStep_Type>(OPT_interface_type) == HLSFlowStep_Type::AXI4LITE_INTERFACE_GENERATION)
   {
      if(isOption(OPT_evaluation_objectives) && getOption<std::string>(OPT_evaluation_objectives).find("AREA") != std::string::npos)
         THROW_ERROR("AXI-based synthesis has to be performed outside of bambu");
      else
         THROW_WARNING("Note that synthesis and simulation scripts generated by bambu for AXI-based designs are actually only templates not real synthesis scripts\n");
   }
#endif
   if (isOption(OPT_discrepancy) and getOption<bool>(OPT_discrepancy))
   {
      if (false
#if HAVE_I386_GCC45_COMPILER
              or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) == GccWrapper_CompilerTarget::CT_I386_GCC45
#endif
#if HAVE_I386_GCC46_COMPILER
              or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) == GccWrapper_CompilerTarget::CT_I386_GCC46
#endif
#if HAVE_I386_GCC47_COMPILER
              or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) == GccWrapper_CompilerTarget::CT_I386_GCC47
#endif
#if HAVE_I386_GCC48_COMPILER
              or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) == GccWrapper_CompilerTarget::CT_I386_GCC48
#endif
#if HAVE_I386_GCC49_COMPILER
              or getOption<GccWrapper_CompilerTarget>(OPT_default_compiler) == GccWrapper_CompilerTarget::CT_I386_GCC49
#endif
            )
      {
         THROW_WARNING("discrepancy analysis can report false positives with old compilers, use --compiler=I386_GCC5 or higher to avoid them");
      }
   }
   if ((isOption(OPT_generate_vcd) and getOption<bool>(OPT_generate_vcd)) or
      (isOption(OPT_discrepancy) and getOption<bool>(OPT_discrepancy)))
   {
      if (not isOption(OPT_generate_testbench) or not getOption<bool>(OPT_generate_testbench))
         THROW_ERROR("Testbench generation required. (--generate-tb or --simulate undeclared).");
   }

   ///In case copy input files
   if(isOption(OPT_file_input_data))
   {
      std::string input_data = getOption<std::string>(OPT_file_input_data);
      std::vector<std::string> splitted;
      boost::algorithm::split(splitted, input_data, boost::algorithm::is_any_of(","));
      size_t i_end = splitted.size();
      for(size_t i=0; i < i_end; i++)
      {
         std::string command = "cp " + splitted[i] + " .";
         int ret = PandaSystem(ParameterConstRef(this, null_deleter()), command);
         if(IsError(ret))
         {
            THROW_ERROR("cp returns an error");
         }
      }
   }

   if(isOption(OPT_no_parse_c_python) and not isOption(OPT_testbench_extra_gcc_flags))
   {
      THROW_ERROR("Include directories and library directories for Python bindings are missing.\n"
            "use --testbench-extra-gcc-flags=\"string\" to provide them");
   }
   setOption<unsigned int>(OPT_host_compiler, static_cast<unsigned int>(getOption<GccWrapper_CompilerTarget>(OPT_default_compiler)));
}

void BambuParameter::SetDefaults()
{
   // ---------- general options ----------- //
   /// Revision
   setOption(OPT_dot_directory, "HLS_output/dot/");
   setOption(OPT_output_directory, "HLS_output/");
   setOption(OPT_simulation_output, "results.txt");
   setOption(OPT_profiling_output, "profiling_results.txt");
   /// Debugging level
   setOption(OPT_output_level, OUTPUT_LEVEL_MINIMUM);
   setOption(OPT_debug_level, DEBUG_LEVEL_NONE);
   setOption(OPT_circuit_debug_level, DEBUG_LEVEL_NONE);
   setOption(OPT_print_dot, false);
   /// maximum execution time: 0 means no time limit
   setOption(OPT_ilp_max_time, 0);
   setOption(OPT_use_raw, false);

#if HAVE_MAPPING_BUILT
   setOption(OPT_driving_component_type, "ARM");
#endif

   /// pragmas related
   setOption(OPT_parse_pragma, false);
   setOption(OPT_ignore_parallelism, false);

   /// ---------- frontend analysis ----------//   
#if HAVE_FROM_RTL_BUILT
   setOption(OPT_use_rtl, false);
#endif

   //setOption("pdg-reduction", false);

#if HAVE_EXPERIMENTAL
   setOption("compute_unrolling_degree", false);
   setOption("insert_verification_operation",false);
#endif
   setOption(OPT_frontend_statistics, false);

   /// ---------- HLS process options ----------- //
   setOption(OPT_synthesis_flow, HLSFlowStep_Type::CLASSICAL_HLS_SYNTHESIS_FLOW);
   setOption(OPT_hls_flow, HLSFlowStep_Type::STANDARD_HLS_FLOW);

   /// ---------- HLS specification reference ----------- //
   setOption(OPT_generate_testbench, false);
   setOption(OPT_max_sim_cycles, 200000000);
   setOption(OPT_chaining, true);

   /// High-level synthesis contraints dump -- //
   setOption("dumpConstraints", false);
   setOption("dumpConstraints_file", "Constraints.XML");

   /// -- Scheduling -- //
   /// Scheduling algorithm (default is list based one)
   setOption(OPT_scheduling_algorithm, HLSFlowStep_Type::LIST_BASED_SCHEDULING);
   setOption(OPT_scheduling_priority, ParametricListBased_Metric::DYNAMIC_MOBILITY);
   /// ilp solver
#if HAVE_ILP_BUILT
#if HAVE_GLPK
   setOption(OPT_ilp_solver, meilp_solver::GLPK);
#elif HAVE_COIN_OR
   setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
#elif HAVE_LP_SOLVE
   setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
#endif
#endif
   /// speculative execution flag
   setOption(OPT_speculative, false);

   /// -- Module binding -- //
   /// module binding algorithm
   setOption(OPT_fu_binding_algorithm, HLSFlowStep_Type::CDFC_MODULE_BINDING);
   setOption(OPT_cdfc_module_binding_algorithm, CliqueCovering_Algorithm::TS_WEIGHTED_CLIQUE_COVERING);

   /// -- Finite state machine -- //
   /// flag to check if finite state machine has to be created
   setOption(OPT_stg, false);
   /// state transition graph creation algorithm
   setOption(OPT_stg_algorithm, HLSFlowStep_Type::BB_STG_CREATOR);

   /// -- Dataflow analysis -- //
   setOption(OPT_liveness_algorithm, HLSFlowStep_Type::FSM_NI_SSA_LIVENESS);

   /// -- Register allocation -- //
   /// register allocation algorithm
   setOption(OPT_register_allocation_algorithm, HLSFlowStep_Type::WEIGHTED_CLIQUE_REGISTER_BINDING);
   setOption(OPT_weighted_clique_register_algorithm, CliqueCovering_Algorithm::WEIGHTED_COLORING);
   /// storage value insertion algorithm
   setOption(OPT_storage_value_insertion_algorithm, HLSFlowStep_Type::VALUES_SCHEME_STORAGE_VALUE_INSERTION);
   setOption(OPT_sync_reset, "no");
   setOption(OPT_level_reset, false);

   /// Function allocation
   setOption(OPT_function_allocation_algorithm, HLSFlowStep_Type::DOMINATOR_FUNCTION_ALLOCATION);

   /// Enable function proxy by default
   setOption(OPT_disable_function_proxy, false);
   setOption(OPT_disable_bounded_function, false);

   /// Disable memory mapped interface for top function by default
   setOption(OPT_memory_mapped_top, false);

   setOption(OPT_mem_delay_read, 2);
   setOption(OPT_mem_delay_write, 1);


   /// -- Memory allocation -- //
   setOption(OPT_memory_allocation_algorithm, HLSFlowStep_Type::DOMINATOR_MEMORY_ALLOCATION);
   setOption(OPT_memory_allocation_policy, MemoryAllocation_Policy::NONE);
   setOption(OPT_base_address, 1073741824); // 1Gbytes maximum address space reserved for the accelerator
   setOption(OPT_channels_type, MemoryAllocation_ChannelsType::MEM_ACC_11);
   setOption(OPT_memory_controller_type, "D00");
   setOption(OPT_sparse_memory, true);
   setOption(OPT_do_not_expose_globals, false);


   /// -- Datapath -- //
   /// Datapath interconnection architecture
   setOption(OPT_datapath_interconnection_algorithm, HLSFlowStep_Type::MUX_INTERCONNECTION_BINDING);
   /// Datapath architecture
   setOption(OPT_datapath_architecture, HLSFlowStep_Type::CLASSIC_DATAPATH_CREATOR);

   /// -- Controller -- //
   /// target architecture for the controller
   setOption(OPT_controller_architecture, HLSFlowStep_Type::FSM_CONTROLLER_CREATOR);
#if HAVE_CUDD
   setOption("logical_optimization", 0);
#endif


   /// -- top entity -- //
   /// Output file name for top entity
   setOption(OPT_top_file, "top");

   /// backend HDL
   setOption(OPT_writer_language, static_cast<int>(HDLWriter_Language::VERILOG));
   setOption("dynamic_generators_dir", PANDA_DATA_INSTALLDIR "/panda-etc");

   /// -- Module Interfaces -- //
   setOption(OPT_interface, true);
   setOption(OPT_interface_type, HLSFlowStep_Type::MINIMAL_INTERFACE_GENERATION);
   setOption(OPT_additional_top, "");

   /// -- Module Characterization -- //
   setOption(OPT_evaluation, false);
   setOption(OPT_evaluation_mode, Evaluation_Mode::NONE);
   setOption(OPT_evaluation_objectives, "");
#if HAVE_EXPERIMENTAL
   setOption("evaluation_statistic", false);
   setOption("evaluation_reduced", false);
   setOption("evaluation_output", "evaluation.txt");
#endif

   /// -- Module Synthesis -- //
   setOption(OPT_rtl, true); ///the resulting specification will be a RTL description
#if HAVE_MODELSIM
   setOption(OPT_simulator, "MODELSIM"); /// Mixed language simulator
#elif HAVE_XILINX_VIVADO
   setOption(OPT_simulator, "XSIM"); /// Mixed language simulator
#elif HAVE_XILINX
   setOption(OPT_simulator, "ISIM"); /// Mixed language simulator
#elif HAVE_VERILATOR
   setOption(OPT_simulator, "VERILATOR");
#else
   setOption(OPT_simulator, "ICARUS");
#endif
   setOption("device_name", "xc7z020");
   setOption("device_speed", "-1");
   setOption("device_package", "clg484");
   setOption("device_synthesis_tool", "VVD");
   setOption(OPT_timing_simulation, false);
   setOption(OPT_timing_violation_abort, false);
   setOption(OPT_target_device_type, static_cast<int>(TargetDevice_Type::FPGA));
   setOption(OPT_export_core, false);
#if HAVE_EXPERIMENTAL
   setOption("edk_wrapper", false);
#endif
   setOption(OPT_connect_iob, true);

#if (HAVE_EXPERIMENTAL && HAVE_BEAGLE)
   // -- Parameters for the design space exploration -- //
   setOption("exploration_technique", dse_hls::BINDING);
   setOption("to_normalize", 0);
   setOption("seed", 0);
   setOption("run", 1);
   setOption("deme_size", 1);
   setOption("population", 50);
   setOption("GA_generation", 10);
   setOption("max_evaluations", 0);
   setOption("fitness_function", objective_evaluator::LINEAR);
   setOption("fitness_inheritance_rate", 0.0);
   /// if the parameter is even, the cache is analysed. Otherwise only the latest population
   setOption("inheritance_mode", 0);
   setOption("max_for_inheritance", 50);
   setOption("min_for_inheritance", 1);
   setOption("distance_rate", 0.20);
   setOption("weighting_function", FitnessFunction::QUADRATIC);
   setOption("time_weight", 0.5);
   setOption("area_weight", 0.5);
   setOption("remove_duplicated", true);
   setOption("lower", 0.05);
   setOption("upper", 0.05);
   setOption("prob_dupl", 0.50);
#endif

   /// -- GCC options -- //
#if HAVE_I386_GCC49_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC49));
#elif HAVE_I386_GCC47_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC47));
#elif HAVE_I386_GCC46_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC46));
#elif HAVE_I386_GCC45_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC45));
#elif HAVE_I386_GCC48_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC48));
#elif HAVE_I386_GCC5_COMPILER
   setOption(OPT_default_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC5));
#else
   THROW_ERROR("No GCC compiler available");
#endif
   setOption(OPT_compatible_compilers, 0
#if HAVE_I386_GCC45_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC45)
#endif
#if HAVE_I386_GCC46_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC46)
#endif
#if HAVE_I386_GCC47_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC47)
#endif
#if HAVE_I386_GCC48_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC48)
#endif
#if HAVE_I386_GCC49_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC49)
#endif
#if HAVE_I386_GCC5_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC5)
#endif
#if HAVE_ARM_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_ARM_GCC)
#endif
#if HAVE_SPARC_COMPILER
      | static_cast<int>(GccWrapper_CompilerTarget::CT_SPARC_GCC)
#endif
      );
   setOption(OPT_gcc_m32_mx32, "-m32");
   setOption(OPT_without_transformation, true);
   setOption(OPT_compute_size_of, true);
   setOption(OPT_precision, 3);
   setOption(OPT_gcc_c, true);
   setOption(OPT_gcc_config, false);
   setOption(OPT_gcc_costs, false);
   setOption(OPT_gcc_openmp_simd, 0);
   setOption(OPT_gcc_standard, "gnu89");
   setOption(OPT_gcc_optimization_set, GccWrapper_OptimizationSet::OBAMBU);
   setOption(OPT_gcc_include_sysdir, false);
   setOption(OPT_model_costs, false);

   std::string defines;
   if(isOption(OPT_gcc_defines))
      defines = getOption<std::string>(OPT_gcc_defines) + STR_CST_string_separator;
   defines += "__BAMBU__";
   setOption(OPT_gcc_defines, defines);

   setOption(OPT_soft_float, true);
   setOption(OPT_hls_div, false);
   setOption(OPT_max_ulp, 1.0);
   setOption(OPT_skip_pipe_parameter, 0);
   setOption(OPT_unaligned_access, false);
   setOption(OPT_aligned_access, false);
   setOption(OPT_gcc_serialize_memory_accesses, false);
   setOption(OPT_use_asynchronous_memories, true);
   setOption(OPT_do_not_chain_memories, false);
   setOption(OPT_bram_high_latency, false);
   setOption(OPT_experimental_setup, "BAMBU-BALANCED-MP");
   setOption(OPT_DSP_margin_combinational, 1.0);
   setOption(OPT_DSP_margin_pipelined, 1.0);
   setOption(OPT_DSP_allocation_coefficient, 1.0);
   setOption(OPT_estimate_logic_and_connections, true);
   setOption(OPT_registered_inputs, "auto");
   setOption(OPT_scheduling_mux_margins, 0.0);
   setOption(OPT_no_return_zero, false);

#if HAVE_HOST_PROFILING_BUILT
   setOption(OPT_exec_argv, STR_CST_string_separator);
   setOption(OPT_profiling_method, static_cast<int>(HostProfiling_Method::PM_NONE));
#if HAVE_I386_GCC45_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC45));
#elif HAVE_I386_GCC46_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC46));
#elif HAVE_I386_GCC47_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC47));
#elif HAVE_I386_GCC48_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC48));
#elif HAVE_I386_GCC49_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC49));
#elif HAVE_I386_GCC5_COMPILER
   setOption(OPT_host_compiler, static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC5));
#else
   THROW_ERROR("No GCC compiler available");
#endif
#endif
   setOption(OPT_clock_period, 10.0);
   setOption(OPT_add_connection_offset, false);
#if HAVE_EXPERIMENTAL
   setOption(OPT_mixed_design, true);
#endif
#if HAVE_EXPERIMENTAL && HAVE_FROM_PRAGMA_BUILT && HAVE_BAMBU_BUILT
   setOption(OPT_num_threads, 4);
#endif
#if HAVE_EXPERIMENTAL
   setOption(OPT_memory_banks_number, 1);
#endif

   panda_parameters["CSE_size"] = "2";
}

void BambuParameter::add_bambu_library(std::string lib)
{
   unsigned int preferred_compiler = getOption<unsigned int>(OPT_default_compiler);
   std::string archive_files;
   bool isSizeOptimized =
       getOption<GccWrapper_OptimizationSet>(OPT_gcc_opt_level) == GccWrapper_OptimizationSet::Os;
   std::string OsSuffix = "";
   if (isSizeOptimized && lib == "m")
     OsSuffix = "_Os";
   if(isOption(OPT_archive_files))
      archive_files= getOption<std::string>(OPT_archive_files) + STR_CST_string_separator;
#if HAVE_I386_GCC45_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC45))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc45" + OsSuffix + ".a");
   }
#endif
#if HAVE_I386_GCC46_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC46))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc46" + OsSuffix + ".a");
   }
#endif
#if HAVE_I386_GCC47_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC47))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc47" + OsSuffix + ".a");
   }
#endif
#if HAVE_I386_GCC48_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC48))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc48" + OsSuffix + ".a");
   }
#endif
#if HAVE_I386_GCC49_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC49))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc49" + OsSuffix + ".a");
   }
#endif
#if HAVE_I386_GCC5_COMPILER
   if(static_cast<int>(preferred_compiler) & static_cast<int>(GccWrapper_CompilerTarget::CT_I386_GCC5))
   {
      setOption(OPT_archive_files, archive_files + PANDA_LIB_INSTALLDIR "/panda-etc/lib"+lib+"_gcc5" + OsSuffix + ".a");
   }
#endif
}
