/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file evaluation.cpp
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/
///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"

///Header include
#include "evaluation.hpp"

///. include
#include "Parameter.hpp"

///behavior include
#include "call_graph_manager.hpp"

///constants include
#include "bambu_results_xml.hpp"

///hls includes
#include "hls.hpp"
#include "hls_constraints.hpp"
#include "hls_manager.hpp"

///polixml include
#include "xml_document.hpp"

///tree include
#include "behavioral_helper.hpp"

///utility include
#include "xml_helper.hpp"

///HLS include
#include "hls_flow_step_factory.hpp"

///behavior include
#include "call_graph_manager.hpp"

///technology/physical_library/models includes
#include "area_model.hpp"
#include "time_model.hpp"

///technology/physical_library/models/area
#include "clb_model.hpp"

///wrapper/synthesis include
#include "BackendFlow.hpp"

Evaluation::Evaluation(const ParameterConstRef _parameters, const HLS_managerRef _HLSMgr, const DesignFlowManagerConstRef _design_flow_manager):
   HLS_step(_parameters, _HLSMgr, _design_flow_manager, HLSFlowStep_Type::EVALUATION)
{
}

Evaluation::~Evaluation()
{

}

const std::unordered_set<std::tuple<HLSFlowStep_Type, HLSFlowStepSpecializationConstRef, HLSFlowStep_Relationship> > Evaluation::ComputeHLSRelationships(const DesignFlowStep::RelationshipType relationship_type) const
{
   std::string objective_string = parameters->getOption<std::string>(OPT_evaluation_objectives);
   std::vector<std::string> objective_vector = convert_string_to_vector<std::string>(objective_string, ",");
   std::unordered_set<std::tuple<HLSFlowStep_Type, HLSFlowStepSpecializationConstRef, HLSFlowStep_Relationship> > ret;
   switch(relationship_type)
   {
      case DEPENDENCE_RELATIONSHIP:
         {
            switch(parameters->getOption<Evaluation_Mode>(OPT_evaluation_mode))
            {
#if HAVE_EXPERIMENTAL
               case Evaluation_Mode::ESTIMATION:
                  {
                     for(const auto objective : objective_vector)
                     {
                        if (objective == "AREA")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::AREA_ESTIMATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else if (objective == "CLOCK_SLACK")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::CLOCK_SLACK_ESTIMATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else if (objective == "TIME")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::TIME_ESTIMATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else
                           THROW_ERROR("Estimation objective not yet supported " + objective);
                     }
                     break;
                  }
#endif
               case Evaluation_Mode::EXACT:
                  {
                     for(const auto objective : objective_vector)
                     {
                        if(false)
                        {}
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT
                        else if (objective == "AREA")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
#endif
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT && HAVE_SIMULATION_WRAPPER_BUILT
                        else if (objective == "AREAxTIME")
                        {
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SIMULATION_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        }
#endif
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT
                        else if (objective == "BRAMS")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else if (objective == "CLOCK_SLACK")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
#endif
#if HAVE_SIMULATION_WRAPPER_BUILT
                        else if (objective == "CYCLES" || objective == "TOTAL_CYCLES")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SIMULATION_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
#endif
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT
                        else if (objective == "DSPS")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
#endif
#if HAVE_EXPERIMENTAL
                        else if (objective == "EDGES_REDUCTION_EVALUATION")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::EDGES_REDUCTION_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::ALL_FUNCTIONS));
#endif
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT
                        else if (objective == "FREQUENCY")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else if (objective == "PERIOD")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        else if (objective == "REGISTERS")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
#endif
#if HAVE_EXPERIMENTAL
                        else if (objective == "NUM_AF_EDGES")
                           ret.insert(std::make_tuple(HLSFlowStep_Type::NUM_AF_EDGES_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::ALL_FUNCTIONS));
#endif
#if HAVE_LIBRARY_CHARACTERIZATION_BUILT && HAVE_SIMULATION_WRAPPER_BUILT
                        else if (objective == "TIME" || objective == "TOTAL_TIME")
                        {
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SIMULATION_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                           ret.insert(std::make_tuple(HLSFlowStep_Type::SYNTHESIS_EVALUATION, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::WHOLE_APPLICATION));
                        }
#endif
                        else
                           THROW_ERROR("Evaluaton objective not yet supported " + objective);
                     }
                     break;
                  }
               case Evaluation_Mode::NONE:
                  {
                     THROW_UNREACHABLE("");
                  }
               default:
                  THROW_UNREACHABLE("");
            }
            break;
         }
      case INVALIDATION_RELATIONSHIP:
         {
            break;
         }
      case PRECEDENCE_RELATIONSHIP:
         {
#if HAVE_EXPERIMENTAL && HAVE_VCD_BUILT
            ret.insert(std::make_tuple(HLSFlowStep_Type::VCD_UTILITY, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::TOP_FUNCTION));
#endif
            break;
         }
      default:
         THROW_UNREACHABLE("");
   }
   return ret;
}

DesignFlowStep_Status Evaluation::Exec()
{
   const auto hls_top = HLSMgr->get_HLS(HLSMgr->CGetCallGraphManager()->GetRootFunction());
   bool printed_area = false;
   std::string objective_string = parameters->getOption<std::string>(OPT_evaluation_objectives);
   std::vector<std::string> objective_vector = convert_string_to_vector<std::string>(objective_string, ",");
   for(const auto objective : objective_vector)
   {
      if (objective == "CYCLES" or objective =="TIME" or objective == "TOTAL_CYCLES" or objective =="TOTAL_TIME" or objective == "TIMExAREA")
      {
         unsigned long long int accum_cycles=0;
         unsigned int n_testcases=0;
         unsigned long long int average_execution = HLSMgr->get_backend_flow()->ExecuteSimulation(accum_cycles, n_testcases);
         if(objective == "CYCLES")
         {
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Total cycles             : " + boost::lexical_cast<std::string>(accum_cycles) + " cycles");
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Number of executions     : " + boost::lexical_cast<std::string>(n_testcases));
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Average execution        : " + boost::lexical_cast<std::string>(average_execution) + " cycles");
         }
         HLSMgr->evaluations["TOTAL_CYCLES"] = std::vector<double>(1, static_cast<double>(accum_cycles));
         HLSMgr->evaluations["CYCLES"] = std::vector<double>(1, static_cast<double>(average_execution));
      }
      else if ((objective == "AREA" or objective == "AREAxTIME") and printed_area == false)
      {
         printed_area = true;
         ///get the used resources from the wrapper
         area_modelRef area_m = HLSMgr->get_backend_flow()->get_used_resources();

         if (area_m) {
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Area                     : " + STR(area_m->get_area_value()));
            double slices =
                  GetPointer<clb_model>(area_m) ?
                  GetPointer<clb_model>(area_m)->get_resource_value(clb_model::SLICE) : 0;
            if (slices) {
               INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Slices                   : " + STR(slices));
               HLSMgr->evaluations["SLICE"] = std::vector<double>(1, slices);
            }

            double sliceLuts =
                  GetPointer<clb_model>(area_m) ?
                  GetPointer<clb_model>(area_m)->get_resource_value(clb_model::SLICE_LUTS) : 0;
            if (sliceLuts) {
               INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Luts                     : " + STR(sliceLuts));
               HLSMgr->evaluations["SLICE_LUTS"] = std::vector<double>(1, sliceLuts);
            }
         } else
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---(no area values)");
         HLSMgr->evaluations["AREA"] = std::vector<double>(1, area_m->get_area_value());
      }
      else if (objective == "BRAMS")
      {
         area_modelRef area_m = HLSMgr->get_backend_flow()->get_used_resources();
         double brams = 0;
         if(GetPointer<clb_model>(area_m))
            brams = GetPointer<clb_model>(area_m)->get_resource_value(clb_model::BRAM);
         if (area_m)
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---BRAMs                    : " + STR(brams));
         else
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---(no BRAMs)");
         HLSMgr->evaluations["BRAMS"] = std::vector<double>(1, brams);
      }
      else if (objective == "CLOCK_SLACK")
      {
         ///get the timing information after the synthesis
         time_modelRef time_m = HLSMgr->get_backend_flow()->get_timing_results();
         double minimum_period = time_m->get_execution_time();

         double clock_period = HLSMgr->get_HLS(HLSMgr->GetCallGraphManager()->GetRootFunction())->HLS_C->get_clock_period();
         double slack = clock_period - minimum_period;
         if(parameters->getOption<bool>(OPT_timing_violation_abort) and slack < 0.0)
         {
            THROW_UNREACHABLE("Slack is " + STR(slack));
         }
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Design slack             : " + boost::lexical_cast<std::string>(slack));
         HLSMgr->evaluations["CLOCK_SLACK"] = std::vector<double>(1, slack);
      }
      else if (objective == "PERIOD")
      {
         ///get the timing information after the synthesis
         time_modelRef time_m = HLSMgr->get_backend_flow()->get_timing_results();
         double minimum_period = time_m->get_execution_time();
         double clock_period = HLSMgr->get_HLS(HLSMgr->GetCallGraphManager()->GetRootFunction())->HLS_C->get_clock_period();
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Clock period             : " + boost::lexical_cast<std::string>(clock_period));
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Design minimum period    : " + boost::lexical_cast<std::string>(minimum_period));
         HLSMgr->evaluations["PERIOD"] = std::vector<double>(1, minimum_period);
      }
      else if (objective == "DSPS")
      {
         ///get the used resources from the wrapper
         area_modelRef area_m = HLSMgr->get_backend_flow()->get_used_resources();
         double dsps = 0;
         if(GetPointer<clb_model>(area_m))
            dsps = GetPointer<clb_model>(area_m)->get_resource_value(clb_model::DSP);
         if (area_m)
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---DSPs                     : " + STR(dsps));
         else
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---(no DSPs)");
         HLSMgr->evaluations["DSPS"] = std::vector<double>(1, dsps);
      }
      else if (objective == "FREQUENCY" or objective == "TIME" or objective == "TOTAL_TIME" or objective == "AREAxTIME")
      {
         ///get the timing information after the synthesis
         time_modelRef time_m = HLSMgr->get_backend_flow()->get_timing_results();
         double minimum_period = time_m->get_execution_time();

         double maximum_frequency = 1000.0/minimum_period;
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Frequency                : " + STR(maximum_frequency));
         HLSMgr->evaluations["FREQUENCY"] = std::vector<double>(1, maximum_frequency);
      }
      else if (objective == "REGISTERS")
      {
         ///get the used resources from the wrapper
         area_modelRef area_m = HLSMgr->get_backend_flow()->get_used_resources();
         double reg = 0;
         if(GetPointer<clb_model>(area_m))
            reg = GetPointer<clb_model>(area_m)->get_resource_value(clb_model::REGISTERS);
         if (area_m)
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Registers                : " + STR(reg));
         else
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---(no registers)");
         HLSMgr->evaluations["REGISTERS"] = std::vector<double>(1, reg);
      }
   }

   std::string out_file_name = "bambu_results";

   unsigned int progressive = 0;
   std::string candidate_out_file_name;
   do
   {
      candidate_out_file_name = out_file_name + "_" + boost::lexical_cast<std::string>(progressive++) + ".xml";
   } while (boost::filesystem::exists(candidate_out_file_name));

   out_file_name = candidate_out_file_name;
   xml_document document;
   xml_element* nodeRoot = document.create_root_node("bambu_results");

   std::string bench_name;
   if(parameters->isOption(OPT_configuration_name))
   {
      bench_name += parameters->getOption<std::string>(OPT_configuration_name) + ":";
   }
   if(parameters->isOption(OPT_benchmark_name))
   {
      bench_name += parameters->getOption<std::string>(OPT_benchmark_name);
   }
   if(bench_name == "")
   {
      bench_name += hls_top->FB->CGetBehavioralHelper()->get_function_name();
   }
   else
      bench_name += ":" + hls_top->FB->CGetBehavioralHelper()->get_function_name();

   bench_name += "_" + boost::lexical_cast<std::string>(progressive-1);

   const std::string bambu_args = parameters->getOption<std::string>(OPT_cat_args);
   const std::string bambu_version = parameters->PrintVersion();
   boost::posix_time::ptime local_time = boost::posix_time::second_clock::local_time();
   const std::string current_time = boost::posix_time::to_simple_string(local_time);
   WRITE_XNVM2(STR_XML_bambu_results_bambu_args, bambu_args, nodeRoot);
   WRITE_XNVM2(STR_XML_bambu_results_bambu_version, bambu_version, nodeRoot);
   WRITE_XNVM2(STR_XML_bambu_results_timestamp, current_time, nodeRoot);
   WRITE_XNVM2("benchmark_name", bench_name, nodeRoot);
   xml_element* child_element;

   for(auto const objective : objective_vector)
   {
      std::string value;
      if(objective == "AREAxTIME")
      {
         THROW_ASSERT(HLSMgr->evaluations.find("AREA") != HLSMgr->evaluations.end(), "Area value not found");
         THROW_ASSERT(HLSMgr->evaluations.find("CYCLES") != HLSMgr->evaluations.end(), "Cycles value not found");
         THROW_ASSERT(HLSMgr->evaluations.find("FREQUENCY") != HLSMgr->evaluations.end(), "Frequency value not found");
         value = STR(HLSMgr->evaluations.find("AREA")->second[0] * HLSMgr->evaluations.find("CYCLES")->second[0]  / HLSMgr->evaluations.find("FREQUENCY")->second[0]);
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---AreaxTime                : " + value);
      }
      else if (objective == "TIME")
      {
         THROW_ASSERT(HLSMgr->evaluations.find("CYCLES") != HLSMgr->evaluations.end(), "Cycles value not found");
         THROW_ASSERT(HLSMgr->evaluations.find("FREQUENCY") != HLSMgr->evaluations.end(), "Frequency value not found");
         value = STR(HLSMgr->evaluations.find("CYCLES")->second[0] / HLSMgr->evaluations.find("FREQUENCY")->second[0]);
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Time                     : " + value);
      }
      else if (objective == "TOTAL_TIME")
      {
         THROW_ASSERT(HLSMgr->evaluations.find("TOTAL_CYCLES") != HLSMgr->evaluations.end(), "Cycles value not found");
         THROW_ASSERT(HLSMgr->evaluations.find("FREQUENCY") != HLSMgr->evaluations.end(), "Frequency value not found");
         value = STR(HLSMgr->evaluations.find("TOTAL_CYCLES")->second[0] / HLSMgr->evaluations.find("FREQUENCY")->second[0]);
         INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "---Tot. Time                : " + value);
      }
      else
      {
         THROW_ASSERT(HLSMgr->evaluations.find(objective) != HLSMgr->evaluations.end(), "Value of " + objective + " not found");
         const std::vector<double> & evaluations = HLSMgr->evaluations.find(objective)->second;
         for(auto const single_evaluation : evaluations)
         {
            if(value != "")
               value += ",";
            value += STR(single_evaluation);
         }
      }
      child_element = nodeRoot->add_child_element(objective);
      WRITE_XNVM2("value", value, child_element);

      if (objective == "AREA" &&
          HLSMgr->evaluations.find("SLICE") != HLSMgr->evaluations.end())
      {
         child_element = nodeRoot->add_child_element("SLICE");
         value = STR(HLSMgr->evaluations.find("SLICE")->second[0]);
         WRITE_XNVM2("value", value, child_element);
      }
      if (objective == "AREA" &&
          HLSMgr->evaluations.find("SLICE_LUTS") != HLSMgr->evaluations.end())
      {
         child_element = nodeRoot->add_child_element("SLICE_LUTS");
         value = STR(HLSMgr->evaluations.find("SLICE_LUTS")->second[0]);
         WRITE_XNVM2("value", value, child_element);
      }
   }
   child_element = nodeRoot->add_child_element("HLS_execution_time");
   long double exec_time = HLSMgr->HLS_execution_time;
   WRITE_XNVM2("value", STR(exec_time/1000), child_element);

   document.write_to_file_formatted(out_file_name);
   return DesignFlowStep_Status::SUCCESS;
}

bool Evaluation::HasToBeExecuted() const
{
   return true;
}
