/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file minimal_interface.cpp
 * @brief Class to generate minimal interfaces for high-level synthesis
 *
 * This class generates minimal interfaces for connecting modules to
 * microprocessors or busses
 *
 * @author Marco Minutoli <mminutoli@gmail.com>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/
#include "minimal_interface.hpp"

#include "hls.hpp"
#include "hls_manager.hpp"
#include "hls_target.hpp"
#include "memory.hpp"
#include "memory_symbol.hpp"
#include "memory_allocation.hpp"

#include "structural_manager.hpp"
#include "structural_objects.hpp"
#include "technology_manager.hpp"
#include "tree_helper.hpp"
#include "Parameter.hpp"

// behavior/ include
#include "call_graph_manager.hpp"

// HLS/simulation include
#include "testbench_generation_base_step.hpp"

// technology/physical_library include
#include "technology_node.hpp"

#include "copyrights_strings.hpp"

minimal_interface::minimal_interface(const ParameterConstRef _Param, const HLS_managerRef _HLSMgr, unsigned int _funId, const DesignFlowManagerConstRef _design_flow_manager, const HLSFlowStep_Type _hls_flow_step_type) :
   module_interface(_Param, _HLSMgr, _funId, _design_flow_manager, _hls_flow_step_type)
{
   THROW_ASSERT(funId, "Function not set in minimal interface");
}

minimal_interface::~minimal_interface()
{

}

DesignFlowStep_Status minimal_interface::InternalExec()
{
   const structural_managerRef SM = HLS->top;
   if (!SM) THROW_ERROR("Top component has not been created yet!");

   structural_objectRef wrappedObj = SM->get_circ();
   std::string module_name = wrappedObj->get_id() + "_minimal_interface";

   structural_managerRef SM_minimal_interface =
      structural_managerRef(new structural_manager(parameters));
   structural_type_descriptorRef module_type =
      structural_type_descriptorRef(new structural_type_descriptor(module_name));
   SM_minimal_interface->set_top_info(module_name, module_type);
   structural_objectRef interfaceObj = SM_minimal_interface->get_circ();

   // add the core to the wrapper
   wrappedObj->set_owner(interfaceObj);
   wrappedObj->set_id(wrappedObj->get_id() + "_i0");
   GetPointer<module>(interfaceObj)->add_internal_object(wrappedObj);
   ///Set some descriptions and legal stuff
   GetPointer<module>(interfaceObj)->set_description("Minimal interface for top component: " + wrappedObj->get_typeRef()->id_type);
   GetPointer<module>(interfaceObj)->set_copyright(GENERATED_COPYRIGHT);
   GetPointer<module>(interfaceObj)->set_authors("Component automatically generated by bambu");
   GetPointer<module>(interfaceObj)->set_license(GENERATED_LICENSE);

   build_wrapper(wrappedObj, interfaceObj, SM_minimal_interface);

   memory::propagate_memory_parameters(HLS->top->get_circ(), SM_minimal_interface);
   // Generation completed, the new created module substitutes the current top-level one
   HLS->top = SM_minimal_interface;
   return DesignFlowStep_Status::SUCCESS;
}

static bool compareMemVarsPair(std::pair<unsigned int, memory_symbolRef> & first,
                               std::pair<unsigned int, memory_symbolRef> & second)
{
   bool result = false;
   if (first.second->get_address() < second.second->get_address())
      result = true;

   return result;
}


void minimal_interface::build_wrapper(structural_objectRef wrappedObj,
                          structural_objectRef interfaceObj,
                          structural_managerRef SM_minimal_interface)
{
   std::map<unsigned int, structural_objectRef> null_values;

   unsigned int base_address = HLSMgr->base_address;
   bool Has_extern_allocated_data = HLSMgr->Rmem->get_memory_address()-base_address>0;
   bool Has_unknown_addresses  = HLSMgr->Rmem->has_unknown_addresses();
   bool Has_intern_shared_data =
      (HLSMgr->Rmem->has_intern_shared_data()) || 
      (parameters->getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::EXT_PIPELINED_BRAM) ||
      (parameters->getOption<MemoryAllocation_Policy>(OPT_memory_allocation_policy) == MemoryAllocation_Policy::NO_BRAM) ||
      (HLSMgr->CGetCallGraphManager()->GetRootFunction() == funId && parameters->getOption<HLSFlowStep_Type>(OPT_interface_type) == HLSFlowStep_Type::WB4_INTERFACE_GENERATION) ||
      (HLSMgr->hasToBeInterfaced(funId) && HLSMgr->CGetCallGraphManager()->GetRootFunction() != funId) ||
      parameters->getOption<bool>(OPT_memory_mapped_top);
   bool with_master = false;
   bool with_slave = false;
   for (unsigned int i = 0; i < GetPointer<module>(wrappedObj)->get_in_port_size(); i++)
   {
      const structural_objectRef& port_obj = GetPointer<module>(wrappedObj)->get_in_port(i);
      if(GetPointer<port_o>(port_obj)->get_is_memory())
      {
         if (GetPointer<port_o>(port_obj)->get_is_master())
            with_master = true;
         else if (GetPointer<port_o>(port_obj)->get_is_slave())
            with_slave = true;
      }
   }

   std::set<structural_objectRef> portsToSkip;
   std::set<structural_objectRef> portsToConstant;

   /// list of ports that have to be connected to constant 0
   if(with_slave)
   {
      portsToConstant.insert(
               wrappedObj->find_member("Sin_DataRdy", port_o_K, wrappedObj));
      portsToConstant.insert(
               wrappedObj->find_member("Sin_Rdata_ram", port_o_K, wrappedObj));
   }
   if(with_master)
   {
      portsToConstant.insert(
               wrappedObj->find_member("Min_oe_ram", port_o_K, wrappedObj));
      portsToConstant.insert(
               wrappedObj->find_member("Min_we_ram", port_o_K, wrappedObj));
      portsToConstant.insert(
               wrappedObj->find_member("Min_addr_ram", port_o_K, wrappedObj));
      portsToConstant.insert(
               wrappedObj->find_member("Min_Wdata_ram", port_o_K, wrappedObj));
      portsToConstant.insert(
               wrappedObj->find_member("Min_data_ram_size", port_o_K, wrappedObj));
   }

   std::map<structural_objectRef,structural_objectRef> portsToConnect;


   if(!Has_intern_shared_data)
   {
      if(!Has_extern_allocated_data)
      {
         if(!with_master && with_slave)
         {
            /// slave INs
            portsToConstant.insert(
                     wrappedObj->find_member("S_oe_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_we_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_addr_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_Wdata_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_data_ram_size", port_o_K, wrappedObj));
            /// slave OUTs
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj));
         }
         else if(with_master && !with_slave)
         {
            /// do nothing
            /// it may happen with ALL_BRAM memory allocation policy
         }
         else if(with_master && with_slave && Has_unknown_addresses && HLSMgr->Rmem->get_allocated_intern_memory()==0)
         {
            /// intern data is not externally visible
            /// slave INs
            portsToConstant.insert(
                     wrappedObj->find_member("S_oe_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_we_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_addr_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_Wdata_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_data_ram_size", port_o_K, wrappedObj));
            /// slave OUTs
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj));
         }
         else if(with_master && with_slave && !Has_unknown_addresses)
         {
            /// slave INs connections
            portsToConnect[wrappedObj->find_member("S_oe_ram", port_o_K, wrappedObj)] = wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj);
            portsToConnect[wrappedObj->find_member("S_we_ram", port_o_K, wrappedObj)] = wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj);
            portsToConnect[wrappedObj->find_member("S_addr_ram", port_o_K, wrappedObj)] = wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj);
            portsToConnect[wrappedObj->find_member("S_Wdata_ram", port_o_K, wrappedObj)] = wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj);
            portsToConnect[wrappedObj->find_member("S_data_ram_size", port_o_K, wrappedObj)] = wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj);
            /// master INs connections
            portsToConnect[wrappedObj->find_member("M_Rdata_ram", port_o_K, wrappedObj)] = wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj);
            portsToConnect[wrappedObj->find_member("M_DataRdy", port_o_K, wrappedObj)] = wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj);
            /// slave OUTs skip
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj));
            /// master OUTs skip
            portsToSkip.insert(
                     wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj));
         }
         else
         {
            /// do nothing
         }
      }
      else
      {
         if(with_master && !Has_unknown_addresses)
         {
            THROW_ASSERT(parameters->getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) != MemoryAllocation_ChannelsType::MEM_ACC_P1N, "unexpected condition");
            /// allocate the unique shared memory
            structural_objectRef shared_memory;
            bool is_memory_splitted;
            std::string latency_postfix = "";
            if(parameters->getOption<bool>(OPT_bram_high_latency))
               latency_postfix = std::string("_3");
            if(parameters->getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_11 or
                  parameters->getOption<MemoryAllocation_ChannelsType>(OPT_channels_type) == MemoryAllocation_ChannelsType::MEM_ACC_N1 )
            {
               is_memory_splitted = false;
               shared_memory = SM_minimal_interface->add_module_from_technology_library("shared_memory", STD_BRAM+latency_postfix, LIBRARY_STD, interfaceObj, HLSMgr->get_HLS_target()->get_technology_manager());
            }
            else
            {
               is_memory_splitted = true;
               shared_memory = SM_minimal_interface->add_module_from_technology_library("shared_memory", STD_BRAMN+latency_postfix, LIBRARY_STD, interfaceObj, HLSMgr->get_HLS_target()->get_technology_manager());
            }
            unsigned int bus_data_bitsize = HLSMgr->Rmem->get_bus_data_bitsize();
            unsigned int bus_addr_bitsize = HLSMgr->Rmem->get_bus_addr_bitsize();
            unsigned int bus_size_bitsize = HLSMgr->Rmem->get_bus_size_bitsize();
            unsigned int bram_bitsize = HLSMgr->Rmem->get_bram_bitsize();
            unsigned int n_bytes = HLSMgr->Rmem->get_memory_address()-base_address;
            unsigned int vec_size = n_bytes/(bus_data_bitsize/8);
            std::string init_filename = "shared_memory.mem";
            std::ofstream init_file_a((init_filename).c_str());
            std::ofstream init_file_b((+"0_"+init_filename).c_str());

            module* shared_memory_module = GetPointer<module>(shared_memory);
            shared_memory_module->set_parameter("address_space_begin", boost::lexical_cast<std::string>(base_address));
            shared_memory_module->set_parameter("address_space_rangesize", boost::lexical_cast<std::string>(n_bytes));
            if(parameters->isOption(OPT_sparse_memory) && parameters->getOption<bool>(OPT_sparse_memory))
               shared_memory_module->set_parameter("USE_SPARSE_MEMORY", boost::lexical_cast<std::string>(1));
            else
               shared_memory_module->set_parameter("USE_SPARSE_MEMORY", boost::lexical_cast<std::string>(0));
            if(is_memory_splitted)
            {
               shared_memory_module->set_parameter("INIT_file_a", "\"\""+init_filename+"\"\"");
               shared_memory_module->set_parameter("INIT_file_b", "\"\"0_"+init_filename+"\"\"");
            }
            else
               shared_memory_module->set_parameter("INIT_file", "\"\""+init_filename+"\"\"");
            shared_memory_module->set_parameter("n_elements", boost::lexical_cast<std::string>(vec_size));
            shared_memory_module->set_parameter("data_size", boost::lexical_cast<std::string>(bus_data_bitsize));
            shared_memory_module->set_parameter("BRAM_BITSIZE", boost::lexical_cast<std::string>(bram_bitsize));
            if(Has_extern_allocated_data || Has_unknown_addresses)
               shared_memory_module->set_parameter("BUS_PIPELINED", boost::lexical_cast<std::string>(0));
            else
               shared_memory_module->set_parameter("BUS_PIPELINED", boost::lexical_cast<std::string>(1));
            unsigned int n_ports = parameters->isOption(OPT_channels_number) ? parameters->getOption<unsigned int>(OPT_channels_number) : 0;
            for(unsigned int i = 0; i < shared_memory_module->get_in_port_size(); i++)
            {
               structural_objectRef port = shared_memory_module->get_in_port(i);
               if(GetPointer<port_o>(port)->get_kind() == port_vector_o_K && GetPointer<port_o>(port)->get_ports_size() == 0)
                  GetPointer<port_o>(port)->add_n_ports(n_ports, port);
               if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_data_bus())
                  port->type_resize(bus_data_bitsize);
               else if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_addr_bus())
                  port->type_resize(bus_addr_bitsize);
               else if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_size_bus())
                  port->type_resize(bus_size_bitsize);
            }
            for(unsigned int i = 0; i < shared_memory_module->get_out_port_size(); i++)
            {
               structural_objectRef port = shared_memory_module->get_out_port(i);
               if(GetPointer<port_o>(port)->get_kind() == port_vector_o_K && GetPointer<port_o>(port)->get_ports_size() == 0)
                  GetPointer<port_o>(port)->add_n_ports(n_ports, port);
               if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_data_bus())
                  port->type_resize(bus_data_bitsize);
               else if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_addr_bus())
                  port->type_resize(bus_addr_bitsize);
               else if (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_size_bus())
                  port->type_resize(bus_size_bitsize);
            }
            const std::map<unsigned int, memory_symbolRef>& mem_vars = HLSMgr->Rmem->get_ext_memory_variables();
            unsigned int nbyte_on_memory = (bram_bitsize/8);

            std::string init_v;
            std::vector<std::string> splitted;
            std::string current_bits;
            unsigned int counter=0;
            bool is_even = true;
            std::list<std::pair<unsigned int, memory_symbolRef> > mem_variables;
            for(std::map<unsigned int, memory_symbolRef>::const_iterator m = mem_vars.begin(); m != mem_vars.end(); m++)
               mem_variables.push_back(std::make_pair(m->first, m->second));
            mem_variables.sort(compareMemVarsPair);

            std::list<std::pair<unsigned int, memory_symbolRef> >::const_iterator m_next;
            for(std::list<std::pair<unsigned int, memory_symbolRef> >::const_iterator m = mem_variables.begin(); m != mem_variables.end(); m++)
            {
               init_v = TestbenchGenerationBaseStep::print_var_init(HLSMgr->get_tree_manager(), m->first, HLSMgr->Rmem);
               boost::algorithm::split(splitted, init_v , boost::algorithm::is_any_of(","));
               unsigned int byte_allocated = 0;
               unsigned int actual_byte = tree_helper::size(HLSMgr->get_tree_manager(), m->first)/8;
               std::vector<std::string> eightbit_string;
               for(unsigned int i = 0; i < splitted.size(); i++)
               {
                  current_bits = splitted[i];
                  for(unsigned int base_index = 0; base_index < current_bits.size(); base_index = base_index + 8, ++byte_allocated)
                     eightbit_string.push_back(current_bits.substr(current_bits.size()-8-base_index, 8));
               }
               std::string str_bit;
               str_bit ="";
               for(unsigned int l = 0; l < eightbit_string.size();)
               {
                  do
                  {
                     str_bit = eightbit_string[l] + str_bit;
                     ++counter;
                     ++l;
                     if(counter%nbyte_on_memory == 0)
                     {
                        if(is_even || !is_memory_splitted)
                           init_file_a << str_bit << std::endl;
                        else
                           init_file_b << str_bit << std::endl;
                        str_bit ="";
                        is_even = !is_even;
                     }
                  }
                  while(l < eightbit_string.size());
               }
               /// data padding
               if(actual_byte>byte_allocated)
               {
                  for(unsigned int base_index = 0; base_index < actual_byte-byte_allocated; ++base_index)
                  {
                     str_bit = "00000000" + str_bit;
                     counter++;
                     if(counter%nbyte_on_memory == 0)
                     {
                        if(is_even || !is_memory_splitted)
                           init_file_a << str_bit << std::endl;
                        else
                           init_file_b << str_bit << std::endl;
                        str_bit ="";
                        is_even = !is_even;
                     }
                  }
               }
               /// alignment padding
               m_next = m;
               ++m_next;
               unsigned int object_offset;
               if(m_next != mem_variables.end())
               {
                  object_offset = HLSMgr->Rmem->get_base_address(m_next->first, HLS->functionId)-HLSMgr->Rmem->get_base_address(m->first, HLS->functionId);
               }
               else
               {
                  object_offset = HLSMgr->Rmem->get_memory_address()-HLSMgr->Rmem->get_base_address(m->first, HLS->functionId);
               }
               THROW_ASSERT(object_offset >= actual_byte, "more allocated memory than expected");
               if(object_offset > actual_byte)
               {
                  for(unsigned int base_index = 0; base_index < object_offset-actual_byte; ++base_index)
                  {
                     str_bit = "00000000" + str_bit;
                     counter++;
                     if(counter%nbyte_on_memory == 0)
                     {
                        if(is_even || !is_memory_splitted)
                           init_file_a << str_bit << std::endl;
                        else
                           init_file_b << str_bit << std::endl;
                        str_bit ="";
                        is_even = !is_even;
                     }
                  }
               }
            }

            structural_objectRef clock_port, reset_port;
            structural_objectRef port_ck = shared_memory->find_member(CLOCK_PORT_NAME, port_o_K, shared_memory);
            clock_port= SM_minimal_interface->add_port(GetPointer<port_o>(port_ck)->get_id(), port_o::IN,
                                                       interfaceObj, port_ck->get_typeRef());
            SM_minimal_interface->add_connection(clock_port, port_ck);

            structural_objectRef port_rst = shared_memory->find_member(RESET_PORT_NAME, port_o_K, shared_memory);
            reset_port= SM_minimal_interface->add_port(GetPointer<port_o>(port_rst)->get_id(), port_o::IN,
                                                       interfaceObj, port_rst->get_typeRef());
            SM_minimal_interface->add_connection(reset_port, port_rst);

            if(!with_slave)
            {
               structural_objectRef port1,port2;
               structural_objectRef sign;
               /// slave INs connections
               port1 = shared_memory->find_member("S_oe_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = shared_memory->find_member("S_we_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = shared_memory->find_member("S_addr_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = shared_memory->find_member("S_Wdata_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = shared_memory->find_member("S_data_ram_size", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);

               port1 = shared_memory->find_member("Sin_DataRdy", port_o_K, shared_memory);
               if (null_values.find(GET_TYPE_SIZE(port1)) == null_values.end())
               {
                  structural_objectRef const_obj =
                        SM_minimal_interface->add_constant(
                           "null_value_" + STR(GET_TYPE_SIZE(port1)),
                           interfaceObj, port1->get_typeRef(), STR(0));
                  null_values[GET_TYPE_SIZE(port1)] = const_obj;
               }
               if(port1->get_kind() == port_vector_o_K)
               {
                  for(unsigned int p = 0; p < GetPointer<port_o>(port1)->get_ports_size(); ++p)
                  {
                     structural_objectRef port_d = GetPointer<port_o>(port1)->get_port(p);
                     SM_minimal_interface->add_connection(port_d, null_values[GET_TYPE_SIZE(port_d)]);
                  }
               }
               else
                  SM_minimal_interface->add_connection(port1, null_values[GET_TYPE_SIZE(port1)]);

               port1 = shared_memory->find_member("Sin_Rdata_ram", port_o_K, shared_memory);
               if (null_values.find(GET_TYPE_SIZE(port1)) == null_values.end())
               {
                  structural_objectRef const_obj =
                        SM_minimal_interface->add_constant(
                           "null_value_" + STR(GET_TYPE_SIZE(port1)),
                           interfaceObj, port1->get_typeRef(), STR(0));
                  null_values[GET_TYPE_SIZE(port1)] = const_obj;
               }
               if(port1->get_kind() == port_vector_o_K)
               {
                  for(unsigned int p = 0; p < GetPointer<port_o>(port1)->get_ports_size(); ++p)
                  {
                     structural_objectRef port_d = GetPointer<port_o>(port1)->get_port(p);
                     SM_minimal_interface->add_connection(port_d, null_values[GET_TYPE_SIZE(port_d)]);
                  }
               }
               else
                  SM_minimal_interface->add_connection(port1, null_values[GET_TYPE_SIZE(port1)]);


               /// master INs connections
               portsToConnect[wrappedObj->find_member("M_Rdata_ram", port_o_K, wrappedObj)] = shared_memory->find_member("Sout_Rdata_ram", port_o_K, shared_memory);
               portsToConnect[wrappedObj->find_member("M_DataRdy", port_o_K, wrappedObj)] = shared_memory->find_member("Sout_DataRdy", port_o_K, shared_memory);

               /// master OUTs skip
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj));
            }
            else
            {
               structural_objectRef port1,port2;
               structural_objectRef sign;
               /// slave INs connections
               port1 = shared_memory->find_member("S_oe_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = wrappedObj->find_member("S_oe_ram", port_o_K, wrappedObj);
               portsToSkip.insert(port1);
               SM_minimal_interface->add_connection(port1, sign);

               port1 = shared_memory->find_member("S_we_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = wrappedObj->find_member("S_we_ram", port_o_K, wrappedObj);
               portsToSkip.insert(port1);
               SM_minimal_interface->add_connection(port1, sign);

               port1 = shared_memory->find_member("S_addr_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = wrappedObj->find_member("S_addr_ram", port_o_K, wrappedObj);
               portsToSkip.insert(port1);
               SM_minimal_interface->add_connection(port1, sign);

               port1 = shared_memory->find_member("S_Wdata_ram", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = wrappedObj->find_member("S_Wdata_ram", port_o_K, wrappedObj);
               portsToSkip.insert(port1);
               SM_minimal_interface->add_connection(port1, sign);

               port1 = shared_memory->find_member("S_data_ram_size", port_o_K, shared_memory);
               port2 = wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);
               port1 = wrappedObj->find_member("S_data_ram_size", port_o_K, wrappedObj);
               portsToSkip.insert(port1);
               SM_minimal_interface->add_connection(port1, sign);

               port1 = wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj);
               port2 = shared_memory->find_member("Sin_DataRdy", port_o_K, shared_memory);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);

               port1 = wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj);
               port2 = shared_memory->find_member("Sin_Rdata_ram", port_o_K, shared_memory);
               if(port1->get_kind() == port_vector_o_K)
                  sign= SM_minimal_interface->add_sign_vector(port1->get_id() + "_INT", GetPointer<port_o>(port1)->get_ports_size(), interfaceObj, port1->get_typeRef());
               else
                  sign= SM_minimal_interface->add_sign(port1->get_id() + "_INT", interfaceObj, port1->get_typeRef());
               SM_minimal_interface->add_connection(port1, sign);
               SM_minimal_interface->add_connection(sign, port2);

               /// master INs connections
               portsToConnect[wrappedObj->find_member("M_Rdata_ram", port_o_K, wrappedObj)] = shared_memory->find_member("Sout_Rdata_ram", port_o_K, shared_memory);
               portsToConnect[wrappedObj->find_member("M_DataRdy", port_o_K, wrappedObj)] = shared_memory->find_member("Sout_DataRdy", port_o_K, shared_memory);

               /// slave OUTs skip
               portsToSkip.insert(
                        wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj));
               /// master OUTs skip
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_oe_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_we_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_addr_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_Wdata_ram", port_o_K, wrappedObj));
               portsToSkip.insert(
                        wrappedObj->find_member("Mout_data_ram_size", port_o_K, wrappedObj));

            }

         }
         else if(with_slave && HLSMgr->Rmem->get_allocated_intern_memory()==0)
         {
            /// slave INs
            portsToConstant.insert(
                     wrappedObj->find_member("S_oe_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_we_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_addr_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_Wdata_ram", port_o_K, wrappedObj));
            portsToConstant.insert(
                     wrappedObj->find_member("S_data_ram_size", port_o_K, wrappedObj));
            /// slave OUTs
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_Rdata_ram", port_o_K, wrappedObj));
            portsToSkip.insert(
                     wrappedObj->find_member("Sout_DataRdy", port_o_K, wrappedObj));
         }
      }
   }


   // in ports
   for (unsigned int i = 0;
        i < GetPointer<module>(wrappedObj)->get_in_port_size(); ++i)
   {
      structural_objectRef port_in =
            GetPointer<module>(wrappedObj)->get_in_port(i);
      if (portsToSkip.find(port_in) == portsToSkip.end() &&
            portsToConnect.find(port_in) == portsToConnect.end() &&
            portsToConstant.find(port_in) == portsToConstant.end())
      {
         structural_objectRef ext_port = interfaceObj->find_member(GetPointer<port_o>(port_in)->get_id(), port_o_K, interfaceObj);

         if(!ext_port)
         {
            if(port_in->get_kind() == port_vector_o_K)
               ext_port= SM_minimal_interface->add_port_vector(
                                               GetPointer<port_o>(port_in)->get_id(), port_o::IN,
                                               GetPointer<port_o>(port_in)->get_ports_size(),
                                               interfaceObj, port_in->get_typeRef());
            else
               ext_port= SM_minimal_interface->add_port(
                                            GetPointer<port_o>(port_in)->get_id(), port_o::IN,
                                            interfaceObj, port_in->get_typeRef());
         }
         port_o::fix_port_properties(port_in, ext_port);
         SM_minimal_interface->add_connection(port_in, ext_port);
      }
      else if(portsToConnect.find(port_in) != portsToConnect.end())
      {
         structural_objectRef sign;
         if(port_in->get_kind() == port_vector_o_K)
            sign = SM_minimal_interface->add_sign_vector(port_in->get_id() + "_INT", GetPointer<port_o>(port_in)->get_ports_size(), interfaceObj, port_in->get_typeRef());
         else
            sign = SM_minimal_interface->add_sign(port_in->get_id() + "_INT", interfaceObj, port_in->get_typeRef());
         SM_minimal_interface->add_connection(port_in, sign);
         SM_minimal_interface->add_connection(sign, portsToConnect[port_in]);
      }
      else if(portsToConstant.find(port_in) != portsToConstant.end())
      {
         if (null_values.find(GET_TYPE_SIZE(port_in)) == null_values.end())
         {
            structural_objectRef const_obj =
                  SM_minimal_interface->add_constant(
                     "null_value_" + STR(GET_TYPE_SIZE(port_in)),
                     interfaceObj, port_in->get_typeRef(), STR(0));
            null_values[GET_TYPE_SIZE(port_in)] = const_obj;
         }
         if(port_in->get_kind() == port_vector_o_K)
         {
            for(unsigned int p = 0; p < GetPointer<port_o>(port_in)->get_ports_size(); ++p)
            {
               structural_objectRef port_d = GetPointer<port_o>(port_in)->get_port(p);
               SM_minimal_interface->add_connection(port_d, null_values[GET_TYPE_SIZE(port_d)]);
            }
         }
         else
            SM_minimal_interface->add_connection(port_in, null_values[GET_TYPE_SIZE(port_in)]);
      }
   }

   // out ports
   for (unsigned int i = 0;
        i < GetPointer<module>(wrappedObj)->get_out_port_size(); ++i)
   {
      structural_objectRef port_out =
            GetPointer<module>(wrappedObj)->get_out_port(i);
      if (portsToSkip.find(port_out) == portsToSkip.end() && portsToConnect.find(port_out) == portsToConnect.end())
      {
         structural_objectRef ext_port;
         if(port_out->get_kind() == port_vector_o_K)
            ext_port = SM_minimal_interface->add_port_vector(
                                               GetPointer<port_o>(port_out)->get_id(), port_o::OUT,
                                               GetPointer<port_o>(port_out)->get_ports_size(),
                                               interfaceObj, port_out->get_typeRef());
         else
            ext_port = SM_minimal_interface->add_port(
                                            GetPointer<port_o>(port_out)->get_id(), port_o::OUT,
                                            interfaceObj, port_out->get_typeRef());

         port_o::fix_port_properties(port_out, ext_port);
         SM_minimal_interface->add_connection(port_out, ext_port);
      }
      else if (port_out->get_kind() == port_vector_o_K && !GetPointer<port_o>(port_out)->find_bounded_object())
      {
         for(unsigned int p = 0; p < GetPointer<port_o>(port_out)->get_ports_size(); ++p)
         {
            structural_objectRef port_d = GetPointer<port_o>(port_out)->get_port(p);
            if(!GetPointer<port_o>(port_d)->find_bounded_object())
            {
               std::string name = "null_out_signal_" + port_out->get_owner()->get_id() + "_" + port_out->get_id() + "_" + port_d->get_id();
               structural_objectRef sign = SM_minimal_interface->add_sign(name, interfaceObj, port_d->get_typeRef());
               SM_minimal_interface->add_connection(port_d, sign);
            }
         }
      }
   }

   // in-out ports
   for (unsigned int i = 0;
        i < GetPointer<module>(wrappedObj)->get_in_out_port_size(); ++i)
   {
      structural_objectRef port_in_out =
         GetPointer<module>(wrappedObj)->get_in_out_port(i);
      structural_objectRef ext_port;
      if(port_in_out->get_kind() == port_vector_o_K)
         ext_port = SM_minimal_interface->add_port_vector(
                                            GetPointer<port_o>(port_in_out)->get_id(), port_o::IO,
                                            GetPointer<port_o>(port_in_out)->get_ports_size(),
                                            interfaceObj, port_in_out->get_typeRef());
      else
         ext_port = SM_minimal_interface->add_port(
                                         GetPointer<port_o>(port_in_out)->get_id(), port_o::IO,
                                         interfaceObj, port_in_out->get_typeRef());
      port_o::fix_port_properties(port_in_out, ext_port);
      SM_minimal_interface->add_connection(port_in_out, ext_port);
   }
}
