/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file add_library.cpp
 * @brief Implementation of the class to add the current module to the technology library
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/
#include "add_library.hpp"

#include "hls.hpp"
#include "hls_constraints.hpp"
#include "hls_target.hpp"
#include "hls_manager.hpp"
#include "state_transition_graph_manager.hpp"
#include "state_transition_graph.hpp"
#include "memory.hpp"

#include "tree_manager.hpp"
#include "tree_helper.hpp"
#include "structural_manager.hpp"
#include "structural_objects.hpp"
#include "technology_manager.hpp"
#include "time_model.hpp"
#include "area_model.hpp"

#include "Parameter.hpp"

///HLS/module_allocation include
#include "allocation_information.hpp"

///technology/physical_library include
#include "technology_node.hpp"

AddLibrarySpecialization::AddLibrarySpecialization(const bool _interfaced) :
   interfaced(_interfaced)
{}

const std::string AddLibrarySpecialization::GetKindText() const
{
   return interfaced ? "Interfaced" : "";
}

const std::string AddLibrarySpecialization::GetSignature() const
{
   return interfaced ? "Interfaced" : "";
}

add_library::add_library(const ParameterConstRef _parameters, const HLS_managerRef _HLSMgr, unsigned _funId, const DesignFlowManagerConstRef _design_flow_manager, const HLSFlowStepSpecializationConstRef _hls_flow_step_specialization) :
   HLSFunctionStep(_parameters, _HLSMgr, _funId, _design_flow_manager, HLSFlowStep_Type::ADD_LIBRARY, _hls_flow_step_specialization)
{

}

add_library::~add_library()
{

}

const std::unordered_set<std::tuple<HLSFlowStep_Type, HLSFlowStepSpecializationConstRef, HLSFlowStep_Relationship> > add_library::ComputeHLSRelationships(const DesignFlowStep::RelationshipType relationship_type) const
{
   const AddLibrarySpecialization * const  add_library_specialization = GetPointer<const AddLibrarySpecialization>(hls_flow_step_specialization);
   THROW_ASSERT(hls_flow_step_specialization, "Empty specialization type");
   THROW_ASSERT(add_library_specialization, "Wrong specialization type: " + hls_flow_step_specialization->GetKindText());
   std::unordered_set<std::tuple<HLSFlowStep_Type, HLSFlowStepSpecializationConstRef, HLSFlowStep_Relationship> > ret;
   switch(relationship_type)
   {
      case DEPENDENCE_RELATIONSHIP:
         {
            if(add_library_specialization->interfaced)
            {
               ret.insert(std::make_tuple(parameters->getOption<HLSFlowStep_Type>(OPT_interface_type), HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::SAME_FUNCTION));
            }
            else
            {
               const HLSFlowStep_Type top_entity_type =
                  HLSMgr->hasToBeInterfaced(funId) and
                  (HLSMgr->existAdditionalTop() or parameters->getOption<HLSFlowStep_Type>(OPT_interface_type) == HLSFlowStep_Type::WB4_INTERFACE_GENERATION) ?
                  HLSFlowStep_Type::TOP_ENTITY_MEMORY_MAPPED_CREATION :
                  HLSFlowStep_Type::TOP_ENTITY_CREATION;
               ret.insert(std::make_tuple(top_entity_type, HLSFlowStepSpecializationConstRef(), HLSFlowStep_Relationship::SAME_FUNCTION));
            }
            break;
         }
      case INVALIDATION_RELATIONSHIP:
         {
            break;
         }
      case PRECEDENCE_RELATIONSHIP:
         {
            break;
         }
      default:
         THROW_UNREACHABLE("");
   }
   return ret;
}

DesignFlowStep_Status add_library::InternalExec()
{
   /*const tree_managerRef TreeM = HLSMgr->get_tree_manager();*/
   std::string function_name = /*tree_helper::normalized_ID(tree_helper::name_function(TreeM, HLS->functionId));*/ HLS->top->get_circ()->get_typeRef()->id_type;
   const technology_managerRef TM = HLS->HLS_T->get_technology_manager();
   THROW_ASSERT(HLS->top, "Top has not been set");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Adding " + function_name + " to " + WORK_LIBRARY + " - Object is " + HLS->top->get_circ()->get_path());
   TM->add_resource(WORK_LIBRARY, function_name, HLS->top);
   TM->add_operation(WORK_LIBRARY, function_name, function_name);
   double clock_period_value = HLS->HLS_C->get_clock_period();
   double cprf = HLS->HLS_C->get_clock_period_resource_fraction();
   double clk=cprf*clock_period_value;
   const target_deviceRef device = HLS->HLS_T->get_target_device();
   functional_unit* fu = GetPointer<functional_unit>(TM->get_fu(function_name, WORK_LIBRARY));
   fu->set_clock_period(clock_period_value);
   fu->set_clock_period_resource_fraction(cprf);
   fu->CM->add_NP_functionality(HLS->top->get_circ(), NP_functionality::LIBRARY, function_name);
   operation* op = GetPointer<operation>(fu->get_operation(function_name));
   op->time_m = time_model::create_model(device->get_type(), parameters);
   op->primary_inputs_registered = HLS->registered_inputs;
   if(HLS->STG->CGetStg()->CGetStateTransitionGraphInfo()->is_a_dag)
   {
      const structural_objectRef cir = HLS->top->get_circ();
      module *mod = GetPointer<module>(cir);
      bool is_bounded = !HLSMgr->Rmem->has_proxied_internal_variables(funId) &&
                        !parameters->getOption<bool>(OPT_disable_bounded_function);
      for (unsigned int i = 0; i < mod->get_in_port_size() && is_bounded; i++)
      {
         const structural_objectRef& port_obj = mod->get_in_port(i);
         if(GetPointer<port_o>(port_obj)->get_is_memory())
            is_bounded = false; ///functions accessing memory are classified as unbounded
      }
      double call_delay = HLS->allocation_information->estimate_call_delay();
      if(is_bounded)
      {
         unsigned int min_cycles = HLS->STG->CGetStg()->CGetStateTransitionGraphInfo()->min_cycles;
         unsigned int max_cycles = HLS->STG->CGetStg()->CGetStateTransitionGraphInfo()->max_cycles;
         if(max_cycles == min_cycles && min_cycles > 0 && min_cycles < 8)
         {
            op->bounded = true;
            double exec_time;
            if(min_cycles>1)
               exec_time = clk*(min_cycles-1)+call_delay;
            else
               exec_time = call_delay;
            op->time_m->set_execution_time(exec_time, min_cycles);
            op->time_m->set_stage_period(0);
            if(min_cycles<=1)
               fu->logical_type = functional_unit::COMBINATIONAL;
         }
         else
         {
            op->bounded = false;
            op->time_m->set_execution_time(call_delay, 0);
         }
      }
      else
      {
         op->bounded = false;
         op->time_m->set_execution_time(call_delay, 0);
      }
   }
   else
   {
      op->bounded = false;
      op->time_m->set_execution_time(HLS->allocation_information->estimate_call_delay(), 0);
   }
   op->time_m->set_synthesis_dependent(true);
   fu->area_m = area_model::create_model(device->get_type(), parameters);
   fu->area_m->set_area_value(2000);/// fake number to avoid sharing of functions
   INDENT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Added " + function_name + (op->bounded ? "" : "(unbounded)")  + " to WORK_LIBRARY");
   return DesignFlowStep_Status::SUCCESS;
}
