/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file allocation_information.cpp
 * @brief This package is used by all HLS packages to manage resource constraints and characteristics.
 *
 * @author Marco Lattuada <marco.lattuada@polimi.it>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 *
*/

///Header include
#include "allocation_information.hpp"

///. include
#include "Parameter.hpp"

///behavior include
#include "basic_block.hpp"

///HLS include
#include "hls.hpp"
#include "hls_constraints.hpp"
#include "hls_manager.hpp"
#include "hls_target.hpp"

///HLS/binding/module
#include "fu_binding.hpp"

///HLS/memory include
#include "memory.hpp"

///HLS/module_allocation include
#include "allocation.hpp"

///HLS/scheduling include
#include "schedule.hpp"

///HLS/stg include
#include "state_transition_graph_manager.hpp"

///technology includes
#include "technology_manager.hpp"

///technology/physical_library
#include "technology_node.hpp"

///technology/physical_library/models includes
#include "area_model.hpp"
#include "clb_model.hpp"
#include "time_model.hpp"

///tree includes
#include "ext_tree_node.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "behavioral_helper.hpp"

///utility include
#include "math_function.hpp"

AllocationInformation::AllocationInformation(const HLS_managerConstRef _HLS_mgr, const hlsRef _hls, const unsigned int _function_id, const ParameterConstRef _parameters) :
   HLS_mgr(_HLS_mgr),
   op_graph(_HLS_mgr->CGetFunctionBehavior(_function_id)->CGetOpGraph(FunctionBehavior::CFG)),
   hls(_hls),
   HLS_C(hls->HLS_C),
   HLS_T(hls->HLS_T),
   Rmem(_HLS_mgr->Rmem),
   TreeM(_HLS_mgr->get_tree_manager()),
   connection_time_ratio(HLS_T->get_target_device()->has_parameter("connection_time_ratio") ? HLS_T->get_target_device()->get_parameter<double>("connection_time_ratio") : 1),
   controller_delay_multiplier(HLS_T->get_target_device()->has_parameter("controller_delay_multiplier") ? HLS_T->get_target_device()->get_parameter<double>("controller_delay_multiplier") : 1),
   setup_multiplier(HLS_T->get_target_device()->has_parameter("setup_multiplier") ? HLS_T->get_target_device()->get_parameter<double>("setup_multiplier") : 1.0),
   time_multiplier(HLS_T->get_target_device()->has_parameter("time_multiplier") ? HLS_T->get_target_device()->get_parameter<double>("time_multiplier") : 1.0),
   mux_time_multiplier(HLS_T->get_target_device()->has_parameter("mux_time_multiplier") ? HLS_T->get_target_device()->get_parameter<double>("mux_time_multiplier") : 1.0),
   memory_correction_coefficient(HLS_T->get_target_device()->has_parameter("memory_correction_coefficient") ? HLS_T->get_target_device()->get_parameter<double>("memory_correction_coefficient") : 0.7),
   connection_offset(_parameters->isOption(OPT_add_connection_offset) && _parameters->getOption<bool>(OPT_add_connection_offset) ?
                        (HLS_T->get_target_device()->has_parameter("connection_offset") ? HLS_T->get_target_device()->get_parameter<double>("connection_offset") : 0.2)
                      : (HLS_T->get_target_device()->has_parameter("connection_offset") ? HLS_T->get_target_device()->get_parameter<double>("connection_offset") : 0)),
   DSPs_margin(HLS_T->get_target_device()->has_parameter("DSPs_margin") && _parameters->getOption<double>(OPT_DSP_margin_combinational) == 1.0 ? HLS_T->get_target_device()->get_parameter<double>("DSPs_margin") : _parameters->getOption<double>(OPT_DSP_margin_combinational)),
   DSPs_margin_stage(HLS_T->get_target_device()->has_parameter("DSPs_margin_stage") && _parameters->getOption<double>(OPT_DSP_margin_pipelined) == 1.0 ? HLS_T->get_target_device()->get_parameter<double>("DSPs_margin_stage") : _parameters->getOption<double>(OPT_DSP_margin_pipelined)),
   DSP_allocation_coefficient(HLS_T->get_target_device()->has_parameter("DSP_allocation_coefficient") && _parameters->getOption<double>(OPT_DSP_allocation_coefficient) == 1.0 ? HLS_T->get_target_device()->get_parameter<double>("DSP_allocation_coefficient") : _parameters->getOption<double>(OPT_DSP_allocation_coefficient)),
   minimumSlack(std::numeric_limits<double>::max()),
   n_complex_operations(0),
   function_id(_function_id),
   parameters(_parameters),
   debug_level(_parameters->get_class_debug_level(GET_CLASS(*this)))
{
   ///initialize mux DBs
   const technology_managerRef TM = HLS_T->get_technology_manager();
   technology_nodeRef f_unit_mux = TM->get_fu(MUX_N_TO_1, LIBRARY_STD_FU);
   THROW_ASSERT(f_unit_mux, "Library miss component: " + std::string(MUX_N_TO_1));
   functional_unit_template * fu_br= GetPointer<functional_unit_template>(f_unit_mux);
   technology_nodeRef op_mux_node =GetPointer<functional_unit>(fu_br->FU)->get_operation(MUX_N_TO_1);
   operation * op_mux = GetPointer<operation>(op_mux_node);
   std::string temp_portsize_parameters = op_mux->portsize_parameters;
   std::vector<unsigned int> mux_precisions;
   mux_precisions.push_back(1);mux_precisions.push_back(8);mux_precisions.push_back(16);mux_precisions.push_back(32);mux_precisions.push_back(64);
   std::vector<std::string> parameters_split;
   boost::algorithm::split(parameters_split, temp_portsize_parameters, boost::algorithm::is_any_of("|"));
   THROW_ASSERT(parameters_split.size() > 0, "unexpected portsize_parameter format");
   for(auto module_prec : mux_precisions)
   {
      for(size_t el_indx = 0; el_indx < parameters_split.size(); ++el_indx)
      {
         std::vector<std::string> parameters_pairs;
         boost::algorithm::split(parameters_pairs, parameters_split[el_indx], boost::algorithm::is_any_of(":"));
         if(parameters_pairs[0] == "*")
         {
            temp_portsize_parameters = parameters_pairs[1];
            break;
         }
         else if(boost::lexical_cast<unsigned int>(parameters_pairs[0]) == module_prec)
         {
            temp_portsize_parameters = parameters_pairs[1];
            break;
         }
      }
      THROW_ASSERT(temp_portsize_parameters != "", "expected some portsize0_parameters for the the template operation");
      std::vector<std::string> portsize_parameters;
      boost::algorithm::split(portsize_parameters, temp_portsize_parameters, boost::algorithm::is_any_of(","));
      for(auto n_inputs : portsize_parameters)
      {
         const technology_nodeRef fu_cur_obj = HLS_T->get_technology_manager()->get_fu(std::string(MUX_N_TO_1)+"_"+STR(module_prec)+"_"+STR(module_prec)+"_"+STR(module_prec)+"_"+n_inputs, LIBRARY_STD_FU);
         if(fu_cur_obj)
         {
            const functional_unit* fu_cur = GetPointer<functional_unit>(fu_cur_obj);
            area_modelRef a_m = fu_cur->area_m;
            double cur_area = GetPointer<clb_model>(a_m)->get_resource_value(clb_model::SLICE_LUTS);
            if(cur_area == 0.0)
               cur_area = GetPointer<clb_model>(a_m)->get_area_value();
            unsigned int n_inputs_value = boost::lexical_cast<unsigned int>(n_inputs);
            mux_area_db[module_prec][n_inputs_value] = cur_area;
            operation* fu_cur_operation = GetPointer<operation>(fu_cur->get_operation(MUX_N_TO_1));
            mux_timing_db[module_prec][n_inputs_value] = time_m_execution_time(fu_cur_operation) * mux_time_multiplier;
         }
      }
   }
#define MAX_MUX_N_INPUTS 65
   for(auto module_prec : mux_precisions)
   {
      if(mux_area_db.find(module_prec) == mux_area_db.end())
      {
         THROW_ASSERT(mux_timing_db.find(module_prec) == mux_timing_db.end(), "unexpected condition");
         for(unsigned int n_ins=2; n_ins <= MAX_MUX_N_INPUTS; ++n_ins)
         {
            unsigned int n_levels;
            for(n_levels = 1; n_ins > (1ULL<<n_levels) ; ++n_levels);
            mux_area_db[module_prec][n_ins] = (n_ins-1) * mux_area_unit_raw(module_prec);
            mux_timing_db[module_prec][n_ins] =  n_levels * (mux_time_unit_raw(module_prec)+get_setup_hold_time()) * mux_time_multiplier;
         }
      }
      else
      {
         THROW_ASSERT(mux_timing_db.find(module_prec) != mux_timing_db.end(), "unexpected condition");
         THROW_ASSERT(mux_area_db.find(module_prec)->second.find(2) != mux_area_db.find(module_prec)->second.end(), "unexpected condition");
         THROW_ASSERT(mux_timing_db.find(module_prec)->second.find(2) != mux_timing_db.find(module_prec)->second.end(), "unexpected condition");
         unsigned int prev_non_null = 2;
         for(unsigned int n_ins=3; n_ins <= MAX_MUX_N_INPUTS; ++n_ins)
         {
            if(mux_area_db.find(module_prec)->second.find(n_ins) != mux_area_db.find(module_prec)->second.end())
            {
               if(prev_non_null+1 != n_ins)
               {
                  for(; prev_non_null+1 < n_ins; ++prev_non_null)
                  {
                     mux_area_db.find(module_prec)->second[prev_non_null+1] = mux_area_db.find(module_prec)->second.find(prev_non_null)->second + (mux_area_db.find(module_prec)->second.find(n_ins)->second - mux_area_db.find(module_prec)->second.find(prev_non_null)->second)/(n_ins-prev_non_null);
                     mux_timing_db.find(module_prec)->second[prev_non_null+1] = mux_timing_db.find(module_prec)->second.find(prev_non_null)->second + (mux_timing_db.find(module_prec)->second.find(n_ins)->second - mux_timing_db.find(module_prec)->second.find(prev_non_null)->second)/(n_ins-prev_non_null);
                  }
               }
               prev_non_null = n_ins;
            }
         }
      }
   }
   mux_area_db[128] = mux_area_db.find(64)->second;
   mux_timing_db[128] = mux_timing_db.find(64)->second;

   ///initialize DSP x and y db
   if(HLS_T->get_target_device()->has_parameter("DSPs_x_sizes"))
   {
      THROW_ASSERT(HLS_T->get_target_device()->has_parameter("DSPs_y_sizes"), "device description is not complete");
      THROW_ASSERT(HLS_T->get_target_device()->has_parameter("DSPs_ry_sizes"), "device description is not complete");
      THROW_ASSERT(HLS_T->get_target_device()->has_parameter("DSPs_rx_sizes"), "device description is not complete");
      std::string DSPs_x_sizes = HLS_T->get_target_device()->get_parameter<std::string>("DSPs_x_sizes");
      std::string DSPs_y_sizes = HLS_T->get_target_device()->get_parameter<std::string>("DSPs_y_sizes");
      std::string DSPs_rx_sizes = HLS_T->get_target_device()->get_parameter<std::string>("DSPs_rx_sizes");
      std::string DSPs_ry_sizes = HLS_T->get_target_device()->get_parameter<std::string>("DSPs_ry_sizes");
      std::vector<std::string> DSPs_x_sizes_vec;
      std::vector<std::string> DSPs_y_sizes_vec;
      std::vector<std::string> DSPs_rx_sizes_vec;
      std::vector<std::string> DSPs_ry_sizes_vec;
      boost::algorithm::split(DSPs_x_sizes_vec, DSPs_x_sizes, boost::algorithm::is_any_of(","));
      boost::algorithm::split(DSPs_y_sizes_vec, DSPs_y_sizes, boost::algorithm::is_any_of(","));
      boost::algorithm::split(DSPs_rx_sizes_vec, DSPs_rx_sizes, boost::algorithm::is_any_of(","));
      boost::algorithm::split(DSPs_ry_sizes_vec, DSPs_ry_sizes, boost::algorithm::is_any_of(","));
      size_t n_elements = DSPs_x_sizes_vec.size();
      DSP_x_db.resize(n_elements);
      DSP_y_db.resize(n_elements);
      DSP_rx_db.resize(n_elements);
      DSP_ry_db.resize(n_elements);
      for(size_t index = 0; index < n_elements; ++index)
      {
         DSP_x_db[index] = boost::lexical_cast<unsigned int>(DSPs_x_sizes_vec[index]);
         DSP_y_db[index] = boost::lexical_cast<unsigned int>(DSPs_y_sizes_vec[index]);
         DSP_rx_db[index] = boost::lexical_cast<unsigned int>(DSPs_rx_sizes_vec[index]);
         DSP_ry_db[index] = boost::lexical_cast<unsigned int>(DSPs_ry_sizes_vec[index]);
      }
   }
}

AllocationInformation::~AllocationInformation()
{}

double AllocationInformation::time_m_execution_time(operation* op) const
{
   return op->time_m->get_execution_time()*time_multiplier;
}

double AllocationInformation::time_m_stage_period(operation* op) const
{
   return op->time_m->get_stage_period()*time_multiplier;
}



std::pair<std::string, std::string> AllocationInformation::get_fu_name(unsigned int id) const
{
   THROW_ASSERT(id_to_fu_names.find(id) != id_to_fu_names.end(), "Functional unit name not stored!");
   return id_to_fu_names.find(id)->second;
}

unsigned int AllocationInformation::get_number_fu_types() const
{
   return static_cast<unsigned int>(list_of_FU.size());
}

unsigned int AllocationInformation::get_number_fu(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return tech_constraints[fu_name];
}

const std::set<unsigned int> & AllocationInformation::can_implement_set(const vertex v) const
{
   return can_implement_set(op_graph->CGetOpNodeInfo(v)->GetNodeId());
}

const std::set<unsigned int> & AllocationInformation::can_implement_set(const unsigned int v) const
{
   std::unordered_map<unsigned int, std::set<unsigned int> >::const_iterator vtf_it = node_id_to_fus.find(v);
   THROW_ASSERT(vtf_it != node_id_to_fus.end(), "unmapped operation " + TreeM->get_tree_node_const(v)->ToString());
   return vtf_it->second;

}

bool AllocationInformation::CanImplementSetNotEmpty(const unsigned int v) const
{
   return node_id_to_fus.find(v) != node_id_to_fus.end();
}

double AllocationInformation::get_execution_time(const unsigned int fu_name, const vertex v, const OpGraphConstRef g) const
{
   return get_execution_time(fu_name, g->CGetOpNodeInfo(v)->GetNodeId());
}

double AllocationInformation::get_execution_time(const unsigned int fu_name, unsigned int v) const
{
   if(v == ENTRY_ID or v == EXIT_ID)
      return 0.0;
   THROW_ASSERT(can_implement_set(v).find(fu_name) != can_implement_set(v).end(), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + STR(v));
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(tree_helper::normalized_ID(GetPointer<const gimple_node>(TreeM->get_tree_node_const(v))->operation));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   double clock_budget = HLS_C->get_clock_period() * HLS_C->get_clock_period_resource_fraction();
   auto n_cycles = GetPointer<operation>(node_op)->time_m->get_cycles();
   if (n_cycles)
   {
      double stage_time = time_m_stage_period(GetPointer<operation>(node_op));
      if(stage_time<clock_budget && stage_time>0)
         return (n_cycles-1) * clock_budget + time_m_stage_period(GetPointer<operation>(node_op));
      else
      {
         double exec_time = get_execution_time_dsp_modified(fu_name, node_op);
         if(exec_time > (n_cycles-1) * clock_budget && exec_time < n_cycles*clock_budget)
            return exec_time;
         else
            return n_cycles * clock_budget;
      }
   }
   ///DSP based components are underestimated when the RTL synthesis backend converts in LUTs, so we slightly increase the execution time
   return get_execution_time_dsp_modified(fu_name, node_op);
}

double AllocationInformation::get_attribute_of_fu_per_op(const vertex v, const OpGraphConstRef g, Allocation_MinMax allocation_min_max, AllocationInformation::op_target target) const
{
   unsigned int fu_name;
   bool flag;
   double res =  get_attribute_of_fu_per_op(v, g, allocation_min_max, target, fu_name, flag);
   THROW_ASSERT(flag, "something of wrong happen");
   return res;
}


double AllocationInformation::get_attribute_of_fu_per_op(const vertex v, const OpGraphConstRef g, Allocation_MinMax allocation_min_max, AllocationInformation::op_target target, unsigned int &fu_name, bool & flag, const updatecopy_HLS_constraints_functor *CF ) const
{
   const unsigned int node_id = g->CGetOpNodeInfo(v)->GetNodeId();
   const std::set<unsigned int> & fu_set = node_id_to_fus.find(g->CGetOpNodeInfo(v)->GetNodeId())->second;

   std::string op_name = tree_helper::normalized_ID(g->CGetOpNodeInfo(v)->GetOperation());
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();
   std::set<unsigned int>::const_iterator f_i = fu_set.begin();
   double double_value;
   flag = false;
   while (CF && f_i != f_end && ((*CF)(*f_i) <= 0 || (binding.find(node_id) != binding.end()  && binding.find(node_id)->second.second != *f_i))) f_i++;
   if (f_i == f_end) return -1.0;
   flag = true;

   switch (target)
   {
      case initiation_time:
      {
         ControlStep temp(0u);
         fu_name = *f_i;
         if(!has_to_be_synthetized(fu_name))
            return 1.0;

         THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(fu_name)->second.first);
         auto int_value = GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m->get_initiation_time();

         if(binding.find(node_id) != binding.end() && binding.find(node_id)->second.second == fu_name)
            return from_strongtype_cast<double>(int_value);
         f_i++;

         for (; f_i != f_end; f_i++)
         {
            if (CF && (*CF)(*f_i) <= 0) continue;
            switch (allocation_min_max)
            {
               case Allocation_MinMax::MAX:
                  THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
                  temp = MAX(int_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_initiation_time());
                  break;
               case Allocation_MinMax::MIN:
                  THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
                  temp = MIN(int_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_initiation_time());
                  break;
               default:
                  temp = ControlStep(0u);
                  THROW_ERROR(std::string("Not supported AllocationInformation::op_perfomed"));
                  break;
            }
            if (temp != int_value)
            {
               fu_name = *f_i;
               int_value = temp;
            }
         }
         return from_strongtype_cast<double>(int_value);
         break;
      }
      case execution_time:
      {
         double temp;
         fu_name = *f_i;
         if(!has_to_be_synthetized(fu_name))
            return 0.0;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Get Execution time " + GET_NAME(g, v));
         THROW_ASSERT(GetPointer<functional_unit>(list_of_FU[fu_name]), "");
         THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name)), op_name + " not provided by " + list_of_FU[fu_name]->get_name());
         THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(fu_name)->second.first);
         double_value = get_execution_time_dsp_modified(fu_name, GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name));
         if(binding.find(node_id) != binding.end() && binding.find(node_id)->second.second == fu_name)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Got Execution time: " + STR(double_value));
            return double_value;
         }
         f_i++;
         for (; f_i != f_end; f_i++)
         {
            if (CF && (*CF)(*f_i) <= 0) continue;
            switch (allocation_min_max)
            {
               case Allocation_MinMax::MAX:
                  THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
                  temp = MAX(double_value, get_execution_time_dsp_modified(fu_name, GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name)));
                  break;
               case Allocation_MinMax::MIN:
                  THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
                  temp = MIN(double_value, get_execution_time_dsp_modified(fu_name, GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name)));
                  break;
               default:
                  temp = 0;
                  THROW_ERROR(std::string("Not supported AllocationInformation::op_perfomed"));
                  break;
            }
            if (temp != double_value)
            {
               fu_name = *f_i;
               double_value = temp;
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Got Execution time: " + STR(double_value));
         return double_value;
         break;
      }
      case(power_consumption):
      default:
         THROW_ERROR(std::string("Not supported AllocationInformation::op_target"));
         break;
   }
   return -1.0;
}

unsigned int AllocationInformation::min_number_of_resources(const vertex v) const
{
   const std::set<unsigned int> & fu_set = node_id_to_fus.find(op_graph->CGetOpNodeInfo(v)->GetNodeId())->second;

   unsigned int min_num_res = INFINITE_UINT, num_res;
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();

   for (std::set<unsigned int>::const_iterator f_i = fu_set.begin(); f_i != f_end; f_i++)
   {
      num_res = tech_constraints[*f_i];
      THROW_ASSERT(num_res != 0, "something of wrong happen");
      min_num_res = min_num_res > num_res ? num_res : min_num_res;
   }
   return min_num_res;
}

double AllocationInformation::get_setup_hold_time() const
{
   technology_nodeRef f_unit_as = HLS_T->get_technology_manager()->get_fu(ASSIGN_VECTOR_BOOL_STD, LIBRARY_STD_FU);
   THROW_ASSERT(f_unit_as, "Library miss component: " + std::string(ASSIGN_VECTOR_BOOL_STD));
   functional_unit * fu_as= GetPointer<functional_unit>(f_unit_as);
   technology_nodeRef op_as_node =fu_as->get_operation(ASSIGN);
   operation * op_ASSIGN = GetPointer<operation>(op_as_node);
   return time_m_execution_time(op_ASSIGN)*setup_multiplier;
}

bool AllocationInformation::is_indirect_access_memory_unit(unsigned int fu_type) const
{
   technology_nodeRef current_fu = get_fu(fu_type);
   std::string memory_ctrl_type = GetPointer<functional_unit>(current_fu)->memory_ctrl_type;
   return memory_ctrl_type != "" && memory_ctrl_type != MEMORY_CTRL_TYPE_PROXY && memory_ctrl_type != MEMORY_CTRL_TYPE_PROXYN  && memory_ctrl_type != MEMORY_CTRL_TYPE_DPROXY  && memory_ctrl_type != MEMORY_CTRL_TYPE_DPROXYN;
}

double AllocationInformation::get_worst_execution_time(const unsigned int fu_name) const
{
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   const functional_unit::operation_vec node_ops = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operations();
   double max_value = 0.0;
   functional_unit::operation_vec::const_iterator no_it_end = node_ops.end();
   for(functional_unit::operation_vec::const_iterator no_it = node_ops.begin(); no_it != no_it_end; ++no_it)
      max_value = std::max(max_value,  get_execution_time_dsp_modified(fu_name, *no_it));
   return max_value;
}

double AllocationInformation::get_area(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   area_modelRef a_m = GetPointer<functional_unit>(list_of_FU[fu_name])->area_m;
   THROW_ASSERT(a_m, "Area information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   double area = GetPointer<clb_model>(a_m)->get_resource_value(clb_model::SLICE_LUTS);
   if(area == 0.0) area = a_m->get_area_value();
   return area;
}

double AllocationInformation::get_DSPs(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   area_modelRef a_m = GetPointer<functional_unit>(list_of_FU[fu_name])->area_m;
   THROW_ASSERT(a_m, "Area information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   if(GetPointer<clb_model>(a_m))
      return GetPointer<clb_model>(a_m)->get_resource_value(clb_model::DSP);
   else return 0;
}

ControlStep AllocationInformation::get_initiation_time(const unsigned int fu_name, const vertex v) const
{
   return get_initiation_time(fu_name, op_graph->CGetOpNodeInfo(v)->GetNodeId());
}

ControlStep AllocationInformation::get_initiation_time(const unsigned int fu_name, const unsigned int statement_index) const
{
   if(statement_index == ENTRY_ID or statement_index == EXIT_ID)
      return ControlStep(0u);
   const auto operation_name = GetPointer<const gimple_node>(TreeM->get_tree_node_const(statement_index))->operation;
   THROW_ASSERT(can_implement_set(statement_index).find(fu_name) != can_implement_set(statement_index).end(), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + operation_name);
   if(!has_to_be_synthetized(fu_name))
      return ControlStep(0u);
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(tree_helper::normalized_ID(operation_name));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<operation>(node_op)->time_m->get_initiation_time();
}

bool AllocationInformation::is_operation_bounded(const OpGraphConstRef g, const vertex& op, unsigned int fu_type) const
{
   const technology_nodeRef node = get_fu(fu_type);
   std::string op_string = tree_helper::normalized_ID(g->CGetOpNodeInfo(op)->GetOperation());
   const functional_unit* fu = GetPointer<functional_unit>(node);
   const technology_nodeRef op_node = fu->get_operation(op_string);
   THROW_ASSERT(GetPointer<operation>(op_node), "Op node is not an operation");
   return GetPointer<operation>(op_node)->is_bounded();
}

bool AllocationInformation::is_operation_bounded(const unsigned int index, unsigned int fu_type) const
{
   const technology_nodeRef node = get_fu(fu_type);
   std::string op_string = tree_helper::normalized_ID(GetPointer<const gimple_node>(TreeM->get_tree_node_const(index))->operation);
   const functional_unit* fu = GetPointer<functional_unit>(node);
   const technology_nodeRef op_node = fu->get_operation(op_string);
   THROW_ASSERT(op_node, get_fu_name(fu_type).first + " cannot execute " + op_string);
   THROW_ASSERT(GetPointer<operation>(op_node), "Op node is not an operation: " + op_string);
   return GetPointer<operation>(op_node)->is_bounded();
}

bool AllocationInformation::is_operation_PI_registered(const OpGraphConstRef g, const vertex& op, unsigned int fu_type) const
{
   const technology_nodeRef node = get_fu(fu_type);
   std::string op_string = tree_helper::normalized_ID(g->CGetOpNodeInfo(op)->GetOperation());
   const functional_unit* fu = GetPointer<functional_unit>(node);
   const technology_nodeRef op_node = fu->get_operation(op_string);
   THROW_ASSERT(GetPointer<operation>(op_node), "Op node is not an operation");
   return GetPointer<operation>(op_node)->is_primary_inputs_registered();
}

bool AllocationInformation::is_operation_PI_registered(const unsigned int index, unsigned int fu_type) const
{
   const technology_nodeRef node = get_fu(fu_type);
   std::string op_string = tree_helper::normalized_ID(GetPointer<const gimple_node>(TreeM->get_tree_node_const(index))->operation);
   const functional_unit* fu = GetPointer<functional_unit>(node);
   const technology_nodeRef op_node = fu->get_operation(op_string);
   THROW_ASSERT(GetPointer<operation>(op_node), "Op node is not an operation");
   return GetPointer<operation>(op_node)->is_primary_inputs_registered();
}


bool AllocationInformation::is_operation_bounded(const unsigned int index) const
{
   if(CanImplementSetNotEmpty(index))
      return is_operation_bounded(index, GetFuType(index));
   const auto ga = GetPointer<const gimple_assign>(TreeM->get_tree_node_const(index));
   if(ga)
   {
      const auto right = GET_NODE(ga->op1);
      if(right->get_kind() == truth_and_expr_K or right->get_kind() == truth_or_expr_K or right->get_kind() == truth_not_expr_K or right->get_kind() == cond_expr_K or right->get_kind() == ternary_plus_expr_K or right->get_kind() == ternary_mp_expr_K or right->get_kind() == ternary_pm_expr_K or right->get_kind() == ternary_mm_expr_K)
         return true;
   }
   return false;
}

bool AllocationInformation::is_direct_access_memory_unit(unsigned int fu_type) const
{
   technology_nodeRef current_fu = get_fu(fu_type);
   std::string memory_type = GetPointer<functional_unit>(current_fu)->memory_type;
   std::string memory_ctrl_type = GetPointer<functional_unit>(current_fu)->memory_ctrl_type;
   return memory_type != "" || memory_ctrl_type == MEMORY_CTRL_TYPE_PROXY || memory_ctrl_type == MEMORY_CTRL_TYPE_PROXYN  || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXY  || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXYN;
}

bool AllocationInformation::is_direct_proxy_memory_unit(unsigned int fu_type) const
{
   technology_nodeRef current_fu = get_fu(fu_type);
   std::string memory_ctrl_type = GetPointer<functional_unit>(current_fu)->memory_ctrl_type;
   return memory_ctrl_type == MEMORY_CTRL_TYPE_PROXY || memory_ctrl_type == MEMORY_CTRL_TYPE_PROXYN  || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXY  || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXYN;
}

bool AllocationInformation::is_memory_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return memory_units.find(fu_name) != memory_units.end();
}

bool AllocationInformation::is_proxy_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return is_proxy_function_unit(fu_name) or is_proxy_wrapped_unit(fu_name);
}

bool AllocationInformation::is_proxy_function_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return proxy_function_units.find(fu_name) != proxy_function_units.end();
}

bool AllocationInformation::is_proxy_wrapped_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return proxy_wrapped_units.find(fu_name) != proxy_wrapped_units.end();
}

bool AllocationInformation::is_vertex_bounded(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return is_vertex_bounded_rel.find(fu_name) != is_vertex_bounded_rel.end();
}

bool AllocationInformation::is_vertex_bounded_with(const vertex v, unsigned int & fu_name) const
{
   return is_vertex_bounded_with(op_graph->CGetOpNodeInfo(v)->GetNodeId(), fu_name);
}

bool AllocationInformation::is_vertex_bounded_with(const unsigned int v, unsigned int & fu_name) const
{
   if(binding.find(v) == binding.end())
   {
      return false;
   }
   else
   {
      ///If this codition is true, the operation changed type from last time it was performed allocation; we do not invalidate binding since this function is const
      if(v != ENTRY_ID and v != EXIT_ID and GetPointer<const gimple_node>(TreeM->get_tree_node_const(v))->operation != binding.find(v)->second.first)
      {
         return false;
      }
      fu_name = binding.find(v)->second.second;
      return true;
   }
}

bool AllocationInformation::is_artificial_fu(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   std::string fu_string_name = list_of_FU[fu_name]->get_name();
   if(fu_string_name == ASSIGN_UNSIGNED_STD || fu_string_name == ASSIGN_SIGNED_STD || fu_string_name == ASSIGN_REAL_STD || !has_to_be_synthetized(fu_name))
      return true;
   else
      return false;
}

unsigned int AllocationInformation::get_memory_var(const unsigned int fu_name) const
{
   THROW_ASSERT(is_memory_unit(fu_name), "functional unit id not meaningful");
   return memory_units.find(fu_name)->second;
}

std::map<unsigned int, unsigned int> AllocationInformation::get_memory_units() const
{
   return memory_units;
}

const std::map<unsigned int,unsigned int>& AllocationInformation::get_proxy_memory_units() const
{
   return proxy_memory_units;
}

unsigned int AllocationInformation::get_proxy_memory_var(const unsigned int fu_name) const
{
   THROW_ASSERT(proxy_memory_units.find(fu_name) != proxy_memory_units.end(), "functional unit id not meaningful");
   return proxy_memory_units.find(fu_name)->second;
}


const std::map<unsigned int,std::string>& AllocationInformation::get_proxy_function_units() const
{
   return proxy_function_units;
}

const std::map<unsigned int,std::string>& AllocationInformation::get_proxy_wrapped_units() const
{
   return proxy_wrapped_units;
}

bool AllocationInformation::has_to_be_synthetized(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   std::string fu_string_name = list_of_FU[fu_name]->get_name();
   if(fu_string_name == GIMPLE_RETURN_STD || fu_string_name == ENTRY_STD || fu_string_name == EXIT_STD || fu_string_name == NOP_STD ||
         fu_string_name == GIMPLE_PHI_STD || fu_string_name == GIMPLE_ASM_STD || fu_string_name == GIMPLE_LABEL_STD || fu_string_name == GIMPLE_GOTO_STD || fu_string_name == GIMPLE_NOP_STD ||
         fu_string_name == GIMPLE_PRAGMA_STD)
      return false;
   else
      return true;
}

double AllocationInformation::get_stage_period(const unsigned int fu_name, const vertex v, const OpGraphConstRef g) const
{
   return get_stage_period(fu_name, g->CGetOpNodeInfo(v)->GetNodeId());
}

double AllocationInformation::get_stage_period(const unsigned int fu_name, const unsigned int v) const
{
   if(v == ENTRY_ID or v == EXIT_ID)
      return 0.0;
   const std::string operation_t = GetPointer<const gimple_node>(TreeM->get_tree_node_const(v))->operation;
   THROW_ASSERT(can_implement_set(v).find(fu_name) != can_implement_set(v).end(), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + tree_helper::normalized_ID(operation_t));
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(tree_helper::normalized_ID(operation_t));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   ///DSP based components are underestimated when the RTL synthesis backend converts in LUTs, so we slightly increase the stage period
   return get_stage_period_dsp_modified(fu_name, node_op);
}

double AllocationInformation::get_stage_period_raw(const unsigned int fu_name, const vertex v, const OpGraphConstRef g) const
{
   THROW_ASSERT(can_implement_set(v).find(fu_name) != can_implement_set(v).end(), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + tree_helper::normalized_ID(g->CGetOpNodeInfo(v)->GetOperation()));
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(tree_helper::normalized_ID(g->CGetOpNodeInfo(v)->GetOperation()));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return time_m_stage_period(GetPointer<operation>(node_op));
}

double AllocationInformation::estimate_mux_time(unsigned int fu_name) const
{
   unsigned int fu_prec =  get_prec(fu_name);
   fu_prec = resize_to_1_8_16_32_64_128_256_512(fu_prec);
   return mux_time_unit(fu_prec);
}

double AllocationInformation::estimate_muxNto1_delay(unsigned int fu_prec, unsigned int mux_ins) const
{
   if(mux_ins<2) return 0;
   fu_prec = resize_to_1_8_16_32_64_128_256_512(fu_prec);
   if(mux_ins > MAX_MUX_N_INPUTS)
      mux_ins = MAX_MUX_N_INPUTS;
   if(fu_prec > 128)
      fu_prec = 128;
   THROW_ASSERT(mux_timing_db.find(fu_prec) != mux_timing_db.end(), STR(fu_prec));
   THROW_ASSERT(mux_timing_db.find(fu_prec)->second.find(mux_ins) != mux_timing_db.find(fu_prec)->second.end(), "");
   double ret= mux_timing_db.find(fu_prec)->second.find(mux_ins)->second - get_setup_hold_time();
   //INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---delay of MUX with " + STR(mux_ins) + " inputs and with " + STR(fu_prec) + " bits: " + STR(ret));
   return ret;
}

double AllocationInformation::estimate_muxNto1_area(unsigned int fu_prec, unsigned int mux_ins) const
{
   if(mux_ins<2) return 0;
   fu_prec = resize_to_1_8_16_32_64_128_256_512(fu_prec);
   if(mux_ins > MAX_MUX_N_INPUTS)
      mux_ins = MAX_MUX_N_INPUTS;
   double ret= mux_area_db.find(fu_prec)->second.find(mux_ins)->second;
   //INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---area of MUX with " + STR(mux_ins) + " inputs and with " + STR(fu_prec) + " bits: " + STR(ret));
   THROW_ASSERT(ret!= 0.0, "unexpected condition");
   return ret;
}

unsigned int AllocationInformation::get_cycles(const unsigned int fu_name, const vertex v, const OpGraphConstRef g) const
{
   return get_cycles(fu_name, g->CGetOpNodeInfo(v)->GetNodeId());
}

unsigned int AllocationInformation::get_cycles(const unsigned int fu_name, const unsigned int v) const
{
   if(v == ENTRY_ID or v == EXIT_ID)
      return 0;
   const std::string operation_t = GetPointer<const gimple_node>(TreeM->get_tree_node_const(v))->operation;
   THROW_ASSERT(can_implement_set(v).find(fu_name) != can_implement_set(v).end(), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + tree_helper::normalized_ID(operation_t));
   if(!has_to_be_synthetized(fu_name))
      return 0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(operation_t);
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<operation>(node_op)->time_m->get_cycles();
}

technology_nodeRef AllocationInformation::get_fu(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id " + STR(fu_name) + " is not meaningful");
   return list_of_FU[fu_name];
}

unsigned int AllocationInformation::get_number_channels(unsigned int fu_name) const
{
   if(nports_map.find(fu_name) == nports_map.end())
      return 0;
   else
      return nports_map.find(fu_name)->second;
}

///ToBeCompleted
std::string AllocationInformation::get_string_name(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() + "_" + STR(fu_name);
}

bool AllocationInformation::can_implement(const unsigned int fu_id, const vertex v) const
{
   return can_implement_set(v).find(fu_id) != can_implement_set(v).end();
}

bool AllocationInformation::is_read_cond(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == READ_COND_STD;
}

bool AllocationInformation::is_assign(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == ASSIGN_UNSIGNED_STD || list_of_FU[fu_name]->get_name() == ASSIGN_SIGNED_STD || list_of_FU[fu_name]->get_name() == ASSIGN_REAL_STD;
}

bool AllocationInformation::is_return(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == GIMPLE_RETURN_STD;
}

double AllocationInformation::get_execution_time_dsp_modified(const unsigned int fu_name, const technology_nodeRef &node_op) const
{
   if(get_DSPs(fu_name)>0)
      return DSPs_margin*time_m_execution_time(GetPointer<operation>(node_op));
   else
      return time_m_execution_time(GetPointer<operation>(node_op));
}

double AllocationInformation::get_stage_period_dsp_modified(const unsigned int fu_name, const technology_nodeRef &node_op) const
{
   if(get_DSPs(fu_name)>0)
      return DSPs_margin_stage*time_m_stage_period(GetPointer<operation>(node_op));
   else
      return time_m_stage_period(GetPointer<operation>(node_op));
}


double AllocationInformation::get_worst_stage_period(const unsigned int fu_name) const
{
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   const functional_unit::operation_vec node_ops = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operations();
   double max_value = 0.0;
   functional_unit::operation_vec::const_iterator no_it_end = node_ops.end();
   for(functional_unit::operation_vec::const_iterator no_it = node_ops.begin(); no_it != no_it_end; ++no_it)
      max_value = std::max(max_value, get_stage_period_dsp_modified(fu_name, *no_it));
   return max_value;
}

void AllocationInformation::set_number_channels(unsigned int fu_name, unsigned int n_ports)
{
   nports_map[fu_name] = n_ports;
}

unsigned int AllocationInformation::max_number_of_resources(const vertex v) const
{
   const std::set<unsigned int> & fu_set = node_id_to_fus.find(op_graph->CGetOpNodeInfo(v)->GetNodeId())->second;

   unsigned int tot_num_res = 0, num_res;
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();

   for (std::set<unsigned int>::const_iterator f_i = fu_set.begin(); f_i != f_end; f_i++)
   {
      num_res = tech_constraints[*f_i];
      THROW_ASSERT(num_res != 0, "something of wrong happen");
      if (num_res == INFINITE_UINT)
         return num_res;
      else
         tot_num_res += num_res;
   }
   return tot_num_res;
}

unsigned int AllocationInformation::max_number_of_operations(unsigned int fu) const
{
   THROW_ASSERT(fu < get_number_fu_types(), "functional unit id not meaningful");
   THROW_ASSERT(fus_to_node_id.find(fu) != fus_to_node_id.end(), "no operation can be mapped on the given functional unit");
   return static_cast<unsigned int>(fus_to_node_id.find(fu)->second.size());
}

bool AllocationInformation::is_one_cycle_direct_access_memory_unit(unsigned int fu_type) const
{
   technology_nodeRef current_fu = get_fu(fu_type);
   return GetPointer<functional_unit>(current_fu)->memory_type == MEMORY_TYPE_ASYNCHRONOUS || GetPointer<functional_unit>(current_fu)->memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXY;
}

void AllocationInformation::GetNodeTypePrec(const vertex node, const OpGraphConstRef g, node_kind_prec_infoRef info, HLS_manager::io_binding_type &constant_id, bool is_constrained, const HLS_managerConstRef HLSMgr) const
{
   std::vector<HLS_manager::io_binding_type> vars_read = HLSMgr->get_required_values(function_id, node);
   unsigned int first_valid_id = 0;
   unsigned int index=0;
   constant_id = HLS_manager::io_binding_type(0,0);
   if (vars_read.size() == 0)
      return;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Getting node type precision of " + GET_NAME(g, node));
   std::string current_op = tree_helper::normalized_ID(g->CGetOpNodeInfo(node)->GetOperation());

   long long int vec_size = 0;
   bool is_a_pointer = false;
   bool is_a_function = false;
   unsigned int type_index = 0;
   bool is_second_constant = false;
   bool has_formal_parameter = false;
   unsigned int formal_parameter_type = 0;
   unsigned int max_size_in = 0;
   bool is_cond_expr_bool_test = false;
   for (std::vector<HLS_manager::io_binding_type>::iterator itr = vars_read.begin(), end = vars_read.end(); itr != end; ++itr, ++index)
   {
      unsigned int id = std::get<0>(*itr);
      if (id && !first_valid_id)
         first_valid_id = id;
      if(current_op == "cond_expr" && id && !tree_helper::is_constant(TreeM, id))
      {
         if(tree_helper::size(TreeM, id) == 1)
            is_cond_expr_bool_test = true;
      }
      if((current_op == "cond_expr" || current_op == "vec_cond_expr") && index != 0 && id)
         first_valid_id = id;
      if(current_op == "cond_expr" || current_op == "vec_cond_expr") /// no constant characterization for cond expr
          is_second_constant=true;
      if (id==0 || ((tree_helper::is_constant(TreeM, id) || tree_helper::is_concat_bit_ior_expr(TreeM, g->CGetOpNodeInfo(node)->GetNodeId())) &&
                    !is_constrained && !is_second_constant && vars_read.size() != 1 && current_op != "mult_expr" && current_op != "widen_mult_expr"))
      {
         info->input_prec.push_back(0);
         info->real_input_nelem.push_back(0);
         info->base128_input_nelem.push_back(0);
         is_second_constant=true;
         constant_id = *itr;
         if(id)
         {
            type_index = tree_helper::get_type_index(TreeM, id);
            if(tree_helper::is_a_vector(TreeM, type_index))
            {
               const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
               const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
               max_size_in = std::max(max_size_in, element_size);
            }
            else
               max_size_in = std::max(max_size_in, tree_helper::size(TreeM, id));
         }
      }
      else
      {
         type_index = tree_helper::get_type_index(TreeM, id);
         if(tree_helper::is_an_array(TreeM, id) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) /*|| tree_helper::is_a_complex(TreeM, type_index)*/)
         {
            info->input_prec.push_back(32);
            info->real_input_nelem.push_back(0);
            info->base128_input_nelem.push_back(0);
         }
         else
         {
            unsigned int node_id = g->CGetOpNodeInfo(node)->GetNodeId();
            unsigned int form_par_type = tree_helper::get_formal_ith(TreeM, node_id, index);
            unsigned int size_tree_var = id == 0 ? 0 : tree_helper::size(TreeM, id);
            unsigned int size_form_par = form_par_type == 0 ? 0 : tree_helper::size(TreeM, form_par_type);
            unsigned int size_value = size_form_par ? size_form_par : size_tree_var;
            if(form_par_type && index==0)
            {
               formal_parameter_type = form_par_type;
               has_formal_parameter = true;
            }
            if(tree_helper::is_a_vector(TreeM, type_index))
            {
               const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
               const unsigned int vector_size = static_cast<unsigned int>(tree_helper::size(TreeM, type_index));
               const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
               info->real_input_nelem.push_back(vector_size/element_size);
               info->base128_input_nelem.push_back(128/element_size);
               info->input_prec.push_back(element_size);
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Type is " + STR(type_index) + " " + STR(TreeM->CGetTreeNode(type_index)) + " - Number of input elements (base128): " + STR(128/element_size) + " - Number of real input elements: " + STR(vector_size/element_size) + " - Input precision: " + STR(element_size));
            }
            else
            {
               info->real_input_nelem.push_back(0);
               info->base128_input_nelem.push_back(0);
               info->input_prec.push_back(size_value);
            }
         }
      }
   }

   THROW_ASSERT(first_valid_id, "Unexpected pattern");
   if(has_formal_parameter)
   {
      is_a_pointer = tree_helper::is_a_pointer(TreeM, formal_parameter_type);
      type_index = formal_parameter_type;
   }
   else
      type_index = tree_helper::get_type_index(TreeM, first_valid_id, vec_size, is_a_pointer, is_a_function);

   if(is_a_pointer || tree_helper::is_an_array(TreeM, type_index) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) || tree_helper::is_a_complex(TreeM, type_index))
   {
      info->node_kind = "VECTOR_BOOL";
   }
   else if (tree_helper::is_int(TreeM, type_index))
   {
      info->node_kind = "INT";
   }
   else if (tree_helper::is_real(TreeM, type_index))
   {
      info->node_kind = "REAL";
   }
   else if (tree_helper::is_unsigned(TreeM, type_index))
   {
      info->node_kind = "UINT";
   }
   else if (tree_helper::is_bool(TreeM, type_index))
   {
      info->node_kind = "VECTOR_BOOL";
   }
   else if(tree_helper::is_a_vector(TreeM, type_index))
   {
      const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
      if (tree_helper::is_int(TreeM, element_type))
         info->node_kind = "VECTOR_INT";
      else if(tree_helper::is_unsigned(TreeM, element_type))
         info->node_kind = "VECTOR_UINT";
      else if(tree_helper::is_real(TreeM, element_type))
         info->node_kind = "VECTOR_REAL";
   }
   else
   {
      THROW_UNREACHABLE("not supported type: " + STR(type_index) + " - " + TreeM->get_tree_node_const(type_index)->ToString());
   }


   /// Now we need to normalize the size to be compliant with the technology library assumptions
   unsigned int max_size_in_true = std::max(max_size_in, *std::max_element(info->input_prec.begin(), info->input_prec.end()));
   if(is_cond_expr_bool_test) info->is_single_bool_test_cond_expr = true;
   max_size_in =  resize_to_1_8_16_32_64_128_256_512(max_size_in_true);
   unsigned int min_n_elements = 0;
   std::vector<unsigned int>::const_iterator n_el_it_end = info->base128_input_nelem.end();
   for(std::vector<unsigned int>::const_iterator n_el_it = info->base128_input_nelem.begin(); n_el_it != n_el_it_end; ++n_el_it)
   {
      if(*n_el_it)
      {
         if(min_n_elements)
            min_n_elements = std::min(min_n_elements, *n_el_it);
         else
            min_n_elements = *n_el_it;
      }
   }
   /// DSPs based components have to be managed in a different way
   if(current_op == "widen_mult_expr" ||
      current_op == "mult_expr")
   {
      unsigned int nodeOutput_id = HLSMgr->get_produced_value(function_id, node);
      type_index = tree_helper::get_type_index(TreeM, nodeOutput_id);
      if(tree_helper::is_a_vector(TreeM, type_index))
      {
         const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
         const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
         unsigned int output_size = resize_to_1_8_16_32_64_128_256_512(tree_helper::size(TreeM, nodeOutput_id));
         info->real_output_nelem = output_size/element_size;
         info->base128_output_nelem = output_size/element_size;
         info->output_prec = element_size;
      }
      else
      {
          unsigned int output_size = static_cast<unsigned int>(tree_helper::size(TreeM, nodeOutput_id));
          THROW_ASSERT(info->input_prec.size()==2, "unexpected number of inputs");
          unsigned int x_actual_bitsize = info->input_prec[0];
          unsigned int y_actual_bitsize = info->input_prec[1];
          if(x_actual_bitsize>y_actual_bitsize)
             std::swap(x_actual_bitsize,y_actual_bitsize);
          unsigned int ind = 0;
#define DSP_MODEL1
#ifdef DSP_MODEL1
          unsigned int hits = 0;
#endif
          for(auto x : DSP_x_db)
          {
             if(x>=x_actual_bitsize && x_actual_bitsize>=DSP_rx_db[ind])
             {
                x_actual_bitsize = DSP_rx_db[ind];
#ifdef DSP_MODEL1
                ++hits;
#endif
             }
             if(DSP_y_db[ind]>=y_actual_bitsize && y_actual_bitsize>=DSP_ry_db[ind])
             {
                y_actual_bitsize = DSP_ry_db[ind];
#ifdef DSP_MODEL1
                ++hits;
#endif
             }
             ++ind;
          }
          max_size_in = std::max(x_actual_bitsize, y_actual_bitsize);
          max_size_in=resize_to_1_8_16_32_64_128_256_512(max_size_in);
#ifdef DSP_MODEL1
          if(hits==2)
             output_size = max_size_in;
          else
             output_size = resize_to_1_8_16_32_64_128_256_512(output_size);
#else
          output_size = max_size_in;
#endif
          if(current_op == "widen_mult_expr")
          {
             output_size = std::max(max_size_in*2, output_size);
             max_size_in = output_size/2;
          }
          else
             max_size_in = output_size = std::max(max_size_in, output_size);
          info->real_output_nelem = info->base128_output_nelem = 0;
          info->output_prec = output_size;
      }
   }
   else if(current_op.find( "float_expr_") == 0 ||
         current_op.find("fix_trunc_expr_") == 0 ||
         current_op == "dot_prod_expr" ||
         current_op == "widen_sum_expr" ||
         current_op == "widen_mult_hi_expr" ||
         current_op == "widen_mult_lo_expr" ||
         current_op == "vec_unpack_hi_expr" ||
         current_op == "vec_unpack_lo_expr"
         )
   {
      /// ad hoc correction for float_expr conversion
      if(current_op.find( "float_expr_") == 0 && max_size_in < 32)
         max_size_in = 32;
      unsigned int nodeOutput_id = HLSMgr->get_produced_value(function_id, node);
      if (nodeOutput_id)
      {
         type_index = tree_helper::get_type_index(TreeM, nodeOutput_id);
         if(tree_helper::is_an_array(TreeM, type_index) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) /*|| tree_helper::is_a_complex(TreeM, type_index)*/)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Output precision is 32");
            info->output_prec = 32;
         }
         else
         {
            info->output_prec = resize_to_1_8_16_32_64_128_256_512(tree_helper::size(TreeM, nodeOutput_id));
            if(tree_helper::is_a_vector(TreeM, type_index))
            {
               const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
               const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
               info->base128_output_nelem = info->output_prec/element_size;
               info->real_output_nelem = info->output_prec/element_size;
               info->output_prec = element_size;
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Number of output elements (base128): " + STR(info->base128_output_nelem) + " - Number of real output elements: " + STR(info->real_output_nelem) + " - Output precision: " + STR(info->output_prec));
            }
            else
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Output is not a vector");
               info->real_output_nelem = 0;
               info->base128_output_nelem = 0;
            }
         }

         /// ad hoc correction for fix_trunc_expr
         if (current_op.find( "fix_trunc_expr_") == 0 && info->output_prec < 32)
           info->output_prec = 32;
      }
      if(current_op == "dot_prod_expr")
      {
         max_size_in = info->output_prec/2;
         min_n_elements = info->base128_output_nelem*2;
      }
   }
   else if(current_op == "plus_expr" || current_op == "minus_expr"||
           current_op == "pointer_plus_expr" || current_op == "ternary_plus_expr" || current_op == "ternary_pm_expr" || current_op == "ternary_mp_expr" || current_op == "ternary_mm_expr" || current_op == "negate_expr" ||
           current_op == "bit_and_expr" || current_op == "bit_ior_expr" || current_op == "bit_xor_expr" || current_op == "bit_not_expr" || current_op == "bit_ior_concat_expr" ||
           current_op == "cond_expr" || current_op == "vec_cond_expr" /// these ops never have info->output_prec > max_size_in
           )
   {
      unsigned int nodeOutput_id = HLSMgr->get_produced_value(function_id, node);
      THROW_ASSERT(nodeOutput_id, "unexpected condition");
      type_index = tree_helper::get_type_index(TreeM, nodeOutput_id);
      info->output_prec = resize_to_1_8_16_32_64_128_256_512(tree_helper::size(TreeM, nodeOutput_id));
      if(tree_helper::is_a_vector(TreeM, type_index))
      {
         const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
         const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
         info->real_output_nelem = info->output_prec/element_size;
         info->base128_output_nelem = info->output_prec/element_size;
         info->output_prec = element_size;
      }
      else
      {
         info->real_output_nelem = 0;
         info->base128_output_nelem = 0;
      }
      if(current_op == "cond_expr" && max_size_in > 64 && info->node_kind == "VECTOR_BOOL")
         max_size_in = 64;

      if(info->output_prec >= max_size_in)
      {
         info->output_prec = max_size_in;
         info->base128_output_nelem = min_n_elements;
         info->real_output_nelem = min_n_elements;
      }
      else
      {
         max_size_in = info->output_prec;
         min_n_elements = info->base128_output_nelem;
         ///NOT really managed real_output_nelem
      }
   }
   else if(current_op == "lshift_expr")
   {
      unsigned int nodeOutput_id = HLSMgr->get_produced_value(function_id, node);
      THROW_ASSERT(nodeOutput_id, "unexpected condition");
      type_index = tree_helper::get_type_index(TreeM, nodeOutput_id);
      info->output_prec = resize_to_1_8_16_32_64_128_256_512(tree_helper::size(TreeM, nodeOutput_id));
      if(tree_helper::is_a_vector(TreeM, type_index))
      {
         const unsigned int element_type = tree_helper::GetElements(TreeM, type_index);
         const unsigned int element_size = static_cast<unsigned int>(tree_helper::size(TreeM, element_type));
         info->real_output_nelem = info->output_prec/element_size;
         info->base128_output_nelem = info->output_prec/element_size;
         info->output_prec = element_size;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Type is " + STR(type_index) + " " + STR(TreeM->CGetTreeNode(type_index)) + " - Number of output elements (base128): " + STR(info->base128_output_nelem) + " - Number of real output elements: " + STR(info->real_output_nelem) + " - Output precision: " + STR(info->output_prec));
      }
      else
      {
         info->real_output_nelem = 0;
         info->base128_output_nelem = 0;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Output is not a vector");
      }
      if(info->output_prec >= max_size_in)
      {
         max_size_in = info->output_prec;
         min_n_elements = info->base128_output_nelem;
         ///NOT really managed real_output_nelem
      }
      else
      {
         info->output_prec = max_size_in;
         info->base128_output_nelem = min_n_elements;
         info->real_output_nelem = min_n_elements;
      }
   }
   else
   {
      info->output_prec = max_size_in;
      info->base128_output_nelem = min_n_elements;
      info->real_output_nelem = min_n_elements;
   }
   size_t n_inputs = info->input_prec.size();
   for (unsigned int i = 0; i < n_inputs; ++i)
      if(info->input_prec[i] != 0)
         info->input_prec[i] = max_size_in;
   for(std::vector<unsigned int>::iterator n_el_it = info->base128_input_nelem.begin(); n_el_it != n_el_it_end; ++n_el_it)
   {
      if(*n_el_it)
         *n_el_it = min_n_elements;
   }

   /// fix for vec_perm_expr
   if(current_op == "vec_perm_expr")
   {
      if(info->input_prec[2] == 0)
      {
         std::swap(info->input_prec[2], info->input_prec[1]);
         std::swap(info->base128_input_nelem[2], info->base128_input_nelem[1]);
         std::swap(info->real_input_nelem[2], info->real_input_nelem[1]);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Got node type precision of " + GET_NAME(g, node));

}

unsigned int updatecopy_HLS_constraints_functor::operator() (const unsigned int name) const
{
   return tech[name];
}

void updatecopy_HLS_constraints_functor::update(const unsigned int name, int delta)
{
   if (tech[name] == INFINITE_UINT)
      return;
   tech[name] = static_cast<unsigned int>(static_cast<int>(tech[name]) + delta);
}

updatecopy_HLS_constraints_functor::updatecopy_HLS_constraints_functor(const AllocationInformationRef allocation_information) :
   tech(allocation_information->tech_constraints)
{}


unsigned int AllocationInformation::get_prec(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   THROW_ASSERT(precision_map.find(fu_name) != precision_map.end(), "missing the precision of " + STR(fu_name));
   return precision_map.find(fu_name)->second != 0 ? precision_map.find(fu_name)->second : 32;
}

double AllocationInformation::mux_time_unit(unsigned int fu_prec) const
{
   return estimate_muxNto1_delay(fu_prec,2);
}

double AllocationInformation::mux_time_unit_raw(unsigned int fu_prec) const
{
   const technology_managerRef TM = HLS_T->get_technology_manager();
   technology_nodeRef f_unit_mux = TM->get_fu(MUX_GATE_STD + STR("_1_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
   THROW_ASSERT(f_unit_mux, "Library miss component: " + std::string(MUX_GATE_STD)+ STR("_1_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
   functional_unit * fu_br= GetPointer<functional_unit>(f_unit_mux);
   technology_nodeRef op_mux_node =fu_br->get_operation(MUX_GATE_STD);
   operation * op_mux = GetPointer<operation>(op_mux_node);
   double mux_delay = time_m_execution_time(op_mux)-get_setup_hold_time();
   if(mux_delay<=0.0) mux_delay = get_setup_hold_time()/2;
//#define MUX_MARGIN 1.3
//   return mux_delay*ALLOCATION_MUX_MARGIN+get_setup_hold_time();
//   return mux_delay+get_setup_hold_time()*ALLOCATION_MUX_MARGIN;
   return mux_delay;
}

void AllocationInformation::print(std::ostream& os) const
{
   std::vector<technology_nodeRef>::const_iterator fu_end = list_of_FU.end();
   unsigned int index = 0;
   for (std::vector<technology_nodeRef>::const_iterator fu =  list_of_FU.begin(); fu != fu_end; fu++)
   {
      os << index << " ";
      index++;
      (*fu)->print(os);
   }
   if (!node_id_to_fus.empty())
   {
      os << "Op_name relation with functional unit name and operations.\n";
      for(const auto node_id : node_id_to_fus)
      {
         for(const auto fu : node_id.second)
            os << "  [" << STR(node_id.first) << ", <" << list_of_FU[fu]->get_name() << ">]" << std::endl;
      }
   }
}

#ifndef NDEBUG
void AllocationInformation::print_allocated_resources() const
{
   if (debug_level >= DEBUG_LEVEL_VERBOSE)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "\nDumping the list of all the fixed bindings FU <-> node");
      for(const auto bind : binding)
      {
         if(bind.first == ENTRY_ID or bind.first == EXIT_ID)
            continue;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  Vertex " + STR(bind.first));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    Corresponding operation: " + tree_helper::normalized_ID(GetPointer<const gimple_node>(TreeM->get_tree_node_const(bind.first))->operation) + "(" + STR(bind.second.second) + ")");
         functional_unit* fu = dynamic_cast<functional_unit*>(GetPointer<functional_unit>(list_of_FU[bind.second.second]));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    Vertex bound to: " + fu->get_name());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Dumping the list of all the possible bindings FU <-> node");
      for(const auto bind : node_id_to_fus)
      {
         if(bind.first == ENTRY_ID or bind.first == EXIT_ID or bind.first)
            continue;
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  Vertex " + STR(bind.first) + "(" + GetPointer<const gimple_node>(TreeM->get_tree_node_const(bind.first))->operation + ")");
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "    Operation can be implemented by the following FUs:");
         for(const auto fu_id : bind.second)
         {
            functional_unit* fu = dynamic_cast<functional_unit*>(GetPointer<functional_unit>(list_of_FU[fu_id]));
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "      FU name: " + fu->get_name()+ "(" + STR(fu_id) + ")");
         }
      }
   }
}
#endif

technology_nodeRef AllocationInformation::get_fu(const std::string& fu_name, const HLS_managerConstRef HLSMgr)
{
   const HLS_targetRef HLS_T = HLSMgr->get_HLS_target();
   std::string library_name = HLS_T->get_technology_manager()->get_library(fu_name);
   if(library_name == "") return technology_nodeRef();
   return HLS_T->get_technology_manager()->get_fu(fu_name, library_name);
}

unsigned int AllocationInformation::GetCycleLatency(const vertex operation) const
{
   return GetCycleLatency(op_graph->CGetOpNodeInfo(operation)->GetNodeId());
}

unsigned int AllocationInformation::GetCycleLatency(const unsigned int operation) const
{
   if(CanImplementSetNotEmpty(operation))
   {
      const unsigned int actual_latency = get_cycles(GetFuType(operation), operation);
      return actual_latency != 0 ? actual_latency : 1;
   }
   else
   {
      const auto tn = TreeM->get_tree_node_const(operation);
      const auto ga = GetPointer<const gimple_assign>(tn);
      if(ga)
      {
         const auto right = GET_NODE(ga->op1);
         if(right->get_kind() == truth_and_expr_K or right->get_kind() == truth_or_expr_K or right->get_kind() == truth_not_expr_K or
                 right->get_kind() == cond_expr_K or right->get_kind() == ternary_plus_expr_K or right->get_kind() == ternary_mp_expr_K or
                 right->get_kind() == ternary_pm_expr_K or right->get_kind() == ternary_mm_expr_K or right->get_kind() == ssa_name_K or
                 right->get_kind() == integer_cst_K or right->get_kind() == rshift_expr_K  or right->get_kind() == lshift_expr_K or
                 right->get_kind() == plus_expr_K or right->get_kind() == minus_expr_K or right->get_kind() == bit_and_expr_K or
                 right->get_kind() == bit_ior_concat_expr_K)
            return 1;
         else
            THROW_UNREACHABLE("Unsupported right part (" + right->get_kind_text() + ") of gimple assignment " + ga->ToString());
      }
      if(tn->get_kind() == gimple_multi_way_if_K or tn->get_kind() == gimple_cond_K or tn->get_kind() == gimple_phi_K or tn->get_kind() == gimple_nop_K or tn->get_kind() == gimple_return_K )
      {
         return 1;
      }
      THROW_UNREACHABLE("Unsupported gimple node (" + tn->ToString() + "):" + tn->get_kind_text());
   }
   return 0;
}

std::pair<double,double>  AllocationInformation::GetTimeLatency(const vertex operation, const AbsControlStep control_step, const unsigned int functional_unit, const unsigned int stage, const bool ignore_connection) const
{
   return GetTimeLatency(op_graph->CGetOpNodeInfo(operation)->GetNodeId(), control_step, functional_unit, stage, ignore_connection);
}

std::pair<double,double> AllocationInformation::GetTimeLatency(const unsigned int operation_index, const AbsControlStep control_step, const unsigned int functional_unit_type, const unsigned int stage, const bool ignore_connection) const
{
   if(operation_index == ENTRY_ID || operation_index == EXIT_ID)
      return std::pair<double,double> (0.0, 0.0);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing time latency of " + STR(operation_index));
   const double epsilon = 0.000000001;
   ///For the intermediate stage of multicycle the latency is the clock cycle
   const unsigned int num_cycles = GetCycleLatency(operation_index);
   if(stage > 0 and stage < num_cycles - 1)
   {
      const double ret_value = HLS_C->get_clock_period_resource_fraction() * HLS_C->get_clock_period();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(ret_value) + "," + STR(ret_value));
      return std::pair<double,double> (ret_value,ret_value);
   }

   if(CanImplementSetNotEmpty(operation_index))
   {
      unsigned int fu_type;
      if(functional_unit_type  != fu_binding::UNKNOWN)
          fu_type = functional_unit_type;
      else
         fu_type = GetFuType(operation_index);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Functional unit name is " + get_fu_name(fu_type).first);
      double connection_contribute = 0;
      bool estimate_net_logic_delays = parameters->getOption<bool>(OPT_estimate_logic_and_connections);
      if(not ignore_connection and estimate_net_logic_delays)
      {
         bool add_delay1 = false;
         bool add_delay2 = false;
         size_t memory_operations = 0;
         CheckDelays(operation_index, control_step, functional_unit_type, add_delay1, add_delay2, memory_operations);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Delays to be added " + std::string(add_delay1 ? "true" : "false") + " " + std::string(add_delay2 ? "true" : "false"));
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Number of memory operations " + STR(memory_operations));
         connection_contribute = get_connection_time(fu_type, add_delay1, add_delay2, get_n_complex_operations(), memory_operations);
      }
      ///The operation execution  time
      double actual_execution_time = get_execution_time(fu_type, operation_index);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Initial execution time " + STR(actual_execution_time));
      double initial_execution_time = actual_execution_time - get_correction_time(fu_type, GetPointer<const gimple_node>(TreeM->get_tree_node_const(operation_index))->operation);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Initial corrected execution time " + STR(initial_execution_time));
      double op_execution_time = initial_execution_time;
      if(op_execution_time <= 0.0)
         op_execution_time = epsilon;
      /// try to take into account the controller delay
      const auto tn = TreeM->get_tree_node_const(operation_index);
      if(not ignore_connection and estimate_net_logic_delays and (GetPointer<const gimple_multi_way_if>(tn) || GetPointer<const gimple_switch>(tn) || GetPointer<const gimple_cond>(tn)))
      {
         op_execution_time += estimate_controller_delay_fb();
      }

      ///The stage period
      double actual_stage_period;
      actual_stage_period = get_stage_period(fu_type, operation_index);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---actual_stage_period="+ STR(actual_stage_period));
      double initial_stage_period=0.0;
      if(get_initiation_time(fu_type, operation_index)>0)
         initial_stage_period = actual_stage_period-get_correction_time(fu_type, GetPointer<const gimple_node>(TreeM->get_tree_node_const(operation_index))->operation);
      double stage_period = initial_stage_period;

      THROW_ASSERT(get_initiation_time(fu_type, operation_index)==0 || stage_period>0.0, "unexpected condition: " + get_fu_name(fu_type).first + "-" +STR(stage_period));

      if(stage_period>0)
         stage_period += connection_contribute;
      else
         op_execution_time += connection_contribute;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + "," + STR(stage_period));
      return std::pair<double,double>(op_execution_time, stage_period);
   }
   else
   {
      const auto ga = GetPointer<const gimple_assign>(TreeM->get_tree_node_const(operation_index));
      if(ga and (GET_NODE(ga->op1)->get_kind() == ssa_name_K or GET_NODE(ga->op1)->get_kind() == integer_cst_K))
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0,0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == cond_expr_K)
      {
#ifndef NDEBUG
         const auto ce = GetPointer<const cond_expr>(GET_NODE(ga->op1));
         THROW_ASSERT(tree_helper::Size(ce->op0) == 1, "Cond expr not allocated " + ga->op1->ToString());
#endif
         ///Computing time of cond_expr as time of cond_expr_FU - setup_time
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto op_execution_time = mux_time_unit(fu_prec);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is mux time (precision is " + STR(fu_prec) + ") " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time,0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == truth_and_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto fu_name = "truth_and_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu(fu_name, LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit " + fu_name + " not found");
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("truth_and_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time,0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == truth_or_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto fu = "truth_or_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec);;
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu(fu, LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit " + fu + " not found");
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("truth_or_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == truth_not_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("truth_not_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: truth_not_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("truth_not_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == ternary_plus_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("ternary_plus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: ternary_plus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("ternary_plus_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == ternary_mp_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("ternary_mp_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: ternary_mp_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("ternary_mp_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == ternary_pm_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("ternary_pm_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: ternary_pm_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("ternary_pm_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == ternary_mm_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("ternary_mm_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: ternary_mm_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("ternary_mm_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == plus_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("plus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: plus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("plus_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == minus_expr_K)
      {
         const auto data_bitsize = tree_helper::Size(ga->op0);
         const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
         const auto new_stmt_temp = HLS_T->get_technology_manager()->get_fu("minus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
         THROW_ASSERT(new_stmt_temp, "Functional unit not found: minus_expr_FU_" + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
         const auto new_stmt_fu = GetPointer<const functional_unit>(new_stmt_temp);
         const auto new_stmt_op_temp = new_stmt_fu->get_operation("minus_expr");
         const auto new_stmt_op = GetPointer<operation>(new_stmt_op_temp);
         auto op_execution_time = time_m_execution_time(new_stmt_op);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uncorrected execution time is " + STR(op_execution_time));
         op_execution_time = op_execution_time - get_setup_hold_time();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is " + STR(op_execution_time) + ",0.0");
         return std::pair<double,double> (op_execution_time, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == bit_ior_concat_expr_K)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == rshift_expr_K)
      {
#ifndef NDEBUG
         const auto re = GetPointer<const rshift_expr>(GET_NODE(ga->op1));
         THROW_ASSERT(GET_NODE(re->op1)->get_kind() ==integer_cst_K, " unexpected pattern " + ga->op1->ToString());
#endif
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == lshift_expr_K)
      {
#ifndef NDEBUG
         const auto le = GetPointer<const lshift_expr>(GET_NODE(ga->op1));
         THROW_ASSERT(GET_NODE(le->op1)->get_kind() ==integer_cst_K, " unexpected pattern " + ga->op1->ToString());
#endif
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      if(ga and GET_NODE(ga->op1)->get_kind() == bit_and_expr_K)
      {
#ifndef NDEBUG
         const auto re = GetPointer<const bit_and_expr>(GET_NODE(ga->op1));
         THROW_ASSERT(GET_NODE(re->op1)->get_kind() ==integer_cst_K, " unexpected pattern " + ga->op1->ToString());
#endif
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      const auto gmwi = GetPointer<const gimple_multi_way_if>(TreeM->get_tree_node_const(operation_index));
      if(gmwi)
      {
         auto controller_delay = estimate_controller_delay_fb();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is "+ STR(controller_delay) + ",0.0");
         return std::pair<double,double> (controller_delay, 0.0);
      }
      const auto gp = GetPointer<const gimple_phi>(TreeM->get_tree_node_const(operation_index));
      if(gp)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      const auto gn = GetPointer<const gimple_nop>(TreeM->get_tree_node_const(operation_index));
      if(gn)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      const auto gr = GetPointer<const gimple_return>(TreeM->get_tree_node_const(operation_index));
      if(gr)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Time is 0.0,0.0");
         return std::pair<double,double> (0.0, 0.0);
      }
      THROW_UNREACHABLE("Latency of " + TreeM->get_tree_node_const(operation_index)->ToString() + " cannot be computed");
      return std::pair<double,double> (0.0, 0.0);
   }
}

double AllocationInformation::GetPhiConnectionLatency(const unsigned int statement_index) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing phi connection delay of " + STR(statement_index));
   ///Checking for output phi
   const auto phi_in_degree = [&] () -> size_t
   {
      size_t ret_value = 0;
      if(statement_index == ENTRY_ID or statement_index == EXIT_ID)
      {
         return 0;
      }
      const auto tn = TreeM->get_tree_node_const(statement_index);
      if(tn->get_kind() != gimple_assign_K)
         return 0;
      const auto ga = GetPointer<const gimple_assign>(tn);
      if(GET_NODE(ga->op0)->get_kind() != ssa_name_K)
         return 0;
      const auto sn = GetPointer<const ssa_name>(GET_NODE(ga->op0));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Analyzing uses of " + sn->ToString());
      for(const auto use : sn->CGetUseStmts())
      {
         const auto target = GET_NODE(use.first);
         if(target->get_kind() == gimple_phi_K)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Phi: " + target->ToString());
            const auto gp = GetPointer<const gimple_phi>(target);
            std::set<unsigned int> phi_inputs;
            for(const auto def_edge : gp->CGetDefEdgesList())
            {
               const BehavioralHelperConstRef behavioral_helper = hls->FB->CGetBehavioralHelper();
               if(def_edge.first->index && !behavioral_helper->is_a_constant(def_edge.first->index))
                  phi_inputs.insert(def_edge.first->index);
            }
            size_t curr_in_degree = phi_inputs.size();
            if(curr_in_degree>4)
               curr_in_degree = 4;
            ret_value = std::max(ret_value, curr_in_degree);
         }
      }
      return ret_value;
   }();
   if(phi_in_degree == 0)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Delay is 0.0");
      return 0.0;
   }
   const auto statement = TreeM->get_tree_node_const(statement_index);
   THROW_ASSERT(statement->get_kind() == gimple_assign_K, statement->ToString());
   const auto sn = GET_NODE(GetPointer<gimple_assign>(statement)->op0);
   THROW_ASSERT(sn, GetPointer<gimple_assign>(statement)->op0->ToString());
   const auto precision = resize_to_1_8_16_32_64_128_256_512(tree_helper::size(TreeM, sn->index));
   const auto mux_time = estimate_muxNto1_delay(precision, static_cast<unsigned int>(phi_in_degree));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Delay (" + STR(phi_in_degree) + " with " + STR(precision) + " bits) is " + STR(mux_time));
   return mux_time;
}

double AllocationInformation::GetCondExprTimeLatency(const unsigned int operation_index) const
{
   const auto tn = TreeM->get_tree_node_const(operation_index);
   const auto gp = GetPointer<const gimple_phi>(tn);
   THROW_ASSERT(gp, "Tree node is " + tn->ToString());
   ///Computing time of cond_expr as time of cond_expr_FU - setup_time
   ///In this way we are correctly estimating only phi with two inputs
   const auto type = tree_helper::get_type_index(TreeM, gp->index);
   const auto data_bitsize = tree_helper::size(TreeM, type);
   const auto fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
   return mux_time_unit(fu_prec);
}

unsigned int AllocationInformation::GetFuType(const vertex operation) const
{
   return GetFuType(op_graph->CGetOpNodeInfo(operation)->GetNodeId());
}

unsigned int AllocationInformation::GetFuType(const unsigned int operation) const
{
   unsigned int fu_type = 0;
   bool single_fu_type = is_vertex_bounded_with(operation, fu_type);
   if(not single_fu_type)
   {
      const std::set<unsigned int> & fu_set = can_implement_set(operation);
      if(fu_set.size() > 1)
      {
         THROW_UNREACHABLE("Multiple fus not supported");
      }
      else
         return *(fu_set.begin());
   }
   return fu_type;
}

double AllocationInformation::mux_area_unit_raw(unsigned int fu_prec) const
{
   const technology_managerRef TM = HLS_T->get_technology_manager();
   technology_nodeRef f_unit_mux = TM->get_fu(MUX_GATE_STD + STR("_1_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
   THROW_ASSERT(f_unit_mux, "Library miss component: " + std::string(MUX_GATE_STD) + STR("_1_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec));
   functional_unit * fu_mux= GetPointer<functional_unit>(f_unit_mux);
   double area = GetPointer<clb_model>(fu_mux->area_m)->get_resource_value(clb_model::SLICE_LUTS);
   if(area == 0.0)
      area = fu_mux->area_m->get_area_value()-1.0;
   return area;
}

double AllocationInformation::estimate_mux_area(unsigned int fu_name) const
{
  unsigned int fu_prec =  get_prec(fu_name);
  fu_prec = resize_to_1_8_16_32_64_128_256_512(fu_prec);
  return estimate_muxNto1_area(fu_prec,2);
}

double AllocationInformation::estimate_controller_delay_fb() const
{
   return 0.5*EstimateControllerDelay();
}

double AllocationInformation::EstimateControllerDelay() const
{
   if(not parameters->getOption<bool>(OPT_estimate_logic_and_connections))
   {
      return 0.0;
   }
   size_t n_states = boost::num_vertices(*hls->FB->CGetBBGraph(FunctionBehavior::BB));
   double n_states_factor = static_cast<double>(n_states)/40.0;
   if(hls->STG && hls->STG->get_number_of_states())
   {
      n_states = hls->STG->get_number_of_states();
      if(n_states==1)
          return 0.0;
      n_states_factor = static_cast<double>(n_states)/80;
   }
   unsigned int fu_prec =  16;
   const technology_managerRef TM = HLS_T->get_technology_manager();
   technology_nodeRef f_unit = TM->get_fu(MULTIPLIER_STD + std::string("_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_0", LIBRARY_STD_FU);
   THROW_ASSERT(f_unit, "Library miss component: " + std::string(MULTIPLIER_STD)+ std::string("_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_0");
   functional_unit * fu= GetPointer<functional_unit>(f_unit);
   technology_nodeRef op_node =fu->get_operation("mult_expr");
   operation * op = GetPointer<operation>(op_node);
   double delay = time_m_execution_time(op);
   delay = delay*controller_delay_multiplier*1.1*(1-exp(-n_states_factor));
   if(delay<2*get_setup_hold_time())
   {
      delay = 2*get_setup_hold_time();
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Controller delay is " + STR(delay) + " while n_states is " + STR(n_states));
   return delay;
}

double AllocationInformation::get_correction_time(unsigned int fu, const std::string &operation_name) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing correction time of '" + operation_name + "'");
   double res_value = get_setup_hold_time();
   technology_nodeRef current_fu = get_fu(fu);
   std::string memory_type = GetPointer<functional_unit>(current_fu)->memory_type;
   std::string memory_ctrl_type = GetPointer<functional_unit>(current_fu)->memory_ctrl_type;
   unsigned int elmt_bitsize = 0;
   bool is_read_only_correction = false;
   bool is_proxied_correction = false;
   bool is_private_correction = false;

   /// first check for component_timing_alias
   if(GetPointer<functional_unit>(current_fu)->component_timing_alias != "")
   {
      std::string component_name = GetPointer<functional_unit>(current_fu)->component_timing_alias;
      std::string library = HLS_T->get_technology_manager()->get_library(component_name);
      technology_nodeRef f_unit_alias = HLS_T->get_technology_manager()->get_fu(component_name, library);
      THROW_ASSERT(f_unit_alias, "Library miss component: " + component_name);
      functional_unit * fu_alias= GetPointer<functional_unit>(f_unit_alias);
      technology_nodeRef op_alias_node =fu_alias->get_operation(operation_name);
      operation * op_alias = op_alias_node ? GetPointer<operation>(op_alias_node) : GetPointer<operation>(fu_alias->get_operations().front());
      double alias_exec_time = op_alias->time_m->get_initiation_time() != 0u ? time_m_stage_period(op_alias) : time_m_execution_time(op_alias);
      functional_unit * fu_cur= GetPointer<functional_unit>(current_fu);
      technology_nodeRef op_cur_node =fu_cur->get_operation(operation_name);
      operation * op_cur = GetPointer<operation>(op_cur_node);
      double cur_exec_time = op_cur->time_m->get_initiation_time() != 0u ? time_m_stage_period(op_cur) : time_m_execution_time(op_cur);
      res_value += cur_exec_time - alias_exec_time;
   }

   if (memory_type == MEMORY_TYPE_SYNCHRONOUS_UNALIGNED)
   {
      unsigned var = get_memory_var(fu);
      if(!Rmem->is_a_proxied_variable(var))
         is_proxied_correction = true;
      else if(Rmem->is_private_memory(var))
          is_private_correction = true;
      if(Rmem->is_read_only_variable(var))
          is_read_only_correction = true;

      elmt_bitsize = Rmem->get_bram_bitsize();
      unsigned int type_index = tree_helper::get_type_index(TreeM, var);
      if(tree_helper::is_an_array(TreeM, type_index))
      {
         std::vector<unsigned int> dims;
         tree_helper::get_array_dimensions(TreeM, type_index, dims);
         unsigned int n_not_power_of_two = 0;
         for(auto idx : dims)
            if(idx & (idx - 1))
               ++n_not_power_of_two;
         if(dims.size()>1 && n_not_power_of_two>0)
         {
            const technology_managerRef TM = HLS_T->get_technology_manager();
            unsigned int bus_addr_bitsize = resize_to_1_8_16_32_64_128_256_512(Rmem->get_bus_addr_bitsize());
            if(bus_addr_bitsize>32)
               bus_addr_bitsize = 32;
            technology_nodeRef f_unit = TM->get_fu(ADDER_STD + std::string("_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)), LIBRARY_STD_FU);
            THROW_ASSERT(f_unit, "Library miss component: " + std::string(ADDER_STD) + std::string("_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)));
            functional_unit * Fu= GetPointer<functional_unit>(f_unit);
            technology_nodeRef op_node =Fu->get_operation("plus_expr");
            operation * op = GetPointer<operation>(op_node);
            double delay = time_m_execution_time(op)-get_setup_hold_time();
            unsigned int n_levels = 0;
            for(; dims.size() >= (1u << n_levels); ++n_levels);
            res_value -= (n_levels-1)*delay;
         }
      }
   }
   else if (memory_type == MEMORY_TYPE_SYNCHRONOUS_SDS || memory_type == MEMORY_TYPE_SYNCHRONOUS_SDS_BUS || memory_type == MEMORY_TYPE_ASYNCHRONOUS)
   {
      unsigned var = get_memory_var(fu);
      if(!Rmem->is_a_proxied_variable(var))
         is_proxied_correction = true;
      if(Rmem->is_read_only_variable(var))
          is_read_only_correction = true;

      elmt_bitsize=1;
      unsigned int type_index = tree_helper::get_type_index(TreeM, var);
      tree_nodeRef type_node = TreeM->get_tree_node_const(type_index);
      tree_helper::accessed_greatest_bitsize(TreeM, type_node, type_index, elmt_bitsize);
      if(tree_helper::is_an_array(TreeM, type_index))
      {
         std::vector<unsigned int> dims;
         tree_helper::get_array_dimensions(TreeM, type_index, dims);
         unsigned int n_not_power_of_two = 0;
         for(auto idx : dims)
            if(idx & (idx - 1))
               ++n_not_power_of_two;
         if((dims.size()>1 && n_not_power_of_two>0))
         {
            const technology_managerRef TM = HLS_T->get_technology_manager();
            unsigned int bus_addr_bitsize = resize_to_1_8_16_32_64_128_256_512(Rmem->get_bus_addr_bitsize());
            if(bus_addr_bitsize>32)
               bus_addr_bitsize = 32;
            technology_nodeRef f_unit = TM->get_fu(ADDER_STD + std::string("_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)), LIBRARY_STD_FU);
            THROW_ASSERT(f_unit, "Library miss component: " + std::string(ADDER_STD) + std::string("_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)));
            functional_unit * Fu= GetPointer<functional_unit>(f_unit);
            technology_nodeRef op_node =Fu->get_operation("plus_expr");
            operation * op = GetPointer<operation>(op_node);
            double delay = time_m_execution_time(op)-get_setup_hold_time();
            unsigned int n_levels = 0;
            for(; dims.size() >= (1u << n_levels); ++n_levels);
            res_value -= (n_levels-1)*delay;
         }
      }
   }
   else if (memory_ctrl_type == MEMORY_CTRL_TYPE_PROXY || memory_ctrl_type == MEMORY_CTRL_TYPE_PROXYN || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXY || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXYN)
   {
      unsigned var = proxy_memory_units.find(fu)->second;
      if(Rmem->is_read_only_variable(var))
          is_read_only_correction = true;
      functional_unit * fu_cur= GetPointer<functional_unit>(current_fu);
      technology_nodeRef op_cur_node =fu_cur->get_operation(operation_name);
      std::string latency_postfix = "";
      if(fu_cur->bram_load_latency == "3")
         latency_postfix = std::string("_3");
      else if(fu_cur->bram_load_latency != "")
         THROW_ASSERT(fu_cur->bram_load_latency == "2", "unexpected condition for bram latency");

      operation * op_cur = GetPointer<operation>(op_cur_node);
      double cur_exec_time = op_cur->time_m->get_initiation_time() != 0u ? time_m_stage_period(op_cur) : time_m_execution_time(op_cur);
      double cur_exec_delta;
      technology_nodeRef f_unit_sds;
      bool unaligned_access_p = parameters->isOption(OPT_unaligned_access) && parameters->getOption<bool>(OPT_unaligned_access);
      if(Rmem->is_sds_var(var))
      {
         if(memory_ctrl_type == MEMORY_CTRL_TYPE_PROXY || memory_ctrl_type == MEMORY_CTRL_TYPE_DPROXY)
         {
            if(Rmem->is_private_memory(var))
            {
               if(parameters->getOption<bool>(OPT_use_asynchronous_memories) &&
                     AllocationInformation::is_asynchronous_ram(TreeM, var, parameters->getOption<unsigned int>(OPT_distram_threshold)))
                  f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_DISTRAM_SDS, LIBRARY_STD_FU);
               else
                  f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_SDS+latency_postfix, LIBRARY_STD_FU);
            }
            else
               f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_SDS_BUS+latency_postfix, LIBRARY_STD_FU);
         }
         else
         {
            if(Rmem->is_private_memory(var))
            {
               if(parameters->getOption<bool>(OPT_use_asynchronous_memories) &&
                     AllocationInformation::is_asynchronous_ram(TreeM, var, parameters->getOption<unsigned int>(OPT_distram_threshold)))
                  f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_DISTRAM_NN_SDS, LIBRARY_STD_FU);
               else
                  f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_NN_SDS+latency_postfix, LIBRARY_STD_FU);
            }
            else
               f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_NN_SDS_BUS+latency_postfix, LIBRARY_STD_FU);
         }

         elmt_bitsize=1;
         unsigned int type_index = tree_helper::get_type_index(TreeM, var);
         tree_nodeRef type_node = TreeM->get_tree_node_const(type_index);
         tree_helper::accessed_greatest_bitsize(TreeM, type_node, type_index, elmt_bitsize);
      }
      else if(!unaligned_access_p && Rmem->get_bram_bitsize()==8 && Rmem->get_bus_data_bitsize() == 8)
      {
         f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_NN_SDS_BUS+latency_postfix, LIBRARY_STD_FU);
         if(Rmem->is_private_memory(var))
            is_private_correction = true;
         elmt_bitsize = Rmem->get_bram_bitsize();
      }
      else
      {
         f_unit_sds = HLS_T->get_technology_manager()->get_fu(ARRAY_1D_STD_BRAM_NN+latency_postfix, LIBRARY_STD_FU);
         if(Rmem->is_private_memory(var))
            is_private_correction = true;
         elmt_bitsize = Rmem->get_bram_bitsize();
      }
      THROW_ASSERT(f_unit_sds, "Library miss component");
      functional_unit * fu_sds= GetPointer<functional_unit>(f_unit_sds);
      technology_nodeRef op_sds_node =fu_sds->get_operation(operation_name);
      operation * op_sds = GetPointer<operation>(op_sds_node);
      double cur_sds_exec_time = op_sds->time_m->get_initiation_time() != 0u ? time_m_stage_period(op_sds) : time_m_execution_time(op_sds);
      cur_exec_delta = cur_exec_time - cur_sds_exec_time;
      res_value = res_value + cur_exec_delta;

      unsigned int type_index = tree_helper::get_type_index(TreeM, var);
      if(tree_helper::is_an_array(TreeM, type_index))
      {
         std::vector<unsigned int> dims;
         tree_helper::get_array_dimensions(TreeM, type_index, dims);
         unsigned int n_not_power_of_two = 0;
         for(auto idx : dims)
            if(idx & (idx - 1))
               ++n_not_power_of_two;
         if(dims.size()>1 && n_not_power_of_two>0)
         {
            const technology_managerRef TM = HLS_T->get_technology_manager();
            unsigned int bus_addr_bitsize = resize_to_1_8_16_32_64_128_256_512(Rmem->get_bus_addr_bitsize());
            if(bus_addr_bitsize>32)
               bus_addr_bitsize = 32;
            technology_nodeRef f_unit = TM->get_fu(ADDER_STD + std::string("_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)+"_"+STR(bus_addr_bitsize)), LIBRARY_STD_FU);
            functional_unit * Fu= GetPointer<functional_unit>(f_unit);
            technology_nodeRef op_node =Fu->get_operation("plus_expr");
            operation * op = GetPointer<operation>(op_node);
            double delay = time_m_execution_time(op)-get_setup_hold_time();
            unsigned int n_levels = 0;
            for(; dims.size() >= (1u << n_levels); ++n_levels);
            res_value -= (n_levels-1)*delay;
         }
      }
   }
   else if(memory_ctrl_type == MEMORY_CTRL_TYPE_D00)
   {
      elmt_bitsize = Rmem->get_bram_bitsize();
   }
   else if(is_single_bool_test_cond_expr_units(fu))
   {
      unsigned int prec = get_prec(fu);
      unsigned int fu_prec = resize_to_1_8_16_32_64_128_256_512(prec);
      double true_delay = mux_time_unit(fu_prec);
      const technology_managerRef TM = HLS_T->get_technology_manager();
      if(fu_prec>1)
      {
         technology_nodeRef f_unit_ce = TM->get_fu(get_fu_name(fu).first, LIBRARY_STD_FU);
         functional_unit * fu_ce= GetPointer<functional_unit>(f_unit_ce);
         technology_nodeRef op_ce_node =fu_ce->get_operation("cond_expr");
         operation * op_ce = GetPointer<operation>(op_ce_node);
         double setup_time=get_setup_hold_time();
         double ce_delay = time_m_execution_time(op_ce)-setup_time;
         double correction = ce_delay-true_delay;
         res_value = res_value + correction;
      }
   }

   double bus_multiplier = 0;
   if(elmt_bitsize==128)
      bus_multiplier = -1.0;
   else if(elmt_bitsize==64)
      bus_multiplier = -0.5;
   else if(elmt_bitsize==32)
      bus_multiplier = 0;
   else if(elmt_bitsize==16)
      bus_multiplier = +0;
   else if(elmt_bitsize==8)
      bus_multiplier = +0;
   res_value = res_value + bus_multiplier*(get_setup_hold_time()/time_multiplier);
   if(is_read_only_correction)
       res_value = res_value + memory_correction_coefficient*0.5*(get_setup_hold_time()/time_multiplier);
   if(is_proxied_correction)
       res_value = res_value + memory_correction_coefficient*(estimate_mux_time(fu)/(mux_time_multiplier * time_multiplier));
   if(is_private_correction)
       res_value = res_value + memory_correction_coefficient*(estimate_mux_time(fu)/(mux_time_multiplier * time_multiplier));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Correction is " + STR(res_value));
   return res_value;
}

double AllocationInformation::estimate_call_delay() const
{
   double clock_budget = HLS_C->get_clock_period_resource_fraction()*HLS_C->get_clock_period();
   double scheduling_mux_margins = parameters->getOption<double>(OPT_scheduling_mux_margins) * mux_time_unit(32);
   double call_delay = clock_budget;
   call_delay -= minimumSlack > 0.0 && minimumSlack != std::numeric_limits<double>::max() ?
                   minimumSlack:
                   0;
   if(call_delay<0.0)
      call_delay = get_setup_hold_time();
   else
      call_delay += get_setup_hold_time();
   auto ctrl_delay = EstimateControllerDelay();
   if(call_delay<ctrl_delay)
      call_delay = ctrl_delay;
   if(call_delay >= clock_budget-scheduling_mux_margins)
      call_delay = clock_budget-scheduling_mux_margins;
   //std::cerr << "estimate call_delay=" << call_delay << " clock_budget="<< clock_budget << "-" << 3*mux_time_unit(32) << "-" << get_setup_hold_time() << std::endl;
   return call_delay;
}

double AllocationInformation::get_connection_time(unsigned /*fu_type*/, bool add_delay1, bool add_delay2, size_t n_complex_ops, size_t n_mem_ops) const
{
   double design_complexity = 1.7*(1-exp(static_cast<double>(n_complex_ops)/-70.0));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Design complexity " + STR(design_complexity));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---add_delay1 " + STR(add_delay1));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---add_delay2 " + STR(add_delay2));
   double load_store_complexity = 1.1*(1-exp(static_cast<double>(n_mem_ops)/-20.0));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Load/Store complexity " + STR(load_store_complexity));
   double delay1_contribute = add_delay1 ? get_setup_hold_time() : 0.0;
   double delay2_contribute = add_delay2 ? get_setup_hold_time() : 0.0;
   double delay_contribute=0;
   if(add_delay1 && add_delay2)
      delay_contribute = (design_complexity+load_store_complexity)*(delay1_contribute+delay2_contribute);
   else if(add_delay1)
      delay_contribute = (design_complexity+load_store_complexity)*delay1_contribute;
   else if(add_delay2)
      delay_contribute = (design_complexity+load_store_complexity)*delay2_contribute;
   double op_connection_offset = 0.0;
   if(delay_contribute==0)
   {
         op_connection_offset = connection_offset;
   }
   return connection_time_ratio*(delay_contribute + op_connection_offset/*+ asynch_memory_contribute*/);
}

double AllocationInformation::compute_normalized_area(unsigned int fu_s1) const
{
   double mux_area = estimate_mux_area(fu_s1);
   double resource_area = is_single_bool_test_cond_expr_units(fu_s1) ? (mux_area>1 ? (mux_area-1) : 0) : get_area(fu_s1);
   if(resource_area > mux_area && resource_area-mux_area < 4)
      resource_area = mux_area;
   return (resource_area/mux_area)+get_DSPs(fu_s1);
}

unsigned int AllocationInformation::get_n_complex_operations() const
{
   return n_complex_operations;
}

std::string AllocationInformation::extract_bambu_provided_name(unsigned int prec_in, unsigned int prec_out, const HLS_managerConstRef HLSMgr, technology_nodeRef &current_fu)
{
   std::string unit_name;
   if(prec_in == 32 && prec_out == 64)
      unit_name = SF_FFDATA_CONVERTER_32_64_STD;
   else if(prec_in == 64 && prec_out == 32)
      unit_name = SF_FFDATA_CONVERTER_64_32_STD;
   else
      THROW_ERROR("not supported float to float conversion: " + STR(prec_in) + " " + STR(prec_out));
   current_fu = get_fu(unit_name, HLSMgr);
   return unit_name;
}

bool AllocationInformation::has_constant_in(unsigned int fu_name) const
{
   if(!has_to_be_synthetized(fu_name))
      return false;
   return GetPointer<functional_unit>(list_of_FU[fu_name])->characterizing_constant_value != "";
}

bool AllocationInformation::is_proxy_memory_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return proxy_memory_units.find(fu_name) != proxy_memory_units.end();
}

bool AllocationInformation::is_readonly_memory_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return (is_memory_unit(fu_name) && Rmem->is_read_only_variable(get_memory_var(fu_name)));
}

bool AllocationInformation::is_single_bool_test_cond_expr_units(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return single_bool_test_cond_expr_units.find(fu_name) != single_bool_test_cond_expr_units.end();
}

unsigned int AllocationInformation::get_worst_number_of_cycles(const unsigned int fu_name) const
{
   if(!has_to_be_synthetized(fu_name))
      return 0;
   const functional_unit::operation_vec node_ops = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operations();
   unsigned int max_value = 0;
   functional_unit::operation_vec::const_iterator no_it_end = node_ops.end();
   for(functional_unit::operation_vec::const_iterator no_it = node_ops.begin(); no_it != no_it_end; ++no_it) max_value = std::max(max_value, GetPointer<operation>(*no_it)->time_m->get_cycles());
   return max_value;
}

double AllocationInformation::GetClockPeriodMargin() const
{
   auto clock_period = HLS_C->get_clock_period();
   auto clock_period_resource_fraction = HLS_C->get_clock_period_resource_fraction();
   auto scheduling_mux_margins = parameters->getOption<double>(OPT_scheduling_mux_margins) * mux_time_unit(32);
   auto setup_hold_time = get_setup_hold_time();

   return clock_period - ((clock_period * clock_period_resource_fraction) - scheduling_mux_margins - setup_hold_time);
}
/// - add_delay1 is true when a load is connected to the inputs of another load or of a
///      single cycle functional unit having DSPs (delay added checking inputs of current_v).
/// - add_delay1 is true when the current vertex current_v is a load or a functional units having DSPs
///      taking less than 3 cycles is chained with more than 5 functional units having DSPs and
///      taking a single cycle (delay added when the DSPs or LOADs are connetted to
///      other DSPs with a high fanout).
///      This analysis try to take into account even if DSPs are pipelined or not.
/// - add_delay2 is true when a functional units having DSPs taking less than 3 cycles is
///      chained with a single cycle functional units having DSPs or with a load
///      (delay added checking inputs of current_v).
/// - add_delay2 is true when the current vertex current_v is a load or a functional unit
///      having DSPs taking less than 3 cycles and its outputs are chained with functional
///      units having DSPs and taking a single cycle (delay added when LOADs
///      or DSPs are chained with other DSPs).
///      This analysis try to take into account even if DSPs are pipelined or not.
///
#define COMPLEX_FIRST_THRESHOLD 1
#define COMPLEX_SECOND_THRESHOLD 5
#define COMPLEX_THIRD_THRESHOLD 5
void AllocationInformation::CheckDelays(const unsigned int statement_index, const AbsControlStep schedule_step, const unsigned int functional_unit, bool & add_delay1, bool & add_delay2, size_t & memory_operations) const
{
   bool is_DSP = false;
   add_delay1 = false;
   add_delay2 = false;
#ifndef NDEBUG
   const auto function_debug_level = GET_FUNCTION_DEBUG_LEVEL();
#endif
   if(statement_index == ENTRY_ID or statement_index == EXIT_ID)
      return;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "-->Checking if delays have to be added to execution time of " + TreeM->get_tree_node_const(statement_index)->ToString());
   unsigned int DSP_n_stages = 0;
   const auto schedule = hls->Rsch;
   std::vector<unsigned int> vars;
   std::map<std::pair<unsigned int, AbsControlStep>, bool> is_complex;
   bool curr_is_a_load_store = GetPointer<const gimple_node>(TreeM->get_tree_node_const(statement_index))->operation == LOAD || GetPointer<const gimple_node>(TreeM->get_tree_node_const(statement_index))->operation == STORE;

#if 0
   if(starting_cs!=static_cast<unsigned int>(-1))
      cs = starting_cs;
#endif
   if(functional_unit != fu_binding::UNKNOWN)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Assigned to " + get_string_name(functional_unit));
      is_DSP = get_DSPs(functional_unit) > 0;
      DSP_n_stages = get_worst_number_of_cycles(functional_unit);
   }
   else if(CanImplementSetNotEmpty(statement_index))
   {
      for(const auto candidate_functional_unit : can_implement_set(statement_index))
      {
         if(get_DSPs(candidate_functional_unit))
         {
            is_DSP = get_DSPs(candidate_functional_unit) > 0;
            DSP_n_stages = get_worst_number_of_cycles(candidate_functional_unit);
         }
      }
   }
   else
   {
      ///Operation not allocated; it has been created by frontend, so it is not a DSP operation
   }
   std::set<unsigned int> complex_vertices;
   const auto tn = TreeM->get_tree_node_const(statement_index);
   if(tn->get_kind() != gimple_assign_K)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "<--Skipped because it is not a gimple assignment");
      return;
   }
   const auto ga = GetPointer<const gimple_assign>(tn);
   const auto rhs_ssa_uses = tree_helper::ComputeSsaUses(ga->op1);
   TreeNodeConstSet zero_chain_operations;
   for(const auto ssa : rhs_ssa_uses)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "-->Considering used ssa " + ssa.first->ToString());
      const auto def = [&]() -> tree_nodeConstRef
      {
         auto temp_def = GET_NODE(GetPointer<const ssa_name>(GET_NODE(ssa.first))->CGetDefStmt());
         zero_chain_operations.insert(temp_def);
         while(true)
         {
            ///If it is not an assignment
            if(temp_def->get_kind() != gimple_assign_K)
            {
               break;
            }
            const double current_delay = GetTimeLatency(temp_def->index, schedule_step, CanImplementSetNotEmpty(temp_def->index) ? GetFuType(temp_def->index) : fu_binding::UNKNOWN, 0, true).first;
            if(current_delay >= 0.01)
            {
               break;
            }
            ///If it uses multiple data
            std::set<ssa_name*> current_ssa_uses;
            const auto current_ga = GetPointer<const gimple_assign>(temp_def);
            tree_helper::compute_ssa_uses_rec_ptr(current_ga->op1, current_ssa_uses);
            if(current_ssa_uses.size() != 1)
            {
               break;
            }
            temp_def = GET_NODE((*(current_ssa_uses.begin()))->CGetDefStmt());
            zero_chain_operations.insert(temp_def);
         }
         return temp_def;
      }();

      if(def->index == statement_index)
         continue;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Predecessor to be considered " + def->ToString());
      if(CanImplementSetNotEmpty(def->index) and (not schedule or not schedule->is_scheduled(def->index) or schedule_step == schedule->get_cstep_end(def->index) or (schedule_step.second == AbsControlStep::UNKNOWN and schedule_step.first == schedule->get_cstep_end(def->index).first)))
      {
         unsigned int src_fu_type = hls->Rfu->is_assigned(def->index) ? hls->Rfu->get_assign(def->index) : GetFuType(def->index);
         if((get_DSPs(src_fu_type)>0 and get_worst_number_of_cycles(src_fu_type)<3) && ((is_DSP && DSP_n_stages<=1)||curr_is_a_load_store))
         {
            add_delay2 = true;
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Chain between DSP and DSP/LOAD");
         }

         bool is_a_LOAD = GetPointer<const gimple_node>(TreeM->get_tree_node_const(def->index))->operation == LOAD;
         add_delay1 |= is_a_LOAD && ((is_DSP && DSP_n_stages<=1)||curr_is_a_load_store)/*||(op_flow_graph->CGetOpNodeInfo(src)->GetOperation() == "lshift_expr" && !allocation_information->has_constant_in(src_fu_type))*/;
         if(is_a_LOAD and is_memory_unit(src_fu_type))
         {
            vars.push_back(get_memory_var(src_fu_type));
         }
         else if(is_a_LOAD and is_proxy_memory_unit(src_fu_type))
         {
            vars.push_back(get_proxy_memory_var(src_fu_type));
         }
      }
      if((is_DSP && DSP_n_stages<3)||curr_is_a_load_store)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Current is a load or a dsp operation");
         for(const auto zero_chain_operation : zero_chain_operations)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Definition is " + zero_chain_operation->ToString());
            const auto current_ga = GetPointer<const gimple_assign>(zero_chain_operation);
            if(current_ga and GET_NODE(current_ga->op0)->get_kind() == ssa_name_K)
            {
               for(const auto statement : GetPointer<const ssa_name>(GET_NODE(current_ga->op0))->CGetUseStmts())
               {
                  const auto tgt = statement.first;
                  const auto tgt_gn = GetPointer<const gimple_node>(TreeM->get_tree_node_const(tgt->index));
                  if(tgt->index != statement_index and tgt_gn->get_kind() == gimple_assign_K)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used in " + tgt->ToString() + " - Type is " + tgt_gn->operation);
                     if(is_complex.find(std::make_pair(tgt->index, schedule_step)) == is_complex.end())
                     {
                        if(schedule and schedule->is_scheduled(tgt->index) && schedule_step.second != AbsControlStep::UNKNOWN)
                        {
                           if(check_output_DSP_chaining(tgt->index, schedule_step))
                           {
                              complex_vertices.insert(tgt->index);
                              is_complex[std::make_pair(tgt->index, schedule_step)] = true;
                           }
                           else
                              is_complex[std::make_pair(tgt->index, schedule_step)] = false;
                        }
                        else if(CanImplementSetNotEmpty(tgt->index))
                        {
                            is_complex[std::make_pair(tgt->index, schedule_step)] = false;
                            for(const auto candidate_functional_unit : can_implement_set(statement_index))
                            {
                               if(get_DSPs(candidate_functional_unit) && get_worst_number_of_cycles(candidate_functional_unit)<=1)
                               {
                                  complex_vertices.insert(tgt->index);
                                  is_complex[std::make_pair(tgt->index, schedule_step)] = true;
                                  break;
                               }
                            }
                        }
                        else
                            is_complex[std::make_pair(tgt->index, schedule_step)] = false;
                     }
                     else if(is_complex.find(std::make_pair(tgt->index, schedule_step))->second)
                        complex_vertices.insert(tgt->index);
                  }
               }
            }
         }
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "<--Considered used ssa " + ssa.first->ToString());
   }

   size_t load_fanout_size = 0;

   ///Checking fanout of a load
   if(GetPointer<const gimple_node>(TreeM->get_tree_node_const(statement_index))->operation == LOAD)
   {
      bool added_register = false;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "-->Checking fanout of load " + STR(ga));
      TreeNodeConstSet already_analyzed_ssas, to_be_analyzed;
      const auto load_sn = GetPointer<const ssa_name>(GET_CONST_NODE(ga->op0));
      if(load_sn)
      {
         to_be_analyzed.insert(GET_CONST_NODE(ga->op0));
      }
      while(to_be_analyzed.size())
      {
         const auto temp = *(to_be_analyzed.begin());
         already_analyzed_ssas.insert(temp);
         to_be_analyzed.erase(to_be_analyzed.begin());
         const auto sn = GetPointer<const ssa_name>(temp);
         if(sn->virtual_flag)
         {
            continue;
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "-->Analyzing uses of " + STR(sn));
         for(const auto use_stmt : sn->CGetUseStmts())
         {
            const auto next_ga = GetPointer<const gimple_assign>(GET_NODE(use_stmt.first));
            if(not next_ga)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used " + STR(use_stmt.second) + " in " + STR(use_stmt.first));
               load_fanout_size += use_stmt.second;
               continue;
            }
            ///Note that we do not use schedule_step since we need end step of current operation
            if(schedule and schedule->is_scheduled(next_ga->index) and schedule->is_scheduled(statement_index) and schedule->get_cstep_end(statement_index) != schedule->get_cstep(next_ga->index))
            {
               if(not added_register)
               {
                  added_register = true;
                  load_fanout_size++;
               }
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used " + STR(use_stmt.second) + " in " + STR(use_stmt.first) + "(register)");
               continue;
            }
            const auto next_sn = GetPointer<const ssa_name>(GET_CONST_NODE(next_ga->op0));
            if(not next_sn)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used " + STR(use_stmt.second) + " in " + STR(use_stmt.first));
               load_fanout_size += use_stmt.second;
               continue;
            }
            const double current_delay = GetTimeLatency(next_ga->index, schedule_step, CanImplementSetNotEmpty(next_ga->index) ? GetFuType(next_ga->index) : fu_binding::UNKNOWN, 0, true).first;
            if(current_delay >= 0.01)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used " + STR(use_stmt.second) + " in " + STR(use_stmt.first));
               load_fanout_size += use_stmt.second;
               continue;
            }
            ///If it uses multiple data
            std::set<ssa_name*> current_ssa_uses;
            tree_helper::compute_ssa_uses_rec_ptr(next_ga->op1, current_ssa_uses);
            if(current_ssa_uses.size() != 1)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used " + STR(use_stmt.second) + " in " + STR(use_stmt.first));
               load_fanout_size += use_stmt.second;
               continue;
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "---Used in zero delay operation " + STR(use_stmt.first));
            if(already_analyzed_ssas.find(GET_CONST_NODE(next_ga->op0)) == already_analyzed_ssas.end())
            {
               to_be_analyzed.insert(GET_CONST_NODE(next_ga->op0));
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "<--Fanout is " + STR(load_fanout_size));
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "<--");
   }

   ///Checking ternary or binary with DSPs as inputs
   size_t dsp_inputs = 0;

   //if(GetPointer<binary_expr>(GET_NODE(ga->op1)) || GetPointer<ternary_expr>(GET_NODE(ga->op1)))
   /// done for any operation
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Checking inputs for DSPs " + ga->ToString());
      for(const auto ssa : rhs_ssa_uses)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Checking used " + ssa.first->ToString());
         unsigned int n_iterations = 0;
         const auto def = [&]() -> tree_nodeConstRef
         {
            auto temp_def = GET_NODE(GetPointer<const ssa_name>(GET_NODE(ssa.first))->CGetDefStmt());
            while(true)
            {
               ///If it is not an assignment
               if(temp_def->get_kind() != gimple_assign_K)
               {
                  break;
               }
               const double current_delay = GetTimeLatency(temp_def->index, schedule_step, CanImplementSetNotEmpty(temp_def->index) ? GetFuType(temp_def->index) : fu_binding::UNKNOWN, 0, true).first;
               if(current_delay >= 0.01)
               {
                  break;
               }
               ///If it uses multiple data
               std::set<ssa_name*> current_ssa_uses;
               const auto current_ga = GetPointer<const gimple_assign>(temp_def);
               tree_helper::compute_ssa_uses_rec_ptr(current_ga->op1, current_ssa_uses);
               if(current_ssa_uses.size() != 1)
               {
                  break;
               }
               n_iterations++;
               temp_def = GET_NODE((*(current_ssa_uses.begin()))->CGetDefStmt());
            }
            return temp_def;
         }();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Definition is " + def->ToString());
         if(schedule and schedule->is_scheduled(def->index) and schedule_step.second != AbsControlStep::UNKNOWN and schedule_step != schedule->get_cstep_end(def->index))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Checked used " + ssa.first->ToString());
            continue;
         }
         if(hls->Rfu->is_assigned(def->index))
         {
            unsigned int candidate_functional_unit = hls->Rfu->get_assign(def->index);
            if(get_DSPs(candidate_functional_unit) > 0)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Assigned to dsp");
               dsp_inputs++;
               if(n_iterations==0 && get_cycles(candidate_functional_unit, def->index)<3)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Assigned to dsp with less than 3 cycles");
                  add_delay1 = true;
                  //add_delay2 = true;
                  //memory_operations += 2;///This is done to increase the operation delay
               }
            }
            else
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Not assigned to dsp");
            }
         }
         else if(CanImplementSetNotEmpty(def->index))
         {
            for(const auto candidate_functional_unit : can_implement_set(def->index))
            {
               if(get_DSPs(candidate_functional_unit))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Assigned to dsp");
                  dsp_inputs++;
                  if(n_iterations==0 && get_cycles(candidate_functional_unit, def->index)<3)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Assigned to dsp with less than 3 cycles");
                     add_delay1 = true;
                     //add_delay2 = true;
                     //memory_operations += 2;///This is done to increase the operation delay
                  }
                  break;
               }
            }
         }
         else
         {
            ///Operation not allocated; it has been created by frontend, so it is not a DSP operation
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Checked used " + ssa.first->ToString());
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Checked inputs of ternary/binary expr " + ga->ToString());
   }

   for(const auto var : vars)
   {
      memory_operations += HLS_mgr->Rmem->get_n_mem_operations(var);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Number of complex vertices " + STR(complex_vertices.size()));
   if(complex_vertices.size() >= COMPLEX_FIRST_THRESHOLD)
      add_delay2 = true;
   if(complex_vertices.size()>COMPLEX_SECOND_THRESHOLD)
      add_delay1 = true;
   if(load_fanout_size >= COMPLEX_THIRD_THRESHOLD)
      add_delay2 = true;
   if(dsp_inputs >= 2)
      add_delay2 = true;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, function_debug_level, "<--Checked delays of " + TreeM->get_tree_node_const(statement_index)->ToString() + ": " + std::string(add_delay1 ? "true" : "false") + " " + std::string(add_delay2 ? "true" : "false"));
}

#define FIRST_LEVEL 1
#define MORE_LEVELS 0
bool AllocationInformation::check_output_DSP_chaining(const unsigned int statement_index, const AbsControlStep cs) const
{
   if(cs != hls->Rsch->get_cstep(statement_index))
      return false;
   if(not hls->Rfu->is_assigned(statement_index))
      return false;
   unsigned int cur_fu_type = hls->Rfu->get_assign(statement_index);
   unsigned int n_cycles = get_worst_number_of_cycles(cur_fu_type);
   if(get_DSPs(cur_fu_type)>0 and n_cycles<=1)
      return true;
   const auto tn = TreeM->get_tree_node_const(statement_index);
   if(tn->get_kind() != gimple_assign_K)
   {
      return false;
   }
   const auto ga = GetPointer<const gimple_assign>(tn);
   std::set<ssa_name*> rhs_ssa_uses;
   tree_helper::compute_ssa_uses_rec_ptr(ga->op1, rhs_ssa_uses);
   if(n_cycles != 0 or rhs_ssa_uses.size() > 1 or ((get_execution_time(cur_fu_type, statement_index) - get_correction_time(cur_fu_type, GetPointer<const gimple_node>(TreeM->get_tree_node_const(statement_index))->operation))>0.01))
      return false;
#if FIRST_LEVEL
   /// compute the set of operations on the control step frontier
   std::queue<unsigned int> fifo;
#if MORE_LEVELS
   std::unordered_set<unsigned int> visited;
#endif
   if(GET_NODE(ga->op0)->get_kind() != ssa_name_K)
      return false;
   for(const auto use_stmt : GetPointer<const ssa_name>(GET_NODE(ga->op0))->CGetUseStmts())
   {
      fifo.push(use_stmt.first->index);
   }
   while(!fifo.empty())
   {
      const auto current_op = fifo.front();
      fifo.pop();
      if(
#if MORE_LEVELS
            visited.find(current_op) == visited.end() &&
#endif
            hls->Rsch->is_scheduled(current_op) && cs == hls->Rsch->get_cstep(current_op))
      {
#if MORE_LEVELS
         visited.insert(current_op);
#endif
         const auto dsp = [&] () -> bool
         {
            if(not hls->Rfu->is_assigned(current_op))
               return false;
            cur_fu_type = hls->Rfu->get_assign(current_op);
            if(get_DSPs(cur_fu_type)>0 and get_worst_number_of_cycles(cur_fu_type)<=1)
               return true;
            else
               return false;
         }();
         if(dsp)
            return true;
         else
         {
            if(cs == hls->Rsch->get_cstep_end(current_op))
            {
#if MORE_LEVELS
               for(boost::tie(eo, eo_end) = boost::out_edges(current_op, *op_flow_graph); eo != eo_end; eo++)
                  fifo.push(boost::target(*eo, *op_flow_graph));
#endif
            }
         }
      }
   }
#endif
   return false;
}

double AllocationInformation::GetConnectionTime(const vertex first_operation, const vertex second_operation, const AbsControlStep cs) const
{
   const auto first_operation_index = op_graph->CGetOpNodeInfo(first_operation)->GetNodeId();
   const auto second_operation_index = second_operation ? op_graph->CGetOpNodeInfo(second_operation)->GetNodeId() : 0;
   return GetConnectionTime(first_operation_index, second_operation_index, cs);
}

double AllocationInformation::GetConnectionTime(const unsigned int first_operation, const unsigned int second_operation, const AbsControlStep cs) const
{
   if(second_operation == 0)
   {
      return GetPhiConnectionLatency(first_operation);
   }
   else
   {
      ///cond_expr additional delay
      double fo_correction = 0;
      const auto second_operation_tn = TreeM->CGetTreeNode(second_operation);
      bool is_cond_expr = second_operation_tn->get_kind() == gimple_assign_K and GET_NODE(GetPointer<const gimple_assign>(second_operation_tn)->op1)->get_kind() == cond_expr_K;
      tree_nodeRef cond_def;

      if(is_cond_expr)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
         size_t cond_expr_var_root_uses = 0;
         unsigned int var_read = GetPointer<const cond_expr>(GET_NODE(GetPointer<const gimple_assign>(TreeM->CGetTreeNode(second_operation))->op1))->op0->index;
         if(hls->FB->CGetBehavioralHelper()->is_a_constant(var_read))
            is_cond_expr = false;
         else
         {
            const auto schedule = hls->Rsch;
            const auto sn = GetPointer<const ssa_name>(TreeM->CGetTreeNode(var_read));
            cond_def = sn->CGetDefStmt();
            if(((schedule->is_scheduled(cond_def->index) and schedule->get_cstep(cond_def->index) == cs) or (cs.second == AbsControlStep::UNKNOWN and cs.first == GetPointer<const gimple_node>(GET_NODE(cond_def))->bb_index)))
            {
               const auto cond_def_uses = tree_helper::ComputeSsaUses(cond_def);
               for(const auto cond_def_use : cond_def_uses)
               {
                   const auto sn2 = GetPointer<const ssa_name>(GET_NODE(cond_def_use.first));
                   if(not sn2)
                       continue;
                   unsigned int cond_expr_multiplier = 1;
                   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---cond_def " + sn2->CGetDefStmt()->ToString());
                   if(GetPointer<const gimple_assign>(GET_NODE(sn2->CGetDefStmt())) and GET_NODE(GetPointer<const gimple_assign>(GET_NODE(sn2->CGetDefStmt()))->op1)->get_kind() == cond_expr_K)
                   {
                       cond_expr_multiplier = tree_helper::Size(GET_NODE(cond_def_use.first));
                   }
                   for(const auto use : sn2->CGetUseStmts())
                   {
                      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Considering use in " + use.first->ToString());
                      if(use.first->index != cond_def->index)
                      {
                         const auto use_stmt = GetPointer<const gimple_assign>(GET_NODE(use.first));
                         if(not use_stmt)
                         {
                            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Not in a gimple assignment");
                            continue;
                         }
                         const auto right_part = GET_NODE(use_stmt->op1);
                         if(right_part->get_kind() == lt_expr_K or
                               right_part->get_kind() == gt_expr_K or
                               right_part->get_kind() == ge_expr_K or
                               right_part->get_kind() == le_expr_K or
                               right_part->get_kind() == eq_expr_K or
                               right_part->get_kind() == ne_expr_K or
                               right_part->get_kind() == truth_and_expr_K or
                               right_part->get_kind() == truth_or_expr_K or
                               right_part->get_kind() == truth_not_expr_K)
                         {
                            if(((schedule->is_scheduled(use.first->index) and schedule->get_cstep(use.first->index) == cs) or (cs.second == AbsControlStep::UNKNOWN and cs.first == GetPointer<const gimple_node>(GET_NODE(use.first))->bb_index)))
                            {
                               for(const auto comparison_result_use : GetPointer<const ssa_name>(GET_NODE(use_stmt->op0))->CGetUseStmts())
                               {
                                  if(((schedule->is_scheduled(comparison_result_use.first->index) and schedule->get_cstep(comparison_result_use.first->index) == cs) or (cs.second == AbsControlStep::UNKNOWN and cs.first == GetPointer<const gimple_node>(GET_NODE(comparison_result_use.first))->bb_index))
                                        and GetPointer<const gimple_assign>(GET_NODE(comparison_result_use.first))
                                        and GET_NODE( GetPointer<const gimple_assign>(GET_NODE(comparison_result_use.first))->op1)->get_kind() == cond_expr_K)
                                  {
                                     const auto tgt_ga = GetPointer<const gimple_assign>(GET_NODE(comparison_result_use.first));
                                     if(tgt_ga)
                                     {
                                        const auto tgt_sn = GetPointer<const ssa_name>(GET_NODE(tgt_ga->op0));
                                        if(tgt_sn)
                                        {
                                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Incrementing cond_expr_var_root_uses of " + STR(tree_helper::Size(GET_NODE(tgt_ga->op0))) + " * " + STR(tgt_sn->CGetNumberUses()) + " because of " + tgt_ga->ToString());
                                           cond_expr_var_root_uses += cond_expr_multiplier*tree_helper::Size(GET_NODE(tgt_ga->op0)) * tgt_sn->CGetNumberUses();
                                        }
                                     }
                                  }
                                  else
                                  {
                                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Incrementing (2) cond_expr_var_root_uses of " + STR(cond_expr_multiplier) + " * " + STR(tree_helper::Size(GET_NODE(cond_def_use.first))) + " because of " + comparison_result_use.first->ToString());
                                     cond_expr_var_root_uses += cond_expr_multiplier*tree_helper::Size(GET_NODE(cond_def_use.first));
                                  }
                               }
                            }
                            else
                            {
                               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Incrementing cond_expr_var_root_uses of " + STR(cond_expr_multiplier) + " * " + STR(tree_helper::Size(GET_NODE(cond_def_use.first))) + " because of " + use.first->ToString());
                               cond_expr_var_root_uses += cond_expr_multiplier*tree_helper::Size(GET_NODE(cond_def_use.first));
                            }
                         }
                      }
                      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Considered use in " + use.first->ToString());
                   }
               }
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Estimated roots of " + STR(cond_def) + "-->" + " cond_expr_var_root_uses="+STR(cond_expr_var_root_uses));
         }
         if(is_cond_expr and cond_def->index == first_operation)
         {
            size_t n_fo = cond_expr_var_root_uses;
            const auto cond_def_ga = GetPointer<const gimple_assign>(GET_NODE(cond_def));
            if(cond_def_ga)
            {
               const auto sn = GetPointer<const ssa_name>(GET_NODE(cond_def_ga->op0));
               if(sn)
               {
                  for(const auto use : sn->CGetUseStmts())
                  {
                     const auto tgt = use.first;
                     if(tgt->index == second_operation
                             or (GetPointer<const gimple_assign>(GET_NODE(use.first))
                                 and GET_NODE((GetPointer<const gimple_assign>(GET_NODE(use.first))->op1))->get_kind() == cond_expr_K))
                     {
                        const auto tgt_ga = GetPointer<const gimple_assign>(GET_NODE(tgt));
                        if(tgt_ga)
                        {
                           const auto tgt_sn = GetPointer<const ssa_name>(GET_NODE(tgt_ga->op0));
                           if(tgt_sn)
                           {
                              for(const auto use_stmt : tgt_sn->CGetUseStmts())
                              {
                                 if(GetPointer<const gimple_node>(GET_NODE(tgt))->bb_index == GetPointer<const gimple_node>(GET_NODE(use_stmt.first))->bb_index)
                                 {
                                    INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Incrementing fan out of " + STR(tree_helper::Size(GET_NODE(tgt_ga->op0))) + " * " + STR(use_stmt.second) + " because of " + tgt_sn->ToString());
                                    n_fo += tree_helper::Size(GET_NODE(tgt_ga->op0)) * use_stmt.second;
                                 }
                                 else
                                 {
                                    INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Incrementing fan out of " + STR(tree_helper::Size(GET_NODE(tgt_ga->op0))) + " because of " + tgt_sn->ToString());
                                    n_fo += tree_helper::Size(GET_NODE(tgt_ga->op0));
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Estimated FO of " + STR(first_operation) + "-->" + STR(second_operation) +" = " + STR(n_fo));
            if(n_fo)
            {
               fo_correction = 0.003 * get_setup_hold_time()*static_cast<double>(n_fo);
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(fo_correction));
               return fo_correction;
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": 0.0");
      }
      const bool is_load_store = GetPointer<const gimple_node>(second_operation_tn)->operation == "LOAD" or GetPointer<const gimple_node>(second_operation_tn)->operation == "STORE";
      if(is_load_store)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
         const auto fu_type = GetFuType(second_operation);
         bool is_array = is_direct_access_memory_unit(fu_type);
         unsigned var = is_array ? (is_memory_unit(fu_type) ? get_memory_var(fu_type) : get_proxy_memory_var(fu_type)) : 0;
         if(var && HLS_mgr->Rmem->get_maximum_references(var)>(2*get_number_channels(fu_type)))
         {
            const auto ret = estimate_muxNto1_delay(get_prec(fu_type), static_cast<unsigned int>(HLS_mgr->Rmem->get_maximum_references(var))/(2*get_number_channels(fu_type)));
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(ret) + " var=" + STR(var));
            return ret;
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": 0.0");
      }
      if(first_operation != ENTRY_ID)
      {
         const auto first_operation_tn = TreeM->CGetTreeNode(first_operation);
         const bool is_first_load = GetPointer<const gimple_node>(first_operation_tn)->operation == "LOAD";
         if(is_first_load)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
            const auto fu_type = GetFuType(first_operation);
            bool is_array = is_direct_access_memory_unit(fu_type);
            unsigned var = is_array ? (is_memory_unit(fu_type) ? get_memory_var(fu_type) : get_proxy_memory_var(fu_type)) : 0;
            if(var && HLS_mgr->Rmem->get_maximum_loads(var)>(get_number_channels(fu_type)))
            {
               auto ret = estimate_muxNto1_delay(get_prec(fu_type), static_cast<unsigned int>(HLS_mgr->Rmem->get_maximum_loads(var))/(get_number_channels(fu_type)));
               if(ret>(2.5*get_setup_hold_time()))
                   ret = 2.5*get_setup_hold_time();
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(ret) + " var=" + STR(var));
               return ret;
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": 0.0");
         }
         else if (GetPointer<const gimple_node>(first_operation_tn)->operation != "STORE")
         {
             if(CanImplementSetNotEmpty(first_operation))
             {
                const auto fu_type = GetFuType(first_operation);
                const auto n_resources = get_number_fu(fu_type);
                if(n_resources != INFINITE_UINT)
                {
                   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
                   auto ret = estimate_muxNto1_delay(get_prec(fu_type), max_number_of_operations(fu_type)/n_resources);
                   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(ret) + " fu_type=" + STR(fu_type) + " n_resources="+STR(n_resources) + " number_of_operations="+STR(max_number_of_operations(fu_type)));
                   if(ret != 0.0) return ret;
                }
             }
         }
      }
      const bool is_second_ternary_plus =
              GetPointer<const gimple_node>(second_operation_tn)->operation == "ternary_plus_expr" ||
              GetPointer<const gimple_node>(second_operation_tn)->operation == "ternary_pm_expr" ||
              GetPointer<const gimple_node>(second_operation_tn)->operation == "ternary_mp_expr" ||
              GetPointer<const gimple_node>(second_operation_tn)->operation == "ternary_mm_expr";
      if(is_second_ternary_plus)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
         auto ret = get_setup_hold_time()/3;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(ret));
         return ret;
      }
      const bool is_second_plus_minus = GetPointer<const gimple_node>(second_operation_tn)->operation == "plus_expr" ||  GetPointer<const gimple_node>(second_operation_tn)->operation == "minus_expr";
      if(is_second_plus_minus)
      {
          const auto fu_type = GetFuType(second_operation);
          if(get_prec(fu_type)>32)
          {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing connection time " + STR(first_operation) + "-->" + STR(second_operation));
            auto ret = get_setup_hold_time()/3;
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed connection time " + STR(first_operation) + "-->" + STR(second_operation) + ": " + STR(ret));
            return ret;
          }
      }
      return 0.0;
   }
}

bool AllocationInformation::is_asynchronous_ram(tree_managerConstRef TM, unsigned int var, unsigned int threshold)
{
   tree_nodeRef var_node = TM->get_tree_node_const(var);
   var_decl* vd = GetPointer<var_decl>(var_node);
   unsigned int var_bitsize = tree_helper::Size(var_node);
   if(vd)
   {
      unsigned int type_index;
      tree_nodeRef array_type_node = tree_helper::get_type_node(var_node, type_index);
      if (GetPointer<array_type>(array_type_node))
      {
         std::vector<unsigned int> dims;
         unsigned int elts_size;
         tree_helper::get_array_dim_and_bitsize(TM, type_index, dims, elts_size);
         unsigned int meaningful_bits = 0;
         if(vd->bit_values.size() != 0)
         {
            for(auto bit_el : vd->bit_values)
               if(bit_el != '0' && bit_el != '1')
                  ++meaningful_bits;
         }
         else
            meaningful_bits = elts_size;
         if(meaningful_bits != elts_size)
            return ((var_bitsize/elts_size)*meaningful_bits <= threshold) && (var_bitsize/elts_size < 127);
         else
            return (var_bitsize <= threshold) && (var_bitsize/elts_size < 127);
      }
      else
         return var_bitsize <= threshold;
   }
   else
      return var_bitsize <= threshold;
}

ControlStep AllocationInformation::op_et_to_cycles(double et , double clock_period) const
{
   return ControlStep(static_cast<unsigned int>(ceil( et / clock_period )));
}
