/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/

#include "testbench_memory_allocation.hpp"

///behavior include
#include "behavioral_helper.hpp"
#include "call_graph_manager.hpp"
#include "function_behavior.hpp"
#include "var_pp_functor.hpp"

///HLS include
#include "hls_manager.hpp"

///HLS/memory include
#include "memory.hpp"

///HLS/simulation include
#include "testbench_generation_base_step.hpp"

///tree include
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

TestbenchMemoryAllocation::TestbenchMemoryAllocation
(
   const ParameterConstRef _parameters,
   const HLS_managerRef _HLSMgr,
   const DesignFlowManagerConstRef _design_flow_manager
) :
   HLS_step
   (
      _parameters,
      _HLSMgr,
      _design_flow_manager,
      HLSFlowStep_Type::TESTBENCH_MEMORY_ALLOCATION
   )
{}

TestbenchMemoryAllocation::~TestbenchMemoryAllocation()
{}

DesignFlowStep_Status TestbenchMemoryAllocation::Exec()
{
   AllocTestbenchMemory();
   return DesignFlowStep_Status::SUCCESS;
}

void TestbenchMemoryAllocation::AllocTestbenchMemory(void) const
{
   const tree_managerConstRef TM = HLSMgr->get_tree_manager();
   const unsigned int function_id =
      HLSMgr->CGetCallGraphManager()->GetRootFunction();
   const BehavioralHelperConstRef behavioral_helper =
      HLSMgr->CGetFunctionBehavior(function_id)->CGetBehavioralHelper();

   const std::map<unsigned int, memory_symbolRef>& mem_vars =
      HLSMgr->Rmem->get_ext_memory_variables();
   // get the mapping between variables in external memory and their external
   // base address
   std::map<unsigned int, unsigned int> address;
   for (const auto & m : mem_vars)
      address[HLSMgr->Rmem->get_external_base_address(m.first)] = m.first;

   std::list<unsigned int> mem;
   for (const auto & ma : address)
      mem.push_back(ma.second);

   const std::list<unsigned int>& func_parameters =
      behavioral_helper->get_parameters();
   for(const auto & p : func_parameters)
   {
      // if the function has some pointer func_parameters some memory needs to be
      // reserved for the place where they point to
      if (behavioral_helper->is_a_pointer(p) && mem_vars.find(p) == mem_vars.end())
         mem.push_back(p);
   }

   // loop on the test vectors
   unsigned int v_idx = 0;
   for (const auto & curr_test_vector : HLSMgr->test_vectors)
   {
      // loop on the variables in memory
      for (std::list<unsigned int>::const_iterator l = mem.begin(); l != mem.end(); l++)
      {
         std::string param = behavioral_helper->PrintVariable(*l);
         if (param[0] == '"')
            param = "@"+STR(*l);

         bool is_memory = false;
         std::string test_v = "0";
         if (mem_vars.find(*l) != mem_vars.end() &&
               std::find(func_parameters.begin(), func_parameters.end(), *l) == func_parameters.end())
         {
            is_memory = true;
            test_v = TestbenchGenerationBaseStep::print_var_init(TM, *l, HLSMgr->Rmem);
         }
         else if (curr_test_vector.find(param) != curr_test_vector.end())
         {
            test_v = curr_test_vector.find(param)->second;
         }

         if (v_idx > 0 && is_memory)
            continue;//memory has been already initialized

         unsigned int reserved_bytes = tree_helper::size(TM, *l) / 8;
         if (reserved_bytes == 0)
            reserved_bytes = 1;

         if (tree_helper::is_a_pointer(TM, *l) && !is_memory)
         {
            unsigned int base_type = tree_helper::get_type_index(TM, *l);
            tree_nodeRef pt_node = TM->get_tree_node_const(base_type);
            THROW_ASSERT(pt_node->get_kind() == pointer_type_K, "A pointer type is expected");
            unsigned int ptd_base_type = GET_INDEX_NODE(GetPointer<pointer_type>(pt_node)->ptd);
            unsigned int base_type_byte_size;

            if(behavioral_helper->is_an_array(ptd_base_type))
               base_type_byte_size = tree_helper::get_array_data_bitsize(TM, ptd_base_type) / 8;
            else if(tree_helper::size(TM, ptd_base_type) == 1)
               base_type_byte_size = 1;
            else
               base_type_byte_size = tree_helper::size(TM, ptd_base_type) / 8;

            if (base_type_byte_size == 0)
               base_type_byte_size = 1;

            std::vector<std::string> splitted;
            boost::algorithm::split(splitted, test_v , boost::algorithm::is_any_of(","));
            reserved_bytes = (static_cast<unsigned int>(splitted.size())) * base_type_byte_size;

            if (HLSMgr->param_address[v_idx].find(*l) ==
                  HLSMgr->param_address[v_idx].end())
            {
               HLSMgr->param_address[v_idx][*l] = HLSMgr->Rmem->get_memory_address();
               HLSMgr->param_mem_size[v_idx][*l] = reserved_bytes;
               HLSMgr->Rmem->reserve_space(reserved_bytes);

               INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level,
                     "---Parameter " + param + " (testvector "+ STR(v_idx)+") allocated at " +
                     STR(HLSMgr->param_address.find(v_idx)->second.find(*l)->second) +
                     " : reserved_mem_size = " +
                     STR(HLSMgr->param_mem_size.find(v_idx)->second.find(*l)->second));
            }
         }
         else if(!is_memory)
         {
            THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "not yet supported case");
         }

         std::list<unsigned int>::const_iterator l_next;
         l_next = l;
         ++l_next;
         unsigned int next_object_offset = 0;
         /// check the next free aligned address
         if (l_next != mem.end() &&
               mem_vars.find(*l_next) != mem_vars.end() &&
               mem_vars.find(*l) != mem_vars.end())
         {
            next_object_offset =
               HLSMgr->Rmem->get_base_address(*l_next, function_id) -
               HLSMgr->Rmem->get_base_address(*l, function_id);
         }
         else if (mem_vars.find(*l) != mem_vars.end() &&
               (l_next == mem.end() ||
                  HLSMgr->param_address.find(v_idx)->second.find(*l_next) ==
                  HLSMgr->param_address.find(v_idx)->second.end()))
         {
            next_object_offset =
               HLSMgr->Rmem->get_memory_address() -
               HLSMgr->Rmem->get_base_address(*l, function_id);
         }
         else if (l_next != mem.end() &&
               mem_vars.find(*l) != mem_vars.end() &&
               HLSMgr->param_address.find(v_idx)->second.find(*l_next) !=
               HLSMgr->param_address.find(v_idx)->second.end())
         {
            next_object_offset =
               HLSMgr->param_address.find(v_idx)->second.find(*l_next)->second -
               HLSMgr->Rmem->get_base_address(*l, function_id);
         }
         else if (l_next != mem.end() &&
               HLSMgr->param_address.find(v_idx)->second.find(*l) !=
               HLSMgr->param_address.find(v_idx)->second.end() &&
               HLSMgr->param_address.find(v_idx)->second.find(*l_next) !=
               HLSMgr->param_address.find(v_idx)->second.end())
         {
            next_object_offset =
               HLSMgr->param_address.find(v_idx)->second.find(*l_next)->second -
               HLSMgr->param_address.find(v_idx)->second.find(*l)->second;
         }
         else if (HLSMgr->param_address.find(v_idx)->second.find(*l) !=
               HLSMgr->param_address.find(v_idx)->second.end())
         {
            next_object_offset =
               HLSMgr->Rmem->get_memory_address() -
               HLSMgr->param_address.find(v_idx)->second.find(*l)->second;
         }
         else
         {
            THROW_ERROR("unexpected pattern");
         }

         THROW_ASSERT(next_object_offset >= reserved_bytes, "more allocated memory than expected");
         HLSMgr->param_next_off[v_idx][*l] = next_object_offset;
      }
      v_idx++;
   }
   return;
}

const std::unordered_set
<
   std::tuple
   <
      HLSFlowStep_Type,
      HLSFlowStepSpecializationConstRef,
      HLSFlowStep_Relationship
   >
>
TestbenchMemoryAllocation::ComputeHLSRelationships
(const DesignFlowStep::RelationshipType relationship_type)
const
{
   std::unordered_set
   <
      std::tuple<HLSFlowStep_Type,
      HLSFlowStepSpecializationConstRef,
      HLSFlowStep_Relationship>
   > ret;
   switch(relationship_type)
   {
      case DEPENDENCE_RELATIONSHIP:
         {
            ret.insert
            (
               std::make_tuple
               (
                  HLSFlowStep_Type::TEST_VECTOR_PARSER,
                  HLSFlowStepSpecializationConstRef(),
                  HLSFlowStep_Relationship::TOP_FUNCTION
               )
            );
            break;
         }
      case INVALIDATION_RELATIONSHIP:
         {
            break;
         }
      case PRECEDENCE_RELATIONSHIP:
         {
            break;
         }
      default:
         THROW_UNREACHABLE("");
   }
   return ret;
}

bool TestbenchMemoryAllocation::HasToBeExecuted() const
{
   return true;
}
