/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * @file c_backend.cpp
 * @brief Simple class used to drive the backend in order to be able to print c source code
 *
 * @author Luca Fossati <fossati@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
 */
///Header include
#include  "c_backend.hpp"

///design_flows include
#include "design_flow_manager.hpp"
#include "design_flow_graph.hpp"

///design_flows/backend/ToC
#include "c_backend_step_factory.hpp"

///design_flows/backend/ToC/source_code_writers include
#include "c_writer.hpp"

///frontend_analysis
#include "application_frontend_flow_step.hpp"
#include "frontend_flow_step.hpp"
#include "frontend_flow_step_factory.hpp"

///Behavior include
#include "application_manager.hpp"
#include "behavioral_helper.hpp"
#include "call_graph_manager.hpp"
#include "call_graph.hpp"
#include "function_behavior.hpp"
#include "op_graph.hpp"
#include "prettyPrintVertex.hpp"

///Graph include
#include "graph.hpp"

#if HAVE_BAMBU_BUILT
///HLS include
#include "hls_flow_step_factory.hpp"
#include "hls_step.hpp"
#endif

///Paramter include
#include "Parameter.hpp"

///STD include
#include <fstream>
#include <iostream>
#include <ostream>
#include <sstream>
#include <string>

///STL include
#include <deque>
#include <list>
#include <map>
#include <set>
#include <utility>
#include <vector>

///tree includes
#include "tree_helper.hpp"
#include "var_pp_functor.hpp"

///Utility include
#include <boost/config.hpp>
#include <boost/lexical_cast.hpp>
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include "indented_output_stream.hpp"
#include "refcount.hpp"

CBackend::CBackend(const Type _c_backend_type, const CBackendInformationConstRef c_backend_information, const DesignFlowManagerConstRef _design_flow_manager, const application_managerConstRef _AppM, const std::string _file_name, const ParameterConstRef _parameters) :
   DesignFlowStep(_design_flow_manager, _parameters),
   indented_output_stream(new IndentedOutputStream()),
   writer(CWriter::CreateCWriter(_c_backend_type, c_backend_information, _AppM, indented_output_stream, _parameters, _parameters->getOption<int>(OPT_debug_level) >= DEBUG_LEVEL_VERBOSE)),
   file_name(_file_name),
   AppM(_AppM),
   TM(_AppM->get_tree_manager()),
   c_backend_type(_c_backend_type)
{
   debug_level = parameters->get_class_debug_level(GET_CLASS(*this));
}

CBackend::~CBackend()
{
}

const CWriterRef CBackend::GetCWriter() const
{
   return writer;
}

DesignFlowStep_Status CBackend::Exec()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->CBackend: writing " + file_name);
   WriteHeader();
   writeIncludes();
   WriteGlobalDeclarations();
   writeImplementations();
   writer->WriteFile(file_name);
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Cbackend: writing " + file_name);
   return DesignFlowStep_Status::SUCCESS;
}

bool CBackend::HasToBeExecuted() const
{
   return true;
}

void CBackend::Initialize()
{
   writer->Initialize();
   if (boost::filesystem::exists(file_name))
      boost::filesystem::remove_all(file_name);
}

void CBackend::WriteHeader()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->C backend: writing header");
   writer->WritePandaHeader();
   writer->WriteHeader();
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--C backend: written header");
}

void CBackend::WriteGlobalDeclarations()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->CBackend: Writing global declarations");
   writer->WriteGlobalDeclarations();
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing function prototypes");
   std::list<unsigned int> extFunctions;
   this->AppM->get_functions_without_body(extFunctions);
   std::list<unsigned int>::iterator extBeg, extEnd;

   for (extBeg = extFunctions.begin(), extEnd = extFunctions.end(); extBeg != extEnd; extBeg++)
   {
      const BehavioralHelperConstRef BH = this->AppM->CGetFunctionBehavior(*extBeg)->CGetBehavioralHelper();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing external function prototype: " + BH->get_function_name());
      if (BH->function_has_to_be_printed(*extBeg))
      {
         writer->DeclareFunctionTypes(*extBeg);
         writer->WriteFunctionDeclaration(*extBeg);
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written external function prototype: " + BH->get_function_name());
   }

   std::list<unsigned int> functions;
   AppM->get_functions_with_body(functions);
   THROW_ASSERT(functions.size() > 0, "at least one function is expected");
   std::list<unsigned int>::const_iterator it, it_end = functions.end();
   for(it = functions.begin(); it != it_end; it++)
   {
      const BehavioralHelperConstRef BH = this->AppM->CGetFunctionBehavior(*it)->CGetBehavioralHelper();

#if HAVE_BAMBU_BUILT
      if (parameters->isOption(OPT_pretty_print))
      {
         std::string f_name = BH->get_function_name();
         if(f_name.find("__builtin_") == 0)
         {
            indented_output_stream->Append("#define " + f_name + " _bambu_"+f_name + "\n");
         }
      }
#endif

      if (BH->function_has_to_be_printed(*it))
      {
         writer->DeclareFunctionTypes(*it);
         writer->WriteFunctionDeclaration(*it);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written function prototypes");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--CBackend: Written global declarations");
}

void CBackend::writeImplementations()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing implementations");
   //First of all I declare the functions and then the tasks
   std::list<unsigned int> functions;
   AppM->get_functions_with_body(functions);
   THROW_ASSERT(functions.size() > 0, "at least one function is expected");
   std::list<unsigned int>::const_iterator it, it_end = functions.end();
   for(it = functions.begin(); it != it_end; it++)
   {
      const BehavioralHelperConstRef BH = this->AppM->CGetFunctionBehavior(*it)->CGetBehavioralHelper();
      if (BH->function_has_to_be_printed(*it))
         writer->WriteFunctionImplementation(*it);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written implementations");
}

void CBackend::writeIncludes()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->CBackend: writing includes");
   ///Set of decl_nodes to be examinated
   std::unordered_set<unsigned int> decl_nodes;

   std::set<std::string> writtenIncludes;

   std::list<unsigned int> extFunctions;
   this->AppM->get_functions_without_body(extFunctions);
   std::list<unsigned int>::iterator extBeg, extEnd;
   for (extBeg = extFunctions.begin(), extEnd = extFunctions.end(); extBeg != extEnd; extBeg++)
   {
      const BehavioralHelperConstRef BH = this->AppM->CGetFunctionBehavior(*extBeg)->CGetBehavioralHelper();
      AnalyzeInclude(*extBeg, BH, writtenIncludes);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Written includes for external functions");

   std::list<unsigned int> Functions;
   AppM->get_functions_with_body(Functions);
   for (extBeg = Functions.begin(), extEnd = Functions.end(); extBeg != extEnd; extBeg++)
   {
      const BehavioralHelperConstRef BH = AppM->CGetFunctionBehavior(*extBeg)->CGetBehavioralHelper();
      const CustomSet<unsigned int> & tmp_vars = writer->GetLocalVariables(*extBeg);
      decl_nodes.insert(tmp_vars.begin(), tmp_vars.end());
      const std::list<unsigned int>& funParams = BH->get_parameters();
      decl_nodes.insert(funParams.begin(), funParams.end());
      if (not BH->function_has_to_be_printed(*extBeg))
         AnalyzeInclude(*extBeg, BH, writtenIncludes);
   }

   const BehavioralHelperConstRef  BH = AppM->CGetFunctionBehavior(*(Functions.begin()))->CGetBehavioralHelper();
   CustomSet<unsigned int> vars = AppM->get_global_variables();
   decl_nodes.insert(vars.begin(), vars.end());
   for (std::unordered_set<unsigned int>::iterator v = decl_nodes.begin(); v != decl_nodes.end(); v++)
   {
      const unsigned int variable_type = BH->get_type(*v);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing " + BH->PrintVariable(*v) + " of type " + boost::lexical_cast<std::string>(variable_type));
      AnalyzeInclude(variable_type, BH, writtenIncludes);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   for(std::set<std::string>::iterator s = writtenIncludes.begin(); s != writtenIncludes.end(); s++)
   {
      writer->writeInclude(*s);
   }
   std::vector<std::string> addHeaders = writer->getIncludes();
   for(std::vector<std::string>::iterator s = addHeaders.begin(); s != addHeaders.end(); s++)
   {
      if (writtenIncludes.find(*s) == writtenIncludes.end())
         writer->writeInclude(*s);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Written includes for variables");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--CBackend: written includes");
}

void CBackend::AnalyzeInclude(unsigned int index, const BehavioralHelperConstRef BH, std::set<std::string>& writtenIncludes)
{
   if(already_visited.find(index) != already_visited.end())
   {
      return;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing include for " + boost::lexical_cast<std::string>(index));
   already_visited.insert(index);
   bool is_system;
   const std::string decl = std::get<0>(BH->get_definition(index, is_system));
   if(not decl.empty() and decl != "<built-in>" and is_system)
   {
      writtenIncludes.insert(decl);
   }
   else
   {
      const unsigned int type = tree_helper::get_type_index(TM, index);
      const std::unordered_set<unsigned int> types_to_be_declared = tree_helper::GetTypesToBeDeclared(TM, type, parameters->getOption<bool>(OPT_without_transformation));
      std::unordered_set<unsigned int>::const_iterator type_to_be_declared, type_to_be_declared_end = types_to_be_declared.end();
      for(type_to_be_declared = types_to_be_declared.begin(); type_to_be_declared != type_to_be_declared_end; type_to_be_declared++)
      {
         AnalyzeInclude(*type_to_be_declared, BH, writtenIncludes);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed");
}

void CBackend::compute_variables(const OpGraphConstRef inGraph, const std::unordered_set<unsigned int> &gblVariables, std::list<unsigned int> &funParams, std::unordered_set<unsigned int> & vars)
{
   //I simply have to go over all the vertices and get the used variables;
   //the variables which have to be declared are all those variables but
   //the globals ones
   VertexIterator v, vEnd;
   for(boost::tie(v, vEnd) = boost::vertices(*inGraph); v != vEnd; v++)
   {
      const CustomSet<unsigned int> vars_temp = inGraph->CGetOpNodeInfo(*v)->cited_variables;
      vars.insert(vars_temp.begin(), vars_temp.end());
   }

   //I have to take out the variables global to the whole program and the function parameters
   std::set<unsigned int> varsTemp;
   std::unordered_set<unsigned int>::const_iterator it, it_end = gblVariables.end();
   for(it = gblVariables.begin(); it != it_end; it++)
   {
      vars.erase(*it);
   }
   std::list<unsigned int>::const_iterator it2, it2_end = funParams.end();
   for(it2 = funParams.begin(); it2 != it2_end; it2++)
   {
      vars.erase(*it2);
   }
}

const DesignFlowStepFactoryConstRef CBackend::CGetDesignFlowStepFactory() const
{
   return design_flow_manager.lock()->CGetDesignFlowStepFactory("CBackend");
}

const std::string CBackend::GetSignature() const
{
   return ComputeSignature(c_backend_type);
}

const std::string CBackend::ComputeSignature(const CBackend::Type type)
{
   switch(type)
   {
#if HAVE_HOST_PROFILING_BUILT
      case(CB_BBP) :
         return "CBackend::BasicBlocksProfiling";
#endif
#if HAVE_HOST_PROFILING_BUILT
      case(CB_DATA_MEMORY_PROFILING) :
         return "CBackend::DataMemoryProfiling";
#endif
#if HAVE_HLS_BUILT && HAVE_EXPERIMENTAL
      case(CB_DISCREPANCY_ANALYSIS):
         return "CBackend::DiscrepancyAnalysis";
#endif
#if HAVE_HOST_PROFILING_BUILT
      case(CB_EPP) :
         return "CBackend::EfficientPathProfiling";
#endif
#if HAVE_TARGET_PROFILING
      case(CB_ESCAPED_SEQUENTIAL):
         return "CBackend::Escaped";
#endif
#if HAVE_BAMBU_BUILT
      case(CB_HLS):
         return "CBackend::HighLevelSynthesis";
#endif
#if HAVE_HOST_PROFILING_BUILT
      case(CB_HPP):
         return "CBackend::HierarchicalPathProfiling";
#endif
#if HAVE_GRAPH_PARTITIONING_BUILT && HAVE_TARGET_PROFILING
      case(CB_INSTRUMENTED_PARALLEL):
         return "CBackend::InstrumentedParallel";
#endif
#if HAVE_TARGET_PROFILING
      case(CB_INSTRUMENTED_SEQUENTIAL):
         return "CBackend::InstrumentedSequential";
#endif
#if HAVE_HOST_PROFILING_BUILT
      case(CB_LOOPS_PROFILING):
         return "CBackend::LoopsProfiling";
#endif
#if HAVE_ZEBU_BUILT
      case(CB_POINTED_DATA_EVALUATION):
         return "CBackend::PointedDataEvaluation";
#endif
#if HAVE_GRAPH_PARTITIONING_BUILT
      case(CB_PARALLEL):
         return "CBackend::Parallel";
#endif
      case(CB_SEQUENTIAL):
         return "CBackend::Sequential";
#if HAVE_HOST_PROFILING_BUILT
      case(CB_TPP):
         return "CBackend::TreePathProfiling";
#endif
      default:
      {
         THROW_UNREACHABLE("");
      }
   }
   return "";
}

const std::string CBackend::GetName() const
{
   return GetSignature();
}

void CBackend::ComputeRelationships(DesignFlowStepSet & relationships, const DesignFlowStep::RelationshipType relationship_type)
{
   switch(relationship_type)
   {
      case(DEPENDENCE_RELATIONSHIP) :
      {
         switch(c_backend_type)
         {
#if HAVE_HOST_PROFILING_BUILT
            case(CB_DATA_MEMORY_PROFILING) :
#endif
            case CB_SEQUENTIAL:
#if HAVE_HOST_PROFILING_BUILT
            case(CB_BBP):
            case(CB_TPP):
#endif
            {
               std::unordered_set<std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship> > frontend_relationships;
               frontend_relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(BASIC_BLOCKS_CFG_COMPUTATION, FrontendFlowStep::ALL_FUNCTIONS));
               frontend_relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(DEAD_CODE_ELIMINATION, FrontendFlowStep::ALL_FUNCTIONS));
               frontend_relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(LOOP_COMPUTATION, FrontendFlowStep::ALL_FUNCTIONS));
               frontend_relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(OPERATIONS_CFG_COMPUTATION, FrontendFlowStep::ALL_FUNCTIONS));
               frontend_relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(VAR_ANALYSIS, FrontendFlowStep::ALL_FUNCTIONS));
               FrontendFlowStep::CreateSteps(design_flow_manager.lock(), frontend_relationships, AppM, relationships);
               break;
            }
#if HAVE_HLS_BUILT
#if HAVE_EXPERIMENTAL
            case (CB_DISCREPANCY_ANALYSIS):
#endif
            case(CB_HLS):
            {
               // The first time this step is added, we add the dependence
               // from the complete call graph computation. Ideally we would
               // need also the dependence from the HLS steps,
               // but at this point we don't know the top function yet.
               // The trick is that the dependencies will be recomputed again
               // before this is executed. At that time the top
               // function will be ready. The dependencies from HLS steps are
               // added after the check on the call graph for this reason.
               const FrontendFlowStepFactory * frontend_step_factory =
                  GetPointer<const FrontendFlowStepFactory>
                     (design_flow_manager.lock()->CGetDesignFlowStepFactory("Frontend"));

               const vertex call_graph_computation_step =
                  design_flow_manager.lock()->GetDesignFlowStep
                  (ApplicationFrontendFlowStep::ComputeSignature(FUNCTION_ANALYSIS));

               const DesignFlowGraphConstRef design_flow_graph =
                  design_flow_manager.lock()->CGetDesignFlowGraph();

               const DesignFlowStepRef cg_design_flow_step = call_graph_computation_step ?
                  design_flow_graph->CGetDesignFlowStepInfo(call_graph_computation_step)->design_flow_step :
                  frontend_step_factory->CreateApplicationFrontendFlowStep(FUNCTION_ANALYSIS);

               relationships.insert(cg_design_flow_step);

               // Root function cannot be computed at the beginning so if the
               // call graph is not ready yet we exit. The relationships will
               // be computed again after the call graph computation.
               const CallGraphManagerConstRef call_graph_manager = AppM->CGetCallGraphManager();
               if(boost::num_vertices(*(call_graph_manager->CGetCallGraph())) == 0)
                  break;

               const HLSFlowStepFactory * hls_step_factory = GetPointer<const HLSFlowStepFactory>(design_flow_manager.lock()->CGetDesignFlowStepFactory("HLS"));

               relationships.insert(hls_step_factory->CreateHLSFlowStep(HLSFlowStep_Type::TESTBENCH_MEMORY_ALLOCATION, HLSFlowStepSpecializationConstRef()));

               relationships.insert(hls_step_factory->CreateHLSFlowStep(HLSFlowStep_Type::TEST_VECTOR_PARSER, HLSFlowStepSpecializationConstRef()));
#if HAVE_EXPERIMENTAL
               if(c_backend_type == CB_DISCREPANCY_ANALYSIS)
               {
                  relationships.insert(hls_step_factory->CreateHLSFlowStep(HLSFlowStep_Type::VCD_SIGNAL_SELECTION, HLSFlowStepSpecializationConstRef()));
               }
#endif

               break;
            }
#endif
#if HAVE_HOST_PROFILING_BUILT
            case(CB_EPP) :
#endif
#if HAVE_TARGET_PROFILING
            case(CB_ESCAPED_SEQUENTIAL):
#endif
#if HAVE_HOST_PROFILING_BUILT
            case(CB_HPP):
#endif
#if HAVE_GRAPH_PARTITIONING_BUILT && HAVE_TARGET_PROFILING
            case(CB_INSTRUMENTED_PARALLEL):
#endif
#if HAVE_TARGET_PROFILING
            case(CB_INSTRUMENTED_SEQUENTIAL):
#endif
#if HAVE_HOST_PROFILING_BUILT
            case(CB_LOOPS_PROFILING):
#endif
#if HAVE_ZEBU_BUILT
            case(CB_POINTED_DATA_EVALUATION):
#endif
#if HAVE_GRAPH_PARTITIONING_BUILT
            case(CB_PARALLEL):
#endif
            default:
            {
               THROW_UNREACHABLE("");
            }
         }
         break;
      }
      case(INVALIDATION_RELATIONSHIP):
      {
         break;
      }
      case(PRECEDENCE_RELATIONSHIP):
      {
         break;
      }
      default:
      {
         THROW_UNREACHABLE("");
      }
   }
}
