/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file fix_characterization.cpp
 * @brief Step to fix components characterization
 *
 * @author Marco Lattuada <marco.lattuada@polimi.it>
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 *
*/

///Header include
#include "fix_characterization.hpp"

///. include
#include "Parameter.hpp"

///technology include
#include "technology_manager.hpp"

///technology/physical_library includes
#include "library_manager.hpp"
#include "technology_node.hpp"

///technology/physical_library/models include
#include "time_model.hpp"

///technology/target_device include
#include "target_device.hpp"

FixCharacterization::FixCharacterization(const technology_managerRef _TM, const target_deviceRef _target, const DesignFlowManagerConstRef _design_flow_manager, const ParameterConstRef _parameters) :
   TechnologyFlowStep(_TM, _target, _design_flow_manager, TechnologyFlowStep_Type::FIX_CHARACTERIZATION, _parameters),
   assignment_execution_time(0.0)
{
   debug_level = parameters->get_class_debug_level(GET_CLASS(*this));
}

FixCharacterization::~FixCharacterization()
{}

void FixCharacterization::Initialize()
{
   const technology_nodeConstRef f_unit_as = TM->get_fu(ASSIGN_VECTOR_BOOL_STD, LIBRARY_STD_FU);
   THROW_ASSERT(f_unit_as, "Library miss component: " + std::string(ASSIGN_VECTOR_BOOL_STD));
   const auto * fu_as= GetPointer<const functional_unit>(f_unit_as);
   const technology_nodeConstRef op_as_node = fu_as->get_operation("ASSIGN");
   const auto op_ASSIGN = GetPointer<const operation>(op_as_node);
   assignment_execution_time = op_ASSIGN->time_m->get_execution_time();
   assignment_characterization_timestamp = fu_as->characterization_timestamp;
}

DesignFlowStep_Status FixCharacterization::Exec()
{
   const auto libraries = TM->get_library_list();
   for(const auto library : libraries)
   {
      const auto LM = TM->get_library_manager(library);
      const auto fus = LM->get_library_fu();
      for(const auto fu : fus)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing " + fu.first);
         auto single_fu = GetPointer<functional_unit>(fu.second);
         if(!single_fu) continue;
         auto template_name = single_fu->fu_template_name;
         if(single_fu)
         {
            /// shift ops
            if(template_name == "rshift_expr_FU" ||
               template_name == "ui_rshift_expr_FU" ||
               template_name == "lshift_expr_FU" ||
               template_name == "ui_lshift_expr_FU" ||
               template_name == "ui_lrotate_expr_FU" ||
               template_name == "ui_rrotate_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() == 3, single_fu->fu_template_parameters);
               if(template_parameters[1] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
            /// vectorize shift ops
            if(template_name == "vec1_rshift_expr_FU" ||
               template_name == "ui_vec1_rshift_expr_FU" ||
               template_name == "vec1_lshift_expr_FU" ||
               template_name == "ui_vec1_rshift_expr_FU" ||
               template_name == "vec_rshift_expr_FU" ||
               template_name == "ui_vec_rshift_expr_FU" ||
               template_name == "vec_lshift_expr_FU" ||
               template_name == "ui_vec_rshift_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() >= 4, single_fu->fu_template_parameters);
               if(template_parameters[2] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
            /// bitwise ops
            if(template_name == "bit_and_expr_FU" ||
               template_name == "ui_bit_and_expr_FU" ||
               template_name == "bit_ior_expr_FU" ||
               template_name == "ui_bit_ior_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() == 3, single_fu->fu_template_parameters);
               if(template_parameters[0] == "0" || template_parameters[1] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
            /// vectorize bitwise ops
            if(template_name == "vec_bit_and_expr_FU" ||
               template_name == "ui_vec_bit_and_expr_FU" ||
               template_name == "vec_bit_ior_expr_FU" ||
               template_name == "ui_vec_bit_ior_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() >= 5, single_fu->fu_template_parameters);
               if(template_parameters[0] == "0" || template_parameters[2] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
            /// cond expr ops
            if(template_name == "cond_expr_FU" ||
               template_name == "ui_cond_expr_FU" ||
               template_name == "fp_cond_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() == 4, single_fu->fu_template_parameters);
               if(template_parameters[0] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
            /// vectorize cond expr ops
            if(template_name == "vec_cond_expr_FU" ||
               template_name == "ui_vec_cond_expr_FU"
                    )
            {
               std::vector<std::string> template_parameters;
               boost::algorithm::split(template_parameters, single_fu->fu_template_parameters, boost::algorithm::is_any_of(" "));
               THROW_ASSERT(template_parameters.size() >= 6, single_fu->fu_template_parameters);
               if(template_parameters[0] == "0")
               {
                  for(auto op : single_fu->get_operations())
                  {
                     if(assignment_execution_time != (GetPointer<operation>(op))->time_m->get_execution_time())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Fixing execution time of " + GetPointer<const operation>(op)->operation_name + " on " + fu.first);
                        (GetPointer<operation>(op))->time_m->set_execution_time(assignment_execution_time,0);
                        single_fu->characterization_timestamp = assignment_characterization_timestamp;
                     }
                  }
               }
            }
         }

         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Analyzed " + fu.first);
      }
   }
   return DesignFlowStep_Status::SUCCESS;
}

const std::unordered_set<TechnologyFlowStep_Type> FixCharacterization::ComputeTechnologyRelationships(const DesignFlowStep::RelationshipType relationship_type) const
{
   std::unordered_set<TechnologyFlowStep_Type> relationships;
   switch(relationship_type)
   {
      case(DEPENDENCE_RELATIONSHIP) :
      {
         break;
      }
      case(INVALIDATION_RELATIONSHIP) :
      {
         break;
      }
      case(PRECEDENCE_RELATIONSHIP) :
      {
         relationships.insert(TechnologyFlowStep_Type::LOAD_DEFAULT_TECHNOLOGY);
         relationships.insert(TechnologyFlowStep_Type::LOAD_DEVICE_TECHNOLOGY);
         relationships.insert(TechnologyFlowStep_Type::LOAD_FILE_TECHNOLOGY);
         break;
      }
      default:
      {
         THROW_UNREACHABLE("");
      }
   }
   return relationships;
}



