/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file Bit_Value_opt.cpp
 * @brief
 *
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/

//Header include
#include "Bit_Value_opt.hpp"

///Autoheader include
#include "config_HAVE_FROM_DISCREPANCY_BUILT.hpp"

//Behavior include
#include "application_manager.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "call_graph_manager.hpp"
#include "call_graph.hpp"

///HLS includes
#include "hls.hpp"
#include "hls_manager.hpp"

#if HAVE_BAMBU_BUILT && HAVE_ILP_BUILT
#include "hls_step.hpp"
#endif

// includes from HLS/vcd
#if HAVE_FROM_DISCREPANCY_BUILT
#include "Discrepancy.hpp"
#endif

//Parameter include
#include "Parameter.hpp"

//STD include
#include <fstream>
#include <string>
#include <math.h>
#include <boost/range/adaptor/reversed.hpp>

//Tree include
#include "tree_manager.hpp"
#include "tree_manipulation.hpp"
#include "ext_tree_node.hpp"
#include "tree_basic_block.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "tree_helper.hpp"

Bit_Value_opt::Bit_Value_opt (const ParameterConstRef _parameters, const application_managerRef _AppM, unsigned int _function_id, const DesignFlowManagerConstRef _design_flow_manager) :
   FunctionFrontendFlowStep (_AppM, _function_id, BIT_VALUE_OPT, _design_flow_manager, _parameters),
   restart_dead_code(false)
{
   debug_level = parameters->get_class_debug_level (GET_CLASS(*this),
                                                    DEBUG_LEVEL_NONE);
}

Bit_Value_opt::~Bit_Value_opt ()
{
}

const std::unordered_set<std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship> > Bit_Value_opt::ComputeFrontendRelationships (const DesignFlowStep::RelationshipType relationship_type) const
{
   std::unordered_set<std::pair<FrontendFlowStepType, FunctionRelationship> > relationships;
   switch (relationship_type)
   {
      case (PRECEDENCE_RELATIONSHIP):
      {
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (EXTRACT_PATTERNS, SAME_FUNCTION));
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (EXTRACT_GIMPLE_COND_OP, SAME_FUNCTION));
#if HAVE_ILP_BUILT && HAVE_BAMBU_BUILT
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (SDC_CODE_MOTION, SAME_FUNCTION));
#endif
         break;
      }
      case DEPENDENCE_RELATIONSHIP:
      {
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (BIT_VALUE, SAME_FUNCTION));
         break;
      }
      case(INVALIDATION_RELATIONSHIP) :
      {
         switch(GetStatus())
         {
            case DesignFlowStep_Status::SUCCESS:
               {
                  relationships.insert(std::pair<FrontendFlowStepType, FunctionRelationship>(DEAD_CODE_ELIMINATION, SAME_FUNCTION));
                  break;
               }
            case DesignFlowStep_Status::SKIPPED:
            case DesignFlowStep_Status::UNCHANGED:
            case DesignFlowStep_Status::UNEXECUTED:
            case DesignFlowStep_Status::UNNECESSARY:
               {
                  break;
               }
            case DesignFlowStep_Status::ABORTED:
            case DesignFlowStep_Status::EMPTY:
            case DesignFlowStep_Status::NONEXISTENT:
            default:
               THROW_UNREACHABLE("");
         }
         break;
      }
      default:
         THROW_UNREACHABLE("");
   }
   return relationships;
}

static tree_nodeRef
create_ga(tree_manipulationRef &IRman, tree_nodeRef, tree_nodeRef &type, tree_nodeRef& op, unsigned int bb_index, const std::string & srcp_default)
{
   tree_nodeRef ssa_vd = IRman->create_ssa_name(tree_nodeRef(), type);
   return IRman->create_gimple_modify_stmt(ssa_vd, op, srcp_default, bb_index);
}


void Bit_Value_opt::optimize(statement_list* sl, tree_managerRef TM)
{
   for(auto bb_pair : sl->list_of_bloc)
   {
      blocRef B = bb_pair.second;
      unsigned int B_id = B->number;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining BB" + STR(B_id));
      const auto list_of_stmt = B->CGetStmtList();
      for (const auto stmt : list_of_stmt)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining statement " + GET_NODE(stmt)->ToString());
#ifndef NDEBUG
         if(not function_behavior->ApplyNewTransformation())
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Skipped because reached limit of cfg transformations");
            continue;
         }
#endif
         if(GET_NODE(stmt)->get_kind() == gimple_assign_K)
         {
            gimple_assign * ga =  GetPointer<gimple_assign>(GET_NODE(stmt));
            unsigned int output_uid = GET_INDEX_NODE(ga->op0);
            ssa_name *ssa = GetPointer<ssa_name> (GET_NODE(ga->op0));
            if (ssa)
            {
               if(tree_helper::is_real(TM, output_uid))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---real variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(stmt)->ToString());
                  continue;
               }
               if(tree_helper::is_a_complex(TM, output_uid))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---complex variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(stmt)->ToString());
                  continue;
               }
               if(tree_helper::is_a_vector(TM, output_uid))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---vector variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(stmt)->ToString());
                  continue;
               }
               if(GetPointer<integer_cst>(GET_NODE(ga->op1)) && tree_helper::is_a_pointer(TM, GET_INDEX_NODE(ga->op1)))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---constant pointer value assignements not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(stmt)->ToString());
                  continue;
               }
               if(GetPointer<call_expr>(GET_NODE(ga->op1)) and ga->vdef)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---calls with side effects cannot be optimized" + STR(GET_INDEX_NODE(ga->op1)));
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(stmt)->ToString());
                  continue;
               }
               unsigned int type_index = tree_helper::get_type_index(TM, GET_INDEX_NODE(ga->op0));
               tree_nodeRef ga_op_type = TM->GetTreeReindex(type_index);
               tree_nodeRef Scpe = TM->GetTreeReindex(function_id);
               const std::string & bit_values = ssa->bit_values;
               bool is_constant=bit_values.size()!=0;
               for(auto current_el : bit_values)
               {
                  if(current_el == 'U')
                  {
                     is_constant = false;
                     break;
                  }
               }
               if(is_constant)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,"---Left part is constant");
                  unsigned long long int  const_value = 0;
                  unsigned int index_val = 0;
                  for(auto current_el : boost::adaptors::reverse(bit_values))
                  {
                     if(current_el == '1')
                        const_value |= 1ULL << index_val;
                     ++index_val;
                  }
                  ///in case do sign extension
                  if (tree_helper::is_int(TM, output_uid) && bit_values[0]=='1')
                  {
                     for(; index_val < 64;++index_val)
                        const_value |= 1ULL << index_val;
                  }
                  tree_nodeRef val;
                  if(GetPointer<integer_cst>(GET_NODE(ga->op1)))
                     val = ga->op1;
                  else
                  {
                     tree_manipulationRef IRman=tree_manipulationRef(new tree_manipulation(TM,parameters));
                     unsigned int integer_cst_id = TM->new_tree_node_id();
                     val = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(const_value), integer_cst_id);
                     if (GET_NODE(ga->op1)->get_kind() == call_expr_K)
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC,  debug_level, "---Call expr in the right part can be removed");
                        const CallGraphManagerRef cg_man = AppM->GetCallGraphManager();
                        const vertex fun_cg_vertex = cg_man->GetVertex(function_id);
                        const CallGraphConstRef cg = cg_man->CGetCallGraph();
                        std::set<EdgeDescriptor> to_remove;

                        OutEdgeIterator oei, oei_end;
                        boost::tie(oei, oei_end) = boost::out_edges(fun_cg_vertex, *cg);
                        const unsigned int call_id = GET_INDEX_NODE(stmt);
                        for (; oei != oei_end; oei++)
                        {
                           const std::set<unsigned int> & direct_calls = cg->CGetFunctionEdgeInfo(*oei)->direct_call_points;
                           const std::set<unsigned int>::iterator call_it = direct_calls.find(call_id);
                           if (call_it != direct_calls.end())
                           {
                              to_remove.insert(*oei);
                           }
                        }
                        THROW_ASSERT(to_remove.size(), "Call to be removed not found in call graph");
                        for (const EdgeDescriptor & e : to_remove)
                        {
                           cg_man->remove_call_point(e, call_id);
                        }
                        if(parameters->getOption<bool>(OPT_print_dot))
                        {
                           AppM->CGetCallGraphManager()->CGetCallGraph()->WriteDot("call_graph.dot");
                        }

                     }
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Replacing " + STR(ga->op1) + " with " + STR(val) + " in " + STR(stmt));
                     TM->ReplaceTreeNode(stmt, ga->op1, val);
                     restart_dead_code = true;
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---replace expression with a constant: "+ stmt->ToString());
                  }
                  const TreeNodeMap<size_t> StmtUses = ssa->CGetUseStmts();
                  for(const auto use : StmtUses)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---replace constant usage before: "+ use.first->ToString());
                     TM->ReplaceTreeNode(use.first, ga->op0, val);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---replace constant usage after: "+ use.first->ToString());
                  }
                  if(not StmtUses.empty())
                     restart_dead_code = true;
#ifndef NDEBUG
                  function_behavior->RegisterTransformation(GetName(), stmt);
#endif
               }
               else if(GET_NODE(ga->op1)->get_kind() == mult_expr_K||GET_NODE(ga->op1)->get_kind() == widen_mult_expr_K)
               {
                  binary_expr * me = GetPointer<binary_expr>(GET_NODE(ga->op1));
                  tree_nodeRef op0 = GET_NODE(me->op0);
                  tree_nodeRef op1 = GET_NODE(me->op1);
                  unsigned int trailing_zero_op0 = 0;
                  unsigned int trailing_zero_op1 = 0;
                  if(GetPointer<ssa_name>(op0))
                  {
                     const std::string & bit_values_op0 = GetPointer<ssa_name>(op0)->bit_values;
                     for(auto current_el : boost::adaptors::reverse(bit_values_op0))
                     {
                        if(current_el == '0' || current_el == 'X')
                           ++trailing_zero_op0;
                        else
                           break;
                     }
                  }
                  if(GetPointer<ssa_name>(op1))
                  {
                     const std::string & bit_values_op1 = GetPointer<ssa_name>(op1)->bit_values;
                     for(auto current_el : boost::adaptors::reverse(bit_values_op1))
                     {
                        if(current_el == '0' || current_el == 'X')
                           ++trailing_zero_op1;
                        else
                           break;
                     }
                  }
                  if(trailing_zero_op0 != 0 || trailing_zero_op1 != 0)
                  {
#ifndef NDEBUG
                     function_behavior->RegisterTransformation(GetName(), stmt);
#endif
                     INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->Bit Value Opt: mult_expr/widen_mult_expr optmized, nbits = " + STR(trailing_zero_op0+trailing_zero_op1));
                     INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--");
                     tree_manipulationRef IRman=tree_manipulationRef(new tree_manipulation(TM,parameters));
                     const std::string srcp_default = ga->include_name + ":" + STR(ga->line_number) + ":" + STR(ga->column_number);
                     if(trailing_zero_op0 != 0)
                     {
                        unsigned int op0_const_rshift_id = TM->new_tree_node_id();
                        tree_nodeRef op0_const_node = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(trailing_zero_op0), op0_const_rshift_id);
                        tree_nodeRef op0_expr = IRman->create_binary_operation(ga_op_type, me->op0, op0_const_node, srcp_default, rshift_expr_K);
                        tree_nodeRef op0_ga = create_ga(IRman, Scpe, ga_op_type, op0_expr, B_id, srcp_default);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(op0_ga));
                        B->PushBefore(op0_ga, stmt);
                        tree_nodeRef op0_ga_var = GetPointer<gimple_assign>(GET_NODE(op0_ga))->op0;
                        TM->ReplaceTreeNode(stmt, me->op0, op0_ga_var);
                        ///set the bit_values to the ssa var
                        ssa_name* op0_ssa = GetPointer<ssa_name>(GET_NODE(op0_ga_var));
                        op0_ssa->bit_values = GetPointer<ssa_name>(op0)->bit_values.substr(0, GetPointer<ssa_name>(op0)->bit_values.size()-trailing_zero_op0);
                     }
                     if(trailing_zero_op1 != 0 and op0->index != op1->index)
                     {
                        unsigned int op1_const_rshift_id = TM->new_tree_node_id();
                        tree_nodeRef op1_const_node = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(trailing_zero_op1), op1_const_rshift_id);
                        tree_nodeRef op1_expr = IRman->create_binary_operation(ga_op_type, me->op1, op1_const_node, srcp_default, rshift_expr_K);
                        tree_nodeRef op1_ga = create_ga(IRman, Scpe, ga_op_type, op1_expr, B_id, srcp_default);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(op1_ga));
                        B->PushBefore(op1_ga, stmt);
                        tree_nodeRef op1_ga_var = GetPointer<gimple_assign>(GET_NODE(op1_ga))->op0;
                        TM->ReplaceTreeNode(stmt, me->op1, op1_ga_var);
                        ///set the bit_values to the ssa var
                        ssa_name* op1_ssa = GetPointer<ssa_name>(GET_NODE(op1_ga_var));
                        op1_ssa->bit_values = GetPointer<ssa_name>(op1)->bit_values.substr(0, GetPointer<ssa_name>(op1)->bit_values.size()-trailing_zero_op1);
                     }

                     tree_nodeRef ssa_vd = IRman->create_ssa_name(tree_nodeRef(), ga_op_type);
                     ssa_name * sn = GetPointer<ssa_name> (GET_NODE(ssa_vd));
                     ///set the bit_values to the ssa var
                     sn->bit_values = ssa->bit_values.substr(0, ssa->bit_values.size()-trailing_zero_op0-trailing_zero_op1);
                     unsigned int op_const_lshift_id = TM->new_tree_node_id();
                     tree_nodeRef op_const_node = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(trailing_zero_op0+trailing_zero_op1), op_const_lshift_id);
                     tree_nodeRef op_expr = IRman->create_binary_operation(ga_op_type, ssa_vd, op_const_node, srcp_default, lshift_expr_K);
                     tree_nodeRef curr_ga = create_ga(IRman, Scpe, ga_op_type, op_expr, B_id, srcp_default);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(curr_ga));
                     TM->ReplaceTreeNode(curr_ga, GetPointer<gimple_assign>(GET_NODE(curr_ga))->op0, ga->op0);
                     TM->ReplaceTreeNode(stmt, ga->op0, ssa_vd);
                     B->PushAfter(curr_ga, stmt);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "pushed");
                  }
               }
               else if(GET_NODE(ga->op1)->get_kind() == plus_expr_K || GET_NODE(ga->op1)->get_kind() == minus_expr_K)
               {
                  binary_expr * me = GetPointer<binary_expr>(GET_NODE(ga->op1));
                  if(me->op0->index == me->op1->index)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Skipped plus expr with same operands " + GET_NODE(stmt)->ToString());
                     continue;
                  }
                  tree_nodeRef op0 = GET_NODE(me->op0);
                  tree_nodeRef op1 = GET_NODE(me->op1);
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(output_uid) +" bitstring: " + bit_values);
                  unsigned int trailing_zero_op0 = 0;
                  unsigned int trailing_zero_op1 = 0;
                  bool is_op0_const = GetPointer<integer_cst>(op0);
                  bool is_op1_const = GetPointer<integer_cst>(op1);
                  if(GetPointer<ssa_name>(op0) && GET_NODE(ga->op1)->get_kind() == plus_expr_K)
                  {
                     const std::string & bit_values_op0 = GetPointer<ssa_name>(op0)->bit_values;
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(me->op0)) +" bitstring: " + bit_values_op0);
                     for(auto current_el : boost::adaptors::reverse(bit_values_op0))
                     {
                        if(current_el == '0' || current_el == 'X')
                           ++trailing_zero_op0;
                        else
                           break;
                     }
                  }
                  else if(GetPointer<integer_cst>(op0) && GET_NODE(ga->op1)->get_kind() == plus_expr_K)
                  {
                     integer_cst *int_const= GetPointer<integer_cst>(op0);
                     unsigned long long int value_int = static_cast<unsigned long long int>(int_const->value);
                     for(unsigned int index=0; index<64 && value_int != 0;++index)
                     {
                        if(value_int&(1ULL<<index))
                           break;
                        else
                           ++trailing_zero_op0;
                     }
                  }
                  if(GetPointer<ssa_name>(op1))
                  {
                     const std::string & bit_values_op1 = GetPointer<ssa_name>(op1)->bit_values;
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(me->op1)) +" bitstring: " + bit_values_op1);
                     for(auto current_el : boost::adaptors::reverse(bit_values_op1))
                     {
                        if(current_el == '0' || current_el == 'X')
                           ++trailing_zero_op1;
                        else
                           break;
                     }
                  }
                  else if(GetPointer<integer_cst>(op1))
                  {
                     integer_cst *int_const= GetPointer<integer_cst>(op1);
                     unsigned long long int value_int = static_cast<unsigned long long int>(int_const->value);
                     for(unsigned int index=0; index<64 && value_int != 0;++index)
                     {
                        if(value_int&(1ULL<<index))
                           break;
                        else
                           ++trailing_zero_op1;
                     }
                  }
                  if(trailing_zero_op0 != 0 || trailing_zero_op1 != 0)
                  {
#ifndef NDEBUG
                     function_behavior->RegisterTransformation(GetName(), stmt);
#endif
                     tree_manipulationRef IRman=tree_manipulationRef(new tree_manipulation(TM,parameters));
                     const std::string srcp_default = ga->include_name + ":" + STR(ga->line_number) + ":" + STR(ga->column_number);
                     bool is_first_max = trailing_zero_op0 > trailing_zero_op1;
                     unsigned int shift_const = is_first_max ? trailing_zero_op0 : trailing_zero_op1;
                     INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---Bit Value Opt: " + (GET_NODE(ga->op1)->get_kind() == plus_expr_K?std::string("plus_expr"):std::string("minus_expr")) + ", nbits = " + STR(shift_const));
                     unsigned int op0_const_rshift_id = TM->new_tree_node_id();
                     tree_nodeRef b_node = is_first_max ? me->op1 : me->op0;
                     tree_nodeRef op_const_node = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(shift_const), op0_const_rshift_id);
                     unsigned int bmask_const_id = TM->new_tree_node_id();
                     tree_nodeRef b_mask = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>((1ULL<<shift_const)-1), bmask_const_id);
                     bool is_op0_null = false;
                     bool is_op1_null = false;

                     if(is_op0_const)
                     {
                        integer_cst *int_const= GetPointer<integer_cst>(op0);
                        unsigned int op0_const_id = TM->new_tree_node_id();
                        if(tree_helper::is_int(TM, GET_INDEX_NODE(me->op0)))
                        {
                           if(static_cast<long long int>(int_const->value>>shift_const) == 0)
                              is_op0_null = GET_NODE(ga->op1)->get_kind() == plus_expr_K;
                           TM->ReplaceTreeNode(stmt, me->op0, IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(int_const->value>>shift_const), op0_const_id));
                        }
                        else
                        {
                           if(static_cast<unsigned long long int>(int_const->value>>shift_const) == 0)
                              is_op0_null = GET_NODE(ga->op1)->get_kind() == plus_expr_K;
                           TM->ReplaceTreeNode(stmt, me->op0, IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(static_cast<unsigned long long int>(int_const->value)>>shift_const), op0_const_id));
                        }
                     }
                     else
                     {
                        std::string resulting_bit_values;
                        THROW_ASSERT(GetPointer<ssa_name>(op0), "expected an SSA name");
                        if((GetPointer<ssa_name>(op0)->bit_values.size()-shift_const)>0)
                           resulting_bit_values = GetPointer<ssa_name>(op0)->bit_values.substr(0, GetPointer<ssa_name>(op0)->bit_values.size()-shift_const);
                        else if(tree_helper::is_int(TM, GET_INDEX_NODE(me->op0)))
                           resulting_bit_values = GetPointer<ssa_name>(op0)->bit_values.substr(0, 1);
                        else
                           resulting_bit_values = "0";
                        if(resulting_bit_values == "0" && GET_NODE(ga->op1)->get_kind() == plus_expr_K)
                           is_op0_null = true;
                        else
                        {
                           tree_nodeRef op0_expr = IRman->create_binary_operation(ga_op_type, me->op0, op_const_node, srcp_default, rshift_expr_K);
                           tree_nodeRef op0_ga = create_ga(IRman, Scpe, ga_op_type, op0_expr, B_id, srcp_default);
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(op0_ga));
                           B->PushBefore(op0_ga, stmt);
                           tree_nodeRef op0_ga_var = GetPointer<gimple_assign>(GET_NODE(op0_ga))->op0;
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Replacing " + me->op0->ToString() + " with " + op0_ga_var->ToString() + " in " + stmt->ToString());
                           TM->ReplaceTreeNode(stmt, me->op0, op0_ga_var);
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Replaced " + me->op0->ToString() + " with " + op0_ga_var->ToString() + " in " + stmt->ToString());
#if HAVE_FROM_DISCREPANCY_BUILT
                           /*
                            * for discrepancy analysis, the ssa assigned by this
                            * bitshift must not be checked if it was applied to
                            * a variable marked as address.
                            */
                           if (parameters->isOption(OPT_discrepancy) and
                                 parameters->getOption<bool>(OPT_discrepancy))
                           {
                              AppM->RDiscr->ssa_to_skip_if_address.insert(GET_NODE(op0_ga_var));
                           }
#endif
                           ///set the bit_values to the ssa var
                           ssa_name* op0_ssa = GetPointer<ssa_name>(GET_NODE(op0_ga_var));
                           op0_ssa->bit_values = resulting_bit_values;

                           PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(op0_ga_var)) +" bitstring: " + STR(op0_ssa->bit_values));
                        }
                     }

                     if(is_op1_const)
                     {
                        integer_cst *int_const= GetPointer<integer_cst>(op1);
                        unsigned int op1_const_id = TM->new_tree_node_id();
                        if(tree_helper::is_int(TM, GET_INDEX_NODE(me->op1)))
                        {
                           if(static_cast<long long int>(int_const->value>>shift_const) == 0)
                              is_op1_null = true;
                           TM->ReplaceTreeNode(stmt, me->op1, IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(int_const->value>>shift_const), op1_const_id));
                        }
                        else
                        {
                           if(static_cast<unsigned long long int>(int_const->value>>shift_const) == 0)
                              is_op1_null = true;
                           TM->ReplaceTreeNode(stmt, me->op1, IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(static_cast<unsigned long long int>(int_const->value)>>shift_const), op1_const_id));
                        }
                     }
                     else
                     {
                        std::string resulting_bit_values;
                        THROW_ASSERT(GetPointer<ssa_name>(op1), "expected an SSA name");
                        if((GetPointer<ssa_name>(op1)->bit_values.size()-shift_const)>0)
                           resulting_bit_values = GetPointer<ssa_name>(op1)->bit_values.substr(0, GetPointer<ssa_name>(op1)->bit_values.size()-shift_const);
                        else if(tree_helper::is_int(TM, GET_INDEX_NODE(me->op1)))
                           resulting_bit_values = GetPointer<ssa_name>(op1)->bit_values.substr(0, 1);
                        else
                           resulting_bit_values = "0";

                        if(resulting_bit_values == "0")
                           is_op1_null = true;
                        else
                        {
                           tree_nodeRef op1_expr = IRman->create_binary_operation(ga_op_type, me->op1, op_const_node, srcp_default, rshift_expr_K);
                           tree_nodeRef op1_ga = create_ga(IRman, Scpe, ga_op_type, op1_expr, B_id, srcp_default);
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(op1_ga));
                           B->PushBefore(op1_ga, stmt);
                           tree_nodeRef op1_ga_var = GetPointer<gimple_assign>(GET_NODE(op1_ga))->op0;
                           TM->ReplaceTreeNode(stmt, me->op1, op1_ga_var);
#if HAVE_FROM_DISCREPANCY_BUILT
                           /*
                            * for discrepancy analysis, the ssa assigned by this
                            * bitshift must not be checked if it was applied to
                            * a variable marked as address.
                            */
                           if (parameters->isOption(OPT_discrepancy) and
                                 parameters->getOption<bool>(OPT_discrepancy))
                           {
                              AppM->RDiscr->ssa_to_skip_if_address.insert(GET_NODE(op1_ga_var));
                           }
#endif
                           ///set the bit_values to the ssa var
                           ssa_name* op1_ssa = GetPointer<ssa_name>(GET_NODE(op1_ga_var));
                           op1_ssa->bit_values = resulting_bit_values;
                           PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(op1_ga_var)) +" bitstring: " + STR(op1_ssa->bit_values));
                        }
                     }

                     tree_nodeRef curr_ga;
                     if(is_op0_null)
                     {
                        curr_ga = create_ga(IRman, Scpe, ga_op_type, me->op1, B_id, srcp_default);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(curr_ga));
                     }
                     else if(is_op1_null)
                     {
                        curr_ga = create_ga(IRman, Scpe, ga_op_type, me->op0, B_id, srcp_default);
                     }
                     else
                     {
                        curr_ga = create_ga(IRman, Scpe, ga_op_type, ga->op1, B_id, srcp_default);
                     }
                     B->PushBefore(curr_ga, stmt);
                     tree_nodeRef curr_ga_var = GetPointer<gimple_assign>(GET_NODE(curr_ga))->op0;
#if HAVE_FROM_DISCREPANCY_BUILT
                     /*
                      * for discrepancy analysis, the ssa assigned by this
                      * bitshift must not be checked if it was applied to
                      * a variable marked as address.
                      */
                     if (parameters->isOption(OPT_discrepancy) and
                           parameters->getOption<bool>(OPT_discrepancy))
                     {
                        AppM->RDiscr->ssa_to_skip_if_address.insert(GET_NODE(curr_ga_var));
                     }
#endif
                     ///set the bit_values to the ssa var
                     ssa_name* op_ssa = GetPointer<ssa_name>(GET_NODE(curr_ga_var));
                     op_ssa->bit_values = ssa->bit_values.substr(0, ssa->bit_values.size()-shift_const);
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(curr_ga_var)) +" bitstring: " + STR(op_ssa->bit_values));

                     tree_nodeRef op_expr = IRman->create_binary_operation(ga_op_type, curr_ga_var, op_const_node, srcp_default, lshift_expr_K);
                     tree_nodeRef lshift_ga = create_ga(IRman, Scpe, ga_op_type, op_expr, B_id, srcp_default);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(lshift_ga));
                     B->PushBefore(lshift_ga, stmt);
                     tree_nodeRef lshift_ga_var = GetPointer<gimple_assign>(GET_NODE(lshift_ga))->op0;
                     ///set the bit_values to the ssa var
                     ssa_name* lshift_ssa = GetPointer<ssa_name>(GET_NODE(lshift_ga_var));
                     lshift_ssa->bit_values = ssa->bit_values.substr(0, ssa->bit_values.size()-shift_const);
                     while(lshift_ssa->bit_values.size() < ssa->bit_values.size())
                        lshift_ssa->bit_values.push_back('0');
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(lshift_ga_var)) +" bitstring: " + STR(lshift_ssa->bit_values));

                     bool do_final_or = false;
                     unsigned int n_iter = 0;
                     for(auto cur_bit : boost::adaptors::reverse(ssa->bit_values))
                     {
                        if(cur_bit == '1' || cur_bit == 'U')
                        {
                           do_final_or = true;
                           break;
                        }
                        n_iter++;
                        if(n_iter==shift_const)
                           break;
                     }

                     if(do_final_or)
                     {
#if HAVE_FROM_DISCREPANCY_BUILT
                        /*
                         * for discrepancy analysis, the ssa assigned by this
                         * bitshift must not be checked if it was applied to
                         * a variable marked as address.
                         */
                        if (parameters->isOption(OPT_discrepancy) and
                              parameters->getOption<bool>(OPT_discrepancy))
                        {
                           AppM->RDiscr->ssa_to_skip_if_address.insert(GET_NODE(lshift_ga_var));
                        }
#endif
                        unsigned int size_node_const_id = TM->new_tree_node_id();
                        tree_nodeRef size_node_val = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(shift_const), size_node_const_id);
                        if(GetPointer<integer_cst>(GET_NODE(b_node)))
                        {
                           integer_cst *int_const= GetPointer<integer_cst>(GET_NODE(b_node));
                           unsigned int b_node_const_id = TM->new_tree_node_id();
                           tree_nodeRef b_node_val = IRman->CreateIntegerCst(ga_op_type, static_cast<long long int>(static_cast<unsigned long long int>(int_const->value)&((1ULL<<shift_const)-1)), b_node_const_id);
                           TM->ReplaceTreeNode(stmt, ga->op1, IRman->create_ternary_operation(ga_op_type, lshift_ga_var, b_node_val, size_node_val, srcp_default, bit_ior_concat_expr_K));
                        }
                        else
                        {
                           tree_nodeRef band_expr = IRman->create_binary_operation(ga_op_type, b_node, b_mask, srcp_default, bit_and_expr_K);
                           tree_nodeRef band_ga = create_ga(IRman, Scpe, ga_op_type, band_expr, B_id, srcp_default);
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Created " + STR(band_ga));
                           B->PushBefore(band_ga, stmt);
                           tree_nodeRef band_ga_var = GetPointer<gimple_assign>(GET_NODE(band_ga))->op0;
#if HAVE_FROM_DISCREPANCY_BUILT
                           /*
                            * for discrepancy analysis, the ssa assigned by this
                            * bitshift must not be checked if it was applied to
                            * a variable marked as address.
                            */
                           if (parameters->isOption(OPT_discrepancy) and
                                 parameters->getOption<bool>(OPT_discrepancy))
                           {
                              AppM->RDiscr->ssa_to_skip_if_address.insert(GET_NODE(band_ga_var));
                           }
#endif
                           ///set the bit_values to the ssa var
                           ssa_name* band_ssa = GetPointer<ssa_name>(GET_NODE(band_ga_var));
                           for(auto cur_bit : boost::adaptors::reverse(ssa->bit_values))
                           {
                              band_ssa->bit_values = cur_bit+band_ssa->bit_values;
                              if(band_ssa->bit_values.size()==shift_const)
                                 break;
                           }
                           band_ssa->bit_values ="0"+band_ssa->bit_values;
                           PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Var_uid: "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(GET_INDEX_NODE(band_ga_var)) +" bitstring: " + STR(band_ssa->bit_values));

                           tree_nodeRef res_expr = IRman->create_ternary_operation(ga_op_type, lshift_ga_var, band_ga_var, size_node_val, srcp_default, bit_ior_concat_expr_K);
                           TM->ReplaceTreeNode(stmt, ga->op1, res_expr);
                        }
                     }
                     else
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Final or not performed: ");
                        TM->ReplaceTreeNode(stmt, ga->op1, lshift_ga_var);
                     }
                     ///set uses of stmt
                  }
               }
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Statement analyzed " + GET_NODE(stmt)->ToString());
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--BB analyzed " + STR(B_id));
   }
}

DesignFlowStep_Status
Bit_Value_opt::InternalExec ()
{
   if(debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      PrintTreeManager(true);
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " --------- BIT_VALUE_OPT ---------- ");
   tree_managerRef TM = AppM->get_tree_manager();

   tree_nodeRef tn = TM->get_tree_node_const(function_id);
   //tree_nodeRef Scpe = TM->GetTreeReindex(function_id);
   function_decl * fd = GetPointer<function_decl>(tn);
   THROW_ASSERT(fd && fd->body, "Node is not a function or it hasn't a body");
   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));
   THROW_ASSERT(sl, "Body is not a statement_list");
   /// for each basic block B in CFG do > Consider all blocks successively
   restart_dead_code = false;
   optimize(sl, TM);
   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      PrintTreeManager(false);
   }
   restart_dead_code ? function_behavior->UpdateBBVersion() : 0;
   return restart_dead_code ? DesignFlowStep_Status::SUCCESS : DesignFlowStep_Status::UNCHANGED;

}
