/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file IR_lowering.cpp
 * @brief Decompose some complex gimple statements into set of simple operations.
 *
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/
///Autoheader include
#include "config_HAVE_BAMBU_BUILT.hpp"

///Header include
#include "IR_lowering.hpp"

///. include
#include "Parameter.hpp"

///behavior includes
#include "application_manager.hpp"
#include "function_behavior.hpp"

///design_flows includes
#include "design_flow_graph.hpp"
#include "design_flow_manager.hpp"

///design_flows/technology includes
#include "technology_flow_step.hpp"
#include "technology_flow_step_factory.hpp"

///HLS includes
#include "hls.hpp"
#include "hls_manager.hpp"
#include "hls_target.hpp"
#include "hls_step.hpp"

///HLS/scheduling include
#include "schedule.hpp"

///STD include
#include <fstream>

#if HAVE_BAMBU_BUILT
///technology include
#include "technology_manager.hpp"

///technology/physical_library/modes include
#include "time_model.hpp"
#endif

///tree includes
#include "behavioral_helper.hpp"

///technology/physical_library include
#include "technology_node.hpp"

///tree includes
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_manipulation.hpp"
#include "tree_reindex.hpp"

///utility include
#include "math_function.hpp"

tree_nodeRef IR_lowering::CreateGimpleAssign(const tree_nodeRef type, const tree_nodeRef op, unsigned int bb_index, const std::string & srcp_default)
{
   tree_nodeRef ssa_vd = tree_man->create_ssa_name(tree_nodeRef(), type);
   auto ret_value = tree_man->create_gimple_modify_stmt(ssa_vd, op, srcp_default, bb_index);
   if(schedule)
      schedule->UpdateTime(ret_value->index);
   return ret_value;
}

/// the code for lowering of div, mult and rem comes from GCC sources (expmed.c)

/// Test whether a value is zero of a power of two.
#define EXACT_POWER_OF_2_OR_ZERO_P(x) (((x) & ((x) - 1)) == 0)

/// Given X, an unsigned number, return the largest int Y such that 2**Y <= X. If X is 0, return -1.
static int
floor_log2 (unsigned long long int x)
{
  int t = 0;

  if (x == 0)
    return -1;

  if (x >= (static_cast<unsigned long long int>(1)) << (t + 32))
      t += 32;
  if (x >= (static_cast<unsigned long long int>(1)) << (t + 16))
    t += 16;
  if (x >= (static_cast<unsigned long long int>(1)) << (t + 8))
    t += 8;
  if (x >= (static_cast<unsigned long long int>(1)) << (t + 4))
    t += 4;
  if (x >= (static_cast<unsigned long long int>(1)) << (t + 2))
    t += 2;
  if (x >= (static_cast<unsigned long long int>(1)) << (t + 1))
    t += 1;

  return t;
}
/// Return the smallest n such that 2**n >= X.
static int
ceil_log2 (unsigned long long int x)
{
  return floor_log2 (x - 1) + 1;
}

/** Return the logarithm of X, base 2, considering X unsigned,
   if X is a power of 2.  Otherwise, returns -1.  */
static int
exact_log2 (unsigned long long int x)
{
  if (x != (x & -x))
    return -1;
  return floor_log2 (x);
}


/// Compute the inverse of X mod 2**n, i.e., find Y such that X * Y is
/// congruent to 1 (mod 2**N).
static unsigned long long int
invert_mod2n (unsigned long long int x, unsigned int n)
{
  /* Solve x*y == 1 (mod 2^n), where x is odd.  Return y.  */

  /* The algorithm notes that the choice y = x satisfies
     x*y == 1 mod 2^3, since x is assumed odd.
     Each iteration doubles the number of bits of significance in y.  */

  unsigned long long int mask;
  unsigned long long int y = x;
  unsigned int nbit = 3;

  mask = (n == sizeof(long long int)*8
      ? ~static_cast<unsigned long long int>(0)
      : (static_cast<unsigned long long int>(1) << n) - 1);

  while (nbit < n)
    {
      y = y * (2 - x*y) & mask;		/* Modulo 2^N */
      nbit *= 2;
    }
  return y;
}

enum alg_code {
  alg_unknown,
  alg_zero,
  alg_m, alg_shift,
  alg_add_t_m2,
  alg_sub_t_m2,
  alg_add_factor,
  alg_sub_factor,
  alg_add_t2_m,
  alg_sub_t2_m,
  alg_impossible
};

/** This structure holds the "cost" of a multiply sequence.  The
   "cost" field holds the total cost of every operator in the
   synthetic multiplication sequence, hence cost(a op b) is defined
   as cost(op) + cost(a) + cost(b), where cost(leaf) is zero.
   The "latency" field holds the minimum possible latency of the
   synthetic multiply, on a hypothetical infinitely parallel CPU.
   This is the critical path, or the maximum height, of the expression
   tree which is the sum of costs on the most expensive path from
   any leaf to the root.  Hence latency(a op b) is defined as zero for
   leaves and cost(op) + max(latency(a), latency(b)) otherwise.  */
struct mult_cost {
  short cost;     /** Total cost of the multiplication sequence.  */
  short latency;  /** The latency of the multiplication sequence.  */
};


/** This structure records a sequence of operations.
   `ops' is the number of operations recorded.
   `cost' is their total cost.
   The operations are stored in `op' and the corresponding
   logarithms of the integer coefficients in `log'.

   These are the operations:
   alg_zero		total := 0;
   alg_m		total := multiplicand;
   alg_shift		total := total * coeff
   alg_add_t_m2		total := total + multiplicand * coeff;
   alg_sub_t_m2		total := total - multiplicand * coeff;
   alg_add_factor	total := total * coeff + total;
   alg_sub_factor	total := total * coeff - total;
   alg_add_t2_m		total := total * coeff + multiplicand;
   alg_sub_t2_m		total := total * coeff - multiplicand;

   The first operand must be either alg_zero or alg_m.  */
struct algorithm
{
  struct mult_cost cost;
  short ops;
  /* The size of the OP and LOG fields are not directly related to the
     word size, but the worst-case algorithms will be if we have few
     consecutive ones or zeros, i.e., a multiplicand like 10101010101...
     In that case we will generate shift-by-2, add, shift-by-2, add,...,
     in total wordsize operations.  */
  enum alg_code op[64];
  char log[64];
};

/** This macro is used to compare a pointer to a mult_cost against an
   single integer "cost" value.  This is equivalent to the macro
   CHEAPER_MULT_COST(X,Z) where Z = {Y,Y}.  */
#define MULT_COST_LESS(X,Y) ((X).cost < (Y)	\
                 || ((X).cost == (Y) && (X).latency < (Y)))

/** This macro is used to compare two mult_costs against
   each other.  The macro returns true if X is cheaper than Y.
   Currently, the cheaper of two mult_costs is the one with the
   lower "cost".  If "cost"s are tied, the lower latency is cheaper.  */
#define CHEAPER_MULT_COST(X,Y)  ((X).cost < (Y).cost		\
                 || ((X).cost == (Y).cost	\
                     && (X).latency < (Y).latency))


static std::unordered_map<std::pair<unsigned int,unsigned long long>, std::pair<enum alg_code, struct mult_cost> > alg_hash;

/** Compute and return the best algorithm for multiplying by T.
   The algorithm must cost less than cost_limit
   If retval.cost >= COST_LIMIT, no algorithm was found and all
   other field of the returned struct are undefined.
   MODE is the machine mode of the multiplication.  */
static void
synth_mult (struct algorithm &alg_out, unsigned long long t,
        const struct mult_cost &cost_limit, unsigned data_bitsize, tree_managerRef &TM)
{
   int m;
   struct algorithm alg_in, best_alg;
   struct mult_cost best_cost;
   struct mult_cost new_limit;
   short op_cost, op_latency;
   unsigned long long int orig_t = t;
   unsigned long long int q;
   int maxm = static_cast<int>(std::min (32u, data_bitsize));
   bool cache_hit = false;
   enum alg_code cache_alg = alg_zero;

   /** Indicate that no algorithm is yet found.  If no algorithm
      is found, this value will be returned and indicate failure.  */
   alg_out.cost.cost = static_cast<short>(cost_limit.cost + 1);
   alg_out.cost.latency = static_cast<short>(cost_limit.latency + 1);

   if (cost_limit.cost < 0
       || (cost_limit.cost == 0 && cost_limit.latency <= 0))
      return;

   /** t == 1 can be done in zero cost.  */
   if (t == 1)
   {
      alg_out.ops = 1;
      alg_out.cost.cost = 0;
      alg_out.cost.latency = 0;
      alg_out.op[0] = alg_m;
      return;
   }

   /** t == 0 sometimes has a cost.  If it does and it exceeds our limit,
      fail now.  */
   if (t == 0)
   {
      alg_out.ops = 1;
      alg_out.cost.cost = 0/*zero_cost[speed]*/;
      alg_out.cost.latency = 0/*zero_cost[speed]*/;
      alg_out.op[0] = alg_zero;
      return;
   }
   best_cost = cost_limit;
   if(alg_hash.find(std::make_pair(data_bitsize,t)) != alg_hash.end())
   {
      std::pair<enum alg_code, struct mult_cost> res = alg_hash.find(std::make_pair(data_bitsize,t))->second;
      cache_alg = res.first;
      if (cache_alg == alg_impossible)
      {
         /** The cache tells us that it's impossible to synthesize
               multiplication by T within alg_hash[hash_index].cost.  */
         if (!CHEAPER_MULT_COST (res.second, cost_limit))
            /** COST_LIMIT is at least as restrictive as the one
                 recorded in the hash table, in which case we have no
                 hope of synthesizing a multiplication.  Just
                 return.  */
            return;

         /** If we get here, COST_LIMIT is less restrictive than the
               one recorded in the hash table, so we may be able to
               synthesize a multiplication.  Proceed as if we didn't
               have the cache entry.  */
      }
      else
      {
         if (CHEAPER_MULT_COST (cost_limit, res.second))
            /** The cached algorithm shows that this multiplication
                 requires more cost than COST_LIMIT.  Just return.  This
                 way, we don't clobber this cache entry with
                 alg_impossible but retain useful information.  */
            return;
         cache_hit = true;

         switch (cache_alg)
         {
            case alg_shift:
               goto do_alg_shift;

            case alg_add_t_m2:
            case alg_sub_t_m2:
               goto do_alg_addsub_t_m2;

            case alg_add_factor:
            case alg_sub_factor:
               goto do_alg_addsub_factor;

            case alg_add_t2_m:
               goto do_alg_add_t2_m;

            case alg_sub_t2_m:
               goto do_alg_sub_t2_m;
            case alg_impossible:
            case alg_m:
            case alg_unknown:
            case alg_zero:
            default:
               THROW_ERROR ("condition not expected");
         }
      }
   }

   /** If we have a group of zero bits at the low-order part of T, try
      multiplying by the remaining bits and then doing a shift.  */

   if ((t & 1) == 0)
   {
      do_alg_shift:
      m = floor_log2 (t & -t);	/* m = number of low zero bits */
      if (m < maxm)
      {
         q = t >> m;
         /* The function expand_shift will choose between a shift and
                  a sequence of additions, so the observed cost is given as
                  MIN (m * add_cost[speed][mode], shift_cost[speed][mode][m]).  */
         op_cost = static_cast<short>(m * 1/*add_cost[speed][mode]*/);
         if (0/*shift_cost[speed][mode][m]*/ < op_cost)
            op_cost = 0/*shift_cost[speed][mode][m]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, q, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_shift;
         }

         /** See if treating ORIG_T as a signed number yields a better
          sequence.  Try this sequence only for a negative ORIG_T
          as it would be useless for a non-negative ORIG_T.  */
         if (static_cast<long long int>(orig_t) < 0)
         {
            /** Shift ORIG_T as follows because a right shift of a
          negative-valued signed type is implementation
          defined.  */
            q = ~(~orig_t >> m);
            /** The function expand_shift will choose between a shift
          and a sequence of additions, so the observed cost is
          given as MIN (m * add_cost[speed][mode], shift_cost[speed][mode][m]).  */
            op_cost = static_cast<short>(m * 1)/*add_cost[speed][mode]*/;
            if (0 /*shift_cost[speed][mode][m]*/ < op_cost)
               op_cost = 0/*shift_cost[speed][mode][m]*/;
            new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
            new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
            synth_mult (alg_in, q, new_limit, data_bitsize, TM);

            alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
            alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
            if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
            {
               best_cost = alg_in.cost;
               std::swap(alg_in, best_alg);
               best_alg.log[best_alg.ops] = static_cast<char>(m);
               best_alg.op[best_alg.ops] = alg_shift;
            }
         }
      }
      if (cache_hit)
         goto done;
   }

   /** If we have an odd number, add or subtract one.  */
   if ((t & 1) != 0)
   {
      unsigned long long int w;

      do_alg_addsub_t_m2:
      for (w = 1; (w & t) != 0; w <<= 1)
         ;
      /** If T was -1, then W will be zero after the loop.  This is another
      case where T ends with ...111.  Handling this with (T + 1) and
      subtract 1 produces slightly better code and results in algorithm
      selection much faster than treating it like the ...0111 case
      below.  */
      if (w == 0
          || (w > 2
              /** Reject the case where t is 3.
                  Thus we prefer addition in that case.  */
              && t != 3))
      {
         /** T ends with ...111.  Multiply by (T + 1) and subtract 1.  */

         op_cost = 1/*add_cost[speed][mode]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, t + 1, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = 0;
            best_alg.op[best_alg.ops] = alg_sub_t_m2;
         }
      }
      else
      {
         /** T ends with ...01 or ...011.  Multiply by (T - 1) and add 1.  */

         op_cost = 1/*add_cost[speed][mode]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, t - 1, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = 0;
            best_alg.op[best_alg.ops] = alg_add_t_m2;
         }
      }

      /** We may be able to calculate a * -7, a * -15, a * -31, etc
      quickly with a - a * n for some appropriate constant n.  */
      m = exact_log2 (-orig_t + 1);
      if (m >= 0 && m < maxm)
      {
         op_cost = 1/*shiftsub1_cost[speed][mode][m]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, static_cast<unsigned long long int>(-orig_t + 1) >> m, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_sub_t_m2;
         }
      }

      if (cache_hit)
         goto done;
   }

   /** Look for factors of t of the form
      t = q(2**m +- 1), 2 <= m <= floor(log2(t - 1)).
      If we find such a factor, we can multiply by t using an algorithm that
      multiplies by q, shift the result by m and add/subtract it to itself.

      We search for large factors first and loop down, even if large factors
      are less probable than small; if we find a large factor we will find a
      good sequence quickly, and therefore be able to prune (by decreasing
      COST_LIMIT) the search.  */

   do_alg_addsub_factor:
   for (m = floor_log2 (t - 1); m >= 2; m--)
   {
      unsigned long long int d;

      d = (static_cast<unsigned long long int>(1) << m) + 1;
      if (t % d == 0 && t > d && m < maxm
          && (!cache_hit || cache_alg == alg_add_factor))
      {
         /** If the target has a cheap shift-and-add instruction use
          that in preference to a shift insn followed by an add insn.
          Assume that the shift-and-add is "atomic" with a latency
          equal to its cost, otherwise assume that on superscalar
          hardware the shift may be executed concurrently with the
          earlier steps in the algorithm.  */
         op_cost = 1/*add_cost[speed][mode]*/ + 0/*shift_cost[speed][mode][m]*/;
         if (1/*shiftadd_cost[speed][mode][m]*/ < op_cost)
         {
            op_cost = 1/*shiftadd_cost[speed][mode][m]*/;
            op_latency = op_cost;
         }
         else
            op_latency = 1/*add_cost[speed][mode]*/;

         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_latency);
         synth_mult (alg_in, t / d, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_latency);
         if (alg_in.cost.latency < op_cost)
            alg_in.cost.latency = op_cost;
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_add_factor;
         }
         /** Other factors will have been taken care of in the recursion.  */
         break;
      }

      d = (static_cast<unsigned long long int>(1) << m) - 1;
      if (t % d == 0 && t > d && m < maxm
          && (!cache_hit || cache_alg == alg_sub_factor))
      {
         /** If the target has a cheap shift-and-subtract insn use
          that in preference to a shift insn followed by a sub insn.
          Assume that the shift-and-sub is "atomic" with a latency
          equal to it's cost, otherwise assume that on superscalar
          hardware the shift may be executed concurrently with the
          earlier steps in the algorithm.  */
         op_cost = 1/*add_cost[speed][mode]*/ + 0/*shift_cost[speed][mode][m]*/;
         if (1/*shiftsub0_cost[speed][mode][m]*/ < op_cost)
         {
            op_cost = 1/*shiftsub0_cost[speed][mode][m]*/;
            op_latency = op_cost;
         }
         else
            op_latency = 1/*add_cost[speed][mode]*/;

         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_latency);
         synth_mult (alg_in, t / d, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_latency);
         if (alg_in.cost.latency < op_cost)
            alg_in.cost.latency = op_cost;
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_sub_factor;
         }
         break;
      }
   }
   if (cache_hit)
      goto done;

   /** Try shift-and-add (load effective address) instructions,
      i.e. do a*3, a*5, a*9.  */
   if ((t & 1) != 0)
   {
   do_alg_add_t2_m:
      q = t - 1;
      q = q & -q;
      m = exact_log2 (q);
      if (m >= 0 && m < maxm)
      {
         op_cost = 1/*shiftadd_cost[speed][mode][m]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, (t - 1) >> m, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_add_t2_m;
         }
      }
      if (cache_hit)
         goto done;

      do_alg_sub_t2_m:
      q = t + 1;
      q = q & -q;
      m = exact_log2 (q);
      if (m >= 0 && m < maxm)
      {
         op_cost = 1/*shiftsub0_cost[speed][mode][m]*/;
         new_limit.cost = static_cast<short>(best_cost.cost - op_cost);
         new_limit.latency = static_cast<short>(best_cost.latency - op_cost);
         synth_mult (alg_in, (t + 1) >> m, new_limit, data_bitsize, TM);

         alg_in.cost.cost = static_cast<short>(alg_in.cost.cost + op_cost);
         alg_in.cost.latency = static_cast<short>(alg_in.cost.latency + op_cost);
         if (CHEAPER_MULT_COST (alg_in.cost, best_cost))
         {
            best_cost = alg_in.cost;
            std::swap(alg_in, best_alg);
            best_alg.log[best_alg.ops] = static_cast<char>(m);
            best_alg.op[best_alg.ops] = alg_sub_t2_m;
         }
      }
      if (cache_hit)
         goto done;
   }

   done:
   /** If best_cost has not decreased, we have not found any algorithm.  */
   if (!CHEAPER_MULT_COST (best_cost, cost_limit))
   {
      /** We failed to find an algorithm.  Record alg_impossible for
      this case (that is, <T, MODE, COST_LIMIT>) so that next time
      we are asked to find an algorithm for T within the same or
      lower COST_LIMIT, we can immediately return to the
      caller.  */
      alg_hash[std::make_pair(data_bitsize, t)] = std::make_pair(alg_impossible, cost_limit);
      return;
   }

   /** Cache the result.  */
   if (!cache_hit)
   {
      mult_cost last_limit = {best_cost.cost,best_cost.latency};
      alg_hash[std::make_pair(data_bitsize, t)] = std::make_pair(best_alg.op[best_alg.ops], last_limit);
   }

   /** If we are getting a too long sequence for `struct algorithm'
      to record, make this search fail.  */
   if (best_alg.ops == 64)
      return;

   /** Copy the algorithm from temporary space to the space at alg_out.
      We avoid using structure assignment because the majority of
      best_alg is normally undefined, and this is a critical function.  */
   alg_out = best_alg;
   alg_out.ops = static_cast<short>(best_alg.ops + 1);

}

/** Find the cheapest way of multiplying a value of mode MODE by VAL.
   Try three variations:
       - a shift/add sequence based on VAL itself
       - a shift/add sequence based on -VAL, followed by a negation
       - a shift/add sequence based on VAL - 1, followed by an addition.

   Return true if the cheapest of these cost less than MULT_COST,
   describing the algorithm in *ALG and final fixup in *VARIANT.  */
static bool
choose_mult_variant (unsigned int data_bitsize, long long int val,
             struct algorithm &alg, enum mult_variant &variant, short int Mult_cost, tree_managerRef &TM)
{
   struct algorithm alg2;
   struct mult_cost limit;
   short int op_cost;

   /** Fail quickly for impossible bounds.  */
   if (Mult_cost < 0)
     return false;

   /** Ensure that Mult_cost provides a reasonable upper bound.
      Any constant multiplication can be performed with less
      than 2 * bits additions.  */
   op_cost = static_cast<short>(2 * data_bitsize * 1/*add_cost[speed][mode]*/);
   if (Mult_cost > op_cost)
     Mult_cost = op_cost;

   variant = basic_variant;
   limit.cost = Mult_cost;
   limit.latency = Mult_cost;
   synth_mult (alg, static_cast<unsigned long long int>(val), limit, data_bitsize, TM);

   /* This works only if the inverted value actually fits in an
      `unsigned int' */
   if (8*sizeof(int) >= data_bitsize)
     {
       op_cost = 1/*neg_cost[speed][mode]*/;
       if (MULT_COST_LESS (alg.cost, Mult_cost))
     {
       limit.cost = static_cast<short>(alg.cost.cost - op_cost);
       limit.latency = static_cast<short>(alg.cost.latency - op_cost);
     }
       else
     {
       limit.cost = static_cast<short>(Mult_cost - op_cost);
       limit.latency = static_cast<short>(Mult_cost - op_cost);
     }

       synth_mult (alg2, static_cast<unsigned long long int>(-val), limit, data_bitsize, TM);
       alg2.cost.cost = static_cast<short>(alg2.cost.cost + op_cost);
       alg2.cost.latency = static_cast<short>(alg2.cost.latency + op_cost);
       if (CHEAPER_MULT_COST (alg2.cost, alg.cost))
          alg = alg2, variant = negate_variant;
     }

   /* This proves very useful for division-by-constant.  */
   op_cost = 1/*add_cost[speed][mode]*/;
   if (MULT_COST_LESS (alg.cost, Mult_cost))
     {
       limit.cost = static_cast<short>(alg.cost.cost - op_cost);
       limit.latency = static_cast<short>(alg.cost.latency - op_cost);
     }
   else
     {
       limit.cost = static_cast<short>(Mult_cost - op_cost);
       limit.latency = static_cast<short>(Mult_cost - op_cost);
     }

   synth_mult (alg2, static_cast<unsigned long long int>(val - 1), limit, data_bitsize, TM);
   alg2.cost.cost = static_cast<short>(alg2.cost.cost + op_cost);
   alg2.cost.latency = static_cast<short>(alg2.cost.latency + op_cost);
   if (CHEAPER_MULT_COST (alg2.cost, alg.cost))
     alg = alg2, variant = add_variant;

   return MULT_COST_LESS (alg.cost, Mult_cost);
 }

tree_nodeRef IR_lowering::expand_mult_const (tree_nodeRef op0, unsigned long long int val, const struct algorithm &alg, enum mult_variant &variant, const tree_nodeRef stmt, const blocRef block, tree_nodeRef &type, const std::string srcp_default)
{
   long long int val_so_far = 0;
   tree_nodeRef accum, tem;
   int opno;
   unsigned data_bitsize = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type)));
   unsigned long long int data_mask = data_bitsize >= 64 ? ~ static_cast<unsigned long long int>(0) : (static_cast<unsigned long long int>(1) << data_bitsize) -1;
   tree_nodeRef tem_ga;
   unsigned int integer_cst0_id = TM->new_tree_node_id();
   tree_nodeRef COST0 = tree_man->CreateIntegerCst(type, 0, integer_cst0_id);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Expanding " + op0->ToString());


   /* ACCUM starts out either as OP0 or as a zero, depending on
      the first operation.  */

   if (alg.op[0] == alg_zero)
   {
      accum = COST0;
      val_so_far = 0;
   }
   else if (alg.op[0] == alg_m)
   {
      accum = op0;
      val_so_far = 1;
   }
   else
      THROW_ERROR("condition not expected");

   for (opno = 1; opno < alg.ops; opno++)
   {
      int log = alg.log[opno];
      tree_nodeRef log_node;

      if(log == 0)
         log_node = COST0;
      else
      {
         unsigned int log_id = TM->new_tree_node_id();
         log_node = tree_man->CreateIntegerCst(type, static_cast<long long int>(log), log_id);
      }


      switch (alg.op[opno])
      {
         case alg_shift:
            tem = tree_man->create_binary_operation(type, accum, log_node, srcp_default, lshift_expr_K);
            tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
            block->PushBefore(tem_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            val_so_far <<= log;
            break;

         case alg_add_t_m2:
            if(log_node != COST0)
            {
               tem = tree_man->create_binary_operation(type, op0, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               tem = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               tem = op0;
            if(accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, accum, tem, srcp_default, plus_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               accum = tem;
            val_so_far += static_cast<long long int>(1) << log;
            break;

         case alg_sub_t_m2:
            if(log_node != COST0)
            {
               tem = tree_man->create_binary_operation(type, op0, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               tem = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               tem = op0;
            if(accum != COST0)
               accum = tree_man->create_binary_operation(type, accum, tem, srcp_default, minus_expr_K);
            else
               accum = tree_man->create_unary_operation(type, tem, srcp_default, negate_expr_K);
            tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
            block->PushBefore(tem_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            val_so_far -= static_cast<long long int>(1) << log;
            break;

         case alg_add_t2_m:
            if(log_node != COST0 && accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, accum, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            if(accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, accum, op0, srcp_default, plus_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               accum = op0;
            val_so_far = (val_so_far << log) + 1;
            break;

         case alg_sub_t2_m:
            if(log_node != COST0 && accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, accum, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            if(accum != COST0)
               accum = tree_man->create_binary_operation(type, accum, op0, srcp_default, minus_expr_K);
            else
               accum = tree_man->create_unary_operation(type, op0, srcp_default, negate_expr_K);
            tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
            block->PushBefore(tem_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            val_so_far = (val_so_far << log) - 1;
            break;

         case alg_add_factor:
            if(log_node != COST0 && accum != COST0)
            {
               tem = tree_man->create_binary_operation(type, accum, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               tem = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               tem = accum;
            if(accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, accum, tem, srcp_default, plus_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               accum = tem;
            val_so_far += val_so_far << log;
            break;

         case alg_sub_factor:
            if(log_node != COST0 && accum != COST0)
            {
               tem = tree_man->create_binary_operation(type, accum, log_node, srcp_default, lshift_expr_K);
               tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               tem = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               tem = accum;
            if(accum != COST0)
            {
               accum = tree_man->create_binary_operation(type, tem, accum, srcp_default, minus_expr_K);
               tem_ga = CreateGimpleAssign(type, accum, block->number, srcp_default);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
               block->PushBefore(tem_ga, stmt);
               accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
            }
            else
               accum = tem;
            val_so_far = (val_so_far << log) - val_so_far;
            break;
         case alg_impossible:
         case alg_m:
         case alg_unknown:
         case alg_zero:
         default:
            THROW_UNREACHABLE("");
      }
   }

   if (variant == negate_variant)
   {
      val_so_far = -val_so_far;
      tem = tree_man->create_unary_operation(type, accum, srcp_default, negate_expr_K);
      tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
      block->PushBefore(tem_ga, stmt);
      accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
   }
   else if (variant == add_variant)
   {
      val_so_far = val_so_far + 1;
      tem = tree_man->create_binary_operation(type, accum, op0, srcp_default, plus_expr_K);
      tem_ga = CreateGimpleAssign(type, tem, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), tem_ga);
#endif
      block->PushBefore(tem_ga, stmt);
      accum = GetPointer<gimple_assign>(GET_NODE(tem_ga))->op0;
   }

   /** Compare only the bits of val and val_so_far that are significant
     in the result mode, to avoid sign-/zero-extension confusion.  */
   val = val & data_mask;
   val_so_far = val_so_far & static_cast<long long int>(data_mask);
   THROW_ASSERT (val == static_cast<unsigned long long int>(val_so_far), "unexpected difference");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Expanded " + op0->ToString());

   return accum;

}

tree_nodeRef IR_lowering::expand_smod_pow2 (tree_nodeRef op0, unsigned long long int  d, const tree_nodeRef stmt, const blocRef block, tree_nodeRef &type, const std::string srcp_default)
{
   unsigned long long int  masklow;
   int logd;

   logd = floor_log2 (d);
   unsigned int integer_cst0_id = TM->new_tree_node_id();
   tree_nodeRef const0 = tree_man->CreateIntegerCst(type, 0, integer_cst0_id);

   tree_nodeRef bt =  tree_man->create_boolean_type();
   tree_nodeRef cond_op0 = tree_man->create_binary_operation(bt, op0, const0, srcp_default, lt_expr_K);
   tree_nodeRef signmask_ga = CreateGimpleAssign(type, cond_op0, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), signmask_ga);
#endif
   block->PushBefore(signmask_ga, stmt);
   tree_nodeRef signmask_var = GetPointer<gimple_assign>(GET_NODE(signmask_ga))->op0;

   masklow = (static_cast<unsigned long long int>(1) << logd) - 1;
   unsigned int integer_masklow_id = TM->new_tree_node_id();
   tree_nodeRef Constmasklow = tree_man->CreateIntegerCst(type, static_cast<long long int>(masklow), integer_masklow_id);

   tree_nodeRef temp = tree_man->create_binary_operation(type, op0, signmask_var, srcp_default, bit_xor_expr_K);
   tree_nodeRef temp_ga = CreateGimpleAssign(type, temp, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), temp_ga);
#endif
   block->PushBefore(temp_ga, stmt);
   tree_nodeRef temp_var = GetPointer<gimple_assign>(GET_NODE(temp_ga))->op0;

   temp = tree_man->create_binary_operation(type, temp_var, signmask_var, srcp_default, minus_expr_K);
   temp_ga = CreateGimpleAssign(type, temp, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), temp_ga);
#endif
   block->PushBefore(temp_ga, stmt);
   temp_var = GetPointer<gimple_assign>(GET_NODE(temp_ga))->op0;

   temp = tree_man->create_binary_operation(type, temp_var, Constmasklow, srcp_default, bit_and_expr_K);
   temp_ga = CreateGimpleAssign(type, temp, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), temp_ga);
#endif
   block->PushBefore(temp_ga, stmt);
   temp_var = GetPointer<gimple_assign>(GET_NODE(temp_ga))->op0;

   temp = tree_man->create_binary_operation(type, temp_var, signmask_var, srcp_default, bit_xor_expr_K);
   temp_ga = CreateGimpleAssign(type, temp, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), temp_ga);
#endif
   block->PushBefore(temp_ga, stmt);
   temp_var = GetPointer<gimple_assign>(GET_NODE(temp_ga))->op0;

   return tree_man->create_binary_operation(type, temp_var, signmask_var, srcp_default, minus_expr_K);

}

tree_nodeRef IR_lowering::expand_sdiv_pow2 (tree_nodeRef op0, unsigned long long int d, const tree_nodeRef stmt, const blocRef block, tree_nodeRef &type, const std::string srcp_default)
{
   int logd;

   logd = floor_log2 (d);
   unsigned int integer_cst0_id = TM->new_tree_node_id();
   tree_nodeRef const0 = tree_man->CreateIntegerCst(type, 0, integer_cst0_id);

   tree_nodeRef bt =  tree_man->create_boolean_type();
   tree_nodeRef cond_op0 = tree_man->create_binary_operation(bt, op0, const0, srcp_default, lt_expr_K);
   tree_nodeRef cond_op0_ga = CreateGimpleAssign(bt, cond_op0, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), cond_op0_ga);
#endif
   block->PushBefore(cond_op0_ga, stmt);
   tree_nodeRef cond_op0_ga_var = GetPointer<gimple_assign>(GET_NODE(cond_op0_ga))->op0;
   tree_nodeRef t_ga;
   tree_nodeRef t2_ga;

   if (d == 2)
   {
      unsigned int integer_cst1_id = TM->new_tree_node_id();
      tree_nodeRef const1 = tree_man->CreateIntegerCst(type, 1,
            integer_cst1_id);
      tree_nodeRef cond_op = tree_man->create_ternary_operation(type, cond_op0_ga_var, const1, const0, srcp_default, cond_expr_K);
      t_ga = CreateGimpleAssign(type, cond_op, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
      block->PushBefore(t_ga, stmt);
      tree_nodeRef cond_ga_var = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;

      tree_nodeRef sum_expr = tree_man->create_binary_operation(type, op0, cond_ga_var, srcp_default, plus_expr_K);

      t2_ga = CreateGimpleAssign(type, sum_expr, block->number, srcp_default);

   }
   else
   {
      unsigned int integer_d_m1_id = TM->new_tree_node_id();
      tree_nodeRef d_m1 = tree_man->CreateIntegerCst(type, static_cast<long long int>(d-1), integer_d_m1_id);

      tree_nodeRef t_expr = tree_man->create_binary_operation(type, op0, d_m1, srcp_default, plus_expr_K);
      t_ga = CreateGimpleAssign(type, t_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
      block->PushBefore(t_ga, stmt);
      tree_nodeRef t_ga_var = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;

      tree_nodeRef cond_op = tree_man->create_ternary_operation(type, cond_op0_ga_var, t_ga_var, op0, srcp_default, cond_expr_K);

      t2_ga = CreateGimpleAssign(type, cond_op, block->number, srcp_default);
   }

#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), t2_ga);
#endif
   block->PushBefore(t2_ga, stmt);

   tree_nodeRef t2_ga_var = GetPointer<gimple_assign>(GET_NODE(t2_ga))->op0;

   unsigned int integer_logd_id = TM->new_tree_node_id();
   tree_nodeRef logdConst = tree_man->CreateIntegerCst(type, logd, integer_logd_id);
   return  tree_man->create_binary_operation(type, t2_ga_var, logdConst, srcp_default, rshift_expr_K);

}


tree_nodeRef IR_lowering::expand_MC(tree_nodeRef op0, integer_cst* ic_node, tree_nodeRef old_target, const tree_nodeRef stmt, const blocRef block, tree_nodeRef &type_expr, const std::string & srcp_default)
{
   long long int ext_op1 = tree_helper::get_integer_cst_value(ic_node);
   short int mult_plus_ratio = 3;
   unsigned int data_bitsize = tree_helper::Size(GET_NODE(op0));
#if HAVE_BAMBU_BUILT
   if(GetPointer<HLS_manager>(AppM))
   {
      const HLS_targetRef HLS_T = GetPointer<HLS_manager>(AppM)->get_HLS_target();
      const technology_managerRef TechManager = HLS_T->get_technology_manager();
      unsigned int fu_prec = resize_to_1_8_16_32_64_128_256_512(data_bitsize);
      if(fu_prec == 1)
      {
         fu_prec = 8;
      }
      technology_nodeRef mult_f_unit = TechManager->get_fu(MULTIPLIER_STD + std::string("_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec) + "_0", LIBRARY_STD_FU);
      functional_unit * mult_fu= GetPointer<functional_unit>(mult_f_unit);
      technology_nodeRef mult_op_node =mult_fu->get_operation("mult_expr");
      operation * mult_op = GetPointer<operation>(mult_op_node);
      double mult_delay = mult_op->time_m->get_execution_time();
      technology_nodeRef add_f_unit = TechManager->get_fu(ADDER_STD + std::string("_") + STR(fu_prec) + "_" + STR(fu_prec) + "_" + STR(fu_prec), LIBRARY_STD_FU);
      functional_unit * add_fu= GetPointer<functional_unit>(add_f_unit);
      technology_nodeRef add_op_node =add_fu->get_operation("plus_expr");
      operation * add_op = GetPointer<operation>(add_op_node);
      double add_delay = add_op->time_m->get_execution_time();
      mult_plus_ratio = static_cast<short int>(ceil(mult_delay/add_delay));
   }
#endif

   /// very special case op1 == 0
   if(ext_op1 == 0)
   {
      unsigned int node_nid = TM->new_tree_node_id();
      return tree_man->CreateIntegerCst(type_expr, 0, node_nid);
   }
   /// very special case op1 == 1
   else if(ext_op1 == 1)
   {
      return op0;
   }
   /// very special case op1 == -1
   else if(ext_op1 == -1)
   {
      return tree_man->create_unary_operation(type_expr, op0, srcp_default, negate_expr_K);
   }
   else
   {
      if(ext_op1 < 0)
      {
         struct algorithm alg;
         enum mult_variant variant;
         short int max_cost = mult_plus_ratio;//static_cast<short int>(tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr)))/mult_cost_divisor);
         if(choose_mult_variant (data_bitsize, -ext_op1, alg, variant, max_cost, TM))
         {
            tree_nodeRef temp_expr = expand_mult_const(op0, static_cast<unsigned long long int>(-ext_op1), alg, variant, stmt, block, type_expr, srcp_default);
            tree_nodeRef temp_expr_ga = CreateGimpleAssign(type_expr, temp_expr, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), temp_expr_ga);
#endif
            block->PushBefore(temp_expr_ga, stmt);
            tree_nodeRef temp_expr_var = GetPointer<gimple_assign>(GET_NODE(temp_expr_ga))->op0;
            return tree_man->create_unary_operation(type_expr, temp_expr_var, srcp_default, negate_expr_K);
         }
         else
         {
            /// keep the old assign
            return old_target;
         }
      }
      else
      {
         unsigned long long int coeff = static_cast<unsigned long long int>(ext_op1);
         if(EXACT_POWER_OF_2_OR_ZERO_P (coeff))
         {
            int l_shift = floor_log2 (coeff);
            unsigned int l_shift_nid = TM->new_tree_node_id();
            tree_nodeRef l_shift_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(l_shift), l_shift_nid);
            return tree_man->create_binary_operation(type_expr, op0, l_shift_node, srcp_default, lshift_expr_K);
         }
         else
         {
            struct algorithm alg;
            enum mult_variant variant;
            short int max_cost = mult_plus_ratio;//static_cast<short int>(tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr)))/mult_cost_divisor);

            if(choose_mult_variant (data_bitsize, static_cast<long long int>(coeff), alg, variant, max_cost, TM))
               return expand_mult_const(op0, coeff, alg, variant, stmt, block, type_expr, srcp_default);
            else
               /// keep the old assign
               return old_target;
         }
      }
   }
}

void IR_lowering::expand_target_mem_ref(target_mem_ref461 * tmr, const tree_nodeRef stmt, const blocRef block, const std::string srcp_default)
{
   tree_nodeRef accum;
   tree_nodeRef type_sum;
   unsigned int params = 0;

   if(tmr->idx)
      ++params;
   if(tmr->step)
      ++params;
   if(tmr->offset)
      ++params;
   if(tmr->idx2)
      ++params;

   if(params <=1)
      return; /// nothing to optimize

   if(tmr->idx)
   {
      if(tmr->step)
      {
         integer_cst* ic_step_node = GetPointer<integer_cst>(GET_NODE(tmr->step));
         type_sum = ic_step_node->type;
         accum = expand_MC(tmr->idx, ic_step_node, tree_nodeRef(), stmt, block, type_sum, srcp_default);
         if(accum)
         {
            tree_nodeRef t_ga = CreateGimpleAssign(type_sum, accum, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
            block->PushBefore(t_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;
         }
         else
         {
            tree_nodeRef t_expr = tree_man->create_binary_operation(type_sum, tmr->idx, tmr->step, srcp_default, mult_expr_K);
            tree_nodeRef t_ga = CreateGimpleAssign(type_sum, t_expr, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
            block->PushBefore(t_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;
         }

         tmr->step = tree_nodeRef();

      }
      else
         accum = tmr->idx;
      tmr->idx = tree_nodeRef();
   }
   if(tmr->offset)
   {
      integer_cst* ic_node = GetPointer<integer_cst>(GET_NODE(tmr->offset));
      long long int ic_value = tree_helper::get_integer_cst_value(ic_node);
      if(ic_value!=0)
      {
         if(!type_sum)
            type_sum = tree_man->create_size_type();

         tree_nodeRef casted_offset_ga = CreateGimpleAssign(type_sum, tmr->offset, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), casted_offset_ga);
#endif
         block->PushBefore(casted_offset_ga, stmt);
         tree_nodeRef casted_offset_var = GetPointer<gimple_assign>(GET_NODE(casted_offset_ga))->op0;

         if(accum)
         {
            tree_nodeRef t_expr = tree_man->create_binary_operation(type_sum, casted_offset_var, accum, srcp_default, plus_expr_K);
            tree_nodeRef t_ga;
            t_ga = CreateGimpleAssign(type_sum, t_expr, block->number, srcp_default);
#ifndef NDEBUG
            function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
            block->PushBefore(t_ga, stmt);
            accum = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;
         }
         else
            accum = casted_offset_var;
      }
      tmr->offset = tree_nodeRef();
   }

   if(tmr->idx2)
   {
      if(!type_sum)
         type_sum = tree_man->create_size_type();
      tree_nodeRef casted_idx2_ga = CreateGimpleAssign(type_sum, tmr->idx2, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), casted_idx2_ga);
#endif
      block->PushBefore(casted_idx2_ga, stmt);
      tree_nodeRef casted_idx2_var = GetPointer<gimple_assign>(GET_NODE(casted_idx2_ga))->op0;
      if(accum)
      {
         tree_nodeRef t_expr = tree_man->create_binary_operation(type_sum, accum, casted_idx2_var, srcp_default, plus_expr_K);
         tree_nodeRef t_ga = CreateGimpleAssign(type_sum, t_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), t_ga);
#endif
         block->PushBefore(t_ga, stmt);
         accum = GetPointer<gimple_assign>(GET_NODE(t_ga))->op0;
      }
      else
         accum = casted_idx2_var;
   }

   tmr->idx2 = accum;
}

#define LOWPART(x) \
  (static_cast<unsigned long long int>(x) & ((1ULL << 32) - 1))
#define HIGHPART(x) \
  (static_cast<unsigned long long int>(x) >> 32)
#define BASE (1LL << 32)

/* Unpack a two-word integer into 4 words.
   LOW and HI are the integer, as two `HOST_WIDE_INT' pieces.
   WORDS points to the array of HOST_WIDE_INTs.  */
static void
encode (long long int  *words, unsigned long long int  low, long long int  hi)
{
  words[0] = static_cast<long long int>(LOWPART (low));
  words[1] = static_cast<long long int>(HIGHPART (low));
  words[2] = static_cast<long long int>(LOWPART (hi));
  words[3] = static_cast<long long int>(HIGHPART (hi));
}

/* Pack an array of 4 words into a two-word integer.
   WORDS points to the array of words.
   The integer is stored into *LOW and *HI as two `HOST_WIDE_INT' pieces.  */

static void
decode (long long int  *words, unsigned long long int  *low,
    long long int  *hi)
{
  *low = static_cast<unsigned long long int>(words[0] + words[1] * BASE);
  *hi = words[2] + words[3] * BASE;
}

/* Divide doubleword integer LNUM, HNUM by doubleword integer LDEN, HDEN
   for a quotient (stored in *LQUO, *HQUO) and remainder (in *LREM, *HREM).
   CODE is a tree code for a kind of division, one of
   TRUNC_DIV_EXPR, FLOOR_DIV_EXPR, CEIL_DIV_EXPR, ROUND_DIV_EXPR
   or EXACT_DIV_EXPR
   It controls how the quotient is rounded to an integer.
   Return nonzero if the operation overflows.
   UNS nonzero says do unsigned division.  */
static int
div_and_round_double_cprop (
                      /* num == numerator == dividend */
                      unsigned long long int lnum_orig,
                      long long int hnum_orig,
                      /* den == denominator == divisor */
                      unsigned long long int lden_orig,
                      unsigned long long int *lquo,
                      long long int *hquo)
{
   long long int num[4 + 1];	/* extra element for scaling.  */
   long long int den[4], quo[4];
   int i, j;
   unsigned long long int work;
   unsigned long long int carry = 0;
   unsigned long long int lnum = lnum_orig;
   long long int hnum = hnum_orig;
   unsigned long long int lden = lden_orig;
   int overflow = 0;

   if (lden == 0)
      overflow = 1, lden = 1;


   if (hnum == 0)
   {				/* single precision */
      *hquo = 0;
      /* This unsigned division rounds toward zero.  */
      *lquo = lnum / lden;
      goto finish_up;
   }

   if (hnum == 0)
   {				/* trivial case: dividend < divisor */
      *hquo = 0;
      *lquo = 0;

      goto finish_up;
   }

   memset (quo, 0, sizeof quo);

   memset (num, 0, sizeof num);	/* to zero 9th element */
   memset (den, 0, sizeof den);

   encode (num, lnum, hnum);
   encode (den, lden, 0);

   /* Special code for when the divisor < BASE.  */
   if (lden < static_cast<unsigned long long int>(BASE))
   {
      /* hnum != 0 already checked.  */
      for (i = 4 - 1; i >= 0; i--)
      {
         work = static_cast<unsigned long long int>(num[i]) + carry * BASE;
         quo[i] = static_cast<long long int>(work / lden);
         carry = work % lden;
      }
   }
   else
   {
      /* Full double precision division,
     with thanks to Don Knuth's "Seminumerical Algorithms".  */
      int num_hi_sig, den_hi_sig;
      unsigned long long int quo_est, scale;

      /* Find the highest nonzero divisor digit.  */
      for (i = 4 - 1;; i--)
         if (den[i] != 0)
         {
            den_hi_sig = i;
            break;
         }

      /* Insure that the first digit of the divisor is at least BASE/2.
     This is required by the quotient digit estimation algorithm.  */

      scale = static_cast<unsigned long long int>(BASE / (den[den_hi_sig] + 1));
      if (scale > 1)
      {		/* scale divisor and dividend */
         carry = 0;
         for (i = 0; i <= 4 - 1; i++)
         {
            work = (static_cast<unsigned long long int>(num[i]) * scale) + carry;
            num[i] = LOWPART (work);
            carry = HIGHPART (work);
         }

         num[4] = static_cast<long long int>(carry);
         carry = 0;
         for (i = 0; i <= 4 - 1; i++)
         {
            work = (static_cast<unsigned long long int>(den[i]) * scale) + carry;
            den[i] = LOWPART (work);
            carry = HIGHPART (work);
            if (den[i] != 0) den_hi_sig = i;
         }
      }

      num_hi_sig = 4;

      /* Main loop */
      for (i = num_hi_sig - den_hi_sig - 1; i >= 0; i--)
      {
         /* Guess the next quotient digit, quo_est, by dividing the first
         two remaining dividend digits by the high order quotient digit.
         quo_est is never low and is at most 2 high.  */
         unsigned long long int tmp;

         num_hi_sig = i + den_hi_sig + 1;
         work = static_cast<unsigned long long int>(num[num_hi_sig]) * BASE + static_cast<unsigned long long int>(num[num_hi_sig - 1]);
         if (num[num_hi_sig] != den[den_hi_sig])
            quo_est = work / static_cast<unsigned long long int>(den[den_hi_sig]);
         else
            quo_est = BASE - 1;

         /* Refine quo_est so it's usually correct, and at most one high.  */
         tmp = work - quo_est * static_cast<unsigned long long int>(den[den_hi_sig]);
         if (tmp < BASE
             && (static_cast<unsigned long long int>(den[den_hi_sig - 1]) * quo_est
                 > (tmp * BASE + static_cast<unsigned long long int>(num[num_hi_sig - 2]))))
            quo_est--;

         /* Try QUO_EST as the quotient digit, by multiplying the
         divisor by QUO_EST and subtracting from the remaining dividend.
         Keep in mind that QUO_EST is the I - 1st digit.  */

         carry = 0;
         for (j = 0; j <= den_hi_sig; j++)
         {
            work = quo_est * static_cast<unsigned long long int>(den[j]) + carry;
            carry = HIGHPART (work);
            work = static_cast<unsigned long long int>(num[i + j]) - LOWPART (work);
            num[i + j] = LOWPART (work);
            carry += HIGHPART (work) != 0;
         }

         /* If quo_est was high by one, then num[i] went negative and
         we need to correct things.  */
         if (num[num_hi_sig] < static_cast<long long int>(carry))
         {
            quo_est--;
            carry = 0;		/* add divisor back in */
            for (j = 0; j <= den_hi_sig; j++)
            {
               work = static_cast<unsigned long long int>(num[i + j]) + static_cast<unsigned long long int>(den[j]) + carry;
               carry = HIGHPART (work);
               num[i + j] = LOWPART (work);
            }

            num [num_hi_sig] = num [num_hi_sig] + static_cast<long long int>(carry);
         }

         /* Store the quotient digit.  */
         quo[i] = static_cast<long long int>(quo_est);
      }
   }

   decode (quo, lquo, hquo);

finish_up:


   return overflow;
}

/* Choose a minimal N + 1 bit approximation to 1/D that can be used to
   replace division by D, and put the least significant N bits of the result
   in *MULTIPLIER_PTR and return the most significant bit.

   The width of operations is N (should be <= HOST_BITS_PER_WIDE_INT), the
   needed precision is in PRECISION (should be <= N).

   PRECISION should be as small as possible so this function can choose
   multiplier more freely.

   The rounded-up logarithm of D is placed in *lgup_ptr.  A shift count that
   is to be used for a final right shift is placed in *POST_SHIFT_PTR.

   Using this function, x/D will be equal to (x * m) >> (*POST_SHIFT_PTR),
   where m is the full HOST_BITS_PER_WIDE_INT + 1 bit multiplier.  */
static
unsigned long long int
choose_multiplier (unsigned long long int d, int n, int precision,
                   unsigned long long int *multiplier_ptr,
                   int *post_shift_ptr, int *lgup_ptr)
{
   long long int mhigh_hi, mlow_hi;
   unsigned long long int mhigh_lo, mlow_lo;
   int lgup, post_shift;
   int pow, pow2;
   unsigned long long int nl;
   long long int nh;

   /* lgup = ceil(log2(divisor)); */
   lgup = ceil_log2 (d);

   THROW_ASSERT (lgup <= n, "unexpected condition");

   pow = n + lgup;
   pow2 = n + lgup - precision;

   /* We could handle this with some effort, but this case is much
     better handled directly with a scc insn, so rely on caller using
     that.  */
   THROW_ASSERT (pow != 128, "unexpected condition");

   /* mlow = 2^(N + lgup)/d */
   if (pow >= 64)
   {
      nh =  1LL << (pow - 64);
      nl = 0;
   }
   else
   {
      nh = 0;
      nl = 1ULL << pow;
   }

   div_and_round_double_cprop (nl, nh, d,
                               &mlow_lo, &mlow_hi);


   /* mhigh = (2^(N + lgup) + 2^(N + lgup - precision))/d */
   if (pow2 >= 64)
      nh |= 1LL << (pow2 - 64);
   else
      nl |= 1ULL << pow2;

   div_and_round_double_cprop (nl, nh, d,
                               &mhigh_lo, &mhigh_hi);

   THROW_ASSERT (!mhigh_hi || static_cast<unsigned long long int>(nh) - d < d, "unexpected condition");
   THROW_ASSERT (mhigh_hi <= 1 && mlow_hi <= 1, "unexpected condition");
   /* Assert that mlow < mhigh.  */
   THROW_ASSERT (mlow_hi < mhigh_hi
                 || (mlow_hi == mhigh_hi && mlow_lo < mhigh_lo), "unexpected condition");

   /* If precision == N, then mlow, mhigh exceed 2^N
     (but they do not exceed 2^(N+1)).  */

   /* Reduce to lowest terms.  */
   for (post_shift = lgup; post_shift > 0; post_shift--)
   {
      unsigned long long int ml_lo = static_cast<unsigned long long int>(mlow_hi << (64 - 1)) | (mlow_lo >> 1);
      unsigned long long int mh_lo = static_cast<unsigned long long int>(mhigh_hi << (64 - 1)) | (mhigh_lo >> 1);
      if (ml_lo >= mh_lo)
         break;

      mlow_hi = 0;
      mlow_lo = ml_lo;
      mhigh_hi = 0;
      mhigh_lo = mh_lo;
   }

   *post_shift_ptr = post_shift;
   *lgup_ptr = lgup;
   if (n < 64)
   {
      unsigned long long int mask = (1ULL << n) - 1;
      *multiplier_ptr = mhigh_lo & mask;
      return mhigh_lo >= mask;
   }
   else
   {
      *multiplier_ptr = mhigh_lo;
      return static_cast<unsigned long long int>(mhigh_hi);
   }
}

tree_nodeRef IR_lowering::expand_mult_highpart(tree_nodeRef op0, unsigned long long int ml, tree_nodeRef type_expr, int data_bitsize, const std::list<tree_nodeRef>::const_iterator it_los, const blocRef block, const std::string srcp_default)
{
   /**
    long long int u0, v0, u1, v1, u0v0, u0v0h, u1v0, u0v0hu1v0, u0v1, u0v0hu1v0u0v1, u0v0hu1v0u0v1h, u1v1;
    u0 = u & ((1LL<<32)-1);
    u1 = u >>32;
    v0 = v & ((1LL<<32)-1);
    v1 = v >>32;
    u0v0 = u0 * v0;
    u0v0h = u0v0>>32;
    u0v0hU = u0v0h & ((1LL<<32)-1);///only for signed computation
    u1v0 = u1 * v0;
    u0v0hu1v0 = u0v0hU + u1v0;
    w1 = u0v0hu1v0 & ((1LL<<32)-1);
    w2 = u0v0hu1v0 >>32;
    u0v1 = u0 * v1;
    w1u0v1 = w1+u0v1;
    w1u0v1h = w1u0v1 >> 32;
    u1v1 = u1 * v1;
    w1u0v1hw2 = w1u0v1h + w2;
    return w1u0v1hw2 + u1v1;
   */
   int half_data_bitsize = data_bitsize/2;
   unsigned int mask_nid = TM->new_tree_node_id();
   tree_nodeRef mask_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>((1LL<<half_data_bitsize)-1), mask_nid);
   unsigned int half_data_bitsize_nid = TM->new_tree_node_id();
   tree_nodeRef half_data_bitsize_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(half_data_bitsize), half_data_bitsize_nid);

   tree_nodeRef u0_expr = tree_man->create_binary_operation(type_expr, op0, mask_node, srcp_default, bit_and_expr_K);
   tree_nodeRef u0_ga = CreateGimpleAssign(type_expr, u0_expr, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), u0_ga);
#endif
   block->PushBefore(u0_ga, *it_los);
   tree_nodeRef u0_ga_var = GetPointer<gimple_assign>(GET_NODE(u0_ga))->op0;

   tree_nodeRef u1_expr = tree_man->create_binary_operation(type_expr, op0, half_data_bitsize_node, srcp_default, rshift_expr_K);
   tree_nodeRef u1_ga = CreateGimpleAssign(type_expr, u1_expr, block->number, srcp_default);
#ifndef NDEBUG
   function_behavior->RegisterTransformation(GetName(), u1_ga);
#endif
   block->PushBefore(u1_ga, *it_los);
   tree_nodeRef u1_ga_var = GetPointer<gimple_assign>(GET_NODE(u1_ga))->op0;

   long long int v0 = static_cast<long long int>(ml) & ((1LL<<half_data_bitsize)-1);
   long long int v1;
   bool unsignedp = tree_helper::is_unsigned(TM, GET_INDEX_NODE(type_expr));
   v1 = static_cast<long long int>(ml >> half_data_bitsize);

   tree_nodeRef u0v0_ga_var;
   tree_nodeRef v0_node;
   tree_nodeRef v1_node;
   if(v0 != 0)
   {
      if(v0 == 1)
         u0v0_ga_var = u0_ga_var;
      else
      {
         unsigned int v0_nid = TM->new_tree_node_id();
         v0_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(v0), v0_nid);
         tree_nodeRef u0v0_expr = tree_man->create_binary_operation(type_expr, u0_ga_var, v0_node, srcp_default, mult_expr_K);
         tree_nodeRef u0v0_ga = CreateGimpleAssign(type_expr, u0v0_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), u0v0_ga);
#endif
         block->PushBefore(u0v0_ga, *it_los);
         u0v0_ga_var = GetPointer<gimple_assign>(GET_NODE(u0v0_ga))->op0;
      }
   }
   tree_nodeRef u0v0h_ga_var;
   if(u0v0_ga_var)
   {
      tree_nodeRef u0v0h_expr = tree_man->create_binary_operation(type_expr, u0v0_ga_var, half_data_bitsize_node, srcp_default, rshift_expr_K);
      tree_nodeRef u0v0h_ga = CreateGimpleAssign(type_expr, u0v0h_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), u0v0h_ga);
#endif
      block->PushBefore(u0v0h_ga, *it_los);
      u0v0h_ga_var = GetPointer<gimple_assign>(GET_NODE(u0v0h_ga))->op0;
   }
   tree_nodeRef u0v0hU_ga_var;
   if(u0v0h_ga_var && !unsignedp)
   {
      tree_nodeRef u0v0hU_expr = tree_man->create_binary_operation(type_expr, u0v0h_ga_var, mask_node, srcp_default, bit_and_expr_K);
      tree_nodeRef u0v0hU_ga = CreateGimpleAssign(type_expr, u0v0hU_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), u0v0hU_ga);
#endif
      block->PushBefore(u0v0hU_ga, *it_los);
      u0v0hU_ga_var = GetPointer<gimple_assign>(GET_NODE(u0v0hU_ga))->op0;
   }
   else
      u0v0hU_ga_var = u0v0h_ga_var;
   tree_nodeRef u1v0_ga_var;
   if(v0 != 0)
   {
      if(v0 == 1)
         u1v0_ga_var = u1_ga_var;
      else
      {
         tree_nodeRef u1v0_expr = tree_man->create_binary_operation(type_expr, u1_ga_var, v0_node, srcp_default, mult_expr_K);
         tree_nodeRef u1v0_ga = CreateGimpleAssign(type_expr, u1v0_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), u1v0_ga);
#endif
         block->PushBefore(u1v0_ga, *it_los);
         u1v0_ga_var = GetPointer<gimple_assign>(GET_NODE(u1v0_ga))->op0;
      }
   }
   tree_nodeRef u0v0hu1v0_ga_var;
   if(u0v0hU_ga_var)
   {
      THROW_ASSERT(u1v0_ga_var, "unexpected condition");
      tree_nodeRef u0v0hu1v0_expr = tree_man->create_binary_operation(type_expr, u0v0hU_ga_var, u1v0_ga_var, srcp_default, plus_expr_K);
      tree_nodeRef u0v0hu1v0_ga = CreateGimpleAssign(type_expr, u0v0hu1v0_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), u0v0hu1v0_ga);
#endif
      block->PushBefore(u0v0hu1v0_ga, *it_los);
      u0v0hu1v0_ga_var = GetPointer<gimple_assign>(GET_NODE(u0v0hu1v0_ga))->op0;
   }
   tree_nodeRef w1_ga_var;
   if(u0v0hu1v0_ga_var)
   {
      tree_nodeRef w1_expr = tree_man->create_binary_operation(type_expr, u0v0hu1v0_ga_var, mask_node, srcp_default, bit_and_expr_K);
      tree_nodeRef w1_ga = CreateGimpleAssign(type_expr, w1_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), w1_ga);
#endif
      block->PushBefore(w1_ga, *it_los);
      w1_ga_var = GetPointer<gimple_assign>(GET_NODE(w1_ga))->op0;
   }

   tree_nodeRef w2_ga_var;
   if(u0v0hu1v0_ga_var)
   {
      tree_nodeRef w2_expr = tree_man->create_binary_operation(type_expr, u0v0hu1v0_ga_var, half_data_bitsize_node, srcp_default, rshift_expr_K);
      tree_nodeRef w2_ga = CreateGimpleAssign(type_expr, w2_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), w2_ga);
#endif
      block->PushBefore(w2_ga, *it_los);
      w2_ga_var = GetPointer<gimple_assign>(GET_NODE(w2_ga))->op0;
   }

   tree_nodeRef u0v1_ga_var;
   if(v1 != 0)
   {
      if(v1 == 1)
         u0v1_ga_var = u0_ga_var;
      else
      {
         unsigned int v1_nid = TM->new_tree_node_id();
         v1_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(v1), v1_nid);
         tree_nodeRef u0v1_expr = tree_man->create_binary_operation(type_expr, u0_ga_var, v1_node, srcp_default, mult_expr_K);
         tree_nodeRef u0v1_ga = CreateGimpleAssign(type_expr, u0v1_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), u0v1_ga);
#endif
         block->PushBefore(u0v1_ga, *it_los);
         u0v1_ga_var = GetPointer<gimple_assign>(GET_NODE(u0v1_ga))->op0;
      }
   }
   tree_nodeRef w1u0v1_ga_var;
   if(w1_ga_var)
   {
      if(u0v1_ga_var)
      {
         tree_nodeRef w1u0v1_expr = tree_man->create_binary_operation(type_expr, w1_ga_var, u0v1_ga_var, srcp_default, plus_expr_K);
         tree_nodeRef w1u0v1_ga = CreateGimpleAssign(type_expr, w1u0v1_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), w1u0v1_ga);
#endif
         block->PushBefore(w1u0v1_ga, *it_los);
         w1u0v1_ga_var = GetPointer<gimple_assign>(GET_NODE(w1u0v1_ga))->op0;
      }
      else
         w1u0v1_ga_var = w1_ga_var;
   }
   else if(u0v1_ga_var)
      w1u0v1_ga_var = u0v1_ga_var;

   tree_nodeRef w1u0v1h_ga_var;
   if(w1u0v1_ga_var)
   {
      tree_nodeRef w1u0v1h_expr = tree_man->create_binary_operation(type_expr, w1u0v1_ga_var, half_data_bitsize_node, srcp_default, rshift_expr_K);
      tree_nodeRef w1u0v1h_ga = CreateGimpleAssign(type_expr, w1u0v1h_expr, block->number, srcp_default);
#ifndef NDEBUG
      function_behavior->RegisterTransformation(GetName(), w1u0v1h_ga);
#endif
      block->PushBefore(w1u0v1h_ga, *it_los);
      w1u0v1h_ga_var = GetPointer<gimple_assign>(GET_NODE(w1u0v1h_ga))->op0;
   }
   tree_nodeRef u1v1_ga_var;
   if(v1 != 0)
   {
      if(v1 == 1)
         u1v1_ga_var = u1_ga_var;
      else
      {
         tree_nodeRef u1v1_expr = tree_man->create_binary_operation(type_expr, u1_ga_var, v1_node, srcp_default, mult_expr_K);
         tree_nodeRef u1v1_ga = CreateGimpleAssign(type_expr, u1v1_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), u1v1_ga);
#endif
         block->PushBefore(u1v1_ga, *it_los);
         u1v1_ga_var = GetPointer<gimple_assign>(GET_NODE(u1v1_ga))->op0;
      }
   }
   tree_nodeRef w1u0v1hw2_ga_var;
   if(w1u0v1h_ga_var)
   {
      if(w2_ga_var)
      {
         tree_nodeRef w1u0v1hw2_expr = tree_man->create_binary_operation(type_expr, w1u0v1h_ga_var, w2_ga_var, srcp_default, plus_expr_K);
         tree_nodeRef w1u0v1hw2_ga = CreateGimpleAssign(type_expr, w1u0v1hw2_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), w1u0v1hw2_ga);
#endif
         block->PushBefore(w1u0v1hw2_ga, *it_los);
         w1u0v1hw2_ga_var = GetPointer<gimple_assign>(GET_NODE(w1u0v1hw2_ga))->op0;
      }
      else
         w1u0v1hw2_ga_var = w1u0v1h_ga_var;
   }
   else if(w2_ga_var)
      w1u0v1hw2_ga_var = w2_ga_var;

   tree_nodeRef res_ga_var;
   if(w1u0v1hw2_ga_var)
   {
      if(u1v1_ga_var)
      {
         tree_nodeRef res_expr = tree_man->create_binary_operation(type_expr, w1u0v1hw2_ga_var, u1v1_ga_var, srcp_default, plus_expr_K);
         tree_nodeRef res_ga = CreateGimpleAssign(type_expr, res_expr, block->number, srcp_default);
#ifndef NDEBUG
         function_behavior->RegisterTransformation(GetName(), res_ga);
#endif
         block->PushBefore(res_ga, *it_los);
         res_ga_var = GetPointer<gimple_assign>(GET_NODE(res_ga))->op0;
      }
      else
         res_ga_var = w1u0v1hw2_ga_var;
   }
   else if(u1v1_ga_var)
      res_ga_var = u1v1_ga_var;
   else
   {
      unsigned int const0_nid = TM->new_tree_node_id();
      res_ga_var = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(0), const0_nid);
   }
   return res_ga_var;

}

IR_lowering::IR_lowering(const ParameterConstRef Param, const application_managerRef _AppM, unsigned int _function_id, const DesignFlowManagerConstRef _design_flow_manager) :
   FunctionFrontendFlowStep(_AppM, _function_id, IR_LOWERING, _design_flow_manager, Param),
   restart_check_system_type(false)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}

const std::unordered_set<std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship> > IR_lowering::ComputeFrontendRelationships(const DesignFlowStep::RelationshipType relationship_type) const
{
   std::unordered_set<std::pair<FrontendFlowStepType, FunctionRelationship> > relationships;
   switch(relationship_type)
   {
      case(DEPENDENCE_RELATIONSHIP) :
         {
            relationships.insert(std::pair<FrontendFlowStepType, FunctionRelationship>(BLOCK_FIX, SAME_FUNCTION));
            relationships.insert(std::pair<FrontendFlowStepType, FunctionRelationship>(SWITCH_FIX, SAME_FUNCTION));
            break;
         }
      case(INVALIDATION_RELATIONSHIP) :
         {
            break;
         }
      case(PRECEDENCE_RELATIONSHIP) :
         {
            break;
         }
      default:
         THROW_UNREACHABLE("");
   }
   return relationships;
}


IR_lowering::~IR_lowering()
{

}

DesignFlowStep_Status IR_lowering::InternalExec()
{
   if(debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      PrintTreeManager(true);
   }
   restart_check_system_type = false;
   tree_nodeRef tn = TM->get_tree_node_const(function_id);
   tree_nodeRef Scpe = TM->GetTreeReindex(function_id);
   function_decl * fd = GetPointer<function_decl>(tn);
   THROW_ASSERT(fd && fd->body, "Node is not a function or it hasn't a body");
   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));
   THROW_ASSERT(sl, "Body is not a statement_list");
   /// for each basic block B in CFG do > Consider all blocks successively
   for(auto block : sl->list_of_bloc)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining BB" + STR(block.first));
      const auto & list_of_stmt = block.second->CGetStmtList();
      /// manage capacity
      auto it_los_end = list_of_stmt.end();
      size_t n_stmts = list_of_stmt.size();
      auto it_los = list_of_stmt.begin();
      while(it_los != it_los_end)
      {
#ifndef NDEBUG
         if((not function_behavior->ApplyNewTransformation()) and not (GetPointer<gimple_assign>(GET_NODE(*it_los)) and GetPointer<cond_expr>(GET_NODE(GetPointer<gimple_assign>(GET_NODE(*it_los))->op1))))
         if(not function_behavior->ApplyNewTransformation())
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Reached max cfg transformations");
            it_los++;
            continue;
         }
#endif
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining statement " + GET_NODE(*it_los)->ToString());
         if(GET_NODE(*it_los)->get_kind() == gimple_assign_K)
         {
            gimple_assign * ga =  GetPointer<gimple_assign>(GET_NODE(*it_los));
            const std::string srcp_default = ga->include_name + ":" + STR(ga->line_number) + ":" + STR(ga->column_number);
            enum kind code0 = GET_NODE(ga->op0)->get_kind();
            enum kind code1 = GET_NODE(ga->op1)->get_kind();

            if(code0 == mem_ref_K && code1 == mem_ref_K)
            {
               /// found a memcpy
               /// do nothing
            }
#if 0
            else if(code1 == addr_expr_K)
            {
               addr_expr * ae =  GetPointer<addr_expr>(GET_NODE(ga->op1));
               enum kind ae_code = GET_NODE(ae->op)->get_kind();
               if(ae_code == mem_ref_K)
               {
                  mem_ref * MR =  GetPointer<mem_ref>(GET_NODE(ae->op));
                  tree_nodeRef op1 = MR->op1;
                  long long int op1_val = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(op1)));
                  if(op1_val != 0)
                  {
                     /// check if the pointer plus is resolved as the sum of two constants
                     bool is_op0_addr_expr = GET_NODE(MR->op0)->get_kind() == addr_expr_K;
                     bool is_var_decl_par_decl = is_op0_addr_expr && (GET_NODE(GetPointer<addr_expr>(GET_NODE(MR->op0))->op)->get_kind() == var_decl_K ||GET_NODE(GetPointer<addr_expr>(GET_NODE(MR->op0))->op)->get_kind() == parm_decl_K) ;
                     if(!is_var_decl_par_decl)
                     {
                        unsigned int integer_cst0_id = TM->new_tree_node_id();
                        tree_nodeRef offset = tree_man->CreateIntegerCst(tree_man->create_size_type(), op1_val, integer_cst0_id);
                        tree_nodeRef mr = tree_man->create_binary_operation(ae->type, MR->op0, offset, srcp_default, pointer_plus_expr_K);
                        ga->op1 = mr;
                        ga->temporary_address = false;
                     }
                  }
                  else
                  {
                     ga->op1 = MR->op0;
                     ga->temporary_address = false;
                  }
               }
               else if(ae_code == target_mem_ref461_K)
               {
                  target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(GET_NODE(ae->op));
                  unsigned int idx2_index = tmr->idx2 ? GET_INDEX_NODE(tmr->idx2) : 0;
                  unsigned int base_index = tmr->base ? GET_INDEX_NODE(tmr->base) : 0;
                  unsigned int idx_index = tmr->idx ? GET_INDEX_NODE(tmr->idx) : 0;
                  unsigned int step_index = tmr->step ? GET_INDEX_NODE(tmr->step) : 0;
                  unsigned int offset_index = tmr->offset ? GET_INDEX_NODE(tmr->offset) : 0;
                  if(base_index && offset_index && !idx2_index && !idx_index && !step_index)
                  {
                     long long int op1_val = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(tmr->offset)));
                     if(op1_val != 0)
                     {
                        bool is_op0_addr_expr = GET_NODE(tmr->base)->get_kind() == addr_expr_K;
                        bool is_var_decl_par_decl = is_op0_addr_expr && (GET_NODE(GetPointer<addr_expr>(GET_NODE(tmr->base))->op)->get_kind() == var_decl_K ||GET_NODE(GetPointer<addr_expr>(GET_NODE(tmr->base))->op)->get_kind() == parm_decl_K) ;
                        if(!is_var_decl_par_decl)
                        {
                           unsigned int integer_cst0_id = TM->new_tree_node_id();
                           tree_nodeRef offset = tree_man->CreateIntegerCst(tree_man->create_size_type(), op1_val, integer_cst0_id);
                           ga->op1 = tree_man->create_binary_operation(ae->type, tmr->base, offset, srcp_default, pointer_plus_expr_K);
                           ga->temporary_address = false;
                        }
                     }
                     else
                     {
                        ga->op1 = tmr->base;
                     ga->temporary_address = false;
                     }
                  }
                  else if(base_index && !offset_index && idx2_index && !idx_index && !step_index)
                  {
                     ga->op1 = tree_man->create_binary_operation(ae->type, tmr->base, tmr->idx2, srcp_default, pointer_plus_expr_K);
                     ga->temporary_address = false;
                  }
                  else
                     std::cerr << "not supported " << GET_NODE(ae->op)->get_kind_text() << std::endl;
               }
               else
                  std::cerr << "not supported " << GET_NODE(ae->op)->get_kind_text() << std::endl;
            }
#endif
            else if(code1 == cond_expr_K)
            {
               cond_expr * ce = GetPointer<cond_expr>(GET_NODE(ga->op1));
               THROW_ASSERT(ce->op1 && ce->op2, "expected three parameters");
               if(GetPointer<binary_expr>(GET_NODE(ce->op0)))
               {
                  binary_expr* be = GetPointer<binary_expr>(GET_NODE(ce->op0));
                  tree_nodeRef new_ga = CreateGimpleAssign(be->type, ce->op0, block.first, srcp_default);
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

                  ce->op0 = ssa_vd;
                  block.second->PushBefore(new_ga, *it_los);
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            else if (code1 == pointer_plus_expr_K)
            {
               pointer_plus_expr * ppe = GetPointer<pointer_plus_expr>(GET_NODE(ga->op1));
               THROW_ASSERT(ppe->op0 && ppe->op1, "expected two parameters");
               if(GetPointer<addr_expr>(GET_NODE(ppe->op0)))
               {
                  addr_expr* ae = GetPointer<addr_expr>(GET_NODE(ppe->op0));
                  tree_nodeRef new_ga = CreateGimpleAssign(ae->type, ppe->op0, block.first, srcp_default);
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

                  ppe->op0 = ssa_vd;
#ifndef NDEBUG
                  function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
                  block.second->PushBefore(new_ga, *it_los);
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            else if(code1 == vec_cond_expr_K)
            {
               vec_cond_expr * vce = GetPointer<vec_cond_expr>(GET_NODE(ga->op1));
               THROW_ASSERT(vce->op1 && vce->op2, "expected three parameters");
               if(GetPointer<binary_expr>(GET_NODE(vce->op0)))
               {
                  binary_expr* be = GetPointer<binary_expr>(GET_NODE(vce->op0));

                  tree_nodeRef new_ga = CreateGimpleAssign(be->type, vce->op0, block.first, srcp_default);
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

                  vce->op0 = ssa_vd;
                  block.second->PushBefore(new_ga, *it_los);
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            else if(code1 == view_convert_expr_K)
            {
               view_convert_expr * vce = GetPointer<view_convert_expr>(GET_NODE(ga->op1));
               if(GetPointer<cst_node>(GET_NODE(vce->op)))
               {
                  cst_node* cn = GetPointer<cst_node>(GET_NODE(vce->op));

                  tree_nodeRef new_ga = CreateGimpleAssign(cn->type, vce->op, block.first, srcp_default);
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;
                  vce->op = ssa_vd;
                  block.second->PushBefore(new_ga, *it_los);
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            else if(code1 == target_mem_ref461_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand target_mem_ref 1 " + STR(GET_INDEX_NODE(ga->op1)));
               target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(GET_NODE(ga->op1));
               /// expose some part of the target_mem_ref statement
               expand_target_mem_ref(tmr, *it_los, block.second, srcp_default);

               /// split the target_mem_ref in a address computation statement and in a load statement
               tree_nodeRef type = tmr->type;

               tree_nodeRef pt = tree_man->create_pointer_type(type);
               tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op1, srcp_default, addr_expr_K);
               tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first, srcp_default);
               GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;

               tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

               unsigned int integer_cst0_id = TM->new_tree_node_id();
               tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
               tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

               ga->op1 = mr;
               block.second->PushBefore(new_ga, *it_los);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
               it_los = list_of_stmt.begin();
               it_los_end = list_of_stmt.end();
               continue;
            }
            else if(code1  == mem_ref_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand mem_ref 1 " + STR(GET_INDEX_NODE(ga->op1)));
               mem_ref * MR = GetPointer<mem_ref>(GET_NODE(ga->op1));
               tree_nodeRef op1 = MR->op1;
               long long int op1_val = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(op1)));
               if(op1_val != 0)
               {
                  tree_nodeRef type = MR->type;

                  tree_nodeRef pt = tree_man->create_pointer_type(type);
                  tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op1, srcp_default, addr_expr_K);
                  tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first, srcp_default);
                  GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

                  unsigned int integer_cst0_id = TM->new_tree_node_id();
                  tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
                  tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

                  ga->op1 = mr;
                  block.second->PushBefore(new_ga, *it_los);
#ifndef NDEBUG
                  function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            /*
            else if(code1 == array_ref_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand array_ref 1 " + STR(GET_INDEX_NODE(ga->op1)));
               array_ref * AR = GetPointer<array_ref>(GET_NODE(ga->op1));
               std::string srcp_default = fd->include_name +":0:0";
               tree_nodeRef type = AR->type;

               tree_nodeRef pt = tree_man->create_pointer_type(type);
               tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op1, srcp_default, addr_expr_K);
               tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first);
               GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;
               tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

               unsigned int integer_cst0_id = TM->new_tree_node_id();
               tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
               tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

               ga->op1 = mr;
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
               block->PushBefore(it_los, new_ga);
               it_los = list_of_stmt.begin();
               it_los_end = list_of_stmt.end();
               continue;
            }*/
            else if(code1 == trunc_div_expr_K || code1 == trunc_mod_expr_K || code1 == exact_div_expr_K)
            {
               tree_nodeRef op1 = GetPointer<binary_expr>(GET_NODE(ga->op1))->op1;
               if(GetPointer<cst_node>(GET_NODE(op1)))
               {
                  cst_node* cn = GetPointer<cst_node>(GET_NODE(op1));
                  tree_nodeRef op0 = GetPointer<binary_expr>(GET_NODE(ga->op1))->op0;
                  tree_nodeRef type_expr = GetPointer<binary_expr>(GET_NODE(ga->op1))->type;

                  bool unsignedp = tree_helper::is_unsigned(TM, GET_INDEX_NODE(type_expr));
                  long long int ext_op1 = tree_helper::get_integer_cst_value(static_cast<integer_cst*>(cn));
                  bool rem_flag = code1 == trunc_mod_expr_K;

                  /// very special case op1 == 1
                  if(ext_op1 == 1)
                  {
                     tree_nodeRef new_op1;
                     if(rem_flag)
                     {
                        unsigned int node_nid = TM->new_tree_node_id();
                        new_op1 = tree_man->CreateIntegerCst(type_expr, 0, node_nid);
                     }
                     else
                        new_op1 = op0;
                     ga->op1 = new_op1;
                  }
                  else if(!unsignedp && ext_op1 == -1)
                  {
                     tree_nodeRef new_op1;
                     if(rem_flag)
                     {
                        unsigned int node_nid = TM->new_tree_node_id();
                        new_op1 = tree_man->CreateIntegerCst(type_expr, 0, node_nid);
                     }
                     else
                     {
                        new_op1 = tree_man->create_unary_operation(type_expr, op0, srcp_default, negate_expr_K);
                     }
                     ga->op1 = new_op1;
                  }
                  else
                  {
                     bool op1_is_pow2 = ((EXACT_POWER_OF_2_OR_ZERO_P (ext_op1)
                              || (! unsignedp && EXACT_POWER_OF_2_OR_ZERO_P (-ext_op1))));
                     //long long int last_div_const = ! rem_flag ? ext_op1 : 0;

                     if (code1 == exact_div_expr_K && op1_is_pow2)
                        code1 = trunc_div_expr_K;


                     if(ext_op1 != 0)
                     {
                        switch (code1)
                        {
                           case trunc_div_expr_K:
                           case trunc_mod_expr_K:
                              {
                                 INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Trunc_div or trunc_mod");
                                 if (unsignedp)
                                 {
                                    INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---unsigned");
                                    int pre_shift;
                                    unsigned long long int d = static_cast<unsigned long long int>(ext_op1);
                                    if (EXACT_POWER_OF_2_OR_ZERO_P (d))
                                    {
                                       pre_shift = floor_log2 (d);
                                       tree_nodeRef new_op1;
                                       if (rem_flag)
                                       {

                                          unsigned int node_nid = TM->new_tree_node_id();
                                          tree_nodeRef mask = tree_man->CreateIntegerCst(type_expr, (static_cast<long long int>(1) << pre_shift) - 1, node_nid);
                                          new_op1 = tree_man->create_binary_operation(type_expr, op0, mask, srcp_default, bit_and_expr_K);
                                       }
                                       else
                                       {
                                          unsigned int node_nid = TM->new_tree_node_id();
                                          tree_nodeRef shift = tree_man->CreateIntegerCst(type_expr, pre_shift, node_nid);

                                          new_op1 = tree_man->create_binary_operation(type_expr, op0, shift, srcp_default, rshift_expr_K);
                                       }
                                       ga->op1 = new_op1;
                                    }
                                    else
                                    {
                                       int data_bitsize = static_cast<int>(tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr))));
                                       if(d<(1ull << (data_bitsize-1)))
                                       {
                                          unsigned long long int mh, ml;
                                          int post_shift;
                                          int dummy;
                                          /// Find a suitable multiplier and right shift count
                                          /// instead of multiplying with D.
                                          mh = choose_multiplier (d, data_bitsize, data_bitsize,
                                                                  &ml, &post_shift, &dummy);
                                          /* If the suggested multiplier is more than SIZE bits,
                                                         we can do better for even divisors, using an
                                                         initial right shift.  */
                                          if (mh != 0 && (d & 1) == 0)
                                          {
                                             pre_shift = floor_log2 (d & -d);
                                             mh = choose_multiplier (d >> pre_shift, data_bitsize,
                                                                     data_bitsize - pre_shift,
                                                                     &ml, &post_shift, &dummy);
                                             THROW_ASSERT (!mh, "unexpected condition");
                                          }
                                          else
                                             pre_shift = 0;
                                          tree_nodeRef quotient_expr;
                                          if (mh != 0)
                                          {
                                             THROW_ASSERT (post_shift - 1 < 64, "fail1");

                                             tree_nodeRef t1_ga_var = expand_mult_highpart(op0, ml, type_expr, data_bitsize, it_los, block.second,  srcp_default);
                                             THROW_ASSERT (t1_ga_var, "fail1");

                                             tree_nodeRef t2_expr = tree_man->create_binary_operation(type_expr, op0, t1_ga_var, srcp_default, minus_expr_K);
                                             tree_nodeRef t2_ga = CreateGimpleAssign(type_expr, t2_expr, block.first, srcp_default);
                                             block.second->PushBefore(t2_ga, *it_los);
                                             tree_nodeRef t2_ga_var = GetPointer<gimple_assign>(GET_NODE(t2_ga))->op0;

                                             unsigned int const1_nid = TM->new_tree_node_id();
                                             tree_nodeRef const1_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(1), const1_nid);
                                             tree_nodeRef t3_expr = tree_man->create_binary_operation(type_expr, t2_ga_var, const1_node, srcp_default, rshift_expr_K);
                                             tree_nodeRef t3_ga = CreateGimpleAssign(type_expr, t3_expr, block.first, srcp_default);
                                             block.second->PushBefore(t3_ga, *it_los);
                                             tree_nodeRef t3_ga_var = GetPointer<gimple_assign>(GET_NODE(t3_ga))->op0;

                                             tree_nodeRef t4_expr = tree_man->create_binary_operation(type_expr, t1_ga_var, t3_ga_var, srcp_default, plus_expr_K);
                                             tree_nodeRef t4_ga = CreateGimpleAssign(type_expr, t4_expr, block.first, srcp_default);
                                             block.second->PushBefore(t4_ga, *it_los);
                                             tree_nodeRef t4_ga_var = GetPointer<gimple_assign>(GET_NODE(t4_ga))->op0;

                                             THROW_ASSERT(post_shift>0, "unexpected condition");

                                             if(post_shift >1)
                                             {
                                                unsigned int post_shift_minusone_nid = TM->new_tree_node_id();
                                                tree_nodeRef post_shift_minusone_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(post_shift-1), post_shift_minusone_nid);
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t4_ga_var, post_shift_minusone_node, srcp_default, rshift_expr_K);
                                             }
                                             else
                                                quotient_expr = t4_ga_var;
                                          }
                                          else
                                          {
                                             THROW_ASSERT (pre_shift < 64 && post_shift < 64, "fail1");
                                             tree_nodeRef t1_ga_var;
                                             if(pre_shift!=0)
                                             {
                                                unsigned int pre_shift_nid = TM->new_tree_node_id();
                                                tree_nodeRef pre_shift_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(pre_shift), pre_shift_nid);
                                                tree_nodeRef t1_expr = tree_man->create_binary_operation(type_expr, op0, pre_shift_node, srcp_default, rshift_expr_K);
                                                tree_nodeRef t1_ga = CreateGimpleAssign(type_expr, t1_expr, block.first, srcp_default);
                                                block.second->PushBefore(t1_ga, *it_los);
                                                t1_ga_var = GetPointer<gimple_assign>(GET_NODE(t1_ga))->op0;
                                             }
                                             else
                                                t1_ga_var = op0;

                                             tree_nodeRef t2_ga_var = expand_mult_highpart(t1_ga_var, ml, type_expr, data_bitsize, it_los, block.second, srcp_default);

                                             if(post_shift != 0)
                                             {
                                                unsigned int post_shift_nid = TM->new_tree_node_id();
                                                tree_nodeRef post_shift_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(post_shift), post_shift_nid);
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t2_ga_var, post_shift_node, srcp_default, rshift_expr_K);
                                             }
                                             else
                                                quotient_expr = t2_ga_var;
                                          }
                                          if(rem_flag)
                                          {
                                             tree_nodeRef quotient_ga = CreateGimpleAssign(type_expr, quotient_expr, block.first, srcp_default);
                                             block.second->PushBefore(quotient_ga, *it_los);
                                             tree_nodeRef quotient_ga_var = GetPointer<gimple_assign>(GET_NODE(quotient_ga))->op0;
                                             tree_nodeRef mul_expr = tree_man->create_binary_operation(type_expr, quotient_ga_var, op1, srcp_default, mult_expr_K);
                                             tree_nodeRef mul_ga = CreateGimpleAssign(type_expr, mul_expr, block.first, srcp_default);
                                             block.second->PushBefore(mul_ga, *it_los);
                                             tree_nodeRef mul_ga_var = GetPointer<gimple_assign>(GET_NODE(mul_ga))->op0;
                                             /// restrict if needed the input bit-widths
                                             tree_nodeRef sub_expr;
                                             if(
#ifndef NDEBUG
                                                   function_behavior->ApplyNewTransformation() and
#endif
                                                   static_cast<unsigned int>(data_bitsize) > tree_helper::Size(GET_NODE(ga->op0)))
                                             {
                                                unsigned long long int  masklow = (static_cast<unsigned long long int>(1) << tree_helper::Size(GET_NODE(ga->op0))) - 1;
                                                unsigned int integer_masklow_id = TM->new_tree_node_id();
                                                tree_nodeRef Constmasklow = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(masklow), integer_masklow_id);
                                                tree_nodeRef temp_op0_expr = tree_man->create_binary_operation(type_expr, op0, Constmasklow, srcp_default, bit_and_expr_K);
                                                tree_nodeRef temp_op0_ga = CreateGimpleAssign(type_expr, temp_op0_expr, block.first, srcp_default);
#ifndef NDEBUG
                                                function_behavior->RegisterTransformation(GetName(), temp_op0_ga);
#endif
                                                block.second->PushBefore(temp_op0_ga, *it_los);
                                                tree_nodeRef temp_op0_ga_var = GetPointer<gimple_assign>(GET_NODE(temp_op0_ga))->op0;
                                                tree_nodeRef temp_op1_expr = tree_man->create_binary_operation(type_expr, mul_ga_var, Constmasklow, srcp_default, bit_and_expr_K);
                                                tree_nodeRef temp_op1_ga = CreateGimpleAssign(type_expr, temp_op1_expr, block.first, srcp_default);
#ifndef NDEBUG
                                                function_behavior->RegisterTransformation(GetName(), temp_op1_ga);
#endif
                                                block.second->PushBefore(temp_op1_ga, *it_los);
                                                tree_nodeRef temp_op1_ga_var = GetPointer<gimple_assign>(GET_NODE(temp_op1_ga))->op0;
                                                tree_nodeRef temp_sub_expr = tree_man->create_binary_operation(type_expr, temp_op0_ga_var, temp_op1_ga_var, srcp_default, minus_expr_K);
                                                tree_nodeRef temp_sub_expr_ga = CreateGimpleAssign(type_expr, temp_sub_expr, block.first, srcp_default);
#ifndef NDEBUG
                                                function_behavior->RegisterTransformation(GetName(), temp_sub_expr_ga);
#endif
                                                block.second->PushBefore(temp_sub_expr_ga, *it_los);
                                                tree_nodeRef temp_sub_expr_ga_var = GetPointer<gimple_assign>(GET_NODE(temp_sub_expr_ga))->op0;
                                                sub_expr = tree_man->create_binary_operation(type_expr, temp_sub_expr_ga_var, Constmasklow, srcp_default, bit_and_expr_K);

                                             }
                                             else
                                             {
                                                sub_expr = tree_man->create_binary_operation(type_expr, op0, mul_ga_var, srcp_default, minus_expr_K);
                                             }
                                             ga->op1 = sub_expr;
                                          }
                                          else
                                             ga->op1 = quotient_expr;
#ifndef NDEBUG
                                          function_behavior->RegisterTransformation(GetName(), *it_los);
#endif
                                          INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
                                          it_los = list_of_stmt.begin();
                                          it_los_end = list_of_stmt.end();
                                          continue;

                                       }
                                    }
                                 }
                                 else
                                 {
                                    INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---signed");
                                    tree_nodeRef new_op1;
                                    long long int d = ext_op1;
                                    unsigned long long int abs_d;

                                    /* Since d might be INT_MIN, we have to cast to
                                       unsigned long long before negating to avoid
                                       undefined signed overflow.  */
                                    abs_d = (d >= 0
                                          ? static_cast<unsigned long long int>(d)
                                          : static_cast<unsigned long long int>(-d));

                                    /* n rem d = n rem -d */
                                    if (rem_flag && d < 0)
                                    {
                                       d = static_cast<long long int>(abs_d);
                                    }

                                    unsigned int size = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr)));
                                    if(abs_d == static_cast<unsigned long long int>(1) << (size - 1))
                                    {
                                       tree_nodeRef quotient_expr = tree_man->create_binary_operation(type_expr, op0, op1, srcp_default, eq_expr_K);
                                       if(
#ifndef NDEBUG
                                          function_behavior->ApplyNewTransformation() and
#endif
                                          rem_flag)
                                       {
                                          tree_nodeRef quotient_ga = CreateGimpleAssign(type_expr, quotient_expr, block.first, srcp_default);
#ifndef NDEBUG
                                          function_behavior->RegisterTransformation(GetName(), quotient_ga);
#endif
                                          block.second->PushBefore(quotient_ga, *it_los);
                                          tree_nodeRef quotient_ga_var = GetPointer<gimple_assign>(GET_NODE(quotient_ga))->op0;
                                          tree_nodeRef mul_expr = tree_man->create_binary_operation(type_expr, quotient_ga_var, op1, srcp_default, mult_expr_K);
                                          tree_nodeRef mul_ga = CreateGimpleAssign(type_expr, mul_expr, block.first, srcp_default);
#ifndef NDEBUG
                                          function_behavior->RegisterTransformation(GetName(), mul_ga);
#endif
                                          block.second->PushBefore(mul_ga, *it_los);
                                          tree_nodeRef mul_ga_var = GetPointer<gimple_assign>(GET_NODE(mul_ga))->op0;
                                          tree_nodeRef sub_expr = tree_man->create_binary_operation(type_expr, op0, mul_ga_var, srcp_default, minus_expr_K);
                                          ga->op1 = sub_expr;
#ifndef NDEBUG
                                          function_behavior->RegisterTransformation(GetName(), *it_los);
#endif
                                          INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                                          it_los = list_of_stmt.begin();
                                          it_los_end = list_of_stmt.end();
                                          continue;
                                       }
                                       else
                                          ga->op1 = quotient_expr;

                                    }
                                    else if (EXACT_POWER_OF_2_OR_ZERO_P (abs_d))
                                    {
                                       if (rem_flag)
                                       {
                                          new_op1 = expand_smod_pow2 (op0, abs_d, *it_los, block.second, type_expr, srcp_default);
                                          ga->op1 = new_op1;
                                       }
                                       else
                                       {
                                          new_op1 = expand_sdiv_pow2 (op0, abs_d, *it_los, block.second, type_expr, srcp_default);
                                          /// We have computed OP0 / abs(OP1).  If OP1 is negative, negate the quotient.
                                          if (d < 0)
                                          {
                                             tree_nodeRef sdiv_pow2_ga = CreateGimpleAssign(type_expr, new_op1, block.first, srcp_default);
#ifndef NDEBUG
                                             function_behavior->RegisterTransformation(GetName(), sdiv_pow2_ga);
#endif
                                             block.second->PushBefore(sdiv_pow2_ga, *it_los);
                                             tree_nodeRef sdiv_pow2_ga_var = GetPointer<gimple_assign>(GET_NODE(sdiv_pow2_ga))->op0;
                                             new_op1 = tree_man->create_unary_operation(type_expr, sdiv_pow2_ga_var, srcp_default, negate_expr_K);
                                          }
                                          ga->op1 = new_op1;
                                       }
#ifndef NDEBUG
                                       function_behavior->RegisterTransformation(GetName(), *it_los);
#endif
                                       INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                                       it_los = list_of_stmt.begin();
                                       it_los_end = list_of_stmt.end();
                                       continue;
                                    }
#if 1
                                    else
                                    {
                                       int data_bitsize = static_cast<int>(tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr))));
                                       if(data_bitsize<=64)
                                       {
                                          unsigned long long int ml;
                                          int post_shift;
                                          int lgup;
                                          choose_multiplier (abs_d, data_bitsize, data_bitsize - 1,
                                                     &ml, &post_shift, &lgup);
                                          //std::cerr << "Multiplier " << ml << std::endl;
                                          //std::cerr << "post_shift " << post_shift << std::endl;
                                          //std::cerr << "lgup " << lgup << std::endl;
                                          tree_nodeRef quotient_expr;
                                          THROW_ASSERT (post_shift < 64 && size - 1 < 64, "unexpected condition");
                                          if (ml < 1ULL << (data_bitsize - 1))
                                          {
                                             tree_nodeRef t1_ga_var = expand_mult_highpart(op0, ml, type_expr, data_bitsize, it_los, block.second, srcp_default);
                                             THROW_ASSERT (t1_ga_var, "fail1");

                                             tree_nodeRef t2_ga_var;
                                             if(post_shift != 0)
                                             {
                                                unsigned int post_shift_nid = TM->new_tree_node_id();
                                                tree_nodeRef post_shift_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(post_shift), post_shift_nid);

                                                tree_nodeRef t2_expr = tree_man->create_binary_operation(type_expr, t1_ga_var, post_shift_node, srcp_default, rshift_expr_K);
                                                tree_nodeRef t2_ga = CreateGimpleAssign(type_expr, t2_expr, block.first, srcp_default);
                                                block.second->PushBefore(t2_ga, *it_los);
                                                t2_ga_var = GetPointer<gimple_assign>(GET_NODE(t2_ga))->op0;
                                             }
                                             else
                                                t2_ga_var = t1_ga_var;


                                             unsigned int data_bitsize_minusone_nid = TM->new_tree_node_id();
                                             tree_nodeRef data_bitsize_minusone_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(data_bitsize-1), data_bitsize_minusone_nid);
                                             tree_nodeRef t3_expr = tree_man->create_binary_operation(type_expr, op0, data_bitsize_minusone_node, srcp_default, rshift_expr_K);
                                             tree_nodeRef t3_ga = CreateGimpleAssign(type_expr, t3_expr, block.first, srcp_default);
                                             block.second->PushBefore(t3_ga, *it_los);
                                             tree_nodeRef t3_ga_var = GetPointer<gimple_assign>(GET_NODE(t3_ga))->op0;


                                             if (d < 0)
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t3_ga_var, t2_ga_var, srcp_default, minus_expr_K);
                                             else
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t2_ga_var, t3_ga_var, srcp_default, minus_expr_K);
                                          }
                                          else
                                          {

                                             ml |= (~ 0ULL) << (data_bitsize - 1);
                                             tree_nodeRef t1_ga_var = expand_mult_highpart(op0, ml, type_expr, data_bitsize, it_los, block.second, srcp_default);
                                             THROW_ASSERT (t1_ga_var, "fail1");
                                             tree_nodeRef t2_expr = tree_man->create_binary_operation(type_expr, t1_ga_var, op0, srcp_default, plus_expr_K);
                                             tree_nodeRef t2_ga = CreateGimpleAssign(type_expr, t2_expr, block.first, srcp_default);
                                             block.second->PushBefore(t2_ga, *it_los);
                                             tree_nodeRef t2_ga_var = GetPointer<gimple_assign>(GET_NODE(t2_ga))->op0;

                                             unsigned int post_shift_nid = TM->new_tree_node_id();
                                             tree_nodeRef post_shift_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(post_shift), post_shift_nid);

                                             tree_nodeRef t3_expr = tree_man->create_binary_operation(type_expr, t2_ga_var, post_shift_node, srcp_default, rshift_expr_K);
                                             tree_nodeRef t3_ga = CreateGimpleAssign(type_expr, t3_expr, block.first, srcp_default);
                                             block.second->PushBefore(t3_ga, *it_los);
                                             tree_nodeRef t3_ga_var = GetPointer<gimple_assign>(GET_NODE(t3_ga))->op0;

                                             unsigned int data_bitsize_minusone_nid = TM->new_tree_node_id();
                                             tree_nodeRef data_bitsize_minusone_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(data_bitsize-1), data_bitsize_minusone_nid);
                                             tree_nodeRef t4_expr = tree_man->create_binary_operation(type_expr, op0, data_bitsize_minusone_node, srcp_default, rshift_expr_K);
                                             tree_nodeRef t4_ga = CreateGimpleAssign(type_expr, t4_expr, block.first, srcp_default);
                                             block.second->PushBefore(t4_ga, *it_los);
                                             tree_nodeRef t4_ga_var = GetPointer<gimple_assign>(GET_NODE(t4_ga))->op0;

                                             if (d < 0)
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t4_ga_var, t3_ga_var, srcp_default, minus_expr_K);
                                             else
                                                quotient_expr = tree_man->create_binary_operation(type_expr, t3_ga_var, t4_ga_var, srcp_default, minus_expr_K);

                                          }
                                          if(rem_flag)
                                          {
                                             tree_nodeRef quotient_ga = CreateGimpleAssign(type_expr, quotient_expr, block.first, srcp_default);
                                             block.second->PushBefore(quotient_ga, *it_los);
                                             tree_nodeRef quotient_ga_var = GetPointer<gimple_assign>(GET_NODE(quotient_ga))->op0;
                                             tree_nodeRef mul_expr = tree_man->create_binary_operation(type_expr, quotient_ga_var, op1, srcp_default, mult_expr_K);
                                             tree_nodeRef mul_ga = CreateGimpleAssign(type_expr, mul_expr, block.first, srcp_default);
                                             block.second->PushBefore(mul_ga, *it_los);
                                             tree_nodeRef mul_ga_var = GetPointer<gimple_assign>(GET_NODE(mul_ga))->op0;
                                             tree_nodeRef sub_expr = tree_man->create_binary_operation(type_expr, op0, mul_ga_var, srcp_default, minus_expr_K);
                                             ga->op1 = sub_expr;
                                          }
                                          else
                                             ga->op1 = quotient_expr;
#ifndef NDEBUG
                                          function_behavior->RegisterTransformation(GetName(), *it_los);
#endif
                                          INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
                                          it_los = list_of_stmt.begin();
                                          it_los_end = list_of_stmt.end();
                                          continue;

                                       }
                                    }
#endif
                                 }
                                 break;
                              }
                           case exact_div_expr_K:
                              {
                                 int pre_shift;
                                 long long int d = ext_op1;
                                 unsigned long long int ml;
                                 unsigned int size = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(type_expr)));

                                 pre_shift = floor_log2 (static_cast<unsigned long long int>(d & -d));
                                 ml = invert_mod2n (static_cast<unsigned long long int>(d >> pre_shift), size);
                                 unsigned int pre_shift_nid = TM->new_tree_node_id();
                                 tree_nodeRef pre_shift_node = tree_man->CreateIntegerCst(type_expr, pre_shift, pre_shift_nid);
                                 unsigned int ml_nid = TM->new_tree_node_id();
                                 tree_nodeRef ml_node = tree_man->CreateIntegerCst(type_expr, static_cast<long long int>(ml), ml_nid);
                                 tree_nodeRef t1_expr = tree_man->create_binary_operation(type_expr, op0, pre_shift_node, srcp_default, rshift_expr_K);
                                 tree_nodeRef t1_ga = CreateGimpleAssign(type_expr, t1_expr, block.first, srcp_default);
                                 block.second->PushBefore(t1_ga, *it_los);
                                 tree_nodeRef t1_ga_var = GetPointer<gimple_assign>(GET_NODE(t1_ga))->op0;
                                 tree_nodeRef quotient_expr = tree_man->create_binary_operation(type_expr, t1_ga_var, ml_node, srcp_default, mult_expr_K);

                                 ga->op1 = quotient_expr;
#ifndef NDEBUG
                                 function_behavior->RegisterTransformation(GetName(), *it_los);
#endif

                                 break;
                              }
                           case assert_expr_K:
                           case bit_and_expr_K:
                           case bit_ior_expr_K:
                           case bit_xor_expr_K:
                           case catch_expr_K:
                           case ceil_div_expr_K:
                           case ceil_mod_expr_K:
                           case complex_expr_K:
                           case compound_expr_K:
                           case eh_filter_expr_K:
                           case eq_expr_K:
                           case fdesc_expr_K:
                           case floor_div_expr_K:
                           case floor_mod_expr_K:
                           case ge_expr_K:
                           case gt_expr_K:
                           case goto_subroutine_K:
                           case in_expr_K:
                           case init_expr_K:
                           case le_expr_K:
                           case lrotate_expr_K:
                           case lshift_expr_K:
                           case lt_expr_K:
                           case max_expr_K:
                           case mem_ref_K:
                           case min_expr_K:
                           case minus_expr_K:
                           case modify_expr_K:
                           case mult_expr_K:
                           case mult_highpart_expr_K:
                           case ne_expr_K:
                           case ordered_expr_K:
                           case plus_expr_K:
                           case pointer_plus_expr_K:
                           case postdecrement_expr_K:
                           case postincrement_expr_K:
                           case predecrement_expr_K:
                           case preincrement_expr_K:
                           case range_expr_K:
                           case rdiv_expr_K:
                           case round_div_expr_K:
                           case round_mod_expr_K:
                           case rrotate_expr_K:
                           case rshift_expr_K:
                           case set_le_expr_K:
                           case truth_and_expr_K:
                           case truth_andif_expr_K:
                           case truth_or_expr_K:
                           case truth_orif_expr_K:
                           case truth_xor_expr_K:
                           case try_catch_expr_K:
                           case try_finally_K:
                           case uneq_expr_K:
                           case ltgt_expr_K:
                           case unge_expr_K:
                           case ungt_expr_K:
                           case unle_expr_K:
                           case unlt_expr_K:
                           case unordered_expr_K:
                           case widen_sum_expr_K:
                           case widen_mult_expr_K:
                           case with_size_expr_K:
                           case vec_lshift_expr_K:
                           case vec_rshift_expr_K:
                           case widen_mult_hi_expr_K:
                           case widen_mult_lo_expr_K:
                           case vec_pack_trunc_expr_K:
                           case vec_pack_sat_expr_K:
                           case vec_pack_fix_trunc_expr_K:
                           case vec_extracteven_expr_K:
                           case vec_extractodd_expr_K:
                           case vec_interleavehigh_expr_K:
                           case vec_interleavelow_expr_K:
                           case binfo_K:
                           case block_K:
                           case call_expr_K:
                           case case_label_expr_K:
                           case constructor_K:
                           case identifier_node_K:
                           case ssa_name_K:
                           case statement_list_K:
                           case target_mem_ref_K:
                           case target_mem_ref461_K:
                           case tree_list_K:
                           case tree_vec_K:
                           case CASE_CPP_NODES:
                           case CASE_CST_NODES:
                           case CASE_DECL_NODES:
                           case CASE_FAKE_NODES:
                           case CASE_GIMPLE_NODES:
                           case CASE_PRAGMA_NODES:
                           case CASE_QUATERNARY_EXPRESSION:
                           case CASE_TERNARY_EXPRESSION:
                           case CASE_TYPE_NODES:
                           case CASE_UNARY_EXPRESSION:
                           default:
                              {
                                 THROW_ERROR("not yet supported code: " + GET_NODE(ga->op1)->get_kind_text());
                                 break;
                              }
                        }
                     }
                  }
               }
            }
            else if(code1 == mult_expr_K)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Expanding mult_expr  " + STR(GET_INDEX_NODE(ga->op1)));
               tree_nodeRef op1 = GetPointer<binary_expr>(GET_NODE(ga->op1))->op1;
               if(GetPointer<cst_node>(GET_NODE(op1)) && !GetPointer<vector_cst>(GET_NODE(op1)))
               {
                  cst_node* cn = GetPointer<cst_node>(GET_NODE(op1));
                  tree_nodeRef op0 = GetPointer<binary_expr>(GET_NODE(ga->op1))->op0;
                  tree_nodeRef type_expr = GetPointer<binary_expr>(GET_NODE(ga->op1))->type;

                  bool realp = tree_helper::is_real(TM, GET_INDEX_NODE(type_expr));
                  if(!realp)
                  {
                     ga->op1 = expand_MC(op0, static_cast<integer_cst*>(cn), ga->op1, *it_los, block.second, type_expr, srcp_default);
                  }
               }
               else
               {
                  tree_nodeRef type_expr = GetPointer<binary_expr>(GET_NODE(ga->op1))->type;

                  bool realp = tree_helper::is_real(TM, GET_INDEX_NODE(type_expr));
                  if(!realp)
                  {
                     ///check if a mult_expr may become a widen_mult_expr
                     unsigned int data_bitsize_out = resize_to_1_8_16_32_64_128_256_512(tree_helper::Size(GET_NODE(ga->op0)));
                     unsigned int data_bitsize_in0 = resize_to_1_8_16_32_64_128_256_512(tree_helper::Size(GET_NODE(GetPointer<binary_expr>(GET_NODE(ga->op1))->op0)));
                     unsigned int data_bitsize_in1 = resize_to_1_8_16_32_64_128_256_512(tree_helper::Size(GET_NODE(GetPointer<binary_expr>(GET_NODE(ga->op1))->op1)));
                     if(std::max(data_bitsize_in0, data_bitsize_in1)*2== data_bitsize_out)
                     {
                        unsigned int type_index = tree_helper::get_type_index(TM, GET_INDEX_NODE(ga->op0));
                        tree_nodeRef op0_type= TM->GetTreeReindex(type_index);
                        ga->op1 = tree_man->create_binary_operation(op0_type, GetPointer<binary_expr>(GET_NODE(ga->op1))->op0, GetPointer<binary_expr>(GET_NODE(ga->op1))->op1, srcp_default, widen_mult_expr_K);
                     }
                  }
               }
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Expanded");
            }
            else if(code1 == widen_mult_expr_K)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Expanding widen_mult_expr  " + STR(GET_INDEX_NODE(ga->op1)));
               const auto be = GetPointer<binary_expr>(GET_NODE(ga->op1));
               tree_nodeRef op1 = be->op1;
               if(GetPointer<cst_node>(GET_NODE(op1)) && !GetPointer<vector_cst>(GET_NODE(op1)))
               {
                  cst_node* cn = GetPointer<cst_node>(GET_NODE(op1));
                  tree_nodeRef op0 = GetPointer<binary_expr>(GET_NODE(ga->op1))->op0;
                  tree_nodeRef type_expr = GetPointer<binary_expr>(GET_NODE(ga->op1))->type;

                  bool realp = tree_helper::is_real(TM, GET_INDEX_NODE(type_expr));
                  if(!realp)
                  {
                     ga->op1 = expand_MC(op0, static_cast<integer_cst*>(cn), ga->op1, *it_los, block.second, type_expr, srcp_default);
                  }
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Expanded");
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
               }
               else
               {
                  if(tree_helper::is_unsigned(TM, be->type->index) != tree_helper::is_unsigned(TM, be->op0->index))
                  {
                     THROW_ASSERT(tree_helper::is_unsigned(TM, be->type->index), "Conversion from unsigned to signed is required for first input of " + STR(ga->op1->index));
                     const auto ga_nop = tree_man->CreateNopExpr(be->op0, tree_man->CreateUnsigned(tree_helper::CGetType(GET_NODE(be->op0))));
                     if(ga_nop)
                     {
                        be->op0 = GetPointer<gimple_assign>(GET_NODE(ga_nop))->op0;
#ifndef NDEBUG
                        function_behavior->RegisterTransformation(GetName(), ga_nop);
#endif
                        block.second->PushBefore(ga_nop, *it_los);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Not Expanded");
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
                        it_los = list_of_stmt.begin();
                        it_los_end = list_of_stmt.end();
                        continue;
                     }
                     else
                     {
#ifndef NDEBUG
                        THROW_UNREACHABLE("Conversion of " + be->op0->ToString() + " cannot be created");
#else
                        THROW_WARNING("Implicit type conversion for first input of " + ga->ToString());
#endif
                     }
                  }
                  if(tree_helper::is_unsigned(TM, be->type->index) != tree_helper::is_unsigned(TM, be->op1->index))
                  {
                     THROW_ASSERT(tree_helper::is_unsigned(TM, be->type->index), "Conversion from unsigned to signed is required for first input of " + STR(ga->op1->index));
                     const auto ga_nop = tree_man->CreateNopExpr(be->op1, tree_man->CreateUnsigned(tree_helper::CGetType(GET_NODE(be->op1))));
                     if(ga_nop)
                     {
                        be->op1 = GetPointer<gimple_assign>(GET_NODE(ga_nop))->op0;
                        block.second->PushBefore(ga_nop, *it_los);
#ifndef NDEBUG
                        function_behavior->RegisterTransformation(GetName(), ga_nop);
#endif
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Not Expanded");
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
                        it_los = list_of_stmt.begin();
                        it_los_end = list_of_stmt.end();
                        continue;
                     }
                     else
                     {
#ifndef NDEBUG
                        THROW_UNREACHABLE("Conversion of " + be->op1->ToString() + " cannot be created");
#else
                        THROW_WARNING("Implicit type conversion for first input of " + ga->ToString());
#endif
                     }
                  }
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Not Expanded");
               }
            }
            else if(code1 == var_decl_K)
            {
               /// check if volatile attribute is missing
               /// it may happen with gcc v4.8 or greater
               var_decl *vd = GetPointer<var_decl>(GET_NODE(ga->op1));
               if(vd->scpe && GET_NODE(vd->scpe)->get_kind() != translation_unit_decl_K && !vd->static_flag && !tree_helper::is_volatile(TM,GET_INDEX_NODE(ga->op1)) &&
                     GET_NODE(vd->type)->get_kind() != array_type_K && GET_NODE(vd->type)->get_kind() != record_type_K && GET_NODE(vd->type)->get_kind() != union_type_K)
               {
                  /// w.r.t. bambu this kind of non-ssa variable is equivalent to a static local variable
                  vd->static_flag = true;
               }
            }



            if(code0 == target_mem_ref461_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand target_mem_ref 2 " + STR(GET_INDEX_NODE(ga->op0)));

               target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(GET_NODE(ga->op0));

               /// expose some part of the target_mem_ref statement
               expand_target_mem_ref(tmr, *it_los, block.second, srcp_default);

               /// split the target_mem_ref in a address computation statement and in a store statement
               tree_nodeRef type = tmr->type;

               tree_nodeRef pt = tree_man->create_pointer_type(type);
               tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op0, srcp_default, addr_expr_K);
               tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first, srcp_default);
               GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;
               tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

               unsigned int integer_cst0_id = TM->new_tree_node_id();
               tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
               tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

               ga->op0 = mr;
               block.second->PushBefore(new_ga, *it_los);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
               it_los = list_of_stmt.begin();
               it_los_end = list_of_stmt.end();
               continue;
            }
            else if(code0  == mem_ref_K && code1 != mem_ref_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand mem_ref 2 " + STR(GET_INDEX_NODE(ga->op0)));
               mem_ref * MR = GetPointer<mem_ref>(GET_NODE(ga->op0));

               tree_nodeRef op1 = MR->op1;
               long long int op1_val = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(op1)));
               if(op1_val != 0)
               {
                  tree_nodeRef type = MR->type;

                  tree_nodeRef pt = tree_man->create_pointer_type(type);
                  tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op0, srcp_default, addr_expr_K);
                  tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first, srcp_default);
                  GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;
                  tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

                  unsigned int integer_cst0_id = TM->new_tree_node_id();
                  tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
                  tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

                  ga->op0 = mr;
                  block.second->PushBefore(new_ga, *it_los);
#ifndef NDEBUG
                  function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
                  it_los = list_of_stmt.begin();
                  it_los_end = list_of_stmt.end();
                  continue;
               }
            }
            /*
            else if(code0  == array_ref_K)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "expand array_ref 2 " + STR(GET_INDEX_NODE(ga->op0)));
               array_ref * AR = GetPointer<array_ref>(GET_NODE(ga->op0));

               std::string srcp_default = fd->include_name +":0:0";
               tree_nodeRef type = AR->type;

               tree_nodeRef pt = tree_man->create_pointer_type(type);
               tree_nodeRef ae = tree_man->create_unary_operation(pt,ga->op0, srcp_default, addr_expr_K);
               tree_nodeRef new_ga = CreateGimpleAssign(pt, ae, block.first);
               GetPointer<gimple_assign>(GET_NODE(new_ga))->temporary_address = true;
               tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;

               unsigned int integer_cst0_id = TM->new_tree_node_id();
               tree_nodeRef offset = tree_man->CreateIntegerCst(pt, 0, integer_cst0_id);
               tree_nodeRef mr = tree_man->create_binary_operation(type, ssa_vd, offset, srcp_default, mem_ref_K);

               ga->op0 = mr;
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
               block.second->PushBefore(it_los, new_ga);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
               it_los = list_of_stmt.begin();
               it_los_end = list_of_stmt.end();
               continue;
            }
            */
            else if(code0 == var_decl_K)
            {
               /// check if volatile attribute is missing
               /// it may happen with gcc v4.8 or greater
               var_decl *vd = GetPointer<var_decl>(GET_NODE(ga->op0));
               if(vd->scpe && GET_NODE(vd->scpe)->get_kind() != translation_unit_decl_K && !vd->static_flag && !tree_helper::is_volatile(TM,GET_INDEX_NODE(ga->op0)) &&
                     GET_NODE(vd->type)->get_kind() != array_type_K /*&& GET_NODE(vd->type)->get_kind() != complex_type_K*/ && GET_NODE(vd->type)->get_kind() != record_type_K && GET_NODE(vd->type)->get_kind() != union_type_K)
               {
                  /// w.r.t. bambu this kind of non-ssa variable is equivalent to a static local variable
                  vd->static_flag = true;
               }
            }
            else if(code0 == result_decl_K && code1 == ssa_name_K)
            {
               unsigned type_index;
               tree_nodeRef type_node = tree_helper::get_type_node(GET_NODE(ga->op1), type_index);
               if(type_node->get_kind() == complex_type_K or type_node->get_kind() == vector_type_K)
               {
                  auto it_los_next = std::next(it_los);
                  if(it_los_next != it_los_end && GET_NODE(*it_los_next)->get_kind() == gimple_return_K)
                  {
                     gimple_return *gr = GetPointer<gimple_return>(GET_NODE(*it_los_next));
                     gr->op = ga->op1;
                     block.second->RemoveStmt(*it_los);
                     it_los = list_of_stmt.begin();
                     it_los_end = list_of_stmt.end();
                  }
               }
            }
            else if(code0 == result_decl_K)
            {
               unsigned int type_index = tree_helper::get_type_index(TM, GET_INDEX_NODE(ga->op0));
               tree_nodeRef op_type= TM->GetTreeReindex(type_index);
               tree_nodeRef new_ga = CreateGimpleAssign(op_type, ga->op1, block.first, srcp_default);
               tree_nodeRef ssa_vd = GetPointer<gimple_assign>(GET_NODE(new_ga))->op0;
               ga->op1 = ssa_vd;
               block.second->PushBefore(new_ga, *it_los);
#ifndef NDEBUG
               function_behavior->RegisterTransformation(GetName(), new_ga);
#endif
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Modified statement " + GET_NODE(*it_los)->ToString());
               it_los = list_of_stmt.begin();
               it_los_end = list_of_stmt.end();
               continue;
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examined statement " + GET_NODE(*it_los)->ToString());
         it_los++;
      }

      if(n_stmts != list_of_stmt.size())
         restart_check_system_type = true;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Examining BB" + STR(block.first));
   }
   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      PrintTreeManager(false);
   }
   function_behavior->UpdateBBVersion();
   return DesignFlowStep_Status::SUCCESS;
}

void IR_lowering::Initialize()
{
   TM = AppM->get_tree_manager();
   tree_man = tree_manipulationRef(new tree_manipulation(TM, parameters));
   if(GetPointer<const HLS_manager>(AppM) and GetPointer<const HLS_manager>(AppM)->get_HLS(function_id) and GetPointer<const HLS_manager>(AppM)->get_HLS(function_id)->Rsch)
   {
      schedule = GetPointer<const HLS_manager>(AppM)->get_HLS(function_id)->Rsch;
   }
}

void IR_lowering::ComputeRelationships(DesignFlowStepSet & relationship, const DesignFlowStep::RelationshipType relationship_type)
{
   switch(relationship_type)
   {
      case(PRECEDENCE_RELATIONSHIP):
         {
            break;
         }
      case DEPENDENCE_RELATIONSHIP:
         {
            const DesignFlowGraphConstRef design_flow_graph = design_flow_manager.lock()->CGetDesignFlowGraph();
            const TechnologyFlowStepFactory * technology_flow_step_factory = GetPointer<const TechnologyFlowStepFactory>(design_flow_manager.lock()->CGetDesignFlowStepFactory("Technology"));
            const std::string technology_flow_signature = TechnologyFlowStep::ComputeSignature(TechnologyFlowStep_Type::LOAD_TECHNOLOGY);
            const vertex technology_flow_step = design_flow_manager.lock()->GetDesignFlowStep(technology_flow_signature);
            const DesignFlowStepRef technology_design_flow_step = technology_flow_step ? design_flow_graph->CGetDesignFlowStepInfo(technology_flow_step)->design_flow_step : technology_flow_step_factory->CreateTechnologyFlowStep(TechnologyFlowStep_Type::LOAD_TECHNOLOGY);
            relationship.insert(technology_design_flow_step);
            break;
         }
      case INVALIDATION_RELATIONSHIP:
         {
            break;
         }
      default:
         THROW_UNREACHABLE("");
   }
   FunctionFrontendFlowStep::ComputeRelationships(relationship, relationship_type);
}
