/*
 *                 _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *               _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *              _/      _/    _/ _/    _/ _/   _/ _/    _/
 *             _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *           ***********************************************
 *                            PandA Project
 *                   URL: http://panda.dei.polimi.it
 *                     Politecnico di Milano - DEIB
 *                      System Architectures Group
 *           ***********************************************
 *            Copyright (c) 2004-2016 Politecnico di Milano
 *
 * This file is part of the PandA framework.
 *
 * The PandA framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "call_graph_builtin_call.hpp"

#include <map>
#include <string>

#include "application_manager.hpp"
#include "basic_block.hpp"
#include "call_graph.hpp"
#include "call_graph_manager.hpp"
#include "design_flow_manager.hpp"
#include "ext_tree_node.hpp"
#include "function_behavior.hpp"
#include "Parameter.hpp"
#include "symbolic_application_frontend_flow_step.hpp"
#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "tree_helper.hpp"
#include "function_decl_refs.hpp"
#include "op_graph.hpp"

static tree_nodeRef getFunctionPointerType(tree_nodeRef fptr);
void
CallGraphBuiltinCall::lookForBuiltinCall(const tree_nodeRef TN)
{
   THROW_ASSERT(TN->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,
                  "--> Update recursively " +
                  boost::lexical_cast<std::string>(GET_INDEX_NODE(TN)));

   const tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef currentTreeNode = GET_NODE(TN);
   switch (currentTreeNode->get_kind())
   {
      case call_expr_K:
      {
         const call_expr * CE = GetPointer<call_expr>(currentTreeNode);
         const std::vector<tree_nodeRef> & args = CE->args;

         tree_nodeRef FN = GetPointer<addr_expr>(GET_NODE(CE->fn)) ?
                           GetPointer<addr_expr>(GET_NODE(CE->fn))->op
                           : CE->fn;
         THROW_ASSERT(FN, "Address expression with null op");
         if (GET_NODE(FN)->get_kind() == function_decl_K)
         {
            unsigned int functionDeclIdx = FN->index;
            std::string funName = tree_helper::name_function(TM, functionDeclIdx);
            if (funName == BUILTIN_WAIT_CALL)
            {
               tree_nodeRef builtinArgZero = GetPointer<addr_expr>(GET_NODE(args[0])) ?
                                             GetPointer<addr_expr>(GET_NODE(args[0]))->op
                                             : args[0];
               if (GET_NODE(builtinArgZero)->get_kind() == function_decl_K)
               {
                  unsigned int calledFunctionId = builtinArgZero->index;
                  AppM->add_function(function_id, calledFunctionId,
                                     AppM->GetFunctionBehavior(calledFunctionId)->GetBehavioralHelper(),
                                     GET_INDEX_NODE(TN), false);
                  AppM->addToBeInterfaced(calledFunctionId);
                  modified = true;
               }
               else if (GET_NODE(builtinArgZero)->get_kind() == ssa_name_K)
               {
                  tree_nodeRef funPtrType = getFunctionPointerType(GET_NODE(builtinArgZero));
                  ExtendCallGraph(function_id, funPtrType, CE->index);
               }
            }
         }
         break;
      }
      case gimple_call_K:
      {
         const gimple_call * CE = GetPointer<gimple_call>(currentTreeNode);
         const std::vector<tree_nodeRef> & args = CE->args;

         tree_nodeRef FN = GetPointer<addr_expr>(GET_NODE(CE->fn)) ?
                           GetPointer<addr_expr>(GET_NODE(CE->fn))->op
                           : CE->fn;
         THROW_ASSERT(FN, "Address expression with null op");
         if (GET_NODE(FN)->get_kind() == function_decl_K)
         {
            unsigned int functionDeclIdx = GET_NODE(FN)->index;
            std::string funName = tree_helper::name_function(TM, functionDeclIdx);
            if (funName == BUILTIN_WAIT_CALL)
            {
               tree_nodeRef builtinArgZero = GetPointer<addr_expr>(GET_NODE(args[0])) ?
                                             GetPointer<addr_expr>(GET_NODE(args[0]))->op
                                             : args[0];
               if (GET_NODE(builtinArgZero)->get_kind() == function_decl_K)
               {
                  unsigned int calledFunctionId = builtinArgZero->index;
                  AppM->add_function(function_id, calledFunctionId,
                                     AppM->GetFunctionBehavior(calledFunctionId)->GetBehavioralHelper(),
                                     GET_INDEX_NODE(TN), false);
                  AppM->addToBeInterfaced(calledFunctionId);
                  modified = true;
               }
               else if (GET_NODE(builtinArgZero)->get_kind() == ssa_name_K)
               {
                  // Function pointers case.
                  tree_nodeRef funPtrType = getFunctionPointerType(GET_NODE(builtinArgZero));
                  ExtendCallGraph(function_id, funPtrType, CE->index);
               }
            }
         }
         break;
      }
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(currentTreeNode);
         if (GetPointer<call_expr>(GET_NODE(gm->op1)))
         {
            lookForBuiltinCall(gm->op1);
         }
         else if(GetPointer<addr_expr>(GET_NODE(gm->op1)))
         {
            addr_expr* ae = GetPointer<addr_expr>(GET_NODE(gm->op1));
            if(GetPointer<function_decl>(GET_NODE(ae->op)))
            {
               AppM->addToBeInterfaced(GET_INDEX_NODE(ae->op));
            }
         }
         break;
      }
      case gimple_nop_K:
      case var_decl_K:
      case parm_decl_K:
      case ssa_name_K:
      case tree_list_K:
      case CASE_UNARY_EXPRESSION:
      case CASE_BINARY_EXPRESSION:
      case CASE_TERNARY_EXPRESSION:
      case CASE_QUATERNARY_EXPRESSION:
      case constructor_K:
      case gimple_cond_K:
      case gimple_switch_K:
      case gimple_multi_way_if_K:
      case gimple_return_K:
      case gimple_for_K:
      case gimple_while_K:
      case CASE_TYPE_NODES:
      case type_decl_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
      case result_decl_K:
      case vector_cst_K:
      case tree_vec_K:
      case case_label_expr_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_goto_K:
      case CASE_PRAGMA_NODES:
      case binfo_K:
      case block_K:
      case const_decl_K:
      case CASE_CPP_NODES:
      case gimple_bind_K:
      case gimple_phi_K:
      case gimple_pragma_K:
      case gimple_predict_K:
      case gimple_resx_K:
      case identifier_node_K:
      case last_tree_K:
      case namespace_decl_K:
      case none_K:
      case placeholder_expr_K:
      case statement_list_K:
      case translation_unit_decl_K:
      case tree_reindex_K:
      default:
      {
         THROW_UNREACHABLE("");
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,
                  "<-- Updated recursively " +
                  boost::lexical_cast<std::string>(GET_INDEX_NODE(TN)));
}


void
CallGraphBuiltinCall::ExtendCallGraph(
    unsigned int callerIdx, tree_nodeRef funType,
    unsigned int stmtIdx)
{
  BehavioralHelperRef helper;
  std::string type = tree_helper::getFunctionTypeString(funType);
  for (TypeDeclarationMap::mapped_type::iterator
           Itr = typeToDeclaration[type].begin(),
           End = typeToDeclaration[type].end();
       Itr != End; ++Itr)
  {
    AppM->add_function(callerIdx, *Itr, helper, stmtIdx, false);
    AppM->addToBeInterfaced(*Itr);
    modified = true;
  }
}


const std::unordered_set<std::pair<FrontendFlowStepType, CallGraphBuiltinCall::FunctionRelationship> >
CallGraphBuiltinCall::ComputeFrontendRelationships(
    DesignFlowStep::RelationshipType RT) const
{
   std::unordered_set<std::pair<FrontendFlowStepType, FunctionRelationship> > relationships;
   switch(RT)
   {
      case (DEPENDENCE_RELATIONSHIP) :
      {
         relationships.insert(std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship>(HWCALL_INJECTION, SAME_FUNCTION));
         break;
      }
      case (INVALIDATION_RELATIONSHIP) :
      {
         break;
      }
      case (PRECEDENCE_RELATIONSHIP) :
      {
         break;
      }
      default:
         THROW_UNREACHABLE("HWCallInjection::ComputeFrontendRelationships");
   }
   return relationships;
}


CallGraphBuiltinCall::CallGraphBuiltinCall(
    const application_managerRef AM,
    unsigned int functionId,
    const DesignFlowManagerConstRef DFM,
    const ParameterConstRef P) :
    FunctionFrontendFlowStep(AM, functionId, CALL_GRAPH_BUILTIN_CALL, DFM, P),
    modified(false)
{
   debug_level = P->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}


CallGraphBuiltinCall::~CallGraphBuiltinCall() {}


void
CallGraphBuiltinCall::Initialize()
{
  modified = false;
}


DesignFlowStep_Status
CallGraphBuiltinCall::InternalExec()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->BuiltinWaitCall Analysis");
   tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef currentTreeNode = TM->GetTreeNode(function_id);
   function_decl * functionDecl = GetPointer<function_decl>(currentTreeNode);
   statement_list * stmtList =
         GetPointer<statement_list>(GET_NODE(functionDecl->body));

   if(parameters->getOption<bool>(OPT_print_dot))
   {
      AppM->CGetCallGraphManager()->CGetCallGraph()->WriteDot("builtin-graph-pre" + STR(function_id) +".dot");
   }

   // Build the typeToDeclarationMap
   const CallGraphManagerRef CGM = AppM->GetCallGraphManager();
   unsigned int root_function = CGM->GetRootFunction();
   std::unordered_set<unsigned int> allFunctions;
   function_decl_refs fdr_visitor(allFunctions);
   tree_nodeRef rf = TM->get_tree_node_const(root_function);
   rf->visit(&fdr_visitor);
   for (std::unordered_set<unsigned int>::const_iterator
            Itr = allFunctions.begin(), End = allFunctions.end(); Itr != End; ++Itr)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,
                     "---Analyzing function " + boost::lexical_cast<std::string>(*Itr));
      tree_nodeRef function = TM->get_tree_node_const(*Itr);
      function_decl * funDecl = GetPointer<function_decl>(function);
      std::string type = tree_helper::getFunctionTypeString(GET_NODE(funDecl->type));
      std::string functionName = tree_helper::name_function(TM, *Itr);
      if(funDecl->body && functionName != "__start_pragma__" && functionName != "__close_pragma__" &&
              functionName.find("__pragma__") != 0)
      typeToDeclaration[type].insert(*Itr);
   }
   for(const auto block : stmtList->list_of_bloc)
   {

      for(const auto stmt : block.second->CGetStmtList())
      {
         if(GET_NODE(stmt)->get_kind() == gimple_call_K or GET_NODE(stmt)->get_kind() == gimple_assign_K)
         {
            lookForBuiltinCall(stmt);
         }
      }
   }
   AppM->GetCallGraphManager()->ComputeReachedFunctions();

   if(parameters->getOption<bool>(OPT_print_dot))
   {
      AppM->CGetCallGraphManager()->CGetCallGraph()->WriteDot("builtin-graph-post" + STR(function_id) +".dot");
   }

   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--BuiltinWaitCall Analysis completed");
   return modified ? DesignFlowStep_Status::SUCCESS : DesignFlowStep_Status::UNCHANGED;
}


static tree_nodeRef
getFunctionPointerType(tree_nodeRef fptr)
{
  ssa_name* sa = GetPointer<ssa_name>(fptr);
  THROW_ASSERT(sa, "Function pointer not in SSA-form");
  pointer_type* pt;
  if(sa->var)
  {
    decl_node* var = GetPointer<decl_node>(GET_NODE(sa->var));
    THROW_ASSERT(var, "Call expression does not point to a declaration node");
    pt = GetPointer<pointer_type>(GET_NODE(var->type));
  }
  else
    pt = GetPointer<pointer_type>(GET_NODE(sa->type));

  THROW_ASSERT(pt, "Declaration node has not information about pointer_type");
  THROW_ASSERT(GetPointer<function_type>(GET_NODE(pt->ptd)),
               "Pointer type has not information about pointed function_type");

  return GET_NODE(pt->ptd);
}
