/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file FSM_NI_SSA_liveness.cpp
 * @brief liveness analysis exploiting the SSA form of the IR
 *
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/

///Autoheader include
#include "FSM_NI_SSA_liveness.hpp"

///. include
#include "Parameter.hpp"

///behavior includes
#include "basic_block.hpp"
#include "function_behavior.hpp"
#include "op_graph.hpp"

///hls includes
#include "hls.hpp"
#include "hls_constraints.hpp"
#include "hls_manager.hpp"

///HLS/liveness include
#include "liveness.hpp"

///hls/memory include
#include "memory.hpp"

///hls/stg includes
#include "state_transition_graph.hpp"
#include "state_transition_graph_manager.hpp"

///tree includes
#include "behavioral_helper.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"

///utility include
#include "dbgPrintHelper.hpp"

FSM_NI_SSA_liveness::FSM_NI_SSA_liveness(const ParameterConstRef _parameters, const HLS_managerRef _HLSMgr, unsigned int _funId, const DesignFlowManagerConstRef _design_flow_manager) :
   liveness_computer(_parameters, _HLSMgr, _funId, _design_flow_manager, HLSFlowStep_Type::FSM_NI_SSA_LIVENESS)
{
   debug_level = parameters->get_class_debug_level(GET_CLASS(*this));
}

FSM_NI_SSA_liveness::~FSM_NI_SSA_liveness()
{
}

void FSM_NI_SSA_liveness::Initialize()
{
   HLSFunctionStep::Initialize();
   if(bb_version != HLS->FB->GetBBVersion())
   {
      HLS->Rliv = livenessRef(new liveness(HLSMgr, parameters));
   }
}

static
unsigned int get_bb_index_from_state_info(unsigned int prev_bb_index, const OpGraphConstRef data, StateInfoConstRef state_info)
{
   vertex last_operation;
   if(state_info->executing_operations.empty())
   {
      if(state_info->ending_operations.empty())
         return prev_bb_index; // it may happen with pipelined operations
      else
      {
         last_operation = *(state_info->ending_operations).rbegin();
         return GET_BB_INDEX(data, last_operation);
      }
   }
   else
   {
      last_operation = *(state_info->executing_operations.rbegin());
      return GET_BB_INDEX(data, last_operation);
   }
}

static
void update_liveout_with_prev(const HLS_managerRef HLSMgr, hlsRef HLS, const StateTransitionGraphConstRef stg, const OpGraphConstRef data, vertex current_state, const StateTransitionGraphConstRef astg, vertex prev_state)
{
   const std::list<vertex>& executing_operations = stg->CGetStateInfo(prev_state)->executing_operations;
   std::list<vertex>::const_reverse_iterator exo_it_end = executing_operations.rend();
   for(std::list<vertex>::const_reverse_iterator exo_it = executing_operations.rbegin(); exo_it != exo_it_end; ++exo_it)
   {
      const CustomSet<unsigned int> & scalar_uses = data->CGetOpNodeInfo(*exo_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::USE);
      CustomSet<unsigned int>::const_iterator scalar_use, scalar_use_end = scalar_uses.end();
      for(scalar_use = scalar_uses.begin(); scalar_use != scalar_use_end; scalar_use++)
      {
         if(HLSMgr->is_register_compatible(*scalar_use))
            HLS->Rliv->set_live_out(current_state, *scalar_use);
      }
   }
   const std::list<vertex>& ending_operations = astg->CGetStateInfo(prev_state)->ending_operations;
   std::list<vertex>::const_reverse_iterator eo_it_end = ending_operations.rend();
   for(std::list<vertex>::const_reverse_iterator eo_it = ending_operations.rbegin(); eo_it != eo_it_end; ++eo_it)
   {
      const CustomSet<unsigned int> & scalar_defs = data->CGetOpNodeInfo(*eo_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::DEFINITION);
      if(not scalar_defs.empty())
      {
         CustomSet<unsigned int>::const_iterator it_end = scalar_defs.end();
         for(CustomSet<unsigned int>::const_iterator it = scalar_defs.begin(); it != it_end; ++it)
            if(HLSMgr->is_register_compatible(*it))
               HLS->Rliv->erase_el_live_out(current_state, *it);
      }
   }
   if(stg->CGetStateInfo(prev_state)->moved_op_use_set.begin() != stg->CGetStateInfo(prev_state)->moved_op_use_set.end())
      HLS->Rliv->set_live_out(current_state, stg->CGetStateInfo(prev_state)->moved_op_use_set.begin(), stg->CGetStateInfo(prev_state)->moved_op_use_set.end());
}

DesignFlowStep_Status FSM_NI_SSA_liveness::InternalExec()
{
   const StateTransitionGraphConstRef astg = HLS->STG->CGetAstg();
   const StateTransitionGraphConstRef stg = HLS->STG->CGetStg();
   const BBGraphConstRef fbb = HLS->FB->CGetBBGraph(FunctionBehavior::FBB);
   const OpGraphConstRef data = HLS->FB->CGetOpGraph(FunctionBehavior::DFG);

   /// Map between basic block node index and vertices
   std::unordered_map<unsigned int, vertex> bb_index_map;
   VertexIterator vit, vend;
   for(boost::tie(vit, vend) = boost::vertices(*fbb); vit != vend; ++vit)
   {
      bb_index_map[fbb->CGetBBNodeInfo(*vit)->get_bb_index()] = *vit;
   }

   /// vertex of the STG analyzed in reverse order
   std::deque<vertex> reverse_order_state_list;
   astg->ReverseTopologicalSort(reverse_order_state_list);
   const vertex bb_exit = fbb->CGetBBGraphInfo()->exit_vertex;
   unsigned int prev_bb_index = fbb->CGetBBNodeInfo(bb_exit)->get_bb_index();
   vertex exit_state = HLS->STG->get_exit_state();
   vertex entry_state = HLS->STG->get_entry_state();
   bool second_state_of_BB = false;
   std::deque<vertex> dummy_states;

   /// compute which operation defines a variable (add_op_definition)
   /// compute in which states an operation complete the computation (add_state_for_ending_op)
   /// compute the set of supporting states
   /// compute the set of dummy states
   std::deque<vertex>::const_iterator rosl_it_end = reverse_order_state_list.end();
   for(std::deque<vertex>::const_iterator rosl_it = reverse_order_state_list.begin(); rosl_it != rosl_it_end; ++rosl_it)
   {
      HLS->Rliv->add_name(*rosl_it, HLS->STG->get_state_name(*rosl_it));
      HLS->Rliv->add_support_state(*rosl_it);
      if(*rosl_it == exit_state)
         continue;
      const StateInfoConstRef state_info = astg->CGetStateInfo(*rosl_it);
      const std::list<vertex>& ending_operations_cur = state_info->ending_operations;
      std::list<vertex>::const_reverse_iterator eoc_it_end = ending_operations_cur.rend();
      for(std::list<vertex>::const_reverse_iterator eoc_it = ending_operations_cur.rbegin(); eoc_it != eoc_it_end; ++eoc_it)
      {
         const CustomSet<unsigned int> & scalar_defs = data->CGetOpNodeInfo(*eoc_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::DEFINITION);
         if(not scalar_defs.empty())
         {
            CustomSet<unsigned int>::const_iterator it_end = scalar_defs.end();
            for(CustomSet<unsigned int>::const_iterator it = scalar_defs.begin(); it != it_end; ++it)
               if(HLSMgr->is_register_compatible(*it))
               {
                  HLS->Rliv->add_op_definition(*it, *eoc_it);
               }
         }
         HLS->Rliv->add_state_for_ending_op(*eoc_it, *rosl_it);
      }

      if(state_info->is_dummy)
      {
         dummy_states.push_back(*rosl_it);
         HLS->Rliv->add_dummy_state(*rosl_it);
         continue;
      }
      if(state_info->moved_op_use_set.begin() != state_info->moved_op_use_set.end())
      {
         HLS->Rliv->set_live_in(*rosl_it, state_info->moved_op_use_set.begin(), state_info->moved_op_use_set.end());
         InEdgeIterator e_it, e_it_end;
         for(boost::tie(e_it, e_it_end) = boost::in_edges(*rosl_it, *stg); e_it != e_it_end; ++e_it)
         {
            vertex src_state = boost::source(*e_it, *astg);
            HLS->Rliv->set_live_out(src_state, state_info->moved_op_use_set.begin(), state_info->moved_op_use_set.end());
         }
      }

   }

   /// compute live_out
   vertex prev_state = exit_state;
   for(std::deque<vertex>::const_iterator rosl_it = reverse_order_state_list.begin(); rosl_it != rosl_it_end; ++rosl_it)
   {
      vertex current_state = *rosl_it;
      if(current_state == exit_state) continue;
      StateInfoConstRef state_info = astg->CGetStateInfo(current_state);
      if(state_info->is_dummy)
      {
         continue;
      }
      if(state_info->clonedState != NULL_VERTEX)
      {
         StateInfoConstRef cloned_state_info = astg->CGetStateInfo(state_info->clonedState);
         unsigned int cloned_bb_index = *cloned_state_info->BB_ids.begin();
         OutEdgeIterator oe_it, oe_it_end;
         bool found = false;
         //std::cerr << "Considerign state " << stg->CGetStateInfo(state_info->clonedState)->name <<  std::endl;
         for(boost::tie(oe_it, oe_it_end) = boost::out_edges(state_info->clonedState, *astg); oe_it != oe_it_end; ++oe_it)
         {
            vertex target_state = boost::target(*oe_it, *astg);
            StateInfoConstRef target_state_info = astg->CGetStateInfo(target_state);
            //std::cerr << "target state " << stg->CGetStateInfo(target_state)->name <<  std::endl;
            unsigned int target_bb_index = *target_state_info->BB_ids.begin();
            if(cloned_bb_index == target_bb_index && target_state != current_state)
            {
               HLS->Rliv->set_live_out(state_info->clonedState, HLS->Rliv->get_live_out(target_state).begin(), HLS->Rliv->get_live_out(target_state).end());
               HLS->Rliv->set_live_out(current_state, HLS->Rliv->get_live_out(target_state).begin(), HLS->Rliv->get_live_out(target_state).end());
               update_liveout_with_prev(HLSMgr, HLS, stg, data, state_info->clonedState, astg, target_state);
               update_liveout_with_prev(HLSMgr, HLS, stg, data, current_state, astg, target_state);
               found = true;
            }
         }
         if(!found)
         {
            //std::cerr << "Not found target " << stg->CGetStateInfo(state_info->clonedState)->name <<  std::endl;
            //std::cerr << "associated with state " << stg->CGetStateInfo(current_state)->name << " cloned_bb_index" << cloned_bb_index << std::endl;
            const std::set<unsigned int> & live_out = fbb->CGetBBNodeInfo(bb_index_map[cloned_bb_index])->get_live_out();
            std::set<unsigned int>::const_iterator lo_it_end = live_out.end();
            for(std::set<unsigned int>::const_iterator lo_it = live_out.begin(); lo_it != lo_it_end; ++lo_it)
               if(HLSMgr->is_register_compatible(*lo_it))
               {
                  HLS->Rliv->set_live_out(current_state, *lo_it);
                  HLS->Rliv->set_live_out(state_info->clonedState, *lo_it);
                  //std::cerr << HLS->FB->CGetBehavioralHelper()->PrintVariable(*lo_it) << std::endl;
               }
         }
         prev_state = *rosl_it;
         prev_bb_index = cloned_bb_index;
         continue;
      }
      else if (state_info->is_duplicated)
      {
         StateInfoConstRef cloned_state_info = astg->CGetStateInfo(*rosl_it);
         unsigned int cloned_bb_index = *cloned_state_info->BB_ids.begin();
         prev_state = *rosl_it;
         prev_bb_index = cloned_bb_index;
         continue;
      }
      unsigned int bb_index = get_bb_index_from_state_info(prev_bb_index, data, state_info);

      if(prev_bb_index != bb_index)
      {
         const std::set<unsigned int> & live_out = fbb->CGetBBNodeInfo(bb_index_map[bb_index])->get_live_out();
         std::set<unsigned int>::const_iterator lo_it_end = live_out.end();
         for(std::set<unsigned int>::const_iterator lo_it = live_out.begin(); lo_it != lo_it_end; ++lo_it)
            if(HLSMgr->is_register_compatible(*lo_it))
               HLS->Rliv->set_live_out(current_state, *lo_it);
         second_state_of_BB = true;
      }
      else
      {
         if(second_state_of_BB)
         {
            second_state_of_BB = false;
            std::set<unsigned int>::const_iterator lo_it_end = HLS->Rliv->get_live_out(prev_state).end();
            for(std::set<unsigned int>::const_iterator lo_it = HLS->Rliv->get_live_out(prev_state).begin(); lo_it != lo_it_end; ++lo_it)
            {
               HLS->Rliv->set_live_out(current_state, *lo_it);
            }
         }
         else
            HLS->Rliv->set_live_out(current_state, HLS->Rliv->get_live_out(prev_state).begin(), HLS->Rliv->get_live_out(prev_state).end());
         update_liveout_with_prev(HLSMgr, HLS, stg, data, current_state, astg, prev_state);
      }
      prev_state = *rosl_it;
      prev_bb_index = bb_index;
   }

   /// compute the live in of a state by traversing the state list in topological order
   prev_state = entry_state;
   prev_bb_index = fbb->CGetBBNodeInfo(bb_exit)->get_bb_index();
   std::set<vertex> state_to_skip;
   std::deque<vertex>::const_reverse_iterator osl_it_end = reverse_order_state_list.rend();
   for(std::deque<vertex>::const_reverse_iterator osl_it = reverse_order_state_list.rbegin(); osl_it != osl_it_end; ++osl_it)
   {
      const StateInfoConstRef state_info = astg->CGetStateInfo(*osl_it);
      if(state_to_skip.find(*osl_it) != state_to_skip.end())
      {
         prev_state = *osl_it;
         prev_bb_index = *state_info->BB_ids.begin();
         continue;
      }
      if(state_info->is_dummy)
         continue;
      if(state_info->is_duplicated)
      {
         unsigned int bb_index = *state_info->BB_ids.begin();
         const std::set<unsigned int> & live_in = fbb->CGetBBNodeInfo(bb_index_map[bb_index])->get_live_in();
         std::set<unsigned int>::const_iterator lo_it_end = live_in.end();
         for(std::set<unsigned int>::const_iterator lo_it = live_in.begin(); lo_it != lo_it_end; ++lo_it)
            if(HLSMgr->is_register_compatible(*lo_it))
               HLS->Rliv->set_live_in(*osl_it, *lo_it);

         OutEdgeIterator oe_it, oe_it_end;
         for(boost::tie(oe_it, oe_it_end) = boost::out_edges(*osl_it, *astg); oe_it != oe_it_end; ++oe_it)
         {
            vertex target_state = boost::target(*oe_it, *astg);
            StateInfoConstRef target_state_info = astg->CGetStateInfo(target_state);
            unsigned int target_bb_index = *target_state_info->BB_ids.begin();
            if(bb_index == target_bb_index && !target_state_info->is_duplicated)
            {
               HLS->Rliv->set_live_in(target_state, HLS->Rliv->get_live_out(*osl_it).begin(), HLS->Rliv->get_live_out(*osl_it).end());
               state_to_skip.insert(target_state);
            }
         }
         prev_state = *osl_it;
         prev_bb_index = bb_index;
         continue;
      }
      unsigned int bb_index = get_bb_index_from_state_info(prev_bb_index, data, state_info);

      if(prev_bb_index != bb_index)
      {
         const std::set<unsigned int> & live_in = fbb->CGetBBNodeInfo(bb_index_map[bb_index])->get_live_in();
         std::set<unsigned int>::const_iterator lo_it_end = live_in.end();
         for(std::set<unsigned int>::const_iterator lo_it = live_in.begin(); lo_it != lo_it_end; ++lo_it)
            if(HLSMgr->is_register_compatible(*lo_it))
               HLS->Rliv->set_live_in(*osl_it, *lo_it);
      }
      else
         HLS->Rliv->set_live_in(*osl_it, HLS->Rliv->get_live_out(prev_state).begin(), HLS->Rliv->get_live_out(prev_state).end());

      prev_state = *osl_it;
      prev_bb_index = bb_index;

   }

   /// fix the live in/out of dummy states
   std::deque<vertex>::const_iterator ds_it_end = dummy_states.end();
   for(std::deque<vertex>::const_iterator ds_it = dummy_states.begin(); ds_it != ds_it_end; ++ds_it)
   {
      const StateInfoConstRef state_info = astg->CGetStateInfo(*ds_it);
      InEdgeIterator e_it, e_it_end;
      for(boost::tie(e_it, e_it_end) = boost::in_edges(*ds_it, *astg); e_it != e_it_end; ++e_it)
      {
         vertex src_state = boost::source(*e_it, *astg);
         HLS->Rliv->set_live_out(*ds_it, HLS->Rliv->get_live_out(src_state).begin(), HLS->Rliv->get_live_out(src_state).end());
         HLS->Rliv->set_live_in(*ds_it, HLS->Rliv->get_live_out(src_state).begin(), HLS->Rliv->get_live_out(src_state).end());
         /// add all the uses of *ds_it to src_state
         const std::list<vertex>& executing_operations = state_info->executing_operations;
         std::list<vertex>::const_iterator eo_it_end = executing_operations.end();
         for(std::list<vertex>::const_iterator eo_it = executing_operations.begin(); eo_it != eo_it_end; ++eo_it)
         {
            const CustomSet<unsigned int> & scalar_uses = data->CGetOpNodeInfo(*eo_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::USE);
            CustomSet<unsigned int>::const_iterator scalar_use, scalar_use_end = scalar_uses.end();
            for(scalar_use = scalar_uses.begin(); scalar_use != scalar_use_end; scalar_use++)
            {
               if(HLSMgr->is_register_compatible(*scalar_use))
               {
                  HLS->Rliv->set_live_out(src_state, *scalar_use);
                  HLS->Rliv->set_live_in(*ds_it, *scalar_use);
                  /// extend the lifetime of used variable to reduce the critical path
                  OutEdgeIterator oe_it, oe_it_end;
                  for(boost::tie(oe_it, oe_it_end) = boost::out_edges(*ds_it, *astg); oe_it != oe_it_end; ++oe_it)
                  {
                     vertex target_state = boost::target(*oe_it, *astg);
                     HLS->Rliv->set_live_in(target_state, *scalar_use);
                  }
               }
            }
         }
      }
   }

   /// fix live out of the states adjacent with the exit state
   InEdgeIterator es_it, es_it_end;
   for(boost::tie(es_it, es_it_end) = boost::in_edges(exit_state, *astg); es_it != es_it_end; ++es_it)
   {
      vertex src_state = boost::source(*es_it, *astg);
      HLS->Rliv->set_live_out(src_state, HLS->Rliv->get_live_in(src_state).begin(), HLS->Rliv->get_live_in(src_state).end());
   }

   /// compute in which state an operation is in execution
   /// compute state in relation: on which transition a variable is live in
   /// compute state out relation: on which transition a variable is live out
   for(std::deque<vertex>::const_iterator rosl_it = reverse_order_state_list.begin(); rosl_it != rosl_it_end; ++rosl_it)
   {
      const StateInfoConstRef state_info = astg->CGetStateInfo(*rosl_it);
      INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Analyzing state " + state_info->name);
      const std::list<vertex>& ending_operations_cur = state_info->ending_operations;
      const std::list<vertex>& running_operations_cur = state_info->executing_operations;
      std::list<vertex>::const_iterator roc_it_end = running_operations_cur.end();
      for(std::list<vertex>::const_iterator roc_it = running_operations_cur.begin(); roc_it != roc_it_end; ++roc_it)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Analyzing running operation " + GET_NAME(data, *roc_it));
         HLS->Rliv->add_state_for_running_op(*roc_it, *rosl_it);
         if((GET_TYPE(data, *roc_it) & TYPE_PHI) != 0)
         {
            for(const auto def_edge : GetPointer<const gimple_phi>(HLSMgr->get_tree_manager()->get_tree_node_const(data->CGetOpNodeInfo(*roc_it)->GetNodeId()))->CGetDefEdgesList())
            {
               unsigned int  tree_var = def_edge.first->index;
               unsigned int bb_index = def_edge.second;
               if(state_info->is_duplicated && !state_info->all_paths)
               {
                  if(!state_info->isOriginalState && bb_index != state_info->sourceBb) continue;
                  if(state_info->isOriginalState && bb_index == state_info->sourceBb) continue;
                  unsigned int written_phi = HLSMgr->get_produced_value(HLS->functionId, *roc_it);
                  if(state_info->moved_op_def_set.find(tree_var) != state_info->moved_op_def_set.end() || state_info->moved_op_use_set.find(written_phi) != state_info->moved_op_use_set.end())
                  {
                     HLS->Rliv->set_live_out(*rosl_it, HLS->Rliv->get_live_out(stg->CGetStateInfo(*rosl_it)->clonedState).begin(), HLS->Rliv->get_live_out(stg->CGetStateInfo(*rosl_it)->clonedState).end());

                     if(state_info->moved_op_use_set.find(written_phi) != state_info->moved_op_use_set.end())
                     {
                        HLS->Rliv->set_live_in(*rosl_it, tree_var);
                        InEdgeIterator ie_it, ie_it_end;
                        for(boost::tie(ie_it, ie_it_end) = boost::in_edges(*rosl_it, *stg); ie_it != ie_it_end; ++ie_it)
                        {
                           vertex src_vrtx = boost::source(*ie_it, *stg);
                           HLS->Rliv->set_live_out(src_vrtx, tree_var);
                        }
                     }
                     //std::cerr << "state_out_candidate " << stg->CGetStateInfo(*rosl_it)->name << " for var " << HLS->FB->CGetBehavioralHelper()->PrintVariable(written_phi) <<  std::endl;
                     HLS->Rliv->set_live_out(*rosl_it, written_phi);
                     OutEdgeIterator e_it, e_it_end;
                     for(boost::tie(e_it, e_it_end) = boost::out_edges(*rosl_it, *stg); e_it != e_it_end; ++e_it)
                     {
                        vertex tgt_vrtx = boost::target(*e_it, *stg);
                        HLS->Rliv->set_live_in(tgt_vrtx, written_phi);
                     }
                  }
               }
               /// now we look for the last state with operations belonging to basic block bb_index
#if HAVE_ASSERTS
               bool found_state = false;
#endif
               InEdgeIterator ie, iend;
               for(boost::tie(ie, iend) = boost::in_edges(*rosl_it, *stg); ie != iend; ie++)
               {
                  vertex src_state = boost::source(*ie, *stg);
                  const StateInfoConstRef source_state_info = stg->CGetStateInfo(src_state);
                  //std::cerr << "considering state=" << source_state_info->name << std::endl;
                  const std::set<unsigned int> & BB_ids = source_state_info->BB_ids;
                  if(BB_ids.find(bb_index) != BB_ids.end())
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "---Adding state in  " + source_state_info->name + " " + HLS->FB->CGetBehavioralHelper()->PrintVariable(tree_var) + " in state " + state_info->name);
                     //THROW_ASSERT((HLS->Rliv->get_live_out(src_state).find(tree_var) != HLS->Rliv->get_live_out(src_state).end()), "unexpected live out condition");
                     THROW_ASSERT(src_state != entry_state, "Source state for phi " + STR(data->CGetOpNodeInfo(*roc_it)->GetNodeId()) + " not found");
                     HLS->Rliv->add_state_in_for_var(def_edge.first->index, *roc_it, *rosl_it, src_state);
#if HAVE_ASSERTS
                     found_state = true;
#endif
                  }

               }
               THROW_ASSERT(found_state, "unexpected condition: current state " + state_info->name + " bb_index=" + STR(bb_index));
            }
         }
         else
         {
            OutEdgeIterator oe, oend;
            for(boost::tie(oe, oend) = boost::out_edges(*rosl_it, *stg); oe != oend; oe++)
            {
               vertex tgt_state = boost::target(*oe, *stg);
               const CustomSet<unsigned int> & scalar_defs = data->CGetOpNodeInfo(*roc_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::DEFINITION);
               const CustomSet<unsigned int> & cited_variables = data->CGetOpNodeInfo(*roc_it)->cited_variables;
               const CustomSet<unsigned int> all_but_scalar_defs = cited_variables - scalar_defs;
               CustomSet<unsigned int>::const_iterator all_but_scalar_def, all_but_scalar_def_end = all_but_scalar_defs.end();
               for(all_but_scalar_def = all_but_scalar_defs.begin(); all_but_scalar_def != all_but_scalar_def_end; all_but_scalar_def++)
               {
                  if(HLSMgr->is_register_compatible(*all_but_scalar_def))
                     HLS->Rliv->add_state_in_for_var(*all_but_scalar_def, *roc_it, *rosl_it, tgt_state);
                  else if(HLSMgr->Rmem->has_base_address(*all_but_scalar_def) ||
                       (tree_helper::is_parameter(HLSMgr->get_tree_manager(), *all_but_scalar_def)))
                  {
                     HLS->Rliv->add_state_in_for_var(*all_but_scalar_def, *roc_it, *rosl_it, tgt_state);
                  }
                  else if(tree_helper::is_ssa_name(HLSMgr->get_tree_manager(), *all_but_scalar_def) && tree_helper::is_virtual(HLSMgr->get_tree_manager(), *all_but_scalar_def))
                  {
                     ;
                  }
                  else
                  {
                     THROW_ERROR("unexpected condition " +STR(*all_but_scalar_def));
                  }
               }
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Analyzed running operation " + GET_NAME(data, *roc_it));
      }

      std::list<vertex>::const_iterator eoc_it_end = ending_operations_cur.end();
      for(std::list<vertex>::const_iterator eoc_it = ending_operations_cur.begin(); eoc_it != eoc_it_end; ++eoc_it)
      {
         if((GET_TYPE(data, *eoc_it) & TYPE_PHI) != 0)
         {
            if(state_info->is_duplicated && state_info->clonedState != NULL_VERTEX && !state_info->all_paths)
            {
               unsigned int written_phi = HLSMgr->get_produced_value(HLS->functionId, *eoc_it);
               bool postponed_phi = false;
               for(const auto def_edge : GetPointer<const gimple_phi>(HLSMgr->get_tree_manager()->get_tree_node_const(data->CGetOpNodeInfo(*eoc_it)->GetNodeId()))->CGetDefEdgesList())
               {
                  unsigned int tree_temp = def_edge.first->index;
                  unsigned int bbID = def_edge.second;
                  if(bbID != state_info->sourceBb) continue;
                  else if(state_info->moved_op_def_set.find(tree_temp) != state_info->moved_op_def_set.end())
                  {
                     postponed_phi = true;
                     break;
                  }
                  else if(state_info->moved_op_use_set.find(written_phi) != state_info->moved_op_use_set.end())
                  {
                     postponed_phi = true;
                     break;
                  }
                  else
                     break;
               }
               if(!postponed_phi)
                  continue;
            }
            else
               continue;
         }
         const CustomSet<unsigned int> & scalar_defs = data->CGetOpNodeInfo(*eoc_it)->GetVariables(FunctionBehavior_VariableType::SCALAR, FunctionBehavior_VariableAccessType::DEFINITION);
         if(not scalar_defs.empty())
         {
            CustomSet<unsigned int>::const_iterator it_end = scalar_defs.end();
            for(CustomSet<unsigned int>::const_iterator it = scalar_defs.begin(); it != it_end; ++it)
               if(HLSMgr->is_register_compatible(*it))
               {
                  OutEdgeIterator e_it, e_it_end;
                  for(boost::tie(e_it, e_it_end) = boost::out_edges(*rosl_it, *stg); e_it != e_it_end; ++e_it)
                  {
                     vertex tgt_state = boost::target(*e_it, *stg);
                     if(HLS->Rliv->get_live_in(tgt_state).find(*it) != HLS->Rliv->get_live_in(tgt_state).end())
                        HLS->Rliv->add_state_out_for_var(*it, *eoc_it, *rosl_it, tgt_state);
                  }
               }
         }
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Analyzed state " + state_info->name);
   }


#ifndef NDEBUG
   if(debug_level>= DEBUG_LEVEL_PEDANTIC)
   {
      /// print the analysis result
      const BehavioralHelperConstRef  BH = HLS->FB->CGetBehavioralHelper();
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "STG Liveness for function " + BH->get_function_name());
      VertexIterator vIt, vEnd;
      for(boost::tie(vIt, vEnd) = boost::vertices(*astg); vIt != vEnd; vIt++)
      {
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "Live In for state " + HLS->STG->get_state_name(*vIt) + ": ");
         std::set<unsigned int>::const_iterator li_it_end = HLS->Rliv->get_live_in(*vIt).end();
         for(std::set<unsigned int>::const_iterator li_it = HLS->Rliv->get_live_in(*vIt).begin(); li_it != li_it_end; ++li_it)
         {
            PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level,  BH->PrintVariable(*li_it) + " ");
         }
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "\n");
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "Live Out for state " + HLS->STG->get_state_name(*vIt) + ": ");
         std::set<unsigned int>::const_iterator lo_it_end = HLS->Rliv->get_live_out(*vIt).end();
         for(std::set<unsigned int>::const_iterator lo_it = HLS->Rliv->get_live_out(*vIt).begin(); lo_it != lo_it_end; ++lo_it)
         {
            PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level,  BH->PrintVariable(*lo_it) + " ");
         }
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "\n");
      }
   }
#endif
   return DesignFlowStep_Status::SUCCESS;
}
