/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file Bit_Value.hpp
 * @brief Full implementation of Bit Value analysis as described in
 * BitValue Inference: Detecting and Exploiting Narrow Bitwidth Computations
 * Mihai Budiu Seth Copen Goldstein
 * http://www.cs.cmu.edu/~seth/papers/budiu-tr00.pdf
 *
 * @author Giulio Stramondo
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/

#ifndef BIT_VALUE_HPP
#define BIT_VALUE_HPP

#include "function_frontend_flow_step.hpp"

#include "refcount.hpp"
/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(Bit_Value);
class binary_expr;
class constructor;
class gimple_assign;
class string_cst;
REF_FORWARD_DECL(tree_manager);
//@}

enum bit_lattice
{
   U, ZERO, ONE, X
};

/**
 * @brief Full implementation of Bit Value analysis as described in
 * BitValue Inference: Detecting and Exploiting
 * Narrow Bitwidth Computations
 * Mihai Budiu Seth Copen Goldstein
 * http://www.cs.cmu.edu/~seth/papers/budiu-tr00.pdf
 */
class Bit_Value : public FunctionFrontendFlowStep
{

   private:

      const std::unordered_set< std::pair<FrontendFlowStepType, FunctionRelationship> >
      ComputeFrontendRelationships (const DesignFlowStep::RelationshipType relationship_type) const;

      /**
       * @brief Map of the current bit-values of each variable.
       * Map storing the current bit-values of the variables at the end of each iteration of forward_transfer or backward_transfer.
       */
      std::map<unsigned int, std::deque<bit_lattice>> current;

      /**
       * @brief Map of the best bit-values of each variable.
       * Map storing the best bit-values of the variables at the end of all the iterations of forward_transfer or backward_transfer.
       */
      std::map<unsigned int, std::deque<bit_lattice>> best;

      /// map private variables with their init value
      std::map<unsigned int, std::deque<bit_lattice>> private_variables;

      /**
       * @brief Map storing the implementation of the forward_transfer's plus_expr.
       * It's initialized by initialize_plus_expr_map().
       * @see initialize_plus_expr_map()
       */
      std::map<bit_lattice, std::map<bit_lattice,std::map<bit_lattice ,std::deque<bit_lattice >>>> plus_expr_map;

      /**
       * @brief Map storing the implementation of the forward_transfer's minus_expr.
       * It's initialized by initialize_minus_expr_map().
       * @see initialize_minus_expr_map()
       */
      std::map<bit_lattice, std::map<bit_lattice,std::map<bit_lattice ,std::deque<bit_lattice >>>> minus_expr_map;

      /**
       * @brief Map storing the implementation of the forward_transfer's bit_ior_expr_map.
       * It's initialized by initialize_bit_ior_expr_map()
       * @see initialize_bit_ior_expr_map()
       */
      std::map<bit_lattice, std::map<bit_lattice ,bit_lattice> > bit_ior_expr_map;

      /**
       * Map storing the implementation of the forward_transfer's bit_xor_expr_map.
       * It's initialized by initialize_bit_xor_expr_map()
       * @see initialize_bit_xor_expr_map()
       */
      std::map<bit_lattice, std::map<bit_lattice ,bit_lattice> > bit_xor_expr_map;

      /**
       * Map storing the implementation of the forward_transfer's bit_and_expr_map.
       * It's initialized by initialize_bit_and_expr_map()
       * @see initialize_bit_and_expr_map()
       */
      std::map<bit_lattice, std::map<bit_lattice ,bit_lattice> > bit_and_expr_map;

      /// True if this step is not executed in the frontend
      bool not_frontend;

      /**
       * Initializes plus_expr_map.
       * Implements the forward_transfer's plus_expr.
       * @see plus_expr_map
       */
      void initialize_plus_expr_map();

      /**
       * Initializes xor_expr_map.
       * Implements the forward_transfer's bit_xor_expr.
       * @see xor_expr_map
       */
      void initialize_bit_xor_expr_map();

      /**
       * Initializes minus_expr_map.
       * Implements the forward_transfer's minus_expr.
       * @see minus_expr_map
       */
      void initialize_minus_expr_map();

      /**
       * Initializes bit_and_expr_map.
       * Implements the forward_transfer's bit_and_expr.
       * @see bit_and_expr_map
       */
      void initialize_bit_and_expr_map();

      /**
       * Initializes bit_ior_expr_map.
       * Implements the forward_transfer's bit_ior_expr.
       * @see bit_ior_expr_map
       */
      void initialize_bit_ior_expr_map();

      /**
       * Substitute a bit in a bitstring
       * @param original the original string
       * @param position the index of the bit to substitute in the sting
       * @param value the value to substitute
       * @return the bitstring after the substitution
       */
      std::deque<bit_lattice> substitute_bit (const std::deque<bit_lattice> &original, size_t position, bit_lattice value);

      /**
       * Tests the substitute_bit function
       * @see substitute_bit()
       */
      bool test_substitute_bit();

      /**
       * Contains the input parameters of the function that's being analyzed.
       */
      std::unordered_set<unsigned int> arguments;

      /**
       * Contains the variables that may have negative values.
       * this information is used by sign sign_extend_bitstring.
       * @see sign_extend_bitstring()
       */
      std::unordered_set<unsigned int> signed_var;

      /**
       * Creates a bitstring containing bits initialized at <U>
       * @param lenght the lenght of the bitstring
       * @return a bitstring of the specified length containing <U> values.
       */
      std::deque<bit_lattice>
      create_u_bitstring (unsigned int lenght);


      std::deque<bit_lattice>
      forward_transfer_plus_expr (const std::deque<bit_lattice> &arg1_bitstring, const std::deque<bit_lattice> &arg2_bitstring);

      std::deque<bit_lattice>
      forward_transfer_minus_expr (const std::deque<bit_lattice> &arg1_bitstring, const std::deque<bit_lattice> &arg2_bitstring);

      /**
       * Get a bit from a bitstring
       * @param original the input bitstring
       * @param position the bit position in the bitstring
       * @return the bitvalue at the specified position.
       */
      bit_lattice get_element_from_bitstring(const std::deque<bit_lattice> &original, size_t position);

      /**
       * Tests get_element_from_bitstring().
       * @see get_element_from_bitstring()
       */
      bool test_get_element_from_bitstring();

      /**
       * Creates a bitstring from a constant input
       * @param value_int integer constant
       * @param length the length of the bitstring to be generated
       * @param signed_value specified if this bitstring can have negative values
       * @return bitstring generated from the integer constant
       */
      std::deque<bit_lattice>
      create_bitstring_from_constant (long long int value_int, unsigned int length, bool signed_value);

      /**
       * Checks if the variable with a given id is an input of the analyzed function
       * @param var_uid uid of the variable of interest
       * @return TRUE if the variable was given as program input, FALSE otherwise.
       */
      bool
         is_input (unsigned int var_uid);

      /**
       * Checks if the variable with a given id is signed
       * @param var_uid uid of the variable of interest
       * @return TRUE if the variable is signed, FALSE otherwise.
       */
      bool
         is_signed (unsigned int var_uid);

      /**
       * This function runs all the bit_value tests.
       */
      void
         run_tests();

      /**
       * Creates a bitstring from two bitvalues.
       * @param e1 first bitvalue
       * @param e0 second bitvalue
       * @return bitstring with both bitvalues
       * @see initialize_plus_expr_map()
       * @see initialize_minus_expr_map()
       */
      std::deque<bit_lattice>
         create_bitstring_couple(bit_lattice e1, bit_lattice e0);

      /**
       * Debugging function used to print the contents of the current and best maps.
       * @param map map to be printed
       */
      void
         print_bitstring_map (const std::map<unsigned int, std::deque<bit_lattice> > &map);

      /**
       * Translates a bitstring ( expressed as an std::deque of bit_lattice ) into a string of characters.
       */
      std::string
         bitstring_to_string (const std::deque<bit_lattice> &bitstring) const;

      /**
       * Checks if two bitstrings are equal
       * @param a first bitstring
       * @param b second bitstring
       * @return TRUE if the bitstrings contain the same bits and have the same length, FALSE otherwise.
       */
      bool
         bitstring_equal (const std::deque<bit_lattice> &a, const std::deque<bit_lattice> &b);

      /**
       * Tests the bitstring_equal function.
       * @see bitstring_equal()
       */
      bool
         test_bitstring_equal ();

      /**
       * Computes the sup between two bitstrings
       * @param a first bitstring variable
       * @param b second bitstring variable
       * @return the sup of the two bitstrings.
       */
      std::deque<bit_lattice>
      sup (const std::deque<bit_lattice> &a, const std::deque<bit_lattice> &b, bool is_signed_p);

      /**
       * Tests the sup() function.
       * @see sup()
       */
      bool
         test_sup ();

      /**
       * Tests the sign_extend_bitstring() function.
       * @see sign_extend_bitstring()
       */
      bool
         test_sign_extend_bitstring ();

      /**
       * Tests the inf() function.
       * @see inf()
       */
      bool
         test_inf ();

      /**
       * Computes the inf between two bitstrings
       * @param a first bitstring
       * @param b second bitstring
       * @return inf between the two bitstrings
       */
      std::deque<bit_lattice>
         inf (const std::deque<bit_lattice> &a, const std::deque<bit_lattice> &b, unsigned int a_uid,
               unsigned int b_uid);

      /**
       * Extends a bitstring
       * @param bitstring to extend
       * @param uid of the variable linked to the bitstring
       * @param final_size desired length of the bitstrign
       * @return the extended bitstring
       */
      std::deque<bit_lattice>
      sign_extend_bitstring (const std::deque<bit_lattice> &bitstring, unsigned int bitstring_uid, long unsigned int final_size);

      /**
       * @brief Reduce the size of a bitstring
       * 	erasing all but one most significant zeros in unsigned bitstring and all but one most significant values in signed bitstrings.
       * 	@param bitstring bitstring to reduce.
       * 	@param bitstring_uid uid of the associated variable, used to check if the bitstring is signed.
       */
      void sign_reduce_bitstring(std::deque<bit_lattice> &bitstring,
            unsigned int bitstring_uid);


      /**
       * auxiliary function used to build the bitstring lattice for read-only arrays
       * @param TM  is the tree manager
       * @param c is the contructor
       * @param current_inf is the current bitstring lattice
       * @param ssa_node_id is the ssa node id of the lattice destination
       */
      void constructor_bitstring(const tree_managerRef TM, constructor* c, std::deque<bit_lattice> &current_inf, unsigned int ssa_node_id);

      /**
       * auxiliary function used to build the bitstring lattice for read-only string_cst
       * @param TM  is the tree manager
       * @param sc is the string_cst
       * @param current_inf is the current bitstring lattice
       * @param ssa_node_id is the ssa node id of the lattice destination
       */
      void string_cst_bitstring(const tree_managerRef TM, string_cst* sc, std::deque<bit_lattice> &current_inf, unsigned int ssa_node_id);

      /**
       * Initializes best with C type as bitstring, signed_var and arguments using the information taken from the syntax tree given by the application manager.
       *
       *
       * Scan all the gimple assign statements in each bloc
       * if the left hand side is signed this is inserted into the signed_var set (this information is used by the signextension)
       *
       * for each lhs variable in the gimple_assign instruction and entry is created in the best map
       * the entry in the map is of the type <GET_INDEX_NODE(lhs), bit_string of <U>s of the variable lenght>
       * for each gimple_assign the used variables are checked against the values in the parm set in order to identify the parameters of the function
       * when those are found their ssa index node is added in the best map and in the arguments map ( used by the clear() function )
      *
      *
      * Scans each phi in the bloc (not virtual)
      * res of the phi is added to the signed_var set if it's signed, and to the best map
      *
      * The edge of each phi are scanned,
      * if the edge is carrying a constant of type integer_cst the index of the integer_cst node is used to identify it,
      * one entry is created for the best table mapping the integer_cst index to its bitstring representation
      * if the value of the constant is negative it's added to the signed_var set
      *
      * if the edge is carrying a ssa variable it's checked if its a parameter,
      * if it is, it is added to the arguments set and, it is added to the best map
      */
      void
         initialize ();

	/**
	 * Clears all the entry in the current map, except for the input arguments
	 */
      void
         clear ();

      /**
       * Mixes the content of current and best using the sup operation, storing the result in best
       * the returned value is used to implement the fixed point algorithm.
       * @return TRUE if best was updated during the process, FALSE otherwise
       */
      bool mix();

      /**
       * Applies the forward algorithm, as described in the paper, analyzing each assignment statement following the program order, and each phi.
       * Uses the forward_transfer() function to compute the output's bitstring, that stores in current.
       * The algorithm loops until current is modified.
       * @see forward_transfer()
       */
      void
         forward ();


      /**
       * Applies the backward algorithm, as described in the paper, analyzing each assignment statement starting from the output, going up to the inputs, and each phi.
       * Uses the backward_transfer() function to compute the output's bitstring, that stores in current.
       * The algorithm loops until current is modified.
       * @see backward_transfer()
       */
      void
         backward ();

      /**
       * Takes a gimple assignment, analyzes the operation performed from the rhs and its input bitstring, and generate a bitstring from the output.
       * @param ga assignment to analyze
       * @return output bitstring
       */
      std::deque<bit_lattice>
         forward_transfer (const gimple_assign *ga);


      /**
       * Compute the inputs back propagation values, given a gimple assignment and the uid of the output variable.
       * @param ga gimple assignment that is being analyzed
       * @param output_id uid of the output of the given gimple assignment.
       * @return computed backpropagation bitstring
       */
      std::deque<bit_lattice>
         backward_transfer (const gimple_assign *ga, unsigned int output_id);

      /**
       * Updates the bitvalues of the intermediate representation with the values taken from the input map.
       * @param map contains the pair uid, bitstring, used to update the intermediate representation
       */
      void
         update_IR (const std::map<unsigned int, std::deque<bit_lattice> > &map);

      /**
       * Given a binary operation, fetches the uid of the arguments and the relative bitstrings.
       * @param operation where the inputs are going to be extracted from.
       * @param arg1_uid where the uid of the first argument is going to be stored
       * @param arg2_uid where the uid of the second argument is going to be stored
       * @param arg1_bitstring where the bitstring ( taken from best ) relative to argument 1 is going to be saved.
       * @param arg2_bitstring where the bitstring ( taken from best ) relative to argument 2 is going to be saved.
       * @return TRUE, if the operation was successful, FALSE otherwise.
       */
      bool manage_forward_binary_operands(const binary_expr* operation, unsigned int &arg1_uid, unsigned int &arg2_uid,
            std::deque<bit_lattice> &arg1_bitstring, std::deque<bit_lattice> &arg2_bitstring) const;


   public:
      /**
       * Constructor.
       * @param _Param is the set of the parameters
       * @param _AppM is the application manager
       * @param function_id is the identifier of the function
       * @param design_flow_manager is the design flow manager
       */
      Bit_Value (const ParameterConstRef Param, const application_managerRef _AppM, unsigned int function_id, const DesignFlowManagerConstRef design_flow_manager);

      /**
       *  Destructor
       */
      ~Bit_Value ();
      /**
       * perform the bit value analysis
       * @return the exit status of this step
       */
      DesignFlowStep_Status
         InternalExec ();

      /**
       * Initialize the step (i.e., like a constructor, but executed just before exec
       */
      virtual void Initialize();
};

#endif /* BIT_VALUE_HPP */
