/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEIB
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2016 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file Bit_Value.cpp
 * @brief Full implementation of Bit Value analysis as described in
 * BitValue Inference: Detecting and Exploiting Narrow Bitwidth Computations
 * Mihai Budiu Seth Copen Goldstein
 * http://www.cs.cmu.edu/~seth/papers/budiu-tr00.pdf
 * Created on: May 27, 2014
 *
 * @author Giulio Stramondo
 * @author Fabrizio Ferrandi <fabrizio.ferrandi@polimi.it>
 * $Revision$
 * $Date$
 * Last modified by $Author$
 *
*/

///Header include
#include "Bit_Value.hpp"

///. include
#include "Parameter.hpp"

///boost include
#include <boost/range/adaptors.hpp>

///behavior includes
#include "application_manager.hpp"
#include "function_behavior.hpp"

///design_flows include
#include "design_flow_manager.hpp"

///frontend_analysis
#include "application_frontend_flow_step.hpp"

///HLS include
#include "hls_manager.hpp"

///HLS/memory include
#include "memory.hpp"

///STD include
#include <fstream>
#include <math.h>
#include <string>

///tree include
#include "behavioral_helper.hpp"
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"

///wrapper/treegcc include
#include "gcc_wrapper.hpp"

Bit_Value::Bit_Value (const ParameterConstRef _parameters, const application_managerRef _AppM, unsigned int _function_id, const DesignFlowManagerConstRef _design_flow_manager) :
   FunctionFrontendFlowStep (_AppM, _function_id, BIT_VALUE, _design_flow_manager, _parameters),
   not_frontend(false)
{
   debug_level = parameters->get_class_debug_level (GET_CLASS(*this),
                                                    DEBUG_LEVEL_NONE);
}

Bit_Value::~Bit_Value ()
{
}

const std::unordered_set<std::pair<FrontendFlowStepType, FrontendFlowStep::FunctionRelationship> > Bit_Value::ComputeFrontendRelationships (const DesignFlowStep::RelationshipType relationship_type) const
{
   std::unordered_set<std::pair<FrontendFlowStepType, FunctionRelationship> > relationships;
   switch (relationship_type)
   {
      case (PRECEDENCE_RELATIONSHIP):
      {
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (IR_LOWERING, SAME_FUNCTION));
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (DEAD_CODE_ELIMINATION, SAME_FUNCTION));
         break;
      }
      case DEPENDENCE_RELATIONSHIP:
      {
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (USE_COUNTING, SAME_FUNCTION));
         relationships.insert (std::pair<FrontendFlowStepType, FunctionRelationship> (WRITTEN_OBJECT_COMPUTATION, WHOLE_APPLICATION));
         break;
      }
      case(INVALIDATION_RELATIONSHIP) :
      {
         break;
      }
      default:
         THROW_UNREACHABLE("");
   }
   return relationships;
}

//computes the sup of two bitstring
std::deque<bit_lattice>
Bit_Value::sup (const std::deque<bit_lattice> &_a, const std::deque<bit_lattice> &_b, bool is_signed_p)
{
   std::deque<bit_lattice> res;
   std::deque<bit_lattice> a=_a, b=_b;

   std::deque<bit_lattice>::const_reverse_iterator a_it = a.rbegin ();
   std::deque<bit_lattice>::const_reverse_iterator b_it = b.rbegin ();

   if(is_signed_p)
   {
      bit_lattice sign_bit_a = a.front();
      bit_lattice sign_bit_b = b.front();
      if(sign_bit_a == ZERO && a.size()>b.size())
         b.push_front(sign_bit_a);
      else if(sign_bit_a == ONE && a.size()>b.size())
         b.push_front(sign_bit_a);
      if(sign_bit_b == ZERO && b.size()>a.size())
         a.push_front(sign_bit_b);
      else if(sign_bit_b == ONE && b.size()>a.size())
         a.push_front(sign_bit_b);
      if(sign_bit_a == sign_bit_b)
      {
         while(a.size()>b.size())
            a.pop_front();
         while(a.size()<b.size())
            b.pop_front();
         if(sign_bit_a != a.front())
         {
            a.pop_front();
            a.push_front(sign_bit_a);
         }
         else if(sign_bit_a != b.front())
         {
            b.pop_front();
            b.push_front(sign_bit_a);
         }
      }
      else
      {
         if (a.size () > b.size ())
         {
            while (b.size () < a.size())
               b.push_front (b.front ());
         }
         if (b.size () > a.size ())
         {
            while (a.size () < b.size())
               a.push_front (a.front ());
         }
      }
   }

   for (; a_it != a.rend () && b_it != b.rend (); a_it++, b_it++)
   {
      if (*a_it == *b_it)
      {
         res.push_front (*a_it);
      }
      else if (*a_it == X || *b_it == X)
      {
         res.push_front (X);
      }
      else if (*a_it == U)
      {
         res.push_front (*b_it);
      }
      else if (*b_it == U)
      {
         res.push_front (*a_it);
      }
      else
      {
         res.push_front (X);
      }
   }
   while(!res.empty() && res.front()==X) res.pop_front();

   if(res.size() == 0)
      res.push_front(X);
   while(res.size()>1 && get_element_from_bitstring(res,0)==ZERO && get_element_from_bitstring(res,1) == X)
   {
      res.pop_front();
      res.pop_front();
      res.push_front(ZERO);
   }

   if(is_signed_p)
   {
      if(res.size()>1 && res.front()==get_element_from_bitstring(res,1))
      {
         bit_lattice most_significant = res.front();

         if(most_significant == U) return res;
         while(!res.empty() && res.front() == most_significant)
            res.pop_front();
         if(most_significant != X || res.empty())
            res.push_front(most_significant);
      }
   }
   return res;
}

void
Bit_Value::sign_reduce_bitstring (std::deque<bit_lattice> &bitstring, unsigned int bitstring_uid)
{
   while(!bitstring.empty() && bitstring.front()==X) bitstring.pop_front();
   if(bitstring.empty())
      bitstring.push_front(X);
   while(bitstring.size()>1 && get_element_from_bitstring(bitstring,0)==ZERO && get_element_from_bitstring(bitstring,1) == X)
   {
      bitstring.pop_front();
      bitstring.pop_front();
      bitstring.push_front(ZERO);
   }
   if(is_signed(bitstring_uid))
   {
      if(bitstring.size()>1 && get_element_from_bitstring(bitstring,0)==get_element_from_bitstring(bitstring,1))
      {
         bit_lattice most_significant = get_element_from_bitstring(bitstring,0);

         if(most_significant == U) return ;
         while(!bitstring.empty() && bitstring.front() == most_significant)
            bitstring.pop_front();
         bitstring.push_front(most_significant);
      }
   }
   else
   {
      while(!bitstring.empty() && bitstring.front() == ZERO)
         bitstring.pop_front();
      if(bitstring.size()==0)
         bitstring.push_front(ZERO);
   }
}

std::deque<bit_lattice>
Bit_Value::sign_extend_bitstring (const std::deque<bit_lattice> &bitstring, unsigned int bitstring_uid, long unsigned int final_size)
{
   std::deque<bit_lattice> res = bitstring;
   if (bitstring.size () >= final_size)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::sign_extend_bitstring invocation error final_size is less than current");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "bitstring to extend: " + bitstring_to_string(bitstring)+ " size: " +STR(bitstring.size()));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "final_size: "+ STR(final_size));
      THROW_ERROR("Wrong sign extension");
   }
   else
   {
      if (is_signed (bitstring_uid))
      {
         while (res.size () < final_size)
         {
            res.push_front (res.front ());
         }
      }
      else
      {
         while (res.size () < final_size)
         {
            res.push_front (ZERO);
         }
      }
   }
   return res;
}

bool
Bit_Value::test_sign_extend_bitstring ()
{
   std::deque<bit_lattice> test_bitstring;
   bool correct = true;

   //case unsigned variable ( extend with ZEROs )
   //test_bitstring <11>
   test_bitstring.push_front (ONE);
   test_bitstring.push_front (ONE);

   std::deque<bit_lattice> res = sign_extend_bitstring (test_bitstring, 999, 4);

   //expected res <0011>

   if (res.front () != ZERO)
      correct = false;
   res.pop_front ();

   if (res.front () != ZERO)
      correct = false;
   res.pop_front ();

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   //case signed variable ( extend with most significant bit )
   signed_var.insert (999);

   //test_bitstring <11>

   res = sign_extend_bitstring (test_bitstring, 999, 4);

   //expected res <1111>

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   if (res.front () != ONE)
      correct = false;
   res.pop_front ();

   signed_var.erase (999);

   return correct;

}

std::deque<bit_lattice>
Bit_Value::inf (const std::deque<bit_lattice> &a, const std::deque<bit_lattice> &b, unsigned int a_uid, unsigned int b_uid)
{
   std::deque<bit_lattice> res;
   std::deque<bit_lattice> a_tmp = a;
   std::deque<bit_lattice> b_tmp = b;
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf: a ->" + bitstring_to_string(a_tmp));
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf: b ->" + bitstring_to_string(b_tmp));

   if (a.size () > b.size ())
   {
      if(b_tmp.front() == X && b_tmp.size()==1)
      {
         while(b_tmp.size ()<a.size ())
            b_tmp.push_front(X);
      }
      else
         b_tmp = sign_extend_bitstring (b_tmp, b_uid, a.size ());
   }
   if (b.size () > a.size ())
   {
      if(a_tmp.front() == X && a_tmp.size()==1)
      {
         while(a_tmp.size ()<b.size ())
            a_tmp.push_front(X);
      }
      else
         a_tmp = sign_extend_bitstring (a_tmp, a_uid, b.size ());
   }

     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf: a ext ->" + bitstring_to_string(a_tmp));
    PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf: b ext ->" + bitstring_to_string(b_tmp));

   std::deque<bit_lattice>::const_iterator a_it = a_tmp.begin ();
   std::deque<bit_lattice>::const_iterator b_it = b_tmp.begin ();

   for (; a_it != a_tmp.end () && b_it != b_tmp.end (); a_it++, b_it++)
   {
      if (*a_it == *b_it)
      {
         res.push_back (*a_it);
      }
      else if (*a_it == U || *b_it == U)
      {
         res.push_back (U);
      }
      else if (*a_it == X)
      {
         res.push_back (*b_it);
      }
      else if (*b_it == X)
      {
         res.push_back (*a_it);
      }
      else
      {
         res.push_back (U);
      }

   }
   // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf: res ->" + bitstring_to_string(res));

   sign_reduce_bitstring(res, a_uid);
   return res;

}

bool
Bit_Value::test_inf ()
{
   std::deque<bit_lattice> res1;
   std::deque<bit_lattice> res2;
   std::deque<bit_lattice> a;
   std::deque<bit_lattice> b;
   bool correct = true;

   // a = <1U>
   a.push_front (U);
   a.push_front (ONE);

   // b = <U0X>
   b.push_front (X);
   b.push_front (ZERO);
   b.push_front (U);

   // Expected result <UUU>

   //Check commutativity
   res1 = inf (a, b, 999, 998);

   res2 = inf (b, a, 999, 998);

   if (res1.size() == res2.size() && res1.size() == 3 &&(res1.front () != U || res2.front () != U))
      correct = false;
   else
   {
      res1.pop_front ();
      res2.pop_front ();

      if (res1.front () != U || res2.front () != U)
         correct = false;
      else
      {
         res1.pop_front ();
         res2.pop_front ();

         if (res1.front () != U || res2.front () != U)
            correct = false;
      }
   }

   return correct;
}

bool
Bit_Value::test_sup ()
{
   std::deque<bit_lattice> res1;
   std::deque<bit_lattice> res2;
   std::deque<bit_lattice> a;
   std::deque<bit_lattice> b;
   bool correct = true;

   // a = <1U>
   a.push_front (U);
   a.push_front (ONE);

   // b = <U0X>
   b.push_front (X);
   b.push_front (ZERO);
   b.push_front (U);

   // Expected result <XX>

   //Check commutativity
   res1 = sup (a, b, false);

   res2 = sup (b, a, false);

   if (res1.front () != X || res2.front () != X)
      correct = false;

   if (res1.front () != X || res2.front () != X)
      correct = false;

   return correct;
}

//creates a bitstring (list of bit_lattice) of size lenght initializing all the
// bit_lattice to U

std::deque<bit_lattice>
Bit_Value::create_u_bitstring (unsigned int lenght)
{
   std::deque<bit_lattice> res;
   for (unsigned int i = 0; i < lenght; i++)
   {
      res.push_front (U);
   }
   return res;
}

std::deque<bit_lattice>
Bit_Value::create_bitstring_from_constant (long long int value_int, unsigned int length, bool signed_value)
{
   long long unsigned int value = static_cast<long long unsigned int>(value_int);
   std::deque<bit_lattice> res;
   if(value == 0){
      res.push_front(ZERO);
      return res;
   }
   unsigned int counter = 0;
   while (value > 0)
   {
      if (value % 2)
         res.push_front (ONE);
      else
         res.push_front (ZERO);

      value /= 2;
      ++counter;
      if(counter>= length)
         break;
   }
   if(counter<length && signed_value)
      res.push_front (ZERO);
   return res;
}

//converts a bitstring (list of bit_lattice) into string
std::string
Bit_Value::bitstring_to_string (const std::deque<bit_lattice> &bitstring) const
{
   std::string res;
   std::deque<bit_lattice>::const_iterator it_bitstring;
   for (it_bitstring = bitstring.begin (); it_bitstring != bitstring.end ();
        it_bitstring++)
   {
      switch (*it_bitstring)
      {
         case (U):
            res.append ("U");
            break;
         case (X):
            res.append ("X");
            break;
         case (ZERO):
            res.append ("0");
            break;
         case (ONE):
            res.append ("1");
            break;
         default:
            res.append ("?");
            break;
      }
   }
   return res;

}

//prints the content of a bitstring map
void
Bit_Value::print_bitstring_map (const std::map<unsigned int, std::deque<bit_lattice>> &map)
{
   const BehavioralHelperConstRef  BH = AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper();
   std::map<unsigned int, std::deque<bit_lattice>>::const_iterator map_it;
   for (map_it = map.begin (); map_it != map.end (); map_it++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "var_uid: "+ STR(map_it->first)+ ":" + BH->PrintVariable(map_it->first) +" bitstring: " + STR(bitstring_to_string(map_it->second)));
   }
}

void
Bit_Value::update_IR (const std::map<unsigned int, std::deque<bit_lattice>> &map)
{
   const tree_managerRef TM = AppM->get_tree_manager ();
   for (auto map_pair: map)
   {
      tree_nodeRef tn = TM->get_tree_node_const (map_pair.first);
      ssa_name * ssa = GetPointer<ssa_name>(tn);
      if(ssa && (ssa->bit_values == "" || ssa->bit_values.size() > map_pair.second.size()))
         ssa->bit_values = bitstring_to_string(map_pair.second);
   }
}


//returns true if the uid of the variable passed as argument, is the uid of a function argument
bool
Bit_Value::is_input (unsigned int var_uid)
{
   return arguments.find (var_uid) != arguments.end();
}

bool
Bit_Value::is_signed (unsigned int var_uid)
{
   return signed_var.find (var_uid) != signed_var.end ();
}

void Bit_Value::constructor_bitstring(const tree_managerRef TM, constructor* c, std::deque<bit_lattice> &current_inf, unsigned int ssa_node_id)
{
   std::vector<std::pair< tree_nodeRef, tree_nodeRef> > &list_of_idx_valu = c->list_of_idx_valu;
   std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator vend = list_of_idx_valu.end();
   for (std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::iterator i = list_of_idx_valu.begin(); i != vend; i++)
   {
      tree_nodeRef el = GET_NODE(i->second);
      THROW_ASSERT(el, "unexpected condition");
      if(el->get_kind() == integer_cst_K)
      {
         integer_cst *int_const= GetPointer<integer_cst>(el);
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use constant: "+ STR(GET_INDEX_NODE(i->second)) +" -> " + STR(int_const->value));
         std::deque<bit_lattice> current_el = create_bitstring_from_constant (int_const->value, tree_helper::Size(el), tree_helper::is_int(TM, ssa_node_id));
         current_inf = inf (current_inf, current_el, ssa_node_id, ssa_node_id);
      }
      else if(el->get_kind() == constructor_K)
      {
         constructor_bitstring(TM, GetPointer<constructor>(el), current_inf, ssa_node_id);
      }
      else
      {
         current_inf = inf (current_inf, create_u_bitstring (tree_helper::Size(el)), ssa_node_id, ssa_node_id);
      }
   }
}

void Bit_Value::string_cst_bitstring(const tree_managerRef TM, string_cst* sc, std::deque<bit_lattice> &current_inf, unsigned int ssa_node_id)
{
   const std::string str = sc->strg;
   for (unsigned int index = 0; index < str.size(); ++index)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use constant: " + STR(str[index]));
      std::deque<bit_lattice> current_el = create_bitstring_from_constant (static_cast<long long int>(str[index]), 8, tree_helper::is_int(TM, ssa_node_id));
      current_inf = inf (current_inf, current_el, ssa_node_id, ssa_node_id);
   }
}

static unsigned int pointer_resizing(const application_managerRef AppM, unsigned int output_id, const FunctionBehaviorRef function_behavior, unsigned int function_id, const bool not_frontend, const ParameterConstRef parameters)
{
#ifndef NDEBUG
   const auto debug_level = parameters->get_class_debug_level("Bit_Value");
#endif
   const tree_managerRef TM = AppM->get_tree_manager ();
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Pointer resizing starting from " + STR(TM->CGetTreeNode(output_id)));
   unsigned int var = tree_helper::get_base_index(TM, output_id);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Base variable is " + STR(TM->CGetTreeNode(var)));
   HLS_manager*  hm = GetPointer<HLS_manager>(AppM);
   if(not_frontend and var && hm && hm->Rmem && function_behavior->is_variable_mem(var))
   {
      unsigned int max_addr = hm->Rmem->get_base_address(var, function_id) + tree_helper::size(TM,  var) / 8;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Maximum address is " + STR(max_addr));
      unsigned int address_bitsize;
      for(address_bitsize=1; max_addr>(1u<<address_bitsize); ++address_bitsize);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Memory variable " + STR(address_bitsize));
      return address_bitsize;
   }
   else if (not_frontend and hm && hm->Rmem)
   {
      unsigned int address_bitsize = hm->Rmem->get_bus_addr_bitsize();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Address bus bitsize " + STR(address_bitsize));
      return address_bitsize;
   }
   else if(not_frontend)
   {
      unsigned int address_bitsize = AppM->get_address_bitsize();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Address bitsize " + STR(address_bitsize));
      return address_bitsize;
   }
   else
   {
      unsigned int address_bitsize = static_cast<unsigned int>(GccWrapper::CGetPointerSize(parameters));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Pointer bitsize " + STR(address_bitsize));
      return address_bitsize;
   }

}


void
Bit_Value::initialize ()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Internal initialize");
   const tree_managerRef TM = AppM->get_tree_manager ();
   tree_nodeRef tn = TM->get_tree_node_const (function_id);
   function_decl * fd = GetPointer<function_decl> (tn);
   THROW_ASSERT(fd && fd->body, "Node is not a function or it hasn't a body");
   statement_list * sl = GetPointer<statement_list> (GET_NODE(fd->body));

   std::unordered_set<unsigned int> parm;
   best.clear();
   current.clear();
   arguments.clear();
   signed_var.clear();
   //Memorize the function parameters into the set parm
   //so that the ssa corrispondent variables can be memorized as parameters when they will be met
   //during the scan of the instructions

   std::vector<tree_nodeRef>::const_iterator arg_it;
   for (arg_it = fd->list_of_args.begin (); arg_it != fd->list_of_args.end ();
        arg_it++)
   {
      parm.insert(GET_INDEX_NODE(*arg_it));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Index node of parameter inserted: " + STR(GET_INDEX_NODE(*arg_it)));
   }



   int bloc_num=1;
   for (std::map<unsigned int, blocRef>::const_iterator B_it = sl->list_of_bloc.begin ();
        B_it != sl->list_of_bloc.end (); ++B_it)
   {
      blocRef B = B_it->second;
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "analyzing BB: "+ STR(B->number));

      int statement_num = 1 ;
      for(const auto stmt : B->CGetStmtList())
      {
         // ga->op1 is equal to var_decl when it binds a newly declared variable to an ssa variable. ie. int a;
         // we can skip this assignment and focus on the ssa variable
         if (GET_NODE(stmt)->get_kind () == gimple_assign_K)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "stmt kind: "+  GET_NODE(stmt)->get_kind_text());
            gimple_assign * ga = GetPointer<gimple_assign> (GET_NODE(stmt));
            unsigned int ssa_node_id = GET_INDEX_NODE(ga->op0);
            ssa_name *ssa = GetPointer<ssa_name> (GET_NODE(ga->op0));

            if(ssa)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "assignment index: "+ STR(GET_INDEX_NODE(stmt)));
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "LHS index: "+ STR(ssa_node_id));
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "operation kind: "+ GET_NODE(ga->op1)->get_kind_text());
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "bloc: "+ STR(bloc_num)+" statement: "+STR(statement_num));

               if(tree_helper::is_real(TM, ssa_node_id))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "real variables not considered: "+ STR(ssa_node_id));
               }
               else if(tree_helper::is_a_complex(TM, ssa_node_id))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "complex variables not considered: "+ STR(ssa_node_id));
               }
               else if(tree_helper::is_a_vector(TM, ssa_node_id))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "vector variables not considered: "+ STR(ssa_node_id));
               }
               else
               {
                  ssa->bit_values.clear();
                  if (tree_helper::is_int(TM, ssa_node_id))
                  {
                     signed_var.insert (ssa_node_id);
                  }
                  /// check if this assignment is a load from a constant array
                  if(GET_NODE(ga->op1)->get_kind() == array_ref_K ||
                        GET_NODE(ga->op1)->get_kind() == mem_ref_K ||
                        GET_NODE(ga->op1)->get_kind() == target_mem_ref_K ||
                        GET_NODE(ga->op1)->get_kind() == target_mem_ref461_K ||
                        GET_NODE(ga->op1)->get_kind() == var_decl_K)
                  {
                     unsigned int base_index = tree_helper::get_base_index(TM, GET_INDEX_NODE(ga->op1));
                     HLS_manager*  hm = GetPointer<HLS_manager>(AppM);
                     if(base_index && AppM->get_written_objects().find(base_index) == AppM->get_written_objects().end() &&
                        hm && hm->Rmem && function_behavior->is_variable_mem(base_index) && hm->Rmem->is_sds_var(base_index))
                     {
                        tree_nodeRef var_node = TM->get_tree_node_const(base_index);
                        if(var_node->get_kind() == var_decl_K)
                        {
                           var_decl * vd = GetPointer<var_decl>(var_node);
                           vd->bit_values = "";
                           if(vd->init)
                           {
                              std::deque<bit_lattice> current_inf;
                              current_inf.push_back(ZERO);
                              if(GET_NODE(vd->init)->get_kind() == constructor_K)
                              {
                                 constructor * c = GetPointer<constructor>(GET_NODE(vd->init));
                                 constructor_bitstring(TM, c, current_inf, ssa_node_id);
                              }
                              else if(GET_NODE(vd->init)->get_kind() == integer_cst_K)
                              {
                                 integer_cst *int_const= GetPointer<integer_cst>(GET_NODE(vd->init));
                                 current_inf = create_bitstring_from_constant (int_const->value, tree_helper::Size(GET_NODE(vd->init)), tree_helper::is_int(TM, GET_INDEX_NODE(int_const->type)));
                              }
                              else if(GET_NODE(vd->init)->get_kind() == string_cst_K)
                              {
                                 string_cst * sc = GetPointer<string_cst>(GET_NODE(vd->init));
                                 string_cst_bitstring(TM, sc, current_inf, ssa_node_id);
                              }
                              else
                                 current_inf = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));

                              PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Used the init bitstring " << bitstring_to_string(current_inf));

                              best[ssa_node_id] = current_inf;
                              if(!bitstring_equal (current_inf, create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)))))
                                 vd->bit_values = bitstring_to_string(current_inf);
                           }
                           else
                              best[ssa_node_id] = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
                        }
                        else
                           best[ssa_node_id] = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
                     }
                     else
                        best[ssa_node_id] = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
                     /// and now something for the written variables
                     if(base_index && AppM->get_written_objects().find(base_index) != AppM->get_written_objects().end() && hm && hm->Rmem && function_behavior->is_variable_mem(base_index) && hm->Rmem->is_private_memory(base_index) && hm->Rmem->is_sds_var(base_index))
                     {
                        tree_nodeRef var_node = TM->get_tree_node_const(base_index);
                        if(var_node->get_kind() == var_decl_K)
                        {
                           if(private_variables.find(base_index) == private_variables.end())
                           {
                              var_decl * vd = GetPointer<var_decl>(var_node);
                              std::deque<bit_lattice> current_inf;
                              vd->bit_values = "";
                              if(vd->init)
                              {
                                 current_inf.push_back(ZERO);
                                 if(GET_NODE(vd->init)->get_kind() == constructor_K)
                                 {
                                    constructor * c = GetPointer<constructor>(GET_NODE(vd->init));
                                    constructor_bitstring(TM, c, current_inf, ssa_node_id);
                                 }
                                 else if(GET_NODE(vd->init)->get_kind() == integer_cst_K)
                                 {
                                    integer_cst *int_const= GetPointer<integer_cst>(GET_NODE(vd->init));
                                    current_inf = create_bitstring_from_constant (int_const->value, tree_helper::Size(GET_NODE(vd->init)), tree_helper::is_int(TM, GET_INDEX_NODE(int_const->type)));
                                 }
                                 else if(GET_NODE(vd->init)->get_kind() == string_cst_K)
                                 {
                                    string_cst * sc = GetPointer<string_cst>(GET_NODE(vd->init));
                                    string_cst_bitstring(TM, sc, current_inf, ssa_node_id);
                                 }
                                 else
                                    current_inf = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
                                 private_variables[base_index] = current_inf;
                              }
                              else
                              {
                                 current_inf.push_back(ZERO);
                                 private_variables[base_index] = current_inf;
                              }
                              PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Computed the init bitstring for " + AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(base_index) + "=" + bitstring_to_string(private_variables.find(base_index)->second));
                              for(auto cur_var : hm->Rmem->get_source_values(base_index))
                              {
                                 tree_nodeRef cur_node = TM->get_tree_node_const(cur_var);
                                 current_inf = inf (current_inf, create_u_bitstring (tree_helper::Size(cur_node)), ssa_node_id, ssa_node_id);
                              }
                              private_variables[base_index] = current_inf;
                              if(!bitstring_equal (current_inf, create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)))))
                              {
                                 vd->bit_values = bitstring_to_string(current_inf);
                                 INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "---Bit Value: variable "+ AppM->CGetFunctionBehavior(function_id)->CGetBehavioralHelper()->PrintVariable(base_index) + " trimmed to bitsize: " + STR(vd->bit_values.size()) + " with bit-value pattern: " + vd->bit_values);
                              }
                           }
                           std::deque<bit_lattice> var_inf = private_variables[base_index];
                           PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Init bitstring for a private written memory variable " + bitstring_to_string(var_inf));
                           best[ssa_node_id] = var_inf;
                        }
                     }
                  }
                  /*else if(tree_helper::is_a_pointer(TM, ssa_node_id))
                  {
                     HLS_manager*  hm = GetPointer<HLS_manager>(AppM);
                     unsigned int var = tree_helper::get_base_index(TM, ssa_node_id);
                     if(var && hm && hm->Rmem && function_behavior->is_variable_mem(var))
                        best[ssa_node_id] = create_u_bitstring(pointer_resizing(AppM, ssa_node_id, function_behavior, function_id, not_frontend, parameters));
                     else
                        best[ssa_node_id] = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
                  }*/
                  else
                     best[ssa_node_id] = create_u_bitstring (tree_helper::Size(GET_NODE(ga->op0)));
               }
            }
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "non gimple assignment statement: "+ STR(GET_INDEX_NODE(stmt)));

         std::vector<std::tuple<unsigned int, unsigned int> > vars_read;
         tree_helper::get_required_values(TM, vars_read, GET_NODE(stmt), GET_INDEX_NODE(stmt));
         for(auto var_pair : vars_read)
         {
            unsigned int node_id =std::get<0>(var_pair);
            if(node_id==0) continue;
            tree_nodeRef node = TM->get_tree_node_const(node_id);
            ssa_name * ssa_use = GetPointer<ssa_name>(node);

            if(ssa_use && ssa_use->var != nullptr && parm.find(GET_INDEX_NODE(ssa_use->var))!= parm.end())
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use: " + STR(node_id) + " which is a parameter");
               ssa_use->bit_values.clear();
               arguments.insert(node_id);
               best[node_id] = create_u_bitstring (tree_helper::Size(node));
               if (tree_helper::is_int(TM, node_id))
               {
                  signed_var.insert (node_id);
               }
            }
            else if(GetPointer<integer_cst>(node))
            {
               integer_cst *int_const= GetPointer<integer_cst>(node);
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use constant: "+ STR(node_id) +" -> " + STR(int_const->value));
               best[node_id] = create_bitstring_from_constant (int_const->value, tree_helper::Size(node), tree_helper::is_int(TM, GET_INDEX_NODE(int_const->type)));
               if (tree_helper::is_int(TM, GET_INDEX_NODE(int_const->type)))
               {
                  signed_var.insert (node_id);
               }
               sign_reduce_bitstring(best.find(node_id)->second,node_id);
            }
            else
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use: " + STR(node_id));
            }
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "=========================================");

         statement_num++;
      }

      for(const auto phi : B->CGetPhiList())
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "phi operation");
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "phi index: "+ STR(GET_INDEX_NODE(phi)));
         gimple_phi * pn = GetPointer<gimple_phi>(GET_NODE(phi));
         bool is_virtual = pn->virtual_flag;
         if(!is_virtual)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "LHS: "+ STR(GET_INDEX_NODE(pn->res)));
            unsigned int ssa_node_id = GET_INDEX_NODE(pn->res);

            if(tree_helper::is_real(TM, ssa_node_id))
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "real variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
               continue;
            }
            if(tree_helper::is_a_complex(TM, ssa_node_id))
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "complex variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
               continue;
            }
            if(tree_helper::is_a_vector(TM, ssa_node_id))
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "vector variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
               continue;
            }
            if (tree_helper::is_int(TM, ssa_node_id))
            {
               signed_var.insert (ssa_node_id);
            }
            GetPointer<ssa_name>(GET_NODE(pn->res))->bit_values.clear();

            best[GET_INDEX_NODE(pn->res)] =  create_u_bitstring (tree_helper::Size(GET_NODE(pn->res)));

            for(const auto def_edge : pn->CGetDefEdgesList())
            {
               if(GET_NODE(def_edge.first)->get_kind()==integer_cst_K)
               {
                  integer_cst *int_const= GetPointer<integer_cst>(GET_NODE(def_edge.first));
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use constant: "+ STR(GET_INDEX_NODE(def_edge.first)) +" -> " + STR(int_const->value));
                  best[GET_INDEX_NODE(def_edge.first)] = create_bitstring_from_constant (int_const->value, tree_helper::Size(GET_NODE(def_edge.first)), tree_helper::is_int(TM, GET_INDEX_NODE(def_edge.first)));
                  if (tree_helper::is_int(TM, GET_INDEX_NODE(int_const->type)))
                  {
                     signed_var.insert (GET_INDEX_NODE(def_edge.first));
                  }
                  sign_reduce_bitstring(best.find(GET_INDEX_NODE(def_edge.first))->second,GET_INDEX_NODE(def_edge.first));
               }
               else
               {
                  unsigned int ssa_use_node_id = GET_INDEX_NODE(def_edge.first);
                  ssa_name * ssa_use = GetPointer<ssa_name>(GET_NODE(def_edge.first));

                  if(ssa_use && ssa_use->var != nullptr && parm.find(GET_INDEX_NODE(ssa_use->var))!= parm.end())
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use: " + STR(GET_INDEX_NODE(def_edge.first)) + " which is a parameter");
                     arguments.insert(GET_INDEX_NODE(def_edge.first));
                     ssa_use->bit_values.clear();
                     best[GET_INDEX_NODE(def_edge.first)] = create_u_bitstring (tree_helper::Size(GET_NODE(def_edge.first)));
                     if (tree_helper::is_int(TM, ssa_use_node_id))
                     {
                        signed_var.insert (GET_INDEX_NODE(def_edge.first));
                     }
                  }
                  else
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Use: " + STR(GET_INDEX_NODE(def_edge.first)));
                  }
               }
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "=========================================");
         }
      }
      bloc_num++;
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Initialized best with all variables in the functions:");
   print_bitstring_map (best);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended internal initialize");
}



void
Bit_Value::clear ()
{
   std::map<unsigned int, std::deque<bit_lattice>>::const_iterator best_it;
   for (best_it = best.begin (); best_it != best.end (); best_it++)
   {
      std::map<unsigned int, std::deque<bit_lattice>>::const_iterator position_in_current = current.find (best_it->first);
      if (position_in_current != current.end ())
      {
         current.erase (position_in_current);
      }
      if ((is_input (best_it->first)))
      {
         current[best_it->first] = best_it->second;
      }

   }
   //PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Cleared current:");
   //print_bitstring_map (current);
}

bool
Bit_Value::mix ()
{
   std::map<unsigned int, std::deque<bit_lattice>>::iterator best_it;
   std::map<unsigned int, std::deque<bit_lattice>>::iterator current_it;
   bool updated = false;

   for (best_it = best.begin (); best_it != best.end (); best_it++)
   {
      current_it = current.find (best_it->first);
      if (current_it != current.end () && !bitstring_equal (best_it->second, sup (current_it->second, best_it->second, is_signed(best_it->first))))
      {
         best_it->second = sup (current_it->second, best_it->second, is_signed(best_it->first));
         updated = true;
      }
   }
   return updated;
}


void
Bit_Value::backward()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Performing backward transfer");
   const tree_managerRef TM = AppM->get_tree_manager ();
   tree_nodeRef tn = TM->get_tree_node_const (function_id);
   function_decl * fd = GetPointer<function_decl> (tn);
   THROW_ASSERT(fd && fd->body, "Node is not a function or it hasn't a body");
   statement_list * sl = GetPointer<statement_list> (GET_NODE(fd->body));
   std::map<unsigned int, blocRef>::const_reverse_iterator B_it_end = sl->list_of_bloc.rend ();

   bool current_updated = true;

   while (current_updated)
   {
      current_updated = false;
      // for each basic block B in CFG do > Consider all blocks successively
      for (std::map<unsigned int, blocRef>::const_reverse_iterator B_it = sl->list_of_bloc.rbegin ();
           B_it != B_it_end; ++B_it)
      {
         blocRef B = B_it->second;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing BB" + STR(B->number));
         for(const auto stmt : boost::adaptors::reverse(B->CGetStmtList()))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing statement " + STR(stmt));
            if (GET_NODE(stmt)->get_kind () == gimple_assign_K)
            {
               gimple_assign * ga = GetPointer<gimple_assign> (GET_NODE(stmt));
               unsigned int output_uid = GET_INDEX_NODE(ga->op0);
               ssa_name *ssa = GetPointer<ssa_name> (GET_NODE(ga->op0));

               if(ssa)
               {
                  if(tree_helper::is_real(TM, output_uid))
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Real variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }
                  if(tree_helper::is_a_complex(TM, output_uid))
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--complex variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }
                  if(tree_helper::is_a_vector(TM, output_uid))
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--vector variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }

                  THROW_ASSERT(best.find (output_uid) != best.end(), "unexpected condition");
                  if (current.find (output_uid) == current.end ())
                  {
                     current[output_uid] = best.find (output_uid)->second;
                  }
                  ///check if current is constant
                  bool is_current_constant = true;
                  for(auto curr_el : current[output_uid])
                     if(curr_el == U)
                     {
                        is_current_constant = false;
                        break;
                     }
                  if(is_current_constant)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--variable has been proven to be constant: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }

                  std::deque<bit_lattice> res;
                  res.push_back(X);
                  if(not ssa->CGetUseStmts().empty())
                  {
                     for(auto statement_node : ssa->CGetUseStmts())
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "---Output " + STR(output_uid) + "(" + STR(ga->op0) + ") is used in instruction  " +  STR(GET_INDEX_NODE(statement_node.first)) + "(" + STR(statement_node.first) + ")");
                        std::deque<bit_lattice> res_fanout;
                        gimple_assign * ga_tmp = GetPointer<gimple_assign> (GET_NODE(statement_node.first));
                        if(ga_tmp)
                        {
                           res_fanout = backward_transfer (ga_tmp, output_uid);
                           if(res_fanout.size() > 0 )
                           {
                              res = inf (res,res_fanout,output_uid,output_uid);
                              PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "output_uid: "+ STR(output_uid)+" bitstring: " + STR(bitstring_to_string(res)));
                           }
                           else
                           {
                              res = best.find (output_uid)->second;
                              PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best: "+ STR(output_uid)+" bitstring: " + STR(bitstring_to_string(res)));
                              break;
                           }
                        }
                        else
                        {
                           res = best.find (output_uid)->second;
                           PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best: "+ STR(output_uid)+" bitstring: " + STR(bitstring_to_string(res)));
                           break;
                        }
                     }
                  }
                  else
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "no one use variable: " + STR(output_uid));

                  }

                  if(res.size () > 0)
                  {
                     sign_reduce_bitstring(res,output_uid);
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "new current: "+ STR(output_uid)+" bitstring: " + STR(bitstring_to_string(res)));
                  }
                  if (res.size () > 0 && !bitstring_equal (current.find (output_uid)->second, sup (res, best.find (output_uid)->second, is_signed(output_uid))))
                  {
                     current[output_uid] = sup (res, best.find (output_uid)->second, is_signed(output_uid));
                     current_updated = true;
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "current: updated "+ STR(output_uid)+" bitstring: " + STR(bitstring_to_string(current[output_uid])));
                  }
               }
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Analyzed statement " + STR(stmt));
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Analyzed BB" + STR(B->number));
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Performed backward transfer");
}

std::deque<bit_lattice> Bit_Value::substitute_bit (const std::deque<bit_lattice> &original, size_t position, bit_lattice value)
{
   std::deque<bit_lattice> res = original;
   std::deque<bit_lattice>::iterator res_it;
   size_t counter = 0;

   for(res_it =  res.begin(); res_it != res.end() && counter < position; res_it++,counter++);

   (*res_it) = value;


   return res;
}

bool Bit_Value::test_substitute_bit(){
   std::deque<bit_lattice> test_string;
   std::deque<bit_lattice> res;
   bool correct = true;
   test_string.push_front(ZERO);
   test_string.push_front(ZERO);
   test_string.push_front(ZERO);
   test_string.push_front(ZERO);

   res=substitute_bit(test_string, 1, ONE);


   if(res.size() != 4) correct = false;

   if(res.front()!= ZERO) correct = false;
   res.pop_front();
   if(res.front() != ONE) correct = false;
   res.pop_front();
   if(res.front() != ZERO) correct = false;
   res.pop_front();
   if(res.front() != ZERO) correct = false;
   res.pop_front();

   return correct;
}

bit_lattice Bit_Value::get_element_from_bitstring(const std::deque<bit_lattice> &original, size_t position)
{

   bit_lattice res;
   std::deque<bit_lattice>::const_iterator original_it;
   size_t iterator_position=0;
   for (original_it = original.begin(); original_it != original.end(), iterator_position<position; original_it++, iterator_position++);

   res = *original_it;

   return res;
}

bool Bit_Value::test_get_element_from_bitstring(){
   std::deque<bit_lattice> test_string;

   bool correct = true;
   test_string.push_front(ZERO);
   test_string.push_front(ONE);
   test_string.push_front(U);
   test_string.push_front(X);


   if(get_element_from_bitstring(test_string,0)!= X) correct = false;
   if(get_element_from_bitstring(test_string,1)!= U) correct = false;
   if(get_element_from_bitstring(test_string,2)!= ONE) correct = false;
   if(get_element_from_bitstring(test_string,3)!= ZERO) correct = false;

   return correct;

}


std::deque<bit_lattice>
Bit_Value::backward_transfer (const gimple_assign *ga, unsigned int output_id)
{
   enum kind left_kind = GET_NODE(ga->op0)->get_kind ();
   if(left_kind == mem_ref_K)
   {
      mem_ref *operation = GetPointer<mem_ref> (GET_NODE(ga->op0));

      if(GET_INDEX_NODE(operation->op0) == output_id && current.find(GET_INDEX_NODE(operation->op0)) != current.end())
      {
         return create_u_bitstring(pointer_resizing(AppM, output_id, function_behavior, function_id, not_frontend, parameters));
      }
      else
         return std::deque<bit_lattice>();
   }
   else if(left_kind == target_mem_ref461_K )
   {
      target_mem_ref461 *operation = GetPointer<target_mem_ref461> (GET_NODE(ga->op0));

      if(GET_INDEX_NODE(operation->base) == output_id && current.find(GET_INDEX_NODE(operation->base)) != current.end())
      {
         return create_u_bitstring(pointer_resizing(AppM, output_id, function_behavior, function_id, not_frontend, parameters));
      }
      else
         return std::deque<bit_lattice>();
   }
   else if(left_kind == array_ref_K )
   {
      array_ref *operation = GetPointer<array_ref> (GET_NODE(ga->op0));
      do
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(operation->op0), function_behavior, function_id, not_frontend, parameters));
         }
         operation = GetPointer<array_ref> (GET_NODE(operation->op0));
      } while(operation);
      return std::deque<bit_lattice>();
   }
   else if(current.find (GET_INDEX_NODE(ga->op0))==current.end()){
      return std::deque<bit_lattice>();
   }
   enum kind op_kind = GET_NODE(ga->op1)->get_kind ();
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, GET_NODE(ga->op1)->get_kind_text());
   if (op_kind == ssa_name_K)
   {
      THROW_ASSERT(output_id == GET_INDEX_NODE(ga->op1), "unexpected condition:"+ STR(output_id)+":"+STR(GET_INDEX_NODE(ga->op1)));
      return current.find (GET_INDEX_NODE(ga->op0))->second;
   }
#if 1
   else if (op_kind == mult_expr_K || op_kind == widen_mult_expr_K ||
            op_kind == plus_expr_K ||
            op_kind == minus_expr_K ||
            op_kind == pointer_plus_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      const std::deque<bit_lattice>& output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);
      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition in backward transfering " + ga->ToString());

      unsigned int sign_extension = 0;
      if (is_signed(arg1_uid) && op_kind == plus_expr_K)
         sign_extension=1;

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      while(arg1_bitstring.size()>output_bitstring.size()+sign_extension)
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 1
   else if (op_kind == truth_and_expr_K || op_kind == truth_or_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      while(arg1_bitstring.size()>1)
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
   else if (op_kind == negate_expr_K)
   {
      negate_expr *operation = GetPointer<negate_expr> (GET_NODE(ga->op1));

      const std::deque<bit_lattice>& output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
   else if (op_kind == abs_expr_K)
   {
      abs_expr *operation = GetPointer<abs_expr> (GET_NODE(ga->op1));

      const std::deque<bit_lattice>& output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
   else if (op_kind == truth_not_expr_K)
   {
      truth_not_expr *operation = GetPointer<truth_not_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = GET_INDEX_NODE(operation->op);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      while(arg1_bitstring.size()>1)
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 0
   else if (op_kind == plus_expr_K)
   {
      plus_expr *operation = GetPointer<plus_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();
      if (best.find (arg2_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      std::deque<bit_lattice> arg2_bitstring = best.find (arg2_uid)->second;


      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(output_id) + " = " + STR(arg1_uid) + " + " + STR(arg2_uid));

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      //it may occur that the current output, being updated from the following operation
      //is smaller than the actual output computed by the forward transfer operation
      //to correctly compute the input variables' bit then the  output in current must be
      //sign extended to the dimension of the actual output.
      if(output_bitstring.size() < arg1_bitstring.size()+1)
      {
         //for(long unsigned int i = output_bitstring.size(); i< arg1_bitstring.size()+1; i++)
         //{
         //   output_bitstring.push_front(X);
         //}
         output_bitstring = sign_extend_bitstring(output_bitstring, GET_INDEX_NODE(ga->op0), arg1_bitstring.size()+1);
      }

      std::vector<std::vector <bit_lattice>> input1 (arg1_bitstring.size (),std::vector<bit_lattice>(output_bitstring.size()));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " +");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::iterator output_it;
      std::deque<bit_lattice>::iterator input1_it;
      size_t output_it_counter=0;
      size_t input1_it_counter = 0;

      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "arg1_bitstring size: " + STR (arg1_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "arg2_bitstring size: " + STR (arg2_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "output_bitstring size: " + STR (output_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "input1 size: " + STR (input1.size()) +" X " + STR(input1[0].size()));

      for(output_it = output_bitstring.begin(); output_it != output_bitstring.end() ; output_it++, output_it_counter++ )
      {
         for ( input1_it = arg1_bitstring.begin(), input1_it_counter= 0; input1_it !=  arg1_bitstring.end() ; input1_it ++,input1_it_counter++)
         {
            // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "output_it_counter: " + STR (output_it_counter) +" input1:counter: " + STR(input1_it_counter));
            if(*output_it == X)
            {
               input1[input1_it_counter][output_it_counter]=X;
               continue;
            }
            std::deque<bit_lattice> X_bitstring;
            while(X_bitstring.size()<arg1_bitstring.size())X_bitstring.push_front(X);
            std::deque<bit_lattice> arg1_bitstring_ZERO = substitute_bit(X_bitstring,input1_it_counter,ZERO);
            std::deque<bit_lattice> arg1_bitstring_ONE = substitute_bit(X_bitstring,input1_it_counter,ONE);
            std::deque<bit_lattice> arg2_bitstring_ZERO = substitute_bit(X_bitstring,input1_it_counter,ZERO);
            std::deque<bit_lattice> arg2_bitstring_ONE = substitute_bit(X_bitstring,input1_it_counter,ONE);


            std::deque<bit_lattice> res_ZERO_ZERO = forward_transfer_plus_expr(arg1_bitstring_ZERO,arg2_bitstring_ZERO);
            std::deque<bit_lattice> res_ZERO_ONE = forward_transfer_plus_expr(arg1_bitstring_ZERO,arg2_bitstring_ONE);
            std::deque<bit_lattice> res_ONE_ZERO = forward_transfer_plus_expr(arg1_bitstring_ONE,arg2_bitstring_ZERO);
            std::deque<bit_lattice> res_ONE_ONE = forward_transfer_plus_expr(arg1_bitstring_ONE,arg2_bitstring_ONE);

            if((get_element_from_bitstring(res_ZERO_ZERO,output_it_counter)== get_element_from_bitstring(res_ZERO_ONE,output_it_counter))&&
                  (get_element_from_bitstring(res_ONE_ZERO,output_it_counter)== get_element_from_bitstring(res_ONE_ONE,output_it_counter)) &&
                  (get_element_from_bitstring(res_ZERO_ZERO,output_it_counter)== get_element_from_bitstring(res_ONE_ONE,output_it_counter))
                  // get_element_from_bitstring(res_ONE,output_it_counter)!=U

                  )
            {
               input1[input1_it_counter][output_it_counter]=X;
               continue;
            }
            if(get_element_from_bitstring(arg1_bitstring,input1_it_counter) == ZERO)
            {
               input1[input1_it_counter][output_it_counter]=ZERO;
               continue;
            }

            if(get_element_from_bitstring(arg1_bitstring,input1_it_counter) == ONE)
            {
               input1[input1_it_counter][output_it_counter]=ONE;
               continue;
            }
            input1[input1_it_counter][output_it_counter]=U;

         }
      }

      std::deque<bit_lattice> res_input1;
      std::deque<bit_lattice> tmp_input1;
      for(long unsigned int i=0; i< arg1_bitstring.size(); i++)
         res_input1.push_back(input1[i][0]);

      for (long unsigned int i=0; i<output_it_counter; i++)
      {
         tmp_input1.clear();
         for(long unsigned int j=0; j< arg1_bitstring.size(); j++)
            tmp_input1.push_back(input1[j][i]);
         // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf of res_input1: " +  bitstring_to_string (res_input1) + " and tmp_input1: "  +  bitstring_to_string (tmp_input1));
         res_input1 = inf ( res_input1 , tmp_input1, arg1_uid, arg1_uid );
      }


      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));

      return res_input1;

   }
#endif
#if 0
   else if(op_kind == minus_expr_K)
   {
      minus_expr *operation = GetPointer<minus_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();
      if (best.find (arg2_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      std::deque<bit_lattice> arg2_bitstring = best.find (arg2_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(output_id) + " = " + STR(arg1_uid) + " - " + STR(arg2_uid));

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      const std::deque<bit_lattice> output_bitstring_tmp = forward_transfer_plus_expr(arg1_bitstring,arg2_bitstring);

      //it may occur that the current output, being updated from the following operation
      //is smaller than the actual output computed by the forward transfer operation
      //to correctly compute the input variables' bit then the  output in current must be
      //sign extended to the dimension of the actual output.
      if(output_bitstring.size() < output_bitstring_tmp.size()){
         for(unsigned int i = 0; i< output_bitstring_tmp.size()-output_bitstring.size(); i++)
            output_bitstring.push_front(X);
         //output_bitstring = sign_extend_bitstring(output_bitstring,output_uid, output_bitstring_tmp.size());
      }

      std::vector<std::vector <bit_lattice>> input1 (arg1_bitstring.size (),std::vector<bit_lattice>(output_bitstring.size()));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " -");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_iterator output_it;
      std::deque<bit_lattice>::const_iterator input1_it;
      long unsigned int output_it_counter=0;
      size_t input1_it_counter = 0;

      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "arg1_bitstring size: " + STR (arg1_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "arg2_bitstring size: " + STR (arg2_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "output_bitstring size: " + STR (output_bitstring.size()));
      // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "input1 size: " + STR (input1.size()) +" X " + STR(input1[0].size()));

      for(output_it = output_bitstring.begin(); output_it != output_bitstring.end() ; output_it++, output_it_counter++ )
      {

         for ( input1_it = arg1_bitstring.begin(), input1_it_counter= 0; input1_it !=  arg1_bitstring.end() ; input1_it ++,input1_it_counter++)
         {
            // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "output_it_counter: " + STR (output_it_counter) +" input1:counter: " + STR(input1_it_counter));
            if(*output_it == X)
            {
               input1[input1_it_counter][output_it_counter]=X;
               continue;
            }
            std::deque<bit_lattice> X_bitstring;
            while(X_bitstring.size()<arg1_bitstring.size())X_bitstring.push_front(X);

            std::deque<bit_lattice> ZER0_bitstring;
            while(ZER0_bitstring.size()<arg1_bitstring.size())ZER0_bitstring.push_front(ZERO);

            std::deque<bit_lattice> arg1_bitstring_ZERO = substitute_bit(ZER0_bitstring,input1_it_counter,ZERO);
            std::deque<bit_lattice> arg1_bitstring_ONE = substitute_bit(ZER0_bitstring,input1_it_counter,ONE);
            std::deque<bit_lattice> arg2_bitstring_ZERO = substitute_bit(X_bitstring,input1_it_counter,ZERO);
            std::deque<bit_lattice> arg2_bitstring_ONE = substitute_bit(X_bitstring,input1_it_counter,ONE);


            std::deque<bit_lattice> res_ZERO_ZERO = forward_transfer_minus_expr(arg1_bitstring_ZERO,arg2_bitstring_ZERO);
            std::deque<bit_lattice> res_ZERO_ONE = forward_transfer_minus_expr(arg1_bitstring_ZERO,arg2_bitstring_ONE);
            std::deque<bit_lattice> res_ONE_ZERO = forward_transfer_minus_expr(arg1_bitstring_ONE,arg2_bitstring_ZERO);
            std::deque<bit_lattice> res_ONE_ONE = forward_transfer_minus_expr(arg1_bitstring_ONE,arg2_bitstring_ONE);

            if((get_element_from_bitstring(res_ZERO_ZERO,output_it_counter)== get_element_from_bitstring(res_ZERO_ONE,output_it_counter))&&
                  (get_element_from_bitstring(res_ONE_ZERO,output_it_counter)== get_element_from_bitstring(res_ONE_ONE,output_it_counter)) &&
                  (get_element_from_bitstring(res_ZERO_ZERO,output_it_counter)== get_element_from_bitstring(res_ONE_ONE,output_it_counter))
                  // get_element_from_bitstring(&res_ONE,(int)output_it_counter)!=U

                  )
            {
               input1[input1_it_counter][output_it_counter]=X;
               continue;
            }
            if(get_element_from_bitstring(arg1_bitstring,input1_it_counter) == ZERO)
            {
               input1[input1_it_counter][output_it_counter]=ZERO;
               continue;
            }

            if(get_element_from_bitstring(arg1_bitstring,input1_it_counter) == ONE)
            {
               input1[input1_it_counter][output_it_counter]=ONE;
               continue;
            }
            input1[input1_it_counter][output_it_counter]=U;

         }
      }

      std::deque<bit_lattice> res_input1;
      std::deque<bit_lattice> tmp_input1;
      for(long unsigned int i=0; i< arg1_bitstring.size(); i++)
         res_input1.push_back(input1[i][0]);

      for (long unsigned int i=0; i<output_it_counter; i++)
      {
         tmp_input1.clear();
         for(long unsigned int j=0; j< arg1_bitstring.size(); j++)
            tmp_input1.push_back(input1[j][i]);
         // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "inf of res_input1: " +  bitstring_to_string (res_input1) + " and tmp_input1: "  +  bitstring_to_string (tmp_input1));
         res_input1 = inf ( res_input1 , tmp_input1, arg1_uid, arg1_uid );
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));

      return res_input1;

   }
#endif
#if 1
   else if (op_kind == bit_and_expr_K)
   {
      bit_and_expr *operation = GetPointer<bit_and_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);
      size_t output_bitsize = output_bitstring.size();

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();
      if (best.find (arg2_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      std::deque<bit_lattice> arg2_bitstring = best.find (arg2_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(arg1_uid) + " & " + STR(arg2_uid));

      size_t initial_size = arg1_bitstring.size ();
      if (is_signed(arg1_uid) && initial_size < output_bitsize)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, output_bitsize);
      }
      if (is_signed(arg1_uid) && initial_size > output_bitsize)
      {
         output_bitstring = sign_extend_bitstring (output_bitstring, GET_INDEX_NODE(ga->op0), initial_size);
      }
      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " &");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg2_bitstring = arg2_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;
      for(;
          it_output_bitstring!=  output_bitstring.rend() &&
          it_arg1_bitstring != arg1_bitstring.rend();
          it_output_bitstring++, it_arg1_bitstring++, it_arg2_bitstring++)
      {
         if ( *it_output_bitstring == X)
         {
            res_input1.push_front(X);
         }
         else if(*it_arg1_bitstring == ZERO && *it_arg2_bitstring != ZERO)
         {
            res_input1.push_front(ZERO);
         }
         else if(*it_arg2_bitstring == ZERO)
         {
            res_input1.push_front(X);
         }
         else if (*it_arg1_bitstring == ONE && *it_arg2_bitstring == ONE)
         {
            res_input1.push_front(ONE);
         }
         else
         {
            res_input1.push_front(U);
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
   else if (op_kind == bit_ior_expr_K)
   {
      bit_ior_expr *operation = GetPointer<bit_ior_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);
      size_t output_bitsize = output_bitstring.size();

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();
      if (best.find (arg2_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      std::deque<bit_lattice> arg2_bitstring = best.find (arg2_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(arg1_uid) + " | " + STR(arg2_uid));

      size_t initial_size = arg1_bitstring.size ();
      if (is_signed(arg1_uid) && initial_size < output_bitsize)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, output_bitsize);
      }
      if (is_signed(arg1_uid) && initial_size > output_bitsize)
      {
         output_bitstring = sign_extend_bitstring (output_bitstring, GET_INDEX_NODE(ga->op0), initial_size);
      }
      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " |");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg2_bitstring = arg2_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;

      for(;
          it_output_bitstring!=  output_bitstring.rend() &&
          it_arg1_bitstring != arg1_bitstring.rend();
          it_output_bitstring++, it_arg1_bitstring++, it_arg2_bitstring++)
      {
         if ( *it_output_bitstring == X)
         {
            res_input1.push_front(X);
         }
         else if(*it_arg1_bitstring == ONE && *it_arg2_bitstring != ONE)
         {
            res_input1.push_front(ONE);
         }
         else if(*it_arg2_bitstring == ONE)
         {
            res_input1.push_front(X);
         }
         else if (*it_arg1_bitstring == ZERO && *it_arg2_bitstring == ZERO)
         {
            res_input1.push_front(ZERO);
         }
         else
         {
            res_input1.push_front(U);
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
   else if (op_kind == bit_ior_concat_expr_K)
   {
      bit_ior_concat_expr *operation = GetPointer<bit_ior_concat_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      size_t output_bitsize = output_bitstring.size();

      unsigned int arg1_uid = 0;
      unsigned int arg2_uid = 0;
      if(GET_INDEX_NODE(operation->op0) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op0);

      if(GET_INDEX_NODE(operation->op1) == output_id)
         arg2_uid = arg1_uid = GET_INDEX_NODE(operation->op1);

      if(GET_INDEX_NODE(operation->op2) == output_id)
         return std::deque<bit_lattice>();

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      long long int offset = GetPointer<integer_cst> (GET_NODE(operation->op2))->value;

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " |concat " + STR(GET_INDEX_NODE(operation->op1)) + "("+STR(offset)+")");

      size_t initial_size = arg1_bitstring.size ();
      if (is_signed(arg1_uid) && initial_size < output_bitsize)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, output_bitsize);
      }
      if (is_signed(arg1_uid) && initial_size > output_bitsize)
      {
         output_bitstring = sign_extend_bitstring (output_bitstring, GET_INDEX_NODE(ga->op0), initial_size);
      }


      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;
      long long int index = 0;

      if(arg2_uid)
      {
          for(;
              it_output_bitstring!=  output_bitstring.rend() &&
              it_arg1_bitstring != arg1_bitstring.rend() && index < offset;
              it_output_bitstring++, it_arg1_bitstring++, ++index)
          {
             res_input1.push_front(*it_output_bitstring);
          }
      }
      else
      {
          for(;
              it_output_bitstring!=  output_bitstring.rend() &&
              it_arg1_bitstring != arg1_bitstring.rend();
              it_output_bitstring++, it_arg1_bitstring++, ++index)
          {
             if(index < offset)
                res_input1.push_front(ZERO);
             else
                res_input1.push_front(*it_output_bitstring);
          }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
   else if (op_kind == bit_xor_expr_K)
   {
      bit_xor_expr *operation = GetPointer<bit_xor_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op0);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op1);
      size_t output_bitsize = output_bitstring.size();

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();
      if (best.find (arg2_uid) == best.end ())
         return std::deque<bit_lattice>();

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      std::deque<bit_lattice> arg2_bitstring = best.find (arg2_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(arg1_uid) + " | " + STR(arg2_uid));
      size_t initial_size = arg1_bitstring.size ();
      if (is_signed(arg1_uid) && initial_size < output_bitsize)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, output_bitsize);
      }
      if (is_signed(arg1_uid) && initial_size > output_bitsize)
      {
         output_bitstring = sign_extend_bitstring (output_bitstring, GET_INDEX_NODE(ga->op0), initial_size);
      }
      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " ^");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;

      for(;it_output_bitstring!=  output_bitstring.rend() && it_arg1_bitstring != arg1_bitstring.rend(); it_output_bitstring++, it_arg1_bitstring++)
      {
         if ( *it_output_bitstring == X)
         {
            res_input1.push_front(X);
         }
         else
         {
            res_input1.push_front(*it_arg1_bitstring);
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
#endif
#if 1
   else if (op_kind == bit_not_expr_K)
   {
      bit_not_expr *operation = GetPointer<bit_not_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      unsigned int arg1_uid = GET_INDEX_NODE(operation->op);
      size_t output_bitsize = output_bitstring.size();

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;
      size_t initial_size = arg1_bitstring.size ();
      if (is_signed(arg1_uid) && initial_size < output_bitsize)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, output_bitsize);
      }
      if (is_signed(arg1_uid) && initial_size > output_bitsize)
      {
         output_bitstring = sign_extend_bitstring (output_bitstring, GET_INDEX_NODE(ga->op0), initial_size);
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = ~" + STR(arg1_uid));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "~" + bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;

      for(;
          it_output_bitstring !=  output_bitstring.rend() && it_arg1_bitstring != arg1_bitstring.rend();
          it_output_bitstring++, it_arg1_bitstring++)
      {
         if ( *it_output_bitstring == X)
         {
            res_input1.push_front(X);
         }
         else
         {
            res_input1.push_front(*it_arg1_bitstring);
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
#endif
#if 1
   else if (op_kind == rshift_expr_K)
   {
      rshift_expr *operation = GetPointer<rshift_expr> (GET_NODE(ga->op1));

      if(GET_NODE(operation->op1)->get_kind () != integer_cst_K)
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            bool op_signed_p = is_signed(output_id);
            const tree_managerRef TM = AppM->get_tree_manager ();
            unsigned int precision = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(ga->op0)));
            unsigned int log2;
            for(log2=1; precision>(1u<<log2); ++log2);
            std::deque<bit_lattice> res_input1 = current.find (GET_INDEX_NODE(operation->op1))->second;
            for(unsigned int index = 0; res_input1.size()>index+log2; ++index)
            {
               if(op_signed_p && (res_input1.size()==index+log2+1))
                  res_input1[index] = ZERO;
               else
                  res_input1[index] = X;
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " >> " + STR(GET_INDEX_NODE(operation->op1)));
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
            return res_input1;
         }
         return std::deque<bit_lattice>();
      }
      THROW_ASSERT(output_id == GET_INDEX_NODE(operation->op0), "unexpected condition");

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;

      integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, created bitstring from constant -> " + STR(const2->value));
      if(const2->value < 0)
      {
         std::deque<bit_lattice> res_input1;
         res_input1.push_back (X);
         return res_input1;
      }

      std::deque<bit_lattice> res_input1=output_bitstring;

      while (res_input1.size () < (output_bitstring.size () + (static_cast<long long unsigned int> (const2->value))))
      {
         res_input1.push_back (X);
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " >> " + STR(GET_INDEX_NODE(operation->op1)));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
      return res_input1;
   }
#endif
#if 1
   else if (op_kind == lshift_expr_K)
   {
      lshift_expr *operation = GetPointer<lshift_expr> (GET_NODE(ga->op1));

      if(GET_NODE(operation->op1)->get_kind () != integer_cst_K)
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            bool op_signed_p = is_signed(output_id);
            const tree_managerRef TM = AppM->get_tree_manager ();
            unsigned int precision = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(ga->op0)));
            unsigned int log2;
            for(log2=1; precision>(1u<<log2); ++log2);
            std::deque<bit_lattice> res_input1 = current.find (GET_INDEX_NODE(operation->op1))->second;
            for(unsigned int index = 0; res_input1.size()>index+log2; ++index)
            {
               if(op_signed_p && (res_input1.size()==index+log2+1))
                  res_input1[index] = ZERO;
               else
                  res_input1[index] = X;
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " << " + STR(GET_INDEX_NODE(operation->op1)));
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));
            return res_input1;
         }
         return std::deque<bit_lattice>();
      }
      THROW_ASSERT(output_id == GET_INDEX_NODE(operation->op0), "unexpected condition: " + STR(AppM->get_tree_manager()->CGetTreeNode(output_id)) + " vs. " + STR(operation->op0));

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;

      integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, created bitstring from constant -> " + STR(const2->value) );

      if(const2->value < 0)
      {
         std::deque<bit_lattice> res_input1;
         res_input1.push_back (X);
         return res_input1;
      }

      std::deque<bit_lattice> res_input1=output_bitstring;

      while (res_input1.size () > (output_bitstring.size () - (static_cast<long long unsigned int> (const2->value))))
      {
         res_input1.pop_back();
      }
      while (res_input1.size () < output_bitstring.size ())
      {
         res_input1.push_front(X);
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " << " + STR(GET_INDEX_NODE(operation->op1)));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));

      return res_input1;
   }
#endif
   else if (op_kind == rrotate_expr_K)
   {
      rrotate_expr *operation = GetPointer<rrotate_expr> (GET_NODE(ga->op1));

      if(GET_NODE(operation->op1)->get_kind () != integer_cst_K)
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            bool op_signed_p = is_signed(output_id);
            const tree_managerRef TM = AppM->get_tree_manager ();
            unsigned int precision = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(ga->op0)));
            unsigned int log2;
            for(log2=1; precision>(1u<<log2); ++log2);
            std::deque<bit_lattice> res_input1 = current.find (GET_INDEX_NODE(operation->op1))->second;
            for(unsigned int index = 0; res_input1.size()>index+log2; ++index)
            {
               if(op_signed_p && (res_input1.size()==index+log2+1))
                  res_input1[index] = ZERO;
               else
                  res_input1[index] = X;
            }
            return res_input1;
         }
         return std::deque<bit_lattice>();
      }
      return std::deque<bit_lattice>();
   }
   else if (op_kind == lrotate_expr_K)
   {
      lrotate_expr *operation = GetPointer<lrotate_expr> (GET_NODE(ga->op1));

      if(GET_NODE(operation->op1)->get_kind () != integer_cst_K)
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            bool op_signed_p = is_signed(output_id);
            const tree_managerRef TM = AppM->get_tree_manager ();
            unsigned int precision = tree_helper::size(TM, tree_helper::get_type_index(TM,GET_INDEX_NODE(ga->op0)));
            unsigned int log2;
            for(log2=1; precision>(1u<<log2); ++log2);
            std::deque<bit_lattice> res_input1 = current.find (GET_INDEX_NODE(operation->op1))->second;
            for(unsigned int index = 0; res_input1.size()>index+log2; ++index)
            {
               if(op_signed_p && (res_input1.size()==index+log2+1))
                  res_input1[index] = ZERO;
               else
                  res_input1[index] = X;
            }
            return res_input1;
         }
         return std::deque<bit_lattice>();
      }
      return std::deque<bit_lattice>();
   }
#if 1
   else if (op_kind == nop_expr_K || op_kind == convert_expr_K)
   {
      unary_expr *operation = GetPointer<unary_expr> (GET_NODE(ga->op1));

      THROW_ASSERT(output_id == GET_INDEX_NODE(operation->op), "unexpected condition");

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      const unsigned int arg1_bitsize = tree_helper::Size(GET_NODE(operation->op));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + (is_signed(GET_INDEX_NODE(ga->op0))?"S":"U") + " = cast " + STR(GET_INDEX_NODE(operation->op)) + (is_signed(GET_INDEX_NODE(operation->op))?"S":"U"));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, (op_kind == nop_expr_K?std::string("cast"):std::string("convert")) + " " + bitstring_to_string (best.find (GET_INDEX_NODE(operation->op))->second) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, (op_kind == nop_expr_K?"     ":"        ")+bitstring_to_string (output_bitstring));

      if(is_signed(GET_INDEX_NODE(operation->op))  && !is_signed(GET_INDEX_NODE(ga->op0)))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " +(op_kind != nop_expr_K?std::string("   "):std::string("")) + bitstring_to_string (best.find (GET_INDEX_NODE(operation->op))->second) + "(" + STR(arg1_bitsize) + "->" + STR(tree_helper::Size(GET_NODE(ga->op0))) + ")");
         return best.find (GET_INDEX_NODE(operation->op))->second;
      }
      std::deque<bit_lattice> res_input1;
      res_input1 = output_bitstring;
      while(res_input1.size()> arg1_bitsize)
         res_input1.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " +(op_kind != nop_expr_K?std::string("   "):std::string("")) + bitstring_to_string (res_input1) + "(" + STR(arg1_bitsize) + "->" + STR(tree_helper::Size(GET_NODE(ga->op0))) + ")");

      return res_input1;
   }
#endif
#if 1
   else if(op_kind == cond_expr_K)
   {
      cond_expr *operation = GetPointer<cond_expr> (GET_NODE(ga->op1));

      unsigned int arg0_uid = GET_INDEX_NODE(operation->op0);
      if(arg0_uid == output_id)
         return std::deque<bit_lattice>();

      unsigned int arg1_uid = GET_INDEX_NODE(operation->op1);
      unsigned int arg2_uid = GET_INDEX_NODE(operation->op2);

      if(arg1_uid != output_id)
         std::swap(arg1_uid, arg2_uid);
      THROW_ASSERT(output_id == arg1_uid, "unexpected condition for statement "+ga->ToString());

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      const std::deque<bit_lattice> &arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(arg0_uid) + " - " + STR(arg1_uid) + " - " + STR(arg2_uid));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      std::deque<bit_lattice>::const_reverse_iterator it_output_bitstring = output_bitstring.rbegin();
      std::deque<bit_lattice>::const_reverse_iterator it_arg1_bitstring = arg1_bitstring.rbegin();

      std::deque<bit_lattice> res_input1;

      for(;
          it_output_bitstring !=  output_bitstring.rend() && it_arg1_bitstring != arg1_bitstring.rend();
          it_output_bitstring++, it_arg1_bitstring++)
      {
         if ( *it_output_bitstring == X)
         {
            res_input1.push_front(X);
         }
         else
         {
            res_input1.push_front(*it_arg1_bitstring);
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (res_input1));

      return res_input1;
   }
#endif
#if 1
   else if(op_kind == ternary_plus_expr_K)
   {
      ternary_plus_expr *operation = GetPointer<ternary_plus_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = 0;
      if(GET_INDEX_NODE(operation->op0) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op0);

      if(GET_INDEX_NODE(operation->op1) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op1);

      if(GET_INDEX_NODE(operation->op2) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op2);

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " + " + STR(GET_INDEX_NODE(operation->op1)) + " + " + STR(GET_INDEX_NODE(operation->op2)));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 1
   else if(op_kind == ternary_pm_expr_K)
   {
      ternary_pm_expr *operation = GetPointer<ternary_pm_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = 0;
      if(GET_INDEX_NODE(operation->op0) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op0);

      if(GET_INDEX_NODE(operation->op1) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op1);

      if(GET_INDEX_NODE(operation->op2) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op2);

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " + " + STR(GET_INDEX_NODE(operation->op1)) + " - " + STR(GET_INDEX_NODE(operation->op2)));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 1
   else if(op_kind == ternary_mp_expr_K)
   {
      ternary_mp_expr *operation = GetPointer<ternary_mp_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = 0;
      if(GET_INDEX_NODE(operation->op0) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op0);

      if(GET_INDEX_NODE(operation->op1) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op1);

      if(GET_INDEX_NODE(operation->op2) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op2);

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " - " + STR(GET_INDEX_NODE(operation->op1)) + " + " + STR(GET_INDEX_NODE(operation->op2)));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 1
   else if(op_kind == ternary_mm_expr_K)
   {
      ternary_mm_expr *operation = GetPointer<ternary_mm_expr> (GET_NODE(ga->op1));

      unsigned int arg1_uid = 0;
      if(GET_INDEX_NODE(operation->op0) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op0);

      if(GET_INDEX_NODE(operation->op1) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op1);

      if(GET_INDEX_NODE(operation->op2) == output_id)
         arg1_uid = GET_INDEX_NODE(operation->op2);

      THROW_ASSERT(output_id == arg1_uid, "unexpected condition");

      if (best.find (arg1_uid) == best.end ())
         return std::deque<bit_lattice>();

      const std::deque<bit_lattice> &output_bitstring = current.find (GET_INDEX_NODE(ga->op0))->second;
      std::deque<bit_lattice> arg1_bitstring = best.find (arg1_uid)->second;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer, operation: " + STR(GET_INDEX_NODE(ga->op0)) + " = " + STR(GET_INDEX_NODE(operation->op0)) + " - " + STR(GET_INDEX_NODE(operation->op1)) + " - " + STR(GET_INDEX_NODE(operation->op2)));

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (output_bitstring));

      while(arg1_bitstring.size()>output_bitstring.size())
         arg1_bitstring.pop_front();

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " input1: " + bitstring_to_string (arg1_bitstring));

      return arg1_bitstring;
   }
#endif
#if 1
   else if (op_kind == mem_ref_K )
   {
      mem_ref *operation = GetPointer<mem_ref> (GET_NODE(ga->op1));

      if(GET_INDEX_NODE(operation->op0) == output_id && current.find(GET_INDEX_NODE(operation->op0)) != current.end())
      {
         return create_u_bitstring(pointer_resizing(AppM, output_id, function_behavior, function_id, not_frontend, parameters));
      }
      else
         return std::deque<bit_lattice>();
   }
   else if(op_kind == target_mem_ref461_K )
   {
      target_mem_ref461 *operation = GetPointer<target_mem_ref461> (GET_NODE(ga->op1));

      if(GET_INDEX_NODE(operation->base) == output_id && current.find(GET_INDEX_NODE(operation->base)) != current.end())
      {
         return create_u_bitstring(pointer_resizing(AppM, output_id, function_behavior, function_id, not_frontend, parameters));
      }
      else if(operation->idx2 && GET_INDEX_NODE(operation->idx2) == output_id && current.find(GET_INDEX_NODE(operation->idx2)) != current.end())
      {
         return create_u_bitstring(pointer_resizing(AppM, output_id, function_behavior, function_id, not_frontend, parameters));
      }
      else
         return std::deque<bit_lattice>();
   }
   else if(op_kind == array_ref_K)
   {
      array_ref *operation = GetPointer<array_ref> (GET_NODE(ga->op1));
      do
      {
         if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
         {
            return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(operation->op0), function_behavior, function_id, not_frontend, parameters));
         }
         operation = GetPointer<array_ref> (GET_NODE(operation->op0));
      } while(operation);
      return std::deque<bit_lattice>();
   }
   else if(op_kind == addr_expr_K)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Backward transfer of addr_expr");
      if(ga->temporary_address)
      {
         addr_expr *operand = GetPointer<addr_expr> (GET_NODE(ga->op1));
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Operand is " + GET_NODE(operand->op)->get_kind_text());
         op_kind = GET_NODE(operand->op)->get_kind();
         if (op_kind == mem_ref_K )
         {
            mem_ref *operation = GetPointer<mem_ref> (GET_NODE(operand->op));

            if(GET_INDEX_NODE(operation->op0) == output_id && current.find(GET_INDEX_NODE(operation->op0)) != current.end())
            {
               return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(ga->op0), function_behavior, function_id, not_frontend, parameters));
            }
            else
            {
               return std::deque<bit_lattice>();
            }
         }
         else if(op_kind == target_mem_ref461_K )
         {
            target_mem_ref461 *operation = GetPointer<target_mem_ref461> (GET_NODE(operand->op));

            if(GET_INDEX_NODE(operation->base) == output_id && current.find(GET_INDEX_NODE(operation->base)) != current.end())
            {
               return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(ga->op0), function_behavior, function_id, not_frontend, parameters));
            }
            else if(operation->idx2 && GET_INDEX_NODE(operation->idx2) == output_id && current.find(GET_INDEX_NODE(operation->idx2)) != current.end())
            {
               return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(ga->op0), function_behavior, function_id, not_frontend, parameters));
            }
            else
            {
               return std::deque<bit_lattice>();
            }
         }
         else if(op_kind == array_ref_K)
         {
            array_ref *operation = GetPointer<array_ref> (GET_NODE(operand->op));
            do
            {
               if(GET_INDEX_NODE(operation->op1) == output_id && current.find(GET_INDEX_NODE(operation->op1)) != current.end())
               {
                  return create_u_bitstring(pointer_resizing(AppM, GET_INDEX_NODE(operation->op0), function_behavior, function_id, not_frontend, parameters));
               }
               operation = GetPointer<array_ref> (GET_NODE(operation->op0));
            } while(operation);
            return std::deque<bit_lattice>();
         }
         else
            return std::deque<bit_lattice>();
      }
      else
         return std::deque<bit_lattice>();
   }
   else if(op_kind == lt_expr_K || op_kind == gt_expr_K || op_kind == le_expr_K || op_kind == ge_expr_K || op_kind == eq_expr_K || op_kind == ne_expr_K ||
           op_kind == call_expr_K ||
           op_kind == trunc_div_expr_K || op_kind == exact_div_expr_K ||
           op_kind == trunc_mod_expr_K ||
           op_kind == view_convert_expr_K||
           op_kind == min_expr_K || op_kind == max_expr_K )
   {
      return std::deque<bit_lattice>();
   }
#endif
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "backward_transfer Error: operation unhandled yet -> "+ GET_NODE(ga->op1)->get_kind_text());
   return std::deque<bit_lattice>();

}

std::deque<bit_lattice>
Bit_Value::forward_transfer_plus_expr (const std::deque<bit_lattice> &arg1_bitstring, const std::deque<bit_lattice> &arg2_bitstring)
{
   std::deque<bit_lattice> res;

   //PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " +");
   //PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");
   std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
   std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
   bit_lattice carry1 = ZERO;

   for (;
        arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
        arg1_it++, arg2_it++)
   {
      res.push_front(plus_expr_map[*arg1_it][*arg2_it][carry1].back());
      carry1 = plus_expr_map[*arg1_it][*arg2_it][carry1].front();
   }
   res.push_front(carry1);

   // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (res));
   return res;
}

std::deque<bit_lattice>
Bit_Value::forward_transfer_minus_expr (const std::deque<bit_lattice> &arg1_bitstring, const std::deque<bit_lattice> &arg2_bitstring)
{
   std::deque<bit_lattice> res;

   //PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " +");
   //PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");
   std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
   std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
   bit_lattice borrow1 = ZERO;

   for (;
        arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
        arg1_it++, arg2_it++)
   {
      res.push_front(minus_expr_map[*arg1_it][*arg2_it][borrow1].back());
      borrow1 = minus_expr_map[*arg1_it][*arg2_it][borrow1].front();
   }
   res.push_front(borrow1);

   // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (res));
   return res;
}

std::deque<bit_lattice>
Bit_Value::create_bitstring_couple(bit_lattice e1, bit_lattice e0){
   std::deque<bit_lattice> res;
   res.push_front(e0);
   res.push_front(e1);
   return res;
}

bool Bit_Value::manage_forward_binary_operands(const binary_expr* operation, unsigned int &arg1_uid, unsigned int &arg2_uid,
                                        std::deque<bit_lattice> &arg1_bitstring, std::deque<bit_lattice> &arg2_bitstring) const
{
   if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
   {
      arg1_uid = GET_INDEX_NODE(operation->op0);

      if (current.find (arg1_uid) == current.end ())
      {
         return false;
      }
      arg1_bitstring = current.find (arg1_uid)->second;
   }
   else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
   {
      arg1_uid = GET_INDEX_NODE(operation->op0);
      THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
      arg1_bitstring = best.find (arg1_uid)->second;
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
   }
   else
      return false;

   if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
   {
      arg2_uid = GET_INDEX_NODE(operation->op1);

      if (current.find (arg2_uid) == current.end ())
      {
         return false;
      }
      arg2_bitstring = current.find (arg2_uid)->second;
   }
   else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
   {
      arg2_uid = GET_INDEX_NODE(operation->op1);
      THROW_ASSERT(best.find (arg2_uid) != best.end(), "unexpected condition");
      arg2_bitstring = best.find (arg2_uid)->second;
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op1))->value) + " : " + bitstring_to_string(arg2_bitstring));
   }
   else
      return false;
   return true;
}
std::deque<bit_lattice>
Bit_Value::forward_transfer (const gimple_assign * ga)
{
   std::deque<bit_lattice> res;
   unsigned int output_uid = GET_INDEX_NODE(ga->op0);
   enum kind op_kind = GET_NODE(ga->op1)->get_kind ();
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, GET_NODE(ga->op1)->get_kind_text());

   if(op_kind == ssa_name_K)
   {
      unsigned int arg2_uid;
      arg2_uid = GET_INDEX_NODE(ga->op1);
      if (current.find (arg2_uid) == current.end ())
      {
         return res;
      }
      res = current.find (arg2_uid)->second;
   }
   else if(op_kind == integer_cst_K)
   {
      unsigned int arg2_uid;
      arg2_uid = GET_INDEX_NODE(ga->op1);
      THROW_ASSERT(best.find (arg2_uid) != best.end(), "unexpected condition");
      res = best.find (arg2_uid)->second;
   }
#if 1
   else if (op_kind == plus_expr_K || op_kind == pointer_plus_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " + " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " +");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
      bit_lattice carry1 = ZERO;
      unsigned int max_bitsize = tree_helper::Size(GET_NODE(ga->op0));

      for (unsigned bit_index = 0;
           bit_index< max_bitsize && arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++, ++bit_index)
      {
         // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "arg1: "+STR(*arg1_it) + "arg2: "STR(*arg2_it) + "carry: " + STR(carry1));
         res.push_front(plus_expr_map[*arg1_it][*arg2_it][carry1].back());
         carry1 = plus_expr_map[*arg1_it][*arg2_it][carry1].front();
      }

      if(is_signed(output_uid) && res.size()<max_bitsize)
         res.push_front(plus_expr_map[arg1_bitstring.front()][arg2_bitstring.front()][carry1].back());
      else if(!is_signed(output_uid))
      {
         while(res.size()<max_bitsize)
         {
            res.push_front(plus_expr_map[ZERO][ZERO][carry1].back());
            carry1 = plus_expr_map[ZERO][ZERO][carry1].front();
         }
      }
   }
#endif
#if 1
   else if (op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K||op_kind == ternary_mp_expr_K||op_kind == ternary_mm_expr_K)
   {
      ternary_expr *operation = GetPointer<ternary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      std::deque<bit_lattice> arg3_bitstring;
      unsigned int arg1_uid = 0;
      unsigned int arg2_uid = 0;
      unsigned int arg3_uid = 0;

      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
            return res;
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         if (current.find (arg2_uid) == current.end ())
            return res;
         arg2_bitstring = current.find (arg2_uid)->second;
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         THROW_ASSERT(best.find (arg2_uid) != best.end(), "unexpected condition");
         arg2_bitstring = best.find (arg2_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op1))->value) + " : " + bitstring_to_string(arg2_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op2)->get_kind () == ssa_name_K)
      {
         arg3_uid = GET_INDEX_NODE(operation->op2);
         if (current.find (arg3_uid) == current.end ())
            return res;
         arg3_bitstring = current.find (arg3_uid)->second;
      }
      else if (GET_NODE(operation->op2)->get_kind () == integer_cst_K)
      {
         arg3_uid = GET_INDEX_NODE(operation->op2);
         THROW_ASSERT(best.find (arg3_uid) != best.end(), "unexpected condition");
         arg3_bitstring = best.find (arg3_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op2))->value) + " : " + bitstring_to_string(arg3_bitstring));
      }
      else
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg1_bitstring.size () > arg3_bitstring.size ())
      {
         arg3_bitstring = sign_extend_bitstring (arg3_bitstring, arg3_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg3_bitstring.size ())
      {
         arg3_bitstring = sign_extend_bitstring (arg3_bitstring, arg3_uid, arg2_bitstring.size ());
      }
      if (arg3_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg3_bitstring.size ());
      }
      if (arg3_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg3_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + (op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K ?" + " : " - ") + STR(arg2_uid) + (op_kind == ternary_plus_expr_K||op_kind == ternary_mp_expr_K ? " +" : " - ") + STR(arg3_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + (op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K ?" + " : " - "));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + (op_kind == ternary_plus_expr_K||op_kind == ternary_mp_expr_K ? " +" : " - "));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg3_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
      bit_lattice carry1 = ZERO;
      std::deque<bit_lattice> res_int;
      unsigned int max_bitsize = tree_helper::Size(GET_NODE(ga->op0));

      for (unsigned bit_index = 0;
           bit_index< max_bitsize && arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++, ++bit_index)
      {
         if(op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K)
         {
            res_int.push_front(plus_expr_map[*arg1_it][*arg2_it][carry1].back());
            carry1 = plus_expr_map[*arg1_it][*arg2_it][carry1].front();
         }
         else
         {
            res_int.push_front(minus_expr_map[*arg1_it][*arg2_it][carry1].back());
            carry1 = minus_expr_map[*arg1_it][*arg2_it][carry1].front();
         }
      }

      if(is_signed(output_uid) && res_int.size()<max_bitsize)
      {
         if(op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K)
            res_int.push_front(plus_expr_map[arg1_bitstring.front()][arg2_bitstring.front()][carry1].back());
         else
            res_int.push_front(minus_expr_map[arg1_bitstring.front()][arg2_bitstring.front()][carry1].back());
      }
      else if(!is_signed(output_uid))
      {
         while(res_int.size()<max_bitsize)
         {
            if(op_kind == ternary_plus_expr_K||op_kind == ternary_pm_expr_K)
            {
               res_int.push_front(plus_expr_map[ZERO][ZERO][carry1].back());
               carry1 = plus_expr_map[ZERO][ZERO][carry1].front();
            }
            else
            {
               res_int.push_front(minus_expr_map[ZERO][ZERO][carry1].back());
               carry1 = minus_expr_map[ZERO][ZERO][carry1].front();
            }
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (res_int)+ ":res_int:");

      if (res_int.size () > arg3_bitstring.size ())
      {
         arg3_bitstring = sign_extend_bitstring (arg3_bitstring, arg3_uid, res_int.size ());
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg3_bitstring)+ ":arg3_bitstring:");
      std::deque<bit_lattice>::const_reverse_iterator arg3_it = arg3_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator res_int_it = res_int.rbegin ();
      carry1 = ZERO;
      for (unsigned bit_index = 0;
           bit_index< max_bitsize && arg3_it != arg3_bitstring.rend () && res_int_it != res_int.rend ();
           arg3_it++, res_int_it++, ++bit_index)
      {
         if(op_kind == ternary_plus_expr_K||op_kind == ternary_mp_expr_K)
         {
            res.push_front(plus_expr_map[*arg3_it][*res_int_it][carry1].back());
            carry1 = plus_expr_map[*arg3_it][*res_int_it][carry1].front();
         }
         else
         {
            res.push_front(minus_expr_map[*arg3_it][*res_int_it][carry1].back());
            carry1 = minus_expr_map[*arg3_it][*res_int_it][carry1].front();
         }
      }

      if(is_signed(output_uid) && res.size()<max_bitsize)
      {
         if(op_kind == ternary_plus_expr_K||op_kind == ternary_mp_expr_K)
            res.push_front(plus_expr_map[arg3_bitstring.front()][res_int.front()][carry1].back());
         else
            res.push_front(minus_expr_map[arg3_bitstring.front()][res_int.front()][carry1].back());
      }
      else if(!is_signed(output_uid))
      {
         while(res.size()<max_bitsize)
         {
            if(op_kind == ternary_plus_expr_K||op_kind == ternary_mp_expr_K)
            {
               res.push_front(plus_expr_map[ZERO][ZERO][carry1].back());
               carry1 = plus_expr_map[ZERO][ZERO][carry1].front();
            }
            else
            {
               res.push_front(minus_expr_map[ZERO][ZERO][carry1].back());
               carry1 = minus_expr_map[ZERO][ZERO][carry1].front();
            }
         }
      }
   }
#endif
#if 1
   else if (op_kind == mult_expr_K || op_kind == widen_mult_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " * " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " *");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
      std::deque<bit_lattice>::const_iterator arg1_it_fw = arg1_bitstring.begin ();
      std::deque<bit_lattice>::const_iterator arg2_it_fw = arg2_bitstring.begin ();
      //trailing zeros of a
      unsigned int ta = 0;
      //trailing zeros of b
      unsigned int tb = 0;
      //leading zeros of a
      unsigned int la = 0;
      //leading zeros of b
      unsigned int lb = 0;
      for (;
           arg1_it != arg1_bitstring.rend () && *arg1_it == ZERO;
           arg1_it++, ta++);

      for (;
           arg2_it != arg2_bitstring.rend () && *arg2_it == ZERO;
           arg2_it++,tb++);

      for (;
           arg1_it_fw != arg1_bitstring.end () && *arg1_it_fw == ZERO;
           arg1_it_fw++, la++);


      for (;
           arg2_it_fw != arg2_bitstring.end () && *arg2_it_fw == ZERO;
           arg2_it_fw++,lb++);

      //if one of the two arguments is all zeros returns zero
      if(la == arg1_bitstring.size() || lb == arg2_bitstring.size())
      {
         res.push_back(ZERO);
      }
      else
      {
         size_t lenght_arg1=arg1_bitstring.size();
         size_t lenght_arg2=arg2_bitstring.size();
         unsigned res_bitsize = tree_helper::Size(GET_NODE(ga->op0));
         THROW_ASSERT(static_cast<int>(lenght_arg1+lenght_arg2-ta-tb-la-lb)>0, "unexpected condition");
         res = create_u_bitstring(static_cast<unsigned int>(lenght_arg1+lenght_arg2-ta-tb-la-lb));
         for(unsigned int i =0; i<ta+tb; i++)res.push_back(ZERO);
         if(is_signed(GET_INDEX_NODE(ga->op0)) && la > 0 && lb > 0)
            res.push_front(ZERO);
         while(res.size()>res_bitsize)
            res.pop_front();
      }
   }
#endif
#if 1
   else if(op_kind == trunc_div_expr_K || op_kind == exact_div_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;
      unsigned res_bitsize = tree_helper::Size(GET_NODE(ga->op0));

      if(is_signed(arg1_uid))
         res = create_u_bitstring(1u+static_cast<unsigned int>(arg1_bitstring.size()));
      else
         res = create_u_bitstring(static_cast<unsigned int>(arg1_bitstring.size()));
      while(res.size()>res_bitsize)
         res.pop_front();
   }
   else if(op_kind == trunc_mod_expr_K)
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;
      unsigned res_bitsize = tree_helper::Size(GET_NODE(ga->op0));

      if(is_signed(arg1_uid))
         res = create_u_bitstring(1+static_cast<unsigned int>(std::min(arg1_bitstring.size(), arg2_bitstring.size())));
      else
         res = create_u_bitstring(static_cast<unsigned int>(std::min(arg1_bitstring.size(), arg2_bitstring.size())));
      while(res.size()>res_bitsize)
         res.pop_front();
   }
#endif
#if 1
   else if (op_kind == gt_expr_K)
   {
      gt_expr *operation = GetPointer<gt_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      bool is_signed_var = is_signed(arg1_uid) || is_signed(arg2_uid) ;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " > " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " >");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            else
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            else
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
      }
      if(!computed_result){
         //If the result is not computed until now the bitstring are equal
         // <0> is FALSE
         res.push_front(ZERO);
      }
   }
   else if (op_kind == ge_expr_K)
   {
      ge_expr *operation = GetPointer<ge_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " >= " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " >=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      bool is_signed_var = is_signed(arg1_uid) || is_signed(arg2_uid) ;
      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            else
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            else
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
      }
      if(!computed_result){
         //If the result is not computed until now the bitstring are equal
         // <01> is TRUE
         res.push_front(ONE);
         res.push_front(ZERO);
      }
   }
   else if (op_kind == lt_expr_K)
   {
      lt_expr *operation = GetPointer<lt_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " < " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      bool is_signed_var = is_signed(arg1_uid) || is_signed(arg2_uid) ;
      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            else
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            else
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
      }
      if(!computed_result)
      {
         //If the result is not computed until now the bitstring are equal
         // <0> is FALSE
         res.push_front(ZERO);
      }
   }
   else if (op_kind == le_expr_K)
   {
      le_expr *operation = GetPointer<le_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " <= " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " <=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      bool is_signed_var = is_signed(arg1_uid) || is_signed(arg2_uid) ;
      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            else
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
         else if( *arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO )
         {
            if(arg1_bitstring_it == arg1_bitstring.begin() && is_signed_var)
            {
               // <01> is TRUE
               res.push_front(ONE);
               if(is_signed(GET_INDEX_NODE(ga->op0)))
                  res.push_front(ZERO);
            }
            else
            {
               // <0> is FALSE
               res.push_front(ZERO);
            }
            computed_result = true;
            break;
         }
      }
      if(!computed_result)
      {
         //If the result is not computed until now the bitstring are equal
         // <01> is TRUE
         res.push_front(ONE);
         if(is_signed(GET_INDEX_NODE(ga->op0)))
            res.push_front(ZERO);
      }
   }
   else if (op_kind == eq_expr_K)
   {
      eq_expr *operation = GetPointer<eq_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " == " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " ==");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( (*arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE) ||
                  (*arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO)
                  )
         {
            // <0> is FALSE
            res.push_front(ZERO);
            computed_result = true;
            break;
         }
      }
      if(!computed_result)
      {
         //If the result is not computed until now the bitstring are equal
         // <01> is TRUE
         res.push_front(ONE);
         if(is_signed(GET_INDEX_NODE(ga->op0)))
            res.push_front(ZERO);
      }
   }
   else if (op_kind == ne_expr_K)
   {
      ne_expr *operation = GetPointer<ne_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " != " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " !=");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " ?");

      std::deque<bit_lattice>::const_iterator arg1_bitstring_it = arg1_bitstring.begin();
      std::deque<bit_lattice>::const_iterator arg2_bitstring_it = arg2_bitstring.begin();
      bool computed_result = false;
      for(; arg1_bitstring_it != arg1_bitstring.end() && arg2_bitstring_it != arg2_bitstring.end(); arg1_bitstring_it++,arg2_bitstring_it++)
      {
         if( *arg1_bitstring_it == U || *arg2_bitstring_it==U )
         {
            // <0U> is UNKNOWN
            res.push_front(U);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
         else if( (*arg1_bitstring_it == ZERO && *arg2_bitstring_it==ONE) ||
                  (*arg1_bitstring_it == ONE && *arg2_bitstring_it==ZERO)
                  )
         {
            // <01> is TRUE
            res.push_front(ONE);
            if(is_signed(GET_INDEX_NODE(ga->op0)))
               res.push_front(ZERO);
            computed_result = true;
            break;
         }
      }
      if(!computed_result)
      {
         //If the result is not computed until now the bitstring are equal
         // <0> is FALSE
         res.push_front(ZERO);
      }
   }
#endif
#if 1
   else if (op_kind == minus_expr_K)
   {
      minus_expr *operation = GetPointer<minus_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " - " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " -");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
      bit_lattice borrow = ZERO;

      unsigned int max_bitsize = tree_helper::Size(GET_NODE(ga->op0));
      for (unsigned bit_index = 0;
           bit_index< max_bitsize && arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++)
      {
         res.push_front(minus_expr_map[*arg1_it][*arg2_it][borrow].back());
         borrow = minus_expr_map[*arg1_it][*arg2_it][borrow].front();
      }
      if(is_signed(output_uid) && res.size()<max_bitsize)
         res.push_front(minus_expr_map[arg1_bitstring.front()][arg2_bitstring.front()][borrow].back());
      else if (!is_signed(output_uid))
      {
         while(res.size()<max_bitsize)
                  res.push_front(borrow);
      }
   }
#endif
#if 1
   else if (op_kind == negate_expr_K)
   {
      negate_expr *operation = GetPointer<negate_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg_bitstring;
      unsigned int arg_uid = 0;

      if(GET_NODE(operation->op)->get_kind () == ssa_name_K)
      {
         arg_uid = GET_INDEX_NODE(operation->op);
         if (current.find (arg_uid) == current.end ())
         {
            return res;
         }
         arg_bitstring = current.find (arg_uid)->second;
      }
      else if(GET_NODE(operation->op)->get_kind () == integer_cst_K)
      {
         arg_uid = GET_INDEX_NODE(operation->op);
         THROW_ASSERT(best.find (arg_uid) != best.end(), "unexpected condition");
         arg_bitstring = best.find (arg_uid)->second;
      }
      else
         return res;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = - " + STR(arg_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " -");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg_bitstring) + " =");

      bit_lattice borrow = ZERO;
      unsigned int max_bitsize = tree_helper::Size(GET_NODE(ga->op0));
      if(!is_signed(output_uid) && max_bitsize > arg_bitstring.size ())
      {
         arg_bitstring = sign_extend_bitstring (arg_bitstring, output_uid, max_bitsize);
      }
      std::deque<bit_lattice>::const_reverse_iterator arg_it = arg_bitstring.rbegin ();
      for (unsigned bit_index = 0;
           bit_index< max_bitsize && arg_it != arg_bitstring.rend ();
           arg_it++)
      {
         res.push_front(minus_expr_map[ZERO][*arg_it][borrow].back());
         borrow = minus_expr_map[ZERO][*arg_it][borrow].front();
      }
      if(is_signed(output_uid) && res.size()<max_bitsize)
         res.push_front(minus_expr_map[ZERO][arg_bitstring.front()][borrow].back());

   }
#endif
#if 1
   else if (op_kind == abs_expr_K)
   {
      abs_expr *operation = GetPointer<abs_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg_bitstring;
      unsigned int arg_uid = 0;

      if(GET_NODE(operation->op)->get_kind () == ssa_name_K)
      {
         arg_uid = GET_INDEX_NODE(operation->op);
         if (current.find (arg_uid) == current.end ())
         {
            return res;
         }
         arg_bitstring = current.find (arg_uid)->second;
      }
      else if(GET_NODE(operation->op)->get_kind () == integer_cst_K)
      {
         arg_uid = GET_INDEX_NODE(operation->op);
         THROW_ASSERT(best.find (arg_uid) != best.end(), "unexpected condition");
         arg_bitstring = best.find (arg_uid)->second;
      }
      else
         return res;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = abs " + STR(arg_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " abs");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg_it = arg_bitstring.rbegin ();

      if(!is_signed(arg_uid))
         res = arg_bitstring;
      else
      {
         if(arg_bitstring.front() == ONE)
         {
            bit_lattice borrow = ZERO;

            unsigned int max_bitsize = tree_helper::Size(GET_NODE(ga->op0));
            for (unsigned bit_index = 0;
                 bit_index< max_bitsize && arg_it != arg_bitstring.rend ();
                 arg_it++)
            {
               res.push_front(minus_expr_map[ZERO][*arg_it][borrow].back());
               borrow = minus_expr_map[ZERO][*arg_it][borrow].front();
            }
            if(is_signed(output_uid) && res.size()<max_bitsize)
               res.push_front(minus_expr_map[ZERO][arg_bitstring.front()][borrow].back());
            else
            {
               while(res.size()<max_bitsize)
               {
                  res.push_front(minus_expr_map[ZERO][arg_bitstring.front()][borrow].back());
                  borrow = minus_expr_map[ZERO][arg_bitstring.front()][borrow].front();
               }
            }
         }
         else if(arg_bitstring.front() == ZERO)
         {
            res = arg_bitstring;
         }
         else
         {
            res = arg_bitstring;
         }
      }
   }
#endif
#if 1
   else if (op_kind == bit_and_expr_K)
   {
      bit_and_expr *operation = GetPointer<bit_and_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      unsigned int max_size = tree_helper::Size(GET_NODE(ga->op0));
      if (max_size > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, max_size);
      }
      if (max_size > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, max_size);
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " & " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " &");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();

      for (;
           arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++)
      {
         res.push_front(bit_and_expr_map[*arg1_it][*arg2_it]);
      }
   }
#endif
#if 1
   else if (op_kind == bit_ior_expr_K)
   {
      bit_ior_expr *operation = GetPointer<bit_ior_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      unsigned int max_size = tree_helper::Size(GET_NODE(ga->op0));
      if (max_size > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, max_size);
      }
      if (max_size > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, max_size);
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " | " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " |");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();

      for (;
           arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++)
      {
         res.push_front(bit_ior_expr_map[*arg1_it][*arg2_it]);
      }
   }
#endif
#if 1
   else if (op_kind == bit_ior_concat_expr_K)
   {
      bit_ior_concat_expr *operation = GetPointer<bit_ior_concat_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg1_uid = 0;
      unsigned int arg2_uid = 0;

      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
            return res;
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         if (current.find (arg2_uid) == current.end ())
            return res;
         arg2_bitstring = current.find (arg2_uid)->second;
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         THROW_ASSERT(best.find (arg2_uid) != best.end(), "unexpected condition");
         arg2_bitstring = best.find (arg2_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op1))->value) + " : " + bitstring_to_string(arg2_bitstring));
      }
      else
         return res;

      long long int offset = GetPointer<integer_cst> (GET_NODE(operation->op2))->value;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " |concat " + STR(arg2_uid) + "(" + STR(offset)+")");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " |concat");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();
      long long index = 0;

      for (;
           arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++, ++index)
      {
         if(index<offset)
            res.push_front(*arg2_it);
         else
            res.push_front(*arg1_it);
      }
   }
#endif
#if 1
   else if (op_kind == bit_xor_expr_K)
   {
      bit_xor_expr *operation = GetPointer<bit_xor_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      unsigned int max_size = tree_helper::Size(GET_NODE(ga->op0));
      if (max_size > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, max_size);
      }
      if (max_size > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, max_size);
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " ^ " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " ^");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      std::deque<bit_lattice>::const_reverse_iterator arg2_it = arg2_bitstring.rbegin ();

      for (;
           arg1_it != arg1_bitstring.rend () && arg2_it != arg2_bitstring.rend ();
           arg1_it++, arg2_it++)
      {
         res.push_front(bit_xor_expr_map[*arg1_it][*arg2_it]);
      }
   }
#endif
#if 1
   else if (op_kind == bit_not_expr_K)
   {
      bit_not_expr *operation = GetPointer<bit_not_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      unsigned int arg1_uid = 0;

      if (GET_NODE(operation->op)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;
      unsigned int max_size = tree_helper::Size(GET_NODE(ga->op0));
      if (arg1_bitstring.size () < max_size)
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, max_size);
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = ~" + STR(arg1_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " ~" + bitstring_to_string (arg1_bitstring) + " =");

      std::deque<bit_lattice>::const_reverse_iterator arg1_it = arg1_bitstring.rbegin ();
      for (;
           arg1_it != arg1_bitstring.rend ();
           arg1_it++)
      {
         res.push_front(bit_xor_expr_map[*arg1_it][ONE]);
      }
   }
#endif
#if 1
   else if (op_kind == truth_and_expr_K)
   {
      truth_and_expr *operation = GetPointer<truth_and_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " && " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " &&");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      bit_lattice arg_left = ZERO;
      for(auto current_bit : arg1_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_left = ONE;
            break;
         }
         else if(current_bit == U)
            arg_left = U;
      }
      bit_lattice arg_right = ZERO;
      for(auto current_bit : arg2_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_right = ONE;
            break;
         }
         else if(current_bit == U)
            arg_right = U;
      }
      res.push_front(bit_and_expr_map[arg_left][arg_right]);
      if(is_signed(GET_INDEX_NODE(ga->op0)))
         res.push_front(ZERO);
   }
   else if (op_kind == truth_or_expr_K)
   {
      truth_or_expr *operation = GetPointer<truth_or_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " || " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " ||");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      bit_lattice arg_left = ZERO;
      for(auto current_bit : arg1_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_left = ONE;
            break;
         }
         else if(current_bit == U)
            arg_left = U;
      }
      bit_lattice arg_right = ZERO;
      for(auto current_bit : arg2_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_right = ONE;
            break;
         }
         else if(current_bit == U)
            arg_right = U;
      }
      res.push_front(bit_ior_expr_map[arg_left][arg_right]);
      if(is_signed(GET_INDEX_NODE(ga->op0)))
         res.push_front(ZERO);
   }
   else if (op_kind == truth_xor_expr_K)
   {
      truth_xor_expr *operation = GetPointer<truth_xor_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;

      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " ^ " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " ^");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      bit_lattice arg_left = ZERO;
      for(auto current_bit : arg1_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_left = ONE;
            break;
         }
         else if(current_bit == U)
            arg_left = U;
      }
      bit_lattice arg_right = ZERO;
      for(auto current_bit : arg2_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_right = ONE;
            break;
         }
         else if(current_bit == U)
            arg_right = U;
      }

      res.push_front(bit_xor_expr_map[arg_left][arg_right]);
      if(is_signed(GET_INDEX_NODE(ga->op0)))
         res.push_front(ZERO);
   }
   else if (op_kind == truth_not_expr_K)
   {
      truth_not_expr *operation = GetPointer<truth_not_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      unsigned int arg1_uid = 0;

      if (GET_NODE(operation->op)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = !" + STR(arg1_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " !" + bitstring_to_string (arg1_bitstring) + " =");

      bit_lattice arg_left = ZERO;
      for(auto current_bit : arg1_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_left = ONE;
            break;
         }
         else if(current_bit == U)
            arg_left = U;
      }
      res.push_front(bit_xor_expr_map[arg_left][ONE]);
      if(is_signed(GET_INDEX_NODE(ga->op0)))
         res.push_front(ZERO);
   }
#endif
#if 1
   else if (op_kind == rshift_expr_K)
   {
      rshift_expr *operation = GetPointer<rshift_expr> (GET_NODE(ga->op1));
      std::deque<bit_lattice> arg1_bitstring;
      unsigned int arg1_uid = 0;
      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         res = create_u_bitstring (static_cast<unsigned int>(arg1_bitstring.size ()));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " >> " + STR(GET_INDEX_NODE(operation->op1)));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string (res));
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(const2->value));
         if(const2->value < 0)
         {
            res.push_back (X);
            return res;
         }

         bool is_signed_var = is_signed(arg1_uid);
         if(arg1_bitstring.size () <= static_cast<size_t> (const2->value))
         {
            if(is_signed_var)
               res.push_front(arg1_bitstring.front());
            else
               res.push_front(ZERO);
         }
         else
         {
            size_t new_lenght = arg1_bitstring.size () - static_cast<size_t> (const2->value);
            std::deque<bit_lattice>::const_iterator arg1_it = arg1_bitstring.begin ();
            while (res.size () < new_lenght)
            {
               res.push_back (*arg1_it);
               arg1_it++;
            }
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " >> " + STR(GET_INDEX_NODE(operation->op1)));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string (arg1_bitstring) + ">>" + STR(const2->value) + " => " + bitstring_to_string (res));
      }
      else
         return res;
   }
#endif
   else if (op_kind == rrotate_expr_K)
   {
      rrotate_expr *operation = GetPointer<rrotate_expr> (GET_NODE(ga->op1));
      std::deque<bit_lattice> arg1_bitstring;
      unsigned int arg1_uid = 0;
      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         unsigned int precision = tree_helper::Size(tree_helper::CGetType(GET_NODE(ga->op0)));
         res = create_u_bitstring (precision);
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
         unsigned int arg2_value = static_cast<unsigned int> (const2->value);
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(arg2_value));
         unsigned int precision = tree_helper::Size(tree_helper::CGetType(GET_NODE(ga->op0)));

         if (precision > arg1_bitstring.size ())
         {
            arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, precision);
         }
         res = arg1_bitstring;
         for (unsigned int index = 0; index < arg2_value; ++index)
         {
            bit_lattice cur_bit = res.back();
            res.pop_back();
            res.push_front(cur_bit);
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " rrotate " + STR(GET_INDEX_NODE(operation->op1)));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " rrotate");
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, STR(arg2_value) + " =");
      }
      return res;
   }
   else if (op_kind == lrotate_expr_K)
   {
      lrotate_expr *operation = GetPointer<lrotate_expr> (GET_NODE(ga->op1));
      std::deque<bit_lattice> arg1_bitstring;
      unsigned int arg1_uid = 0;
      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         unsigned int precision = tree_helper::Size(tree_helper::CGetType(GET_NODE(ga->op0)));
         res = create_u_bitstring (precision);
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
         unsigned int arg2_value = static_cast<unsigned int> (const2->value);
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(arg2_value));
         unsigned int precision = tree_helper::Size(tree_helper::CGetType(GET_NODE(ga->op0)));

         if (precision > arg1_bitstring.size ())
         {
            arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, precision);
         }
         res = arg1_bitstring;
         for (unsigned int index = 0; index < arg2_value; ++index)
         {
            bit_lattice cur_bit = res.front();
            res.pop_front();
            res.push_back(cur_bit);
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " lrotate " + STR(GET_INDEX_NODE(operation->op1)));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " lrotate");
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, STR(arg2_value) + " =");
      }
      return res;
   }
#if 1
   else if (op_kind == lshift_expr_K)
   {
      lshift_expr *operation = GetPointer<lshift_expr> (GET_NODE(ga->op1));
      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg1_uid = 0;
      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;

      unsigned int res_bitsize = tree_helper::Size(GET_NODE(ga->op0));
      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         unsigned int arg2_uid;
         arg2_uid = GET_INDEX_NODE(operation->op1);
         if (current.find (arg2_uid) == current.end ())
         {
            return res;
         }
         arg2_bitstring = current.find (arg2_uid)->second;

         long long unsigned int bsize_elev2 = 1ULL<< arg2_bitstring.size ();
         if(res_bitsize < bsize_elev2 || res_bitsize < bsize_elev2+arg1_bitstring.size ())
            res = create_u_bitstring (res_bitsize);
         else
            res = create_u_bitstring (static_cast<unsigned int>(arg1_bitstring.size () + bsize_elev2));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " << " + STR(arg2_uid));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string (arg1_bitstring) + "<<" + bitstring_to_string(arg2_bitstring) + " => " + bitstring_to_string (res));
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         integer_cst *const2 = GetPointer<integer_cst> (GET_NODE(operation->op1));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(const2->value));
         if(const2->value < 0)
         {
            res.push_back (X);
            return res;
         }

         const unsigned int arg1_bitsize = tree_helper::Size(GET_NODE(operation->op0));
         if(res_bitsize <= static_cast<long long unsigned int> (const2->value))
            res.push_front(ZERO);
         else
         {
            res = arg1_bitstring;
            while(res.size()>arg1_bitsize)
               res.pop_front();
            for (int i = 0; i < const2->value; i++)
            {
               res.push_back (ZERO);
               if(res.size()>res_bitsize)
                  res.pop_front ();
            }
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " << " + STR(GET_INDEX_NODE(operation->op1)));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string (arg1_bitstring) + "<<" + STR(const2->value) + " => " + bitstring_to_string (res));
      }
      else
         return res;
   }
#endif
#if 1
   else if (op_kind == nop_expr_K || op_kind == convert_expr_K)
   {
      unary_expr *operation = GetPointer<unary_expr> (GET_NODE(ga->op1));
      const tree_managerRef TM = AppM->get_tree_manager ();
      unsigned int res_bitsize = tree_helper::Size(GET_NODE(ga->op0));
      if(best.find (GET_INDEX_NODE(ga->op0)) != best.end() && best.find(GET_INDEX_NODE(ga->op0))->second.size() < res_bitsize)
         res_bitsize = static_cast<unsigned int>(best.find(GET_INDEX_NODE(ga->op0))->second.size());
      if(tree_helper::is_real(TM, GET_INDEX_NODE(operation->op)) ||
            tree_helper::is_a_complex(TM, GET_INDEX_NODE(operation->op)) ||
            tree_helper::is_a_vector(TM, GET_INDEX_NODE(operation->op)))
      {
         res = create_u_bitstring(res_bitsize);
         return res;
      }
      if (GET_NODE(operation->op)->get_kind () == ssa_name_K)
      {
         unsigned int arg1_uid;
         arg1_uid = GET_INDEX_NODE(operation->op);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         const std::deque<bit_lattice> &op_lattice = current.find (arg1_uid)->second;
         res = op_lattice;
         while(res.size()>res_bitsize)
            res.pop_front();
         unsigned int res_uid = GET_INDEX_NODE(ga->op0);
         if((is_signed(arg1_uid) && is_signed(res_uid)) ||
               (!is_signed(arg1_uid) && !is_signed(res_uid)))
            ;
         else
         {
            if(res.size() == res_bitsize)
               ;
            else
               res = sign_extend_bitstring(op_lattice, arg1_uid, res_bitsize);
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid)+ (is_signed(output_uid)?"S":"U") + " = "+ (op_kind == nop_expr_K?"cast":"convert") + " " + STR(arg1_uid)+ (is_signed(arg1_uid)?"S":"U"));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string(res));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " = op:  " + bitstring_to_string(op_lattice) + "(" + STR(tree_helper::Size(GET_NODE(operation->op))) + "->" + STR(res_bitsize)+ ")");
      }
      else
         return res;
   }
#endif
#if 1
   else if (op_kind == cond_expr_K)
   {
      cond_expr *operation = GetPointer<cond_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      std::deque<bit_lattice> arg3_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;
      unsigned int arg3_uid = 0;

      if (GET_NODE(operation->op0)->get_kind () == ssa_name_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         if (current.find (arg1_uid) == current.end ())
         {
            return res;
         }
         arg1_bitstring = current.find (arg1_uid)->second;
      }
      else if (GET_NODE(operation->op0)->get_kind () == integer_cst_K)
      {
         arg1_uid = GET_INDEX_NODE(operation->op0);
         THROW_ASSERT(best.find (arg1_uid) != best.end(), "unexpected condition");
         arg1_bitstring = best.find (arg1_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op0))->value) + " : " + bitstring_to_string(arg1_bitstring));
      }
      else
         return res;


      if (GET_NODE(operation->op1)->get_kind () == ssa_name_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         if (current.find (arg2_uid) == current.end ())
         {
            return res;
         }
         arg2_bitstring = current.find (arg2_uid)->second;
      }
      else if (GET_NODE(operation->op1)->get_kind () == integer_cst_K)
      {
         arg2_uid = GET_INDEX_NODE(operation->op1);
         THROW_ASSERT(best.find (arg2_uid) != best.end(), "unexpected condition");
         arg2_bitstring = best.find (arg2_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op1))->value) + " : " + bitstring_to_string(arg2_bitstring));
      }
      else
         return res;

      if (GET_NODE(operation->op2)->get_kind () == ssa_name_K)
      {
         arg3_uid = GET_INDEX_NODE(operation->op2);
         if (current.find (arg3_uid) == current.end ())
         {
            return res;
         }
         arg3_bitstring = current.find (arg3_uid)->second;
      }
      else if (GET_NODE(operation->op2)->get_kind () == integer_cst_K)
      {
         arg3_uid = GET_INDEX_NODE(operation->op2);
         THROW_ASSERT(best.find (arg3_uid) != best.end(), "unexpected condition");
         arg3_bitstring = best.find (arg3_uid)->second;
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, created bitstring from constant -> " + STR(GetPointer<integer_cst> (GET_NODE(operation->op2))->value) + " : " + bitstring_to_string(arg3_bitstring));
      }
      else
         return res;

      bit_lattice arg_cond = ZERO;
      for(auto current_bit : arg1_bitstring)
      {
         if(current_bit == ONE)
         {
            arg_cond = ONE;
            break;
         }
         else if(current_bit == U)
            arg_cond = U;
      }

      if(arg_cond==ZERO)
      {
         //Condition is false return second arg (arg3_bitstring)
         res=arg3_bitstring;

      }
      else if(arg_cond==ONE)
      {
         //Condition is false return second arg (arg3_bitstring)
         res=arg2_bitstring;

      }
      else
      {
         //CONDITION IS UNKNOWN RETURN INF OF ARGS
         res = inf (arg2_bitstring,arg3_bitstring,arg2_uid,arg3_uid);
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string(arg1_bitstring) + " ? " + bitstring_to_string(arg2_bitstring) +" : " + bitstring_to_string(arg3_bitstring));
   }
#endif
#if 1
   else if(op_kind == min_expr_K || op_kind == max_expr_K )
   {
      binary_expr *operation = GetPointer<binary_expr> (GET_NODE(ga->op1));

      std::deque<bit_lattice> arg1_bitstring;
      std::deque<bit_lattice> arg2_bitstring;
      unsigned int arg2_uid = 0;
      unsigned int arg1_uid = 0;
      if(!manage_forward_binary_operands(operation, arg1_uid, arg2_uid,
                                              arg1_bitstring, arg2_bitstring))
         return res;

      if (arg1_bitstring.size () > arg2_bitstring.size ())
      {
         arg2_bitstring = sign_extend_bitstring (arg2_bitstring, arg2_uid, arg1_bitstring.size ());
      }
      if (arg2_bitstring.size () > arg1_bitstring.size ())
      {
         arg1_bitstring = sign_extend_bitstring (arg1_bitstring, arg1_uid, arg2_bitstring.size ());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer, operation: " + STR(output_uid) + " = " + STR(arg1_uid) + " " +(op_kind == min_expr_K ? std::string("min") : std::string("max")) +" " + STR(arg2_uid));
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg1_bitstring) + " " +(op_kind == min_expr_K ? std::string("min") : std::string("max")) +"");
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string (arg2_bitstring) + " =");

      res = inf (arg1_bitstring,arg2_bitstring,arg1_uid,arg2_uid);
   }
#endif
#if 1
   else if(op_kind == addr_expr_K || op_kind == mem_ref_K || op_kind == component_ref_K || op_kind == var_decl_K || op_kind == array_ref_K || op_kind == call_expr_K ||
           op_kind == view_convert_expr_K)
   {
      // do nothing
   }
#endif
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "forward_transfer Error: operation unhandled yet -> "+ GET_NODE(ga->op1)->get_kind_text());
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, bitstring_to_string(res));
   return res;
}

bool
Bit_Value::bitstring_equal (const std::deque<bit_lattice> &a, const std::deque<bit_lattice> &b)
{
   if (a.size () != b.size ())
   {
      return false;
   }
   bool equal = true;
   std::deque<bit_lattice>::const_iterator a_it = a.begin ();
   std::deque<bit_lattice>::const_iterator b_it = b.begin ();

   for (; a_it != a.end () && b_it != b.end (); a_it++, b_it++)
   {
      if (*a_it != *b_it)
      {
         equal = false;
         break;
      }
   }
   return equal;
}

bool
Bit_Value::test_bitstring_equal ()
{
   std::deque<bit_lattice> a;
   std::deque<bit_lattice> b;
   bool correct = true;

   a.push_front (ZERO);
   a.push_front (U);

   b = a;

   if (!bitstring_equal (a, b))
      correct = false;
   if (!bitstring_equal (b, a))
      correct = false;

   b.pop_front ();

   if (bitstring_equal (a, b))
      correct = false;
   if (bitstring_equal (b, a))
      correct = false;

   b.pop_front ();

   b.push_front (ONE);
   b.push_front (U);

   if (bitstring_equal (a, b))
      correct = false;
   if (bitstring_equal (b, a))
      correct = false;

   return correct;

}
void
Bit_Value::forward ()
{
   const tree_managerRef TM = AppM->get_tree_manager ();
   tree_nodeRef tn = TM->get_tree_node_const (function_id);
   function_decl * fd = GetPointer<function_decl> (tn);
   THROW_ASSERT(fd && fd->body, "Node is not a function or it hasn't a body");
   statement_list * sl = GetPointer<statement_list> (GET_NODE(fd->body));
   std::map<unsigned int, blocRef>::const_iterator B_it_end = sl->list_of_bloc.end ();

   bool current_updated = true;

   while (current_updated)
   {
      current_updated = false;

      // for each basic block B in CFG do > Consider all blocks successively
      for (std::map<unsigned int, blocRef>::const_iterator B_it = sl->list_of_bloc.begin ();
           B_it != B_it_end; ++B_it)
      {
         blocRef B = B_it->second;
         for(const auto stmt : B->CGetStmtList())
         {
            if (GET_NODE(stmt)->get_kind () == gimple_assign_K)
            {
               gimple_assign * ga = GetPointer<gimple_assign> (GET_NODE(stmt));
               unsigned int output_uid = GET_INDEX_NODE(ga->op0);
               ssa_name *ssa = GetPointer<ssa_name> (GET_NODE(ga->op0));

               if (ssa)
               {
                  if(tree_helper::is_real(TM, output_uid))
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "real variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }
                  if(tree_helper::is_a_complex(TM, output_uid))
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "complex variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }
                  if(tree_helper::is_a_vector(TM, output_uid))
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "vector variables not considered: "+ STR(GET_INDEX_NODE(ga->op0)));
                     continue;
                  }


                  THROW_ASSERT(best.find (output_uid) != best.end(), "unexpected condition");
                  if (current.find (output_uid) == current.end ())
                  {
                     current[output_uid] = best.find (output_uid)->second;
                  }

                  std::deque<bit_lattice> res = forward_transfer (ga);

                  if(res.size () > 0)
                  {
                     sign_reduce_bitstring(res,output_uid);
                  }
                  if ((res.size () > 0 && !bitstring_equal (current.find (output_uid)->second, sup (res, best.find (output_uid)->second, is_signed(output_uid)))))
                  {
                     current[output_uid] = sup (res, best.find (output_uid)->second, is_signed(output_uid));
                     current_updated = true;
                  }

                  // PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "current at the end of forward_transfer:");
                  // print_bitstring_map (current);
               }
            }
         }

         for(const auto phi : B->CGetPhiList())
         {
            gimple_phi * pn = GetPointer<gimple_phi>(GET_NODE(phi));
            bool is_virtual = pn->virtual_flag;
            if(!is_virtual)
            {
               unsigned int output_uid = GET_INDEX_NODE(pn->res);

               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "phi: "+STR(GET_INDEX_NODE(phi)));
               if(tree_helper::is_real(TM, output_uid))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "real variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
                  continue;
               }
               if(tree_helper::is_a_complex(TM, output_uid))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "complex variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
                  continue;
               }
               if(tree_helper::is_a_vector(TM, output_uid))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "vector variables not considered: "+ STR(GET_INDEX_NODE(pn->res)));
                  continue;
               }

               THROW_ASSERT(best.find (output_uid) != best.end(), "unexpected condition");
               if (current.find (output_uid) == current.end ())
               {
                  current[output_uid] = best.find (output_uid)->second;
               }

               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res id: " + STR(output_uid));
               std::deque<bit_lattice>res;
               res.push_back(X);
               for(const auto def_edge : pn->CGetDefEdgesList())
               {
                  if(current.find(GET_INDEX_NODE(def_edge.first)) == current.end())
                  {
                     if(best.find(GET_INDEX_NODE(def_edge.first)) == best.end())
                        current[GET_INDEX_NODE(def_edge.first)] = create_u_bitstring(tree_helper::Size(GET_NODE(pn->res)));
                     else
                        current[GET_INDEX_NODE(def_edge.first)] = best.find(GET_INDEX_NODE(def_edge.first))->second;
                  }
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Edge " + STR(def_edge.second) + ": " + bitstring_to_string(current.find(GET_INDEX_NODE(def_edge.first))->second));

                  res = inf (res,current.find(GET_INDEX_NODE(def_edge.first))->second,output_uid,GET_INDEX_NODE(def_edge.first));
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "current res: " + bitstring_to_string(res));
               }
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "res: " + bitstring_to_string(res));

               if(res.size () > 0)
               {
                  sign_reduce_bitstring(res,output_uid);
               }
               if ((res.size () > 0 && !bitstring_equal (current.find (output_uid)->second, sup (res, best.find (output_uid)->second, is_signed(output_uid)))))
               {
                  current[output_uid] = sup (res, best.find (output_uid)->second, is_signed(output_uid));
                  current_updated = true;
               }
            }
         }

      }



   }
}
void Bit_Value::initialize_minus_expr_map(){
   //          a  b  borrow
   minus_expr_map[X][X][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][X][ONE]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][X][U]= create_bitstring_couple(ZERO,ZERO);

   minus_expr_map[X][ZERO][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][ZERO][ONE]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][ZERO][U]= create_bitstring_couple(ZERO,ZERO);

   minus_expr_map[X][ONE][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][ONE][ONE]= create_bitstring_couple(ONE,ZERO);
   minus_expr_map[X][ONE][U]= create_bitstring_couple(U,U);

   minus_expr_map[X][U][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[X][U][ONE]= create_bitstring_couple(ZERO,U);
   minus_expr_map[X][U][U]= create_bitstring_couple(U,U);



   minus_expr_map[ZERO][X][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ZERO][X][ONE]= create_bitstring_couple(ZERO,ONE);
   minus_expr_map[ZERO][X][U]= create_bitstring_couple(U,U);

   minus_expr_map[ZERO][ZERO][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ZERO][ZERO][ONE]= create_bitstring_couple(ONE,ONE);
   minus_expr_map[ZERO][ZERO][U]= create_bitstring_couple(U,U);

   minus_expr_map[ZERO][ONE][ZERO]= create_bitstring_couple(ONE,ONE);
   minus_expr_map[ZERO][ONE][ONE]= create_bitstring_couple(ONE,ZERO);
   minus_expr_map[ZERO][ONE][U]= create_bitstring_couple(ONE,U);

   minus_expr_map[ZERO][U][ZERO]= create_bitstring_couple(U,U);
   minus_expr_map[ZERO][U][ONE]= create_bitstring_couple(ONE,U);
   minus_expr_map[ZERO][U][U]= create_bitstring_couple(U,U);



   minus_expr_map[ONE][X][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ONE][X][ONE]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ONE][X][U]= create_bitstring_couple(ZERO,U);

   minus_expr_map[ONE][ZERO][ZERO]= create_bitstring_couple(ZERO,ONE);
   minus_expr_map[ONE][ZERO][ONE]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ONE][ZERO][U]= create_bitstring_couple(ZERO,U);

   minus_expr_map[ONE][ONE][ZERO]= create_bitstring_couple(ZERO,ZERO);
   minus_expr_map[ONE][ONE][ONE]= create_bitstring_couple(ONE,ONE);
   minus_expr_map[ONE][ONE][U]= create_bitstring_couple(U,U);

   minus_expr_map[ONE][U][ZERO]= create_bitstring_couple(ZERO,U);
   minus_expr_map[ONE][U][ONE]= create_bitstring_couple(U,U);
   minus_expr_map[ONE][U][U]= create_bitstring_couple(U,U);



   minus_expr_map[U][X][ZERO]= create_bitstring_couple(ZERO,U);
   minus_expr_map[U][X][ONE]= create_bitstring_couple(U,U);
   minus_expr_map[U][X][U]= create_bitstring_couple(U,U);

   minus_expr_map[U][ZERO][ZERO]= create_bitstring_couple(ZERO,U);
   minus_expr_map[U][ZERO][ONE]= create_bitstring_couple(U,U);
   minus_expr_map[U][ZERO][U]= create_bitstring_couple(U,U);

   minus_expr_map[U][ONE][ZERO]= create_bitstring_couple(U,U);
   minus_expr_map[U][ONE][ONE]= create_bitstring_couple(ONE,U);
   minus_expr_map[U][ONE][U]= create_bitstring_couple(U,U);

   minus_expr_map[U][U][ZERO]= create_bitstring_couple(U,U);
   minus_expr_map[U][U][ONE]= create_bitstring_couple(U,U);
   minus_expr_map[U][U][U]= create_bitstring_couple(U,U);
}
void Bit_Value::initialize_plus_expr_map(){
   //      a  b  carry
   plus_expr_map[X][X][ZERO]= create_bitstring_couple(ZERO,X);
   plus_expr_map[X][X][ONE]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[X][X][U]= create_bitstring_couple(ZERO,U);

   plus_expr_map[X][ZERO][ZERO]= create_bitstring_couple(ZERO,ZERO);
   plus_expr_map[X][ZERO][ONE]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[X][ZERO][U]= create_bitstring_couple(ZERO,U);

   plus_expr_map[X][ONE][ZERO]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[X][ONE][ONE]= create_bitstring_couple(ONE,ZERO);
   plus_expr_map[X][ONE][U]= create_bitstring_couple(U,U);

   plus_expr_map[X][U][ZERO]= create_bitstring_couple(ZERO,U);
   plus_expr_map[X][U][ONE]= create_bitstring_couple(U,U);
   plus_expr_map[X][U][U]= create_bitstring_couple(U,U);



   plus_expr_map[ZERO][X][ZERO]= create_bitstring_couple(ZERO,ZERO);
   plus_expr_map[ZERO][X][ONE]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[ZERO][X][U]= create_bitstring_couple(ZERO,U);

   plus_expr_map[ZERO][ZERO][ZERO]= create_bitstring_couple(ZERO,ZERO);
   plus_expr_map[ZERO][ZERO][ONE]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[ZERO][ZERO][U]= create_bitstring_couple(ZERO,U);

   plus_expr_map[ZERO][ONE][ZERO]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[ZERO][ONE][ONE]= create_bitstring_couple(ONE,ZERO);
   plus_expr_map[ZERO][ONE][U]= create_bitstring_couple(U,U);

   plus_expr_map[ZERO][U][ZERO]= create_bitstring_couple(ZERO,U);
   plus_expr_map[ZERO][U][ONE]= create_bitstring_couple(U,U);
   plus_expr_map[ZERO][U][U]= create_bitstring_couple(U,U);


   plus_expr_map[ONE][X][ZERO]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[ONE][X][ONE]= create_bitstring_couple(ONE,ZERO);
   plus_expr_map[ONE][X][U]= create_bitstring_couple(U,U);

   plus_expr_map[ONE][ZERO][ZERO]= create_bitstring_couple(ZERO,ONE);
   plus_expr_map[ONE][ZERO][ONE]= create_bitstring_couple(ONE,ZERO);
   plus_expr_map[ONE][ZERO][U]= create_bitstring_couple(U,U);

   plus_expr_map[ONE][ONE][ZERO]= create_bitstring_couple(ONE,ZERO);
   plus_expr_map[ONE][ONE][ONE]= create_bitstring_couple(ONE,ONE);
   plus_expr_map[ONE][ONE][U]= create_bitstring_couple(ONE,U);

   plus_expr_map[ONE][U][ZERO]= create_bitstring_couple(U,U);
   plus_expr_map[ONE][U][ONE]= create_bitstring_couple(ONE,U);
   plus_expr_map[ONE][U][U]= create_bitstring_couple(U,U);



   plus_expr_map[U][X][ZERO]= create_bitstring_couple(ZERO,U);
   plus_expr_map[U][X][ONE]= create_bitstring_couple(U,U);
   plus_expr_map[U][X][U]= create_bitstring_couple(U,U);

   plus_expr_map[U][ZERO][ZERO]= create_bitstring_couple(ZERO,U);
   plus_expr_map[U][ZERO][ONE]= create_bitstring_couple(U,U);
   plus_expr_map[U][ZERO][U]= create_bitstring_couple(U,U);

   plus_expr_map[U][ONE][ZERO]= create_bitstring_couple(U,U);
   plus_expr_map[U][ONE][ONE]= create_bitstring_couple(ONE,U);
   plus_expr_map[U][ONE][U]= create_bitstring_couple(U,U);

   plus_expr_map[U][U][ZERO]= create_bitstring_couple(U,U);
   plus_expr_map[U][U][ONE]= create_bitstring_couple(U,U);
   plus_expr_map[U][U][U]= create_bitstring_couple(U,U);
}
void Bit_Value::initialize_bit_ior_expr_map(){

   bit_ior_expr_map[ZERO][ZERO] = ZERO;
   bit_ior_expr_map[ZERO][ONE] = ONE;
   bit_ior_expr_map[ZERO][U] = U;
   bit_ior_expr_map[ZERO][X] = X;

   bit_ior_expr_map[ONE][ZERO] = ONE;
   bit_ior_expr_map[ONE][ONE] = ONE;
   bit_ior_expr_map[ONE][U] = ONE;
   bit_ior_expr_map[ONE][X] = ONE;

   bit_ior_expr_map[U][ZERO] = U;
   bit_ior_expr_map[U][ONE] = ONE;
   bit_ior_expr_map[U][U] = U;
   bit_ior_expr_map[U][X] = X;

   bit_ior_expr_map[X][ZERO] = X;
   bit_ior_expr_map[X][ONE] = ONE;
   bit_ior_expr_map[X][U] = X;
   bit_ior_expr_map[X][X] = X;


}
void Bit_Value::initialize_bit_xor_expr_map(){
   bit_xor_expr_map[ZERO][ZERO] = ZERO;
   bit_xor_expr_map[ZERO][ONE] = ONE;
   bit_xor_expr_map[ZERO][U] = U;
   bit_xor_expr_map[ZERO][X] = X;

   bit_xor_expr_map[ONE][ZERO] = ONE;
   bit_xor_expr_map[ONE][ONE] = ZERO;
   bit_xor_expr_map[ONE][U] = U;
   bit_xor_expr_map[ONE][X] = X;

   bit_xor_expr_map[U][ZERO] = U;
   bit_xor_expr_map[U][ONE] = U;
   bit_xor_expr_map[U][U] = U;
   bit_xor_expr_map[U][X] = X;

   bit_xor_expr_map[X][ZERO] = X;
   bit_xor_expr_map[X][ONE] = X;
   bit_xor_expr_map[X][U] = X;
   bit_xor_expr_map[X][X] = X;
}
void Bit_Value::initialize_bit_and_expr_map(){
   bit_and_expr_map[ZERO][ZERO] = ZERO;
   bit_and_expr_map[ZERO][ONE] = ZERO;
   bit_and_expr_map[ZERO][U] = ZERO;
   bit_and_expr_map[ZERO][X] = ZERO;

   bit_and_expr_map[ONE][ZERO] = ZERO;
   bit_and_expr_map[ONE][ONE] = ONE;
   bit_and_expr_map[ONE][U] = U;
   bit_and_expr_map[ONE][X] = X;

   bit_and_expr_map[U][ZERO] = ZERO;
   bit_and_expr_map[U][ONE] = U;
   bit_and_expr_map[U][U] = U;
   bit_and_expr_map[U][X] = X;

   bit_and_expr_map[X][ZERO] = ZERO;
   bit_and_expr_map[X][ONE] = X;
   bit_and_expr_map[X][U] = X;
   bit_and_expr_map[X][X] = X;
}
void Bit_Value::run_tests(){
   if (!test_sup ())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::sup test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::sup test: OK");

   if (!test_sign_extend_bitstring ())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_sign_extend_bitstring test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_sign_extend_bitstring test: OK");

   if (!test_sign_extend_bitstring ())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_inf test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_inf test: OK");

   if (!test_bitstring_equal ())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_bitstring_equal test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_bitstring_equal test: OK");

   if (!test_substitute_bit ())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_substitute_bit test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_substitute_bit test: OK");

   if (!test_get_element_from_bitstring())
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_get_element_from_bitstring test: ERROR");
   else
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Bit_Value::test_get_element_from_bitstring test: OK");
}

DesignFlowStep_Status
Bit_Value::InternalExec ()
{
   if(debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      PrintTreeManager(true);
   }

#ifndef NDEBUG
   run_tests();
#endif

   initialize_plus_expr_map();
   initialize_minus_expr_map();
   initialize_bit_ior_expr_map();
   initialize_bit_xor_expr_map();
   initialize_bit_and_expr_map();
   bool restart;

   // Initialize best map with C type as bitstring
   initialize ();
   backward();
   mix ();
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best at end of initial backward:");
   print_bitstring_map (best);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "");
   do
   {
      // Clears the current map leaving just the inputs argument of the functions with the best's values
      clear ();
      forward ();
      mix ();
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best at end of forward:");
      print_bitstring_map (best);
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "");

      backward();
      restart = mix ();
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best at end of backward:");
      print_bitstring_map (best);
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "");
   } while (restart);

   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "best at end of alg:");
   print_bitstring_map (best);

   update_IR(best);

   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      PrintTreeManager(false);
   }
   return DesignFlowStep_Status::SUCCESS;

}

void Bit_Value::Initialize()
{
   const std::string bambu_frontend_flow_signature = ApplicationFrontendFlowStep::ComputeSignature(FrontendFlowStepType::BAMBU_FRONTEND_FLOW);
   not_frontend = design_flow_manager.lock()->GetStatus(bambu_frontend_flow_signature) == DesignFlowStep_Status::EMPTY;
}
